
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Lindsay Martin and Kobus Barnard (authors). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

/* Kobus: 07-07-99 */
#define POSTERIOR_PLOT_RG_RESOLUTION 50

/*
// Kobus (01/08/99) (Added compile switch--default is as it was before.)
//
*/
#define USE_ONLY_PRESENCE_INFO 

#define POSTERIOR_RESOLUTION 200

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "spectra.h"
#include "illum_data.h"
#include "corr_lib.h"
#include "correlation.h"

/* ------------------------------------------------------------------------- */

#define CORR_01_DATA_FILE               "corr_01.cormat" 
#define CORR_UNIFORM_INV_DIST_DATA_FILE "corr_uniform_inv_dist.cormat"
#define CORR_UNIFORM_GAUSSIAN_DATA_FILE "corr_uniform_gaussian.cormat"
#define CORR_GAUSSIAN_DATA_FILE         "corr_gaussian.cormat"

#define CORR_MASK_SIZE               5
#define CORR_MASK_SIGMA              1.0

#define CORR_INVDIST_BOX_SIZE        5
#define CORR_INVDIST_NUM_NEIGHBOURS  5

/* ------------------------------------------------------------------------- */

static char fs_illum_rgb_index_file[ MAX_FILE_NAME_SIZE ]; 
static char fs_surface_rgb_index_file[ MAX_FILE_NAME_SIZE ]; 

/* 0/1 Correlation Data Matrix.                                              */
static Corr_matrix* fs_cached_01_data_cp[NUM_CORMATS_PER_ALG];
static char fs_z1_file[NUM_CORMATS_PER_ALG][MAX_FILE_NAME_SIZE];
static int  fs_z1_version[NUM_CORMATS_PER_ALG];

/* Log of uniform probability data with zero entries filled using a 
// gaussian mask convolution algorithm.                                      */
static Corr_matrix* fs_cached_log_uniform_gaussian_cp[NUM_CORMATS_PER_ALG];
static char fs_log_uniform_gaussian_file[NUM_CORMATS_PER_ALG][MAX_FILE_NAME_SIZE];
static int  fs_log_uniform_gaussian_version[NUM_CORMATS_PER_ALG];

static Matrix* fs_posterior_dist_mp = NULL; 
static Matrix* fs_mlm_posterior_dist_mp = NULL; 
static int  fs_correlation_mlm_mask_size = 25;
static double fs_correlation_mlm_mask_sigma = 8;

/*---------------------------------------------------------------------------*/

/* User-Settable Options ----------------------------------------------------*/

/* Enable/Disable correlation commands.
//
// Enable to allow the making of correlation matrices and running correlation
// colour constancy algorithms. This switch is global to ALL correlation
// algorithms.                                                               */
static int  fs_settable_corr_enabled  = TRUE;

/* Default number of histogram bins per chromaticity axis.
//
// NOTE: This value applies to all correlation matrices when they are
//       constructed with a "make_*" command.                                */
static int  fs_settable_num_bins = 20;

/* Default mask size for the gaussian smoothing operation
// applied to Uniform Probability (Gaussian Fill) Correlation Matrix.        */
static int  fs_settable_mask_size = CORR_MASK_SIZE;

/* Default standard deviation for the gaussian smoothing operation
// applied to Uniform Probability (Gaussian Fill) Correlation Matrix.        */
static double fs_settable_mask_sigma = CORR_MASK_SIGMA;

/*
// Kobus (28/07/99): A replacement for CORR_THRESHOLD. On second thought, this
// parameter does not do much--any reasonable value should give a similar
// answer.
*/
static double fs_corr_averaging_threshold = 0.95; 

/*---------------------------------------------------------------------------*/

static int initialize_01_data(int index);

static int read_01_data_from_file(int index, const char* file_name);

static int write_01_data(int index, char* file_name);

static int set_01_data_matrix
(
    int          index,
    Corr_matrix* new_01_data_cp,
    const char*  config_file 
);

static int make_01_data(int index, char* file_name);

static int insert_01_histogram
(
    Corr_matrix** cormat_cp_ptr,
    int           cur_illum,
    Matrix*       chrom_mp,
    int           num_bins,
    double        max_projection_coordinate 
);

static int bin_hull_facet
(
    Matrix** hull_bins_mpp,
    Matrix*  cur_facet_mp,
    int      num_bins,
    double   max_projection_coordinate 
);

static int do_01
(
    int      index,
    Matrix*  data_mp,
    int      prompt_to_increase_error_flag,
    Vector** result_vpp 
);

static int do_01_with_error
(
    int      index,       /* Index of matrix to use   */
    Matrix*  rgb_data_mp, /* Input RGB data           */
    Vector** result_vpp   /* Output RGB of illuminant */              
);

static int initialize_uniform_gaussian_fill_data(int index);

static int read_uniform_gaussian_fill_data_from_file
(
    int         index,
    const char* file_name 
);

static int write_uniform_gaussian_fill_data(int index, char* file_name);

static int set_uniform_gaussian_fill_data_matrix
(
    int          index,
    Corr_matrix* new_uniform_gaussian_data_cp,
    const char*  config_file 
);

static int make_log_uniform_gaussian_fill_data
(
    int   index,
    char* file_name 
);

/* Kobus: 08/07/99 | Added plot_enable arg. */
static int do_uniform_gaussian_fill
(
    int      index,
    Matrix*  rgb_data_mp,
    Vector** result_vpp,
    Vector** mlm_result_vpp,
    Vector** mmse_result_vpp,
    double*  cpu_time_ptr,
    double*  mlm_cpu_time_ptr,
    double*  mmse_cpu_time_ptr,
    int      plot_enable 
);

static int make_uniform_probability_data
(
    Corr_matrix** new_unform_data_cp_ptr,
    int*          world_version_ptr,
    int*          illum_spectra_version_ptr 
);

static int insert_uniform_histogram
(
    Matrix** uniform_cormat_mpp,
    int      illum,
    Matrix*  chrom_mp,
    int      num_bins,
    double   max_projection_coordinate 
);

/* Kobus: 08/07/99 | Added id_str and plot_enable args. */
static int do_probability_algorithm
(
    Matrix*      rgb_data_mp,
    Vector**     result_vpp,
    Vector**     mlm_result_vpp,
    Vector**     mmse_result_vpp,
    double*      cpu_time_ptr,
    double*      mlm_cpu_time_ptr,
    double*      mmse_cpu_time_ptr,
    Corr_matrix* source_prob_data_cp,
    int          plot_enable,
    char*        id_str 
);

static int map_2D_to_1D_histogram
(
    Vector** hist_1D_vpp,
    Matrix*  hist_2D_mp 
);

static int map_1D_to_2D_histogram
(
    Matrix** hist_2D_mpp,
    Vector*  hist_1D_vp 
);

static int bin_chromaticity
(
    double chrom_value,
    int    num_bins,
    double max_projection_coordinate 
);

static int get_gaussian_mask
(
    Matrix** mask_mpp,
    int      mask_size,
    double   sigma 
);

static int convolve
(
    Matrix** output_mpp,
    Matrix*  input_mp,
    Matrix*  mask_mp 
);

/* Kobus: 07-07-99 */
static void plot_posterior
(
    const Matrix* illum_chrom_mp,
    Vector* sum_of_log_probs_vp,
    char*   id_str 
);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* --------------------------------------------------------------------------*/
/* EXPORTED CORRELATION FUNCTIONS -------------------------------------------*/
/* --------------------------------------------------------------------------*/

int set_correlation_options(const char* option, const char* value)
{
    int  temp_boolean_value;
    int  temp_int;
    double temp_double;
    char lc_option[ 100 ];
    int  i, index, first_index, second_index;
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    /*
    // Options that are same for ALL variants -------------------------------
    */

    /* Enable or disable ALL correlation commands. */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-enable"))
        || (match_pattern(lc_option, "correlation-enable")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-enable = %s\n", 
                    fs_settable_corr_enabled ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_settable_corr_enabled == TRUE)
            {
                ERE(pso("2D Correlation algorithms enabled.\n"));
            }
            else
            {
                ERE(pso("2D Correlation algorithms DISABLED.\n"));
            }
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_settable_corr_enabled = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    /*
    // Set the number of bins to use along each chromaticiy axis. All
    // algorithms within this group share this value.
    */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-num-bins"))
        || (match_pattern(lc_option, "correlation-num-bins")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-num-bins = %d\n",
                    fs_settable_num_bins));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Number of bins per chromaticity axis is %d.\n",
                    fs_settable_num_bins));
        }
        else
        {
            ERE(ss1pi(value, &temp_int)); 
            fs_settable_num_bins = temp_int;
            
            /* Invalidate ALL the correlation matrices. */
            for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
            {
                (fs_z1_version[i])++;
                (fs_log_uniform_gaussian_version[i])++;
            }
        }
        result = NO_ERROR; 
    }

    /*
    // Set the mask size for gaussian fill function.
    */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-gauss-mask-size"))
        || (match_pattern(lc_option, "correlation-gaussian-mask-size")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-gaussian-mask-size = %d\n",
                    fs_settable_mask_size));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Mask size for gaussian fill function is %d.\n",
                    fs_settable_mask_size));
        }
        else
        {
            ERE(ss1pi(value, &temp_int)); 
            fs_settable_mask_size = temp_int;
            if ( IS_EVEN(fs_settable_mask_size) )
                fs_settable_mask_size += 1;

            for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
                (fs_log_uniform_gaussian_version[i])++;
        }
        
        result = NO_ERROR;
    }

    /*
    // Set a sigma value for gaussian fill function.
    */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-gauss-mask-sigma"))
        || (match_pattern(lc_option, "correlation-gaussian-mask-sigma")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-gaussian-mask-sigma = %f\n",
                    fs_settable_mask_sigma));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Sigma for gaussian fill function is %f.\n",
                    fs_settable_mask_sigma));
        }
        else
        {
            ERE(ss1d(value, &temp_double)); 
            fs_settable_mask_sigma = temp_double;

            for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
                (fs_log_uniform_gaussian_version[i])++;
        }
        
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-mlm-mask-size"))
        || (match_pattern(lc_option, "correlation-mlm-mask-size")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-mlm-mask-size = %d\n",
                    fs_correlation_mlm_mask_size));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Mask size for correlation MLM estimate is %d.\n",
                    fs_correlation_mlm_mask_size));
        }
        else
        {
            ERE(ss1pi(value, &temp_int)); 
            fs_correlation_mlm_mask_size = temp_int;

            if ( IS_EVEN(fs_correlation_mlm_mask_size) )
                fs_correlation_mlm_mask_size += 1;
        }
        
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-mlm-mask-sigma"))
        || (match_pattern(lc_option, "correlation-mlm-mask-sigma")))
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-mlm-mask-sigma = %f\n",
                    fs_correlation_mlm_mask_sigma));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Sigma for correlation MLM estimate is %f.\n",
                    fs_correlation_mlm_mask_sigma));
        }
        else
        {
            ERE(ss1d(value, &temp_double)); 
            fs_correlation_mlm_mask_sigma = temp_double;
        }
        
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "correlation-averaging-threshold"))
        || (match_pattern(lc_option, "correlation-threshold"))
        || (match_pattern(lc_option, "corr-averaging-threshold"))
        || (match_pattern(lc_option, "corr-ave-threshold"))
        || (match_pattern(lc_option, "corr-threshold"))
       )
    {
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("correlation-averaging-threshold = %f\n",
                    fs_corr_averaging_threshold));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Threshold for combining near max O/1 correlation results is %f.\n",
                    fs_corr_averaging_threshold));
        }
        else
        {
            ERE(ss1d(value, &temp_double)); 
            fs_corr_averaging_threshold = temp_double;
        }
        
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "corr-illum-rgb-index"))
         || (match_pattern(lc_option, "corr-illum-rgb-index-file"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("corr-illum-rgb-index-file = %s\n",
                    fs_illum_rgb_index_file[ 0 ] ? fs_illum_rgb_index_file
                                                  : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_illum_rgb_index_file[ 0 ] == '\0') 
            {
                ERE(pso("No correlation illum rgb index file has been set.\n"));
            }
            else 
            {
                ERE(pso("Correlation illum rgb index file is %s.\n", 
                        fs_illum_rgb_index_file));
            }
        }
        else if (is_no_value_word(lc_option))
        {
            fs_illum_rgb_index_file[ 0 ] = '\0'; 
        }
        else
        {
            BUFF_CPY(fs_illum_rgb_index_file, value); 
        }
    dbs(fs_illum_rgb_index_file);
        
        result = NO_ERROR; 
    }    

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "corr-surface-rgb-index"))
         || (match_pattern(lc_option, "corr-surface-rgb-index-file"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("corr-surface-rgb-index-file = %s\n",
                    fs_surface_rgb_index_file[ 0 ] ? fs_surface_rgb_index_file
                                                  : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_surface_rgb_index_file[ 0 ] == '\0') 
            {
                ERE(pso("No correlation surface rgb index file has been set.\n"));
            }
            else 
            {
                ERE(pso("Correlation surface rgb index file is %s.\n", 
                        fs_surface_rgb_index_file));
            }
        }
        else if (is_no_value_word(lc_option))
        {
            fs_surface_rgb_index_file[ 0 ] = '\0'; 
        }
        else
        {
            BUFF_CPY(fs_surface_rgb_index_file, value); 
        }
        
        result = NO_ERROR; 
    }    

    /*
    // Variant (i.e. array element) specific options ------------------------
    */
    ERE(parse_index(lc_option, NUM_CORMATS_PER_ALG,
                    &first_index, &second_index, &index, (int *)NULL));
    
    /* Set the 0/1 correlation matrix data file */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-01-file"))
        || (match_pattern(lc_option, "correlation-01-file")))
    {               
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG,
                              "0/1 Correlation Algorithm"));

        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {    
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_z1_file[i];

                ERE(pso("correlation-01-file:%d = %s\n",
                        i + 1, *file_name ? file_name : "off"));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_z1_file[i];

                ERE(pso("0/1 correlation matrix %d data source is %s\n",
                        i + 1, *file_name ? file_name : "<none>"));
            }
        }
        else
        {
            ERE(read_01_data_from_file(index, value));
        }
        result = NO_ERROR; 
    }
    
    /* Set the uniform probability (gaussian mask fill) matrix data file */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "corr-uniform-gauss-file"))
        || (match_pattern(lc_option, "corr-uniform-gaussian-file"))
        || (match_pattern(lc_option, "correlation-uniform-gauss-file"))
        || (match_pattern(lc_option, "correlation-uniform-gaussian-file"))
      )
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG,
                              "Uniform Probability (Gaussian Mask Fill) Correlation Algorithm"));
        
        if (value == NULL) 
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_log_uniform_gaussian_file[i];

                ERE(pso("correlation-uniform-gaussian-file:%d = %s\n",
                        i + 1, *file_name ? file_name : "off"));
            }            
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_log_uniform_gaussian_file[i];

                ERE(pso("Uniform correlation (gaussian mask fill) matrix %d data source is %s\n",
                        i + 1, *file_name ? file_name : "<none>"));
            }
        }
        else
        {
            ERE(read_uniform_gaussian_fill_data_from_file(index, value));
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

int make_correlation(char* make_directive, char* file_name)
{
    char lc_make_directive[ 100 ];
    int  result = NOT_FOUND;
    /* Kobus: Since we don't have an enable, we don't use these. */
    /* int  i, first_index, second_index; */
    int  index;

    if (fs_settable_corr_enabled == FALSE)
    {
        ERE(verbose_pso(5, "Correlation disabled. Skipping all correlation makes.\n"));
        return NOT_FOUND;
    }

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }
    
    ERE(parse_index(lc_make_directive, NUM_CORMATS_PER_ALG, (int*)NULL, 
                    (int*)NULL, &index, (int*)NULL));

    if (    STRCMP_EQ(lc_make_directive, "all") 
         || STRCMP_EQ(lc_make_directive, "corr-all")
         || STRCMP_EQ(lc_make_directive, "correlation-all")
         || STRCMP_EQ(lc_make_directive, "corr-01")
         || STRCMP_EQ(lc_make_directive, "correlation-01")
       )
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG,
                              "0/1 Correlation"));

        ERE(make_01_data(index, file_name));
        result = NO_ERROR;
    }

    if (    STRCMP_EQ(lc_make_directive, "all") 
         || STRCMP_EQ(lc_make_directive, "corr-all")
         || STRCMP_EQ(lc_make_directive, "correlation-all")
         || STRCMP_EQ(lc_make_directive, "corr-uniform-gauss")
         || STRCMP_EQ(lc_make_directive, "corr-uniform-gaussian")
         || STRCMP_EQ(lc_make_directive, "correlation-uniform-gaussian-fill")
       )
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG,
                              "Uniform Correlation (Gaussian Mask Fill)"));
       
        ERE(make_log_uniform_gaussian_fill_data(index, file_name));

        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Kobus: Added cc_methods_ptr argument for finer tuning of error handling. */
int do_correlation
(
    Colour_constancy_methods* cc_methods_ptr,
    Matrix*                   data_mp,
    int                       prompt_to_increase_error_flag,
    Correlation_result*       corr_result_ptr,
    Correlation_error*        cpu_time_ptr 
)
{
    long       cpu_time_in_ms;
    double     cpu_time; 
    Vector*    illum_rgb_vp          = NULL;
    Vector*    mlm_illum_rgb_vp      = NULL;
    Vector*    mmse_illum_rgb_vp      = NULL;
    Vector*    rgb_map_vp            = NULL;
    int        i;
    int        result = NO_ERROR;

    if (fs_settable_corr_enabled == FALSE)
    {
        ERE(verbose_pso(5, "Correlation disabled. Skipping all algorithms.\n"));
        /*
        // return NOT_AVAILABLE;
        //
        // Kobus: Changed to NO_ERROR;   07/07/99
        */
        return NO_ERROR; 
    }
    
    /*
    // Use 0/1 Correlation Matrix -----------------------------------------
    */
    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        if (result == ERROR) break;

        if (initialize_01_data(i) != ERROR)
        {
            init_cpu_time();

            result = do_01(i, data_mp, prompt_to_increase_error_flag, 
                           &illum_rgb_vp);
            
            if (    (result != ERROR)
                 && (result != NOT_AVAILABLE)
                 && (result != NO_SOLUTION)
               ) 
            {
                cpu_time_in_ms = get_cpu_time(); 
                cpu_time = ((double)cpu_time_in_ms) /
                                             ((double)MILLI_SECONDS_PER_SECOND);

                cpu_time_ptr->correlation_01[i] = cpu_time;

                result = get_map_to_canon_illum(&rgb_map_vp, illum_rgb_vp);
            }
        
            if (    (result != ERROR)
                 && (result != NOT_AVAILABLE)
                 && (result != NO_SOLUTION)
               ) 
            {
                result = get_projection_vector(&(corr_result_ptr->correlation_01_vp[i]),
                                               rgb_map_vp, DIVIDE_BY_BLUE, 
                                               MAX_X_DIV_B);
            }
        }
    }

    if (result != ERROR) result = NO_ERROR;
    
    /*
    // Kobus: Added extra bit of error handling stuff. 
    */
    if (result == ERROR)
    {
        if (cc_methods_ptr->print_algorithm_error) 
        {
            kjb_print_error();
            result = NO_ERROR; 
        }
        else if ( ! cc_methods_ptr->skip_on_error)
        {
            result = NO_ERROR; 
        }
    }
                    
    /*
    // Use Uniform Probability Correlation Matrix (Gaussian Mask Fill) ----
    */
    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        if (result == ERROR) break;

        if (initialize_uniform_gaussian_fill_data(i) != ERROR)
        {
            free_vector(illum_rgb_vp);
            illum_rgb_vp = NULL;
            free_vector(mlm_illum_rgb_vp);
            mlm_illum_rgb_vp = NULL; 
            free_vector(mmse_illum_rgb_vp);
            mmse_illum_rgb_vp = NULL; 

            result = do_uniform_gaussian_fill(i, data_mp, 
                                              &illum_rgb_vp, 
                                              &mlm_illum_rgb_vp, 
                                              &mmse_illum_rgb_vp, 
                    &(cpu_time_ptr->correlation_uniform_MAP_gaussian_fill[i]), 
                    &(cpu_time_ptr->correlation_uniform_MLM_gaussian_fill[i]), 
                    &(cpu_time_ptr->correlation_uniform_MMSE_gaussian_fill[i]), 
                                              cc_methods_ptr->plot_enable);
            
            if ((result != ERROR) && (illum_rgb_vp != NULL)) 
            {
                result = get_map_to_canon_illum(&rgb_map_vp, illum_rgb_vp);

                if (result != ERROR)
                {
                    result = get_projection_vector(&(corr_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[i]),
                                                   rgb_map_vp,
                                                   DIVIDE_BY_BLUE, MAX_X_DIV_B);
                }
            }

            if ((result != ERROR) && (mlm_illum_rgb_vp != NULL)) 
            {
                result = get_map_to_canon_illum(&rgb_map_vp, mlm_illum_rgb_vp);

                if (result != ERROR)
                {
                    result = get_projection_vector(&(corr_result_ptr->correlation_uniform_MLM_gaussian_fill_vp[i]),
                                                   rgb_map_vp,
                                                   DIVIDE_BY_BLUE, MAX_X_DIV_B);
                }
            }

            if ((result != ERROR) && (mmse_illum_rgb_vp != NULL)) 
            {
                result = get_map_to_canon_illum(&rgb_map_vp, mmse_illum_rgb_vp);

                if (result != ERROR)
                {
                    result = get_projection_vector(&(corr_result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[i]),
                                                   rgb_map_vp,
                                                   DIVIDE_BY_BLUE, MAX_X_DIV_B);
                }
            }
        }
    }

    if (result != ERROR) result = NO_ERROR;

    /*
    // Kobus: Added extra bit of error handling stuff. 
    */
    if (result == ERROR)
    {
        if (cc_methods_ptr->print_algorithm_error) 
        {
            kjb_print_error();
            result = NO_ERROR; 
        }
        else if ( ! cc_methods_ptr->skip_on_error)
        {
            result = NO_ERROR; 
        }
    }
                    
    if (result != ERROR) result = NO_ERROR;
    
    free_vector(rgb_map_vp);
    free_vector(illum_rgb_vp);
    free_vector(mlm_illum_rgb_vp);
    free_vector(mmse_illum_rgb_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/*--------------------------------------------------------------------------*/
/* ZERO-ONE CORRELATION DATA FUNCTIONS -------------------------------------*/
/*--------------------------------------------------------------------------*/

static int initialize_01_data(int index)
{
    static int first_time = TRUE;
    int result = NO_ERROR;

    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    kjb_clear_error();

    if (first_time && (fs_cached_01_data_cp[ index ] == NULL))
    {
        first_time = FALSE;

        if (index == 0)
        {
            ERE(read_01_data_from_file(index, CORR_01_DATA_FILE));
        }
    }
    
    if (fs_cached_01_data_cp[ index ] == NULL)
    {
        insert_error("0/1 correlation data %d is not available", index + 1);
        result = ERROR;
    }
    
    return result; 
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int read_01_data_from_file(int index, const char* file_name)
{
    Corr_matrix* new_data_cp               = NULL;
    char         msg[ MAX_FILE_NAME_SIZE ];
    int          result;

    ERE(kjb_sprintf(msg, MAX_FILE_NAME_SIZE,
                    "0/1 correlation matrix %d", index + 1));

    result = read_correlation_matrix(&new_data_cp, file_name, msg);

    if (    (result == NO_ERROR)
         && (set_01_data_matrix(index, new_data_cp, file_name) == ERROR)
        )
    {
        add_error("Unable to update 0/1 correlation matrix %d", index + 1);
        add_error("with data from file  \"%s\"", file_name);
        result =  ERROR;
    }

    free_correlation_matrix(new_data_cp);    

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int write_01_data(int index, char* file_name)
{
    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    return write_correlation_matrix(file_name, fs_cached_01_data_cp[ index ]);
}
    
/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int set_01_data_matrix
(
    int          index,
    Corr_matrix* new_01_data_cp,
    const char*  config_file 
)
{
    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (fs_cached_01_data_cp[index] != NULL)
    {
      free_correlation_matrix(fs_cached_01_data_cp[index]);
      fs_cached_01_data_cp[index] = NULL;
    }

    ERE(copy_correlation_matrix(&(fs_cached_01_data_cp[index]), new_01_data_cp));
    BUFF_CPY(fs_z1_file[index], config_file);

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    (fs_z1_version[index])++;

    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int make_01_data(int index, char* file_name)
{
    static int previous_world_version[ NUM_CORMATS_PER_ALG ];
    static int previous_z1_version[ NUM_CORMATS_PER_ALG ];
    static int first_time    = TRUE;
    int        world_version = NOT_SET;
    int        result        = NO_ERROR;
    int        i;

    if (first_time)
    {
        for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
        {
            previous_world_version[i]           = NOT_SET;
            previous_z1_version[i]              = NOT_SET;
        }
    }
    first_time = FALSE;
    
    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (   (fs_cached_01_data_cp[index] != NULL)
        && (fs_z1_version[index] == previous_z1_version[index])
        && (world_is_up_to_date(previous_world_version[index]))
        && (fs_illum_rgb_index_file[ 0 ] == '\0')     /* HACK--force make if data file. */
        && (fs_surface_rgb_index_file[ 0 ] == '\0')   /* HACK--force make if data file. */
       )
    {
        verbose_pso(5, "0/1 correlation matrix %d up to date. Skipping make.\n",
                    index + 1);
    }
    else 
    {
        Corr_matrix* correl_01_data_cp = NULL;
        Spectra*     illum_sp    = NULL;  /* Illuminant database       */
        Matrix*      rgb_data_mp = NULL;  /* RGBs under current illum. */
        Matrix*      chrom_mp    = NULL;  /* Chromaticities            */
        int          illum_spectra_version;
        int          illum;
        int          num_illum;
        int          num_surfaces;
        Matrix*      illum_rgb_mp = NULL;
        Vector*      illum_rgb_vp = NULL; 
        Vector*      illum_chrom_vp = NULL; 
        char         data_file_name[ MAX_FILE_NAME_SIZE ];
        FILE*        illum_rgb_index_fp = NULL;
        FILE*        surface_rgb_index_fp = NULL;

        verbose_pso(5, "0/1 correlation matrix %d out of date. Remaking it ",
                    index + 1); 

        if (    (fs_illum_rgb_index_file[ 0 ] != '\0')
              || (fs_surface_rgb_index_file[ 0 ] != '\0')
           )
        {
            ERE(verbose_pso(5, "from data files.\n"));

            if (    (fs_illum_rgb_index_file[ 0 ] == '\0')
                 || (fs_surface_rgb_index_file[ 0 ] == '\0')
               )
            {
                set_error("Only one of corr_illum_index_file/corr_surface file set.");
                set_error("Either none or both should be set.");
                return ERROR;
            }

            NRE(illum_rgb_index_fp = open_gamut_config_file((const char*)NULL, 
                                                        fs_illum_rgb_index_file,
                                               "correlation illum rgb index file"));

            surface_rgb_index_fp =  open_gamut_config_file((const char*)NULL,
                                                    fs_surface_rgb_index_file, 
                                               "correlation surface rgb index file");

            if (surface_rgb_index_fp == NULL)
            {
                kjb_fclose(illum_rgb_index_fp); 
                return ERROR;
            }

            num_illum = count_real_lines(illum_rgb_index_fp);

            if (num_illum != count_real_lines(surface_rgb_index_fp))
            {
                set_error("Mismatch in the number of lines in corr_illum_index_filei and corr_surface_index_file.");
                kjb_fclose(illum_rgb_index_fp); 
                kjb_fclose(surface_rgb_index_fp); 
                return ERROR;
            }
        }
        else 
        {
            ERE(verbose_pso(5, "from spectra.\n"));

            ERE(illum_spectra_version = get_illum_spectra(&illum_sp));
            
            if (illum_spectra_version == ERROR)
            {
                result = ERROR;
            }
            else 
            {            
                num_illum = illum_sp->spectra_mp->num_rows;  /* Kobus */
            }
        }

        num_surfaces = NOT_SET;

        ERE(get_target_correlation_matrix(&correl_01_data_cp,
                                          CMA_ZERO_ONE,
                                          CMS_SPECTRA,
                                          CMD_NONE,
                                          num_illum,
                                          num_surfaces,
                                          fs_settable_num_bins,
                                          NOT_SET,
                                          CORR_RG_SPACE,
                                          0.0,
                                          DBL_NOT_SET));

        if (illum_sp != NULL)
        {
            ERE(copy_spectra(&(correl_01_data_cp->illum_sp), illum_sp));
        }

        if (illum_rgb_index_fp != NULL) 
        {
            ERE(get_target_matrix(&(correl_01_data_cp->illum_chrom_mp), num_illum, 2));
        }
        else if (illum_sp != NULL)
        {
            ERE(get_RGB_data_from_spectra(&illum_rgb_mp, illum_sp, (Spectra*)NULL));
            ERE(get_divide_by_sum_projection_matrix(&(correl_01_data_cp->illum_chrom_mp),
                                                    illum_rgb_mp, 
                                                    SKIP_INVALID_CHROMATICITY,
                                                    1000.0, DBL_NOT_SET, DBL_NOT_SET,
                                                    (Matrix**)NULL));    
        }

        for (illum = 0; illum < num_illum; illum++)
        {
            if (illum_rgb_index_fp != NULL) 
            {
                ERE(BUFF_GET_REAL_LINE(illum_rgb_index_fp, data_file_name)); 

                ERE(read_matrix_from_gamut_config_file(&illum_rgb_mp,
                                                       (const char*)NULL,
                                                       data_file_name,
                                                       "indexed surface rgb file",
                                                       (char*)NULL, 0)); 

                ERE(average_matrix_rows(&illum_rgb_vp, illum_rgb_mp)); 

                ERE(project_vector(&illum_chrom_vp, illum_rgb_vp, 
                                   DIVIDE_BY_SUM)); 
                ERE(put_matrix_row(correl_01_data_cp->illum_chrom_mp, 
                                   illum_chrom_vp, illum)); 
                
                ERE(BUFF_GET_REAL_LINE(surface_rgb_index_fp, data_file_name)); 

                ERE(read_matrix_from_gamut_config_file(&rgb_data_mp,
                                                       (const char*)NULL,
                                                       data_file_name,
                                                       "indexed surface rgb file",
                                                       (char*)NULL, 0)); 
            }
            else 
            {
                IMPORT volatile int io_atn_flag;
                IMPORT volatile int halt_all_output;
                    
                world_version = get_synthetic_RGB(&rgb_data_mp, illum);

                if (world_version == ERROR)
                {
                    set_error("Error generating synthetic RGB data from spectra");
                    result = ERROR;
                    break;
                }
                    
                if (io_atn_flag)
                {
                    set_error("Processing interrupted.");
                    halt_all_output = FALSE;
                    result = ERROR;
                    break;
                }
            }

            /* Set the number of surfaces in the correlation matrix. */
            num_surfaces = rgb_data_mp->num_rows;
            correl_01_data_cp->num_surf = num_surfaces;

            ERE(get_divide_by_sum_projection_matrix(&chrom_mp, 
                                                    rgb_data_mp, 
                                                    SKIP_INVALID_CHROMATICITY,
                                                    1000.0,
                                                    DBL_NOT_SET,
                                                    DBL_NOT_SET,
                                                    (Matrix**)NULL));
            
            verbose_pso(300, "0/1 matrix %d - Illuminant %3d/%d\n",
                        index + 1, illum + 1, num_illum);
            
            ERE(insert_01_histogram(&correl_01_data_cp,
                                    illum, chrom_mp,
                                    fs_settable_num_bins,
                                    MAX_RG_CHROM));
            
            ERE(ow_zero_matrix(rgb_data_mp));
            ERE(ow_zero_matrix(chrom_mp));
        }
            
        if (result != ERROR)
        {
            result = set_01_data_matrix(index, correl_01_data_cp, "<make>");
        }
        
        if (result != ERROR) 
        {
            previous_world_version[index]           = world_version;
            previous_z1_version[index]              = fs_z1_version[index];
            
            verbose_pso(10, "0/1 correlation matrix %d remade.\n",
                        index + 1);

        }
        
        free_correlation_matrix(correl_01_data_cp);
        free_spectra(illum_sp); 
        free_matrix(rgb_data_mp); 
        free_matrix(chrom_mp);
        free_matrix(illum_rgb_mp);
        free_vector(illum_chrom_vp); 
        free_vector(illum_rgb_vp); 

        kjb_fclose(illum_rgb_index_fp);
        kjb_fclose(surface_rgb_index_fp);
    } 
    
    if (   (result != ERROR)
        && (file_name != NULL)
        && (file_name[ 0 ] != '\0'))
    {
        result = write_01_data(index, file_name);
    }

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int insert_01_histogram
(
    Corr_matrix** cormat_cp_ptr,
    int           cur_illum,
    Matrix*       chrom_mp,
    int           num_bins,
    double        max_projection_coordinate 
)
{
    int     i;
    int     col, min_col, max_col;
    Hull*   hull_hp;
    Matrix* histogram_2D_mp   = NULL;
    Vector* histogram_1D_vp   = NULL;
    Matrix* hull_bin_range_mp = NULL;
    Vector* hull_min_vp   = NULL; /* Hull chromticity minimum.           */
    Vector* hull_max_vp   = NULL; /* Hull chromticity maximum.           */
    Matrix* cur_facet_mp;         /* Temp ptr to current facet matrix.   */

    if ((chrom_mp == NULL) || (num_bins < 1))
    {
        SET_ARGUMENT_BUG();
        return ERROR;    
    }

    if (max_projection_coordinate == DBL_NOT_SET)
    {
        max_projection_coordinate = 1.0; /* Default for rg-chrom. */
    }

    /* Allocate 1-D and 2-D histograms. */ 
    ERE(get_zero_vector( &histogram_1D_vp, num_bins * num_bins ));
    ERE(get_zero_matrix( &histogram_2D_mp, num_bins, num_bins ));

    /* Find convex hull of chromaticity values and find bin row range. */
    NRE(hull_hp = find_convex_hull(chrom_mp, DEFAULT_HULL_OPTIONS));

    /* Initialize the hull boundary range matrix.
    // 
    // The hull boundary histogram bin range matrix is defined as follows:
    //  index : row index in 2D histogram
    //  col 0 : minimum column index in 2D histogram of hull boundary
    //           at row index
    //  col 1 : maximum column index in 2D histogram of hull boundary
    //           at row index
    */
    ERE(get_zero_matrix(&hull_bin_range_mp, num_bins, 2));
    for(i = 0; i < hull_bin_range_mp->num_rows; i++)
    {
        hull_bin_range_mp->elements[i][MIN_COL_INDEX] = -1.0 * DBL_NOT_SET;
        hull_bin_range_mp->elements[i][MAX_COL_INDEX] = DBL_NOT_SET;
    }

    /* Update the hull boundary range matrix for each facet. */
    for (i = 0; i < hull_hp->facets->length; i++)
    {
        cur_facet_mp = hull_hp->facets->elements[i];
        ERE(bin_hull_facet(&hull_bin_range_mp, 
                           cur_facet_mp, 
                           num_bins,
                           max_projection_coordinate));
    }

    /* Set all histogram bins within the range of the hull boundary to 1. */
    for (i = 0; i < hull_bin_range_mp->num_rows; i++)
    {
        min_col = (int)hull_bin_range_mp->elements[i][MIN_COL_INDEX];
        max_col = (int)hull_bin_range_mp->elements[i][MAX_COL_INDEX];

        if (max_col != NOT_SET)
        {
            for (col = min_col; col <= max_col; col++)
                histogram_2D_mp->elements[i][col] = 1.0;
        }
    }

    /* Add the histogram for the current illuminant to correlation matrix. */
    ERE(map_2D_to_1D_histogram( &histogram_1D_vp, histogram_2D_mp ));

#ifdef ROGRAMMER_IS_colour
    TEST_PSO(("Sum of 1D histogram = %f\n", sum_vector_elements(histogram_1D_vp)));
#endif
    
    ERE(put_matrix_row( (*cormat_cp_ptr)->cormat_mp,
                        histogram_1D_vp, cur_illum) );

    free_hull(hull_hp);
    free_matrix(histogram_2D_mp);
    free_matrix(hull_bin_range_mp);
    free_vector(histogram_1D_vp);
    free_vector(hull_min_vp);
    free_vector(hull_max_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int bin_hull_facet
(
    Matrix** hull_bins_mpp,
    Matrix*  cur_facet_mp,
    int      num_bins,
    double   max_projection_coordinate 
)
{
    int     x1, y1, x2, y2;
    int     d, x, y, ax, ay, dx, dy, sx, sy;
    int     cur_min_col, cur_max_col;
    Matrix* hull_bins_mp = *hull_bins_mpp;

    if (hull_bins_mpp == NULL)      /* Must be preallocated/initialized. */
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (cur_facet_mp == NULL)          /* Must be preallocated. */
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (   cur_facet_mp->num_cols != 2
        || cur_facet_mp->num_rows != 2) /* Must contain two 2D points. */
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    /* Convert facet endpoints from chromaticity to histogram bin coords. */
    x1 = bin_chromaticity(cur_facet_mp->elements[0][R_CHROM_INDEX],
                          num_bins, max_projection_coordinate);
    x2 = bin_chromaticity(cur_facet_mp->elements[1][R_CHROM_INDEX],
                          num_bins, max_projection_coordinate);
    y1 = bin_chromaticity(cur_facet_mp->elements[0][G_CHROM_INDEX],
                          num_bins, max_projection_coordinate);
    y2 = bin_chromaticity(cur_facet_mp->elements[1][G_CHROM_INDEX],
                          num_bins, max_projection_coordinate);

    /* Use midpoint algorithm to scan convert the facet endpoints. */
    dx = x2 - x1;
    dy = y2 - y1;

    ax = ABS_OF(dx) << 1;
    ay = ABS_OF(dy) << 1;

    sx = INT_SIGN_OF(dx);
    sy = INT_SIGN_OF(dy);

    x = x1;
    y = y1;

    if (ax > ay) /* X-axis dominates line slope. */
    {
        d = ay - (ax >> 1);
    
        for (;;)
        {
            cur_min_col = (int)hull_bins_mp->elements[y][MIN_COL_INDEX];
            cur_max_col = (int)hull_bins_mp->elements[y][MAX_COL_INDEX];

            if (x < cur_min_col )
            {
                hull_bins_mp->elements[y][MIN_COL_INDEX] = (double)x;
            }
            if (x > cur_max_col )
            {
                hull_bins_mp->elements[y][MAX_COL_INDEX] = (double)x;
            }
      
            if (x == x2)
            {
                return NO_ERROR;
            }
            if (d>=0)
            {
                y += sy;
                d -= ax;
            }
            x += sx;
            d += ay;
        }
    }
    else        /* Y-axis dominates line slope. */
    {
        d = ax - (ay >> 1);

        for (;;) 
        {
            cur_min_col = (int)hull_bins_mp->elements[y][MIN_COL_INDEX];
            cur_max_col = (int)hull_bins_mp->elements[y][MAX_COL_INDEX];

            if (x < cur_min_col )
            {
                hull_bins_mp->elements[y][MIN_COL_INDEX] = (double)x;
            }
            if (x > cur_max_col )
            {
                hull_bins_mp->elements[y][MAX_COL_INDEX] = (double)x;
            }

            if (y == y2)
            {
                return NO_ERROR;
            }
            if (d >= 0)
            {
                x += sx;
                d -= ay;
            }
            y += sy;
            d += ax;
        }
    }
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* Kobus: Replaced do_01 to handle no solutions. */
static int do_01
(
    int      index,
    Matrix*  rgb_data_mp,
    int      prompt_to_increase_error_flag,
    Vector** result_vpp 
)
{
    IMPORT double min_absolute_pixel_error;
    IMPORT double min_relative_pixel_error;
    int           result;
    double        absolute_pixel_error;
    double        relative_pixel_error;
    Matrix*       rgb_with_error_mp        = NULL;
    int           try                      = 0;

    absolute_pixel_error = min_absolute_pixel_error;
    relative_pixel_error = min_relative_pixel_error;

    while (try < 20)
    {
        try++; 

        result = get_points_with_error(&rgb_with_error_mp, rgb_data_mp, 
                                       absolute_pixel_error, 
                                       relative_pixel_error,
                                       MIN_X_DIV_B);

        if (result != ERROR)
        {
            result = do_01_with_error(index, rgb_with_error_mp, result_vpp);
        }

        if (    (result == NO_SOLUTION)
             && (ok_to_increase_pixel_error(prompt_to_increase_error_flag,
                                            &absolute_pixel_error,
                                            &relative_pixel_error))
           )
        {
            /*EMPTY*/
            ; /* Do nothing */
        }
        else
        {
            break;
        }
    }

    if (result == NO_SOLUTION)
    {
        set_error("Unable to find a correlation 0/1 solution after %d tries.",
                  try);
        
        if (absolute_pixel_error > 0.0)
        {
            add_error("Absolute pixel error of %.1f is not sufficient.",
                      absolute_pixel_error);
        }
        else if (relative_pixel_error > 0.0)
        {
            add_error("Relative pixel error of %.1f%% is not sufficient.", 
                      relative_pixel_error);
        }

        result = ERROR; 
    }

    free_matrix(rgb_with_error_mp); 

    return result;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Kobus: Renamed this routine. */
static int do_01_with_error
(
    int      index,
    Matrix*  rgb_data_mp,
    Vector** result_vpp 
)
{
    Matrix*  input_chrom_mp     = NULL;
    Matrix*  input_2D_hist_mp   = NULL;
    Vector*  input_1D_hist_vp   = NULL;
    
    Matrix*  illum_rgb_mp       = NULL;
    const Matrix* illum_chrom_mp;

    Vector*  colour_count_vp    = NULL;
    Vector*  cur_illum_chrom_vp = NULL;
    Vector*  sum_illum_chrom_vp = NULL;

    Vector*  actual_rgb_vp      = NULL;
    Vector*  actual_chrom_vp    = NULL;
    
    int      cached_num_bins;
    int      i, row, col, num_matching_illum;
    double     max_val, threshold;

    ERE(interactive_pso("Doing 0/1 correlation algorithm with matrix %d .\n",
                        index + 1));
    
    if ( fs_cached_01_data_cp[index] == NULL )
    {
        return NOT_AVAILABLE;
    }

    illum_chrom_mp = (fs_cached_01_data_cp[index])->illum_chrom_mp; 

    cached_num_bins = (fs_cached_01_data_cp[index])->num_bins;
    
    ERE(get_zero_matrix(&input_2D_hist_mp, cached_num_bins, cached_num_bins));
    ERE(get_zero_vector(&input_1D_hist_vp, cached_num_bins * cached_num_bins));
    ERE(get_zero_vector(result_vpp, 3));
    /* New! */
    ERE(get_zero_vector(&colour_count_vp,
                        (fs_cached_01_data_cp[index])->cormat_mp->num_rows));
    
    /* Convert input RGB data to chromaticity space. */
    ERE(get_divide_by_sum_projection_matrix(&input_chrom_mp, rgb_data_mp, 
                                            SKIP_INVALID_CHROMATICITY,
                                            1000.0, DBL_NOT_SET, DBL_NOT_SET,
                                            (Matrix**)NULL));

    /* Convert input chromaticity values to 2D histogram matrix. */
    for (i=0; i<input_chrom_mp->num_rows; i++)
    {
        row = bin_chromaticity(input_chrom_mp->elements[i][G_CHROM_INDEX],
                               cached_num_bins, MAX_RG_CHROM);
        col = bin_chromaticity(input_chrom_mp->elements[i][R_CHROM_INDEX],
                               cached_num_bins, MAX_RG_CHROM);

        /* Kobus (01/08/99) */ 
#ifdef USE_ONLY_PRESENCE_INFO
        input_2D_hist_mp->elements[row][col] = (double)1.0;
#else 
        input_2D_hist_mp->elements[row][col] += (double)1.0;
#endif 
    }
    ERE( map_2D_to_1D_histogram(&input_1D_hist_vp, input_2D_hist_mp) );
    
    /* Mask the correlation matrix data with the input chromaticities. */
    ERE( multiply_matrix_and_vector(&colour_count_vp,
                                    (fs_cached_01_data_cp[index])->cormat_mp,
                                    input_1D_hist_vp) );

    if (   colour_count_vp->length
        != (fs_cached_01_data_cp[index])->cormat_mp->num_rows )
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    verbose_pso(500, "Illuminant rg-chromaticities (in 0/1 matrix %d):\n",
                index + 1);

    for (i = 0; i < illum_chrom_mp->num_rows; i++)
    {
        verbose_pso(500, "     %3d: (%6.4f, %6.4f)", i + 1,
                    illum_chrom_mp->elements[i][RED_INDEX],
                    illum_chrom_mp->elements[i][GREEN_INDEX]);
        verbose_pso(500, " count = %3d\n", (int)colour_count_vp->elements[i]);
    }
    verbose_pso(500,"-----------------------------------------------------\n");
    
    /*
    // Compute the result as the average of all illuminants within the
    //  threshold range of the illuminant with the maximum colour count.
    */
    (void)get_max_vector_element(colour_count_vp, &max_val);

    if ( IS_ZERO_DBL(max_val) )
    {
        /* Kobus: Since this is not a completely unusual return, we need to
        // guard against memory leaks. 
        */
        free_matrix(input_2D_hist_mp);
        free_vector(input_1D_hist_vp);
        free_vector(colour_count_vp);
        free_matrix(input_chrom_mp);
        free_matrix(illum_rgb_mp);
        
        return NO_SOLUTION;
    }
    
    threshold = fs_corr_averaging_threshold * max_val;
    num_matching_illum = 0;
    verbose_pso(50, "Maximum matching colour count for all illums = %d\n",
                (int)max_val);
    verbose_pso(50, "                 Using correlation threshold = %4.2f\n\n",
                fs_corr_averaging_threshold);
    
    verbose_pso(50, "Finding illums with colour counts >= %4.2f (%4.2f * %d):\n",
                threshold, fs_corr_averaging_threshold, (int)max_val);
    
    ERE(get_zero_vector(&sum_illum_chrom_vp, 2));

    for (i = 0; i <illum_chrom_mp ->num_rows; i++)
    {
        if (colour_count_vp->elements[i] >= threshold)
        {
            /* Get rg-chromaticity of matching illuminant */
            ERE(get_matrix_row(&cur_illum_chrom_vp, illum_chrom_mp, i));
            
            /* Sum the matching illuminant chromaticities. */
            ERE(ow_add_vectors(sum_illum_chrom_vp, cur_illum_chrom_vp));
            num_matching_illum++;
            verbose_pso(100, "     %3d: (%6.4f, %6.4f)", i + 1,
                        cur_illum_chrom_vp->elements[RED_INDEX],
                        cur_illum_chrom_vp->elements[GREEN_INDEX]);
            verbose_pso(100, " count = %3d\n",
                        (int)colour_count_vp->elements[i]);
        }    
    }
    verbose_pso(100,"-----------------------------------------------------\n");

    /* Compute the average chromaticity, and convert back to RGB. */
    ERE(ow_divide_vector_by_scalar(sum_illum_chrom_vp,
                                   (double)num_matching_illum));
    ERE(back_project_vector(result_vpp,
                            sum_illum_chrom_vp,
                            DIVIDE_BY_SUM));
    ERE(ow_multiply_vector_by_scalar(*result_vpp, 255.0));
    
    verbose_pso(50, "Average : (%6.4f, %6.4f) (%d illuminants)\n",
                sum_illum_chrom_vp->elements[RED_INDEX],
                sum_illum_chrom_vp->elements[GREEN_INDEX],
                num_matching_illum);
    
    ERE(get_actual_illum(&actual_rgb_vp));
    ERE(get_projection_vector(&actual_chrom_vp, actual_rgb_vp,
                              DIVIDE_BY_SUM, 1000.0));
    
    verbose_pso(50, "Actual  : (%6.4f, %6.4f)\n\n",
                actual_chrom_vp->elements[RED_INDEX],
                actual_chrom_vp->elements[GREEN_INDEX]);
    
    free_matrix(input_chrom_mp);
    free_matrix(input_2D_hist_mp);
    free_vector(input_1D_hist_vp);

    free_matrix(illum_rgb_mp);
    
    free_vector(colour_count_vp);
    free_vector(cur_illum_chrom_vp);
    free_vector(sum_illum_chrom_vp);
    
    free_vector(actual_rgb_vp);
    free_vector(actual_chrom_vp);
    
    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/*--------------------------------------------------------------------------*/
/* UNIFORM PROBABILITY (GAUSSIAN MASK FILL) CORRELATION DATA FUNCTIONS -----*/
/*--------------------------------------------------------------------------*/ 

static int initialize_uniform_gaussian_fill_data(int index)
{
    static int first_time = TRUE;

    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
    
    kjb_clear_error();

    if (first_time && (fs_cached_log_uniform_gaussian_cp[ index ] == NULL))
    {
        first_time = FALSE;

        if (index == 0)
        {
            ERE(read_uniform_gaussian_fill_data_from_file(index,
                                                          CORR_UNIFORM_GAUSSIAN_DATA_FILE));
        }
    }

    if (fs_cached_log_uniform_gaussian_cp[ index ] == NULL)
    {
        insert_error("Uniform correlation (inverse distance fill) data %d is not available",
                     index + 1);
        return ERROR;
    }
    
    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int read_uniform_gaussian_fill_data_from_file
(
    int         index,
    const char* file_name 
)
{
    Corr_matrix* new_data_cp = NULL;
    char         msg[ MAX_FILE_NAME_SIZE ];
    int          result;

    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (void)kjb_sprintf(msg, MAX_FILE_NAME_SIZE,
                      "Log uniform (gaussian mask fill) matrix %d",
                      index + 1);

    result = read_correlation_matrix(&new_data_cp, file_name, msg);

    if (    (result == NO_ERROR)
         && (set_uniform_gaussian_fill_data_matrix(index,
                                                   new_data_cp,
                                                   file_name) == ERROR)
        )
    {
        add_error("Unable to update log uniform (gaussian mask fill) ");
        add_error("correlation matrix %d with data from file  \"%s\"",
                  index + 1, file_name);
        result =  ERROR;
    }

    free_correlation_matrix(new_data_cp);
    
    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int write_uniform_gaussian_fill_data(int index, char* file_name)
{
    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
    
    return write_correlation_matrix(file_name,
                                    fs_cached_log_uniform_gaussian_cp[ index ]);
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int set_uniform_gaussian_fill_data_matrix
(
    int          index,
    Corr_matrix* new_uniform_gaussian_data_cp,
    const char*  config_file 
)
{
    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (fs_cached_log_uniform_gaussian_cp[index] != NULL)
    {
      free_correlation_matrix(fs_cached_log_uniform_gaussian_cp[index]);
      fs_cached_log_uniform_gaussian_cp[index] = NULL;
    }

    ERE(copy_correlation_matrix(&(fs_cached_log_uniform_gaussian_cp[index]),
                                new_uniform_gaussian_data_cp));
    BUFF_CPY(fs_log_uniform_gaussian_file[index], config_file);

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    (fs_log_uniform_gaussian_version[index])++;

    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* ===========================================================================
 *                 make_log_uniform_gaussian_fill_data
 *
 * This function simply converts input uniform probability data to log
 * space. Matrix elements with P = 0 are set to CORRELATION_MINUS_INFINITY.
 * 
 * The input probability data is smoothed using a gaussian filter before
 * computing the log values.
 * 
 */

static int make_log_uniform_gaussian_fill_data
(
    int   index,
    char* file_name 
)
{
    static int   previous_log_uniform_gaussian_version[ NUM_CORMATS_PER_ALG ];
    static int   previous_world_version[ NUM_CORMATS_PER_ALG ];
    static int   previous_illum_spectra_version[ NUM_CORMATS_PER_ALG ];
    
    Corr_matrix* uniform_data_cp       = NULL;  /* Source          */    
    Matrix*      uniform_cormat_mp;
    Matrix*      uniform_histogram_mp  = NULL;    
    Vector*      uniform_histogram_vp  = NULL;

    Matrix*      gaussian_mask_mp      = NULL;  /* Intermediate    */
    Matrix*      smoothed_cormat_mp    = NULL;
    Matrix*      smoothed_histogram_mp = NULL;
    Vector*      smoothed_histogram_vp = NULL;

    Corr_matrix* log_data_cp           = NULL;  /* Target Log Data */
    Matrix*      log_data_mp;
    
    double src_prob;
    int  i, illum, colour, num_bins, num_illum, num_colours;

    int  world_version           = NOT_SET;
    int  illum_spectra_version   = NOT_SET;
    
    static int first_time = TRUE;
    int        result;
    
    double ave; 
    
    if (first_time)
    {
        for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
        {
            previous_world_version[i]                = NOT_SET;
            previous_log_uniform_gaussian_version[i] = NOT_SET;
            previous_illum_spectra_version[i]        = NOT_SET;
        }
    }
    first_time = FALSE;

    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (   (fs_cached_log_uniform_gaussian_cp[index] != NULL)
        && (fs_log_uniform_gaussian_version[index] == previous_log_uniform_gaussian_version[index])
        && (world_is_up_to_date(previous_world_version[index]))
        && (illum_spectra_is_up_to_date(previous_illum_spectra_version[index]))
        && (fs_illum_rgb_index_file[ 0 ] == '\0')  /* HACK--force make if data file. */
        && (fs_surface_rgb_index_file[ 0 ] == '\0')   /* HACK--force make if data file. */
       )
    {
        ERE(verbose_pso(5, "Log uniform (gaussian mask fill) matrix %d up to date. Skipping make.\n",
                        index+1));
        return NO_ERROR;
    }
    else
    {
        ERE(verbose_pso(5, "Log uniform (gaussian mask fill) matrix %d out of date. Remaking it.\n",
                        index+1));
    }
    
    /* Start the make */
    result = make_uniform_probability_data(&uniform_data_cp, &world_version,
                                           &illum_spectra_version);
    
    if (result == ERROR)
    {
        add_error("Failed to generate uniform probability correlation data");
        return ERROR;
    }

    /* Allocate the output log of gaussian filled data */
    if (copy_correlation_matrix(&log_data_cp, uniform_data_cp) == ERROR)
    {
        add_error("Failed to allocate output log probability correlation matrix");
        free_correlation_matrix(uniform_data_cp);
        return ERROR;
    }
    else
    {
        log_data_cp->algorithm = CMA_LOG_PROBABILITY;
        log_data_cp->sigma     = fs_settable_mask_sigma;
        log_data_mp = log_data_cp->cormat_mp;

        if (log_data_mp == NULL)
        {
            /* Leak the correlation matrix memory for debugging */
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }     
    }
    
    uniform_cormat_mp = uniform_data_cp->cormat_mp;
    if (uniform_cormat_mp == NULL)
    {
        add_error("Uniform probability correlation matrix is empty.");
        return ERROR;
    }
    
    num_bins    = uniform_data_cp->num_bins;
    num_illum   = uniform_data_cp->num_illum;
    num_colours = num_bins * num_bins;

    /* Get the temporary smoothed probability correlation matrix. */
    ERE(get_zero_matrix(&smoothed_cormat_mp, num_illum, num_bins * num_bins));

    if (    (uniform_cormat_mp->num_rows !=  smoothed_cormat_mp->num_rows)
         || (uniform_cormat_mp->num_cols !=  smoothed_cormat_mp->num_cols)
       )
    {
        free_correlation_matrix(uniform_data_cp);
        free_matrix(smoothed_cormat_mp);
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    if (   (uniform_cormat_mp->num_rows !=  log_data_mp->num_rows)
        || (uniform_cormat_mp->num_cols !=  log_data_mp->num_cols ))
    {
        free_correlation_matrix(uniform_data_cp);
        free_matrix(smoothed_cormat_mp);
        free_matrix(log_data_mp);
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }  

    /* Get the gaussian smoothing mask. */
    if (get_gaussian_mask(&gaussian_mask_mp, fs_settable_mask_size,
                         (double)fs_settable_mask_sigma)
        == ERROR)
    {
        insert_error("Unable to obtain gaussian smoothing mask.");
        free_correlation_matrix(uniform_data_cp);
        free_matrix(smoothed_cormat_mp);
        free_matrix(log_data_mp);
        free_matrix(gaussian_mask_mp);
        return ERROR;
    }    

    /* Smooth the probability matrix. */
    for (illum = 0; illum < num_illum; illum++)
    {
        IMPORT volatile int io_atn_flag;
        IMPORT volatile int halt_all_output;
        
        verbose_pso(300, "Log uniform (gaussian mask) matrix %d - Illuminant %3d/%d\n",
                    index + 1, illum + 1, num_illum);

        if (io_atn_flag)
        {
            set_error("Processing interrupted during convolution with gaussian mask.");
            halt_all_output = FALSE;
            result = ERROR;
            break;
        }
        
        /* Smooth the probability histogram of the current illuminant. */
        ERE(get_matrix_row(&uniform_histogram_vp, uniform_cormat_mp, illum));
        ERE(map_1D_to_2D_histogram(&uniform_histogram_mp, 
                                   uniform_histogram_vp));
        ERE(convolve(&smoothed_histogram_mp, uniform_histogram_mp, 
                     gaussian_mask_mp));

        /* Insert the histogram into a temp correlation matrix. */
        ERE(map_2D_to_1D_histogram(&smoothed_histogram_vp,
                                   smoothed_histogram_mp));
        ERE(put_matrix_row(smoothed_cormat_mp, smoothed_histogram_vp, illum));
    } /* for current illum */

    ave = average_matrix_elements(smoothed_cormat_mp);
    ERE(ow_divide_matrix_by_scalar(smoothed_cormat_mp, ave)); 

    /* Fill the target correlation matrix with the log of smoothed
    // probability values taken from the source matrix.                */
    for (illum = 0; illum < num_illum; illum++)
    {
        IMPORT volatile int io_atn_flag;
        IMPORT volatile int halt_all_output;
        
        verbose_pso(300, "Log uniform (gaussian mask) matrix %d - Illuminant %3d/%d\n",
                    index + 1, illum + 1, num_illum);

        if (io_atn_flag)
        {
            set_error("Processing interrupted during convolution with gaussian mask.");
            halt_all_output = FALSE;
            result = ERROR;
            break;
        }
        
        /* Compute the log of the smoothed probabilies. */
        for (colour = 0; colour < num_colours; colour++)
        {
            src_prob = smoothed_cormat_mp->elements[illum][colour];
            
            if (src_prob != 0.0)
            {
                log_data_mp->elements[illum][colour] = log(src_prob);
            }
            else
            {
                log_data_mp->elements[illum][colour]=CORRELATION_MINUS_INFINITY;
            }
        }
    } /* for current illum */

    if (result == NO_ERROR)
    {
        result = set_uniform_gaussian_fill_data_matrix(index, log_data_cp, 
                                                       "<make>");
    }

    if (result == NO_ERROR)
    {
        previous_world_version[index]           = world_version;
        previous_illum_spectra_version[index]   = illum_spectra_version;
        previous_log_uniform_gaussian_version[index]
                                        = fs_log_uniform_gaussian_version[index];

        verbose_pso(10, "Log uniform (gaussian mask fill) matrix %d remade.\n",
                    index+1);

        if ( (file_name != NULL) && (file_name[0] != '\0') )
        {
            result = write_uniform_gaussian_fill_data(index, file_name);
        }
    }
    free_correlation_matrix(uniform_data_cp);
    free_correlation_matrix(log_data_cp);
    free_matrix(uniform_histogram_mp);
    free_vector(uniform_histogram_vp);
    free_matrix(smoothed_cormat_mp);
    free_matrix(smoothed_histogram_mp);
    free_vector(smoothed_histogram_vp);
    free_matrix(gaussian_mask_mp);

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* Kobus: 08/07/99 | Added plot_enable arg. */
static int do_uniform_gaussian_fill
(
    int      index,
    Matrix*  rgb_data_mp,
    Vector** result_vpp,
    Vector** mlm_result_vpp,
    Vector** mmse_result_vpp,
    double*  cpu_time_ptr,
    double*  mlm_cpu_time_ptr,
    double*  mmse_cpu_time_ptr,
    int      plot_enable 
)
{
    int result;
    char id_str[ 200 ]; /* Kobus: 08/07/99 */

    ERE(interactive_pso("Doing uniform probability (gaussian mask fill) algorithm with matrix %d.\n",
                        index + 1));

    if ( (index < 0) || (index >= NUM_CORMATS_PER_ALG) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if ( fs_cached_log_uniform_gaussian_cp[ index ] == NULL )
        return NOT_AVAILABLE;

    /* Kobus: 08/07/99 */
    ERE(kjb_sprintf(id_str, sizeof(id_str), "%s (%s) [ %d ]",
                    "uniform probability", "gaussian mask fill", index + 1));

    /* Kobus: 08/07/99 | Added id_str and plot_enable args. */
    result = do_probability_algorithm(rgb_data_mp,
                                      result_vpp, 
                                      mlm_result_vpp, 
                                      mmse_result_vpp, 
                                      cpu_time_ptr, 
                                      mlm_cpu_time_ptr,
                                      mmse_cpu_time_ptr,
                                      fs_cached_log_uniform_gaussian_cp[index],
                                      plot_enable, id_str);
        
    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int make_uniform_probability_data
(
    Corr_matrix** new_uniform_data_cp_ptr,
    int*          world_version_ptr,
    int*          illum_spectra_version_ptr 
)
{
    Corr_matrix* prob_uniform_data_cp = NULL;
    Spectra*     illum_sp             = NULL;
    Matrix*      rgb_data_mp          = NULL;
    Matrix*      chrom_mp             = NULL;
    Matrix*      prob_colour_illum_mp = NULL;  /* P(colour|illum) */
    Vector*      prob_colour_vp       = NULL;  /* P(colour)       */
    Vector*      cur_prob_row_vp      = NULL;
    int          world_version = NOT_SET;
    int          illum_spectra_version = NOT_SET;
    int          num_illum, num_colours, num_surfaces;
    int          illum, colour;
    int          result = NO_ERROR;
    FILE* illum_rgb_index_fp = NULL; 
    FILE* surface_rgb_index_fp = NULL; 
    char file_name[ MAX_FILE_NAME_SIZE ]; 
    Matrix* illum_rgb_mp = NULL; 
    Vector* illum_chrom_vp = NULL; 
    Vector* illum_rgb_vp = NULL; 

    ERE(verbose_pso(5, "Making uniform probability correlation data "));
    
    if (    (fs_illum_rgb_index_file[ 0 ] != '\0')
         || (fs_surface_rgb_index_file[ 0 ] != '\0')
       )
    {
        ERE(verbose_pso(5, "from data files.\n"));

        if (    (fs_illum_rgb_index_file[ 0 ] == '\0')
             || (fs_surface_rgb_index_file[ 0 ] == '\0')
           )
        {
            set_error("Only one of corr_illum_index_file/corr_surface file set.");
            set_error("Either none or both should be set.");
            return ERROR;
        }

        NRE(illum_rgb_index_fp = open_gamut_config_file((const char*)NULL, 
                                                    fs_illum_rgb_index_file,
                                           "correlation illum rgb index file"));

        surface_rgb_index_fp =  open_gamut_config_file((const char*)NULL,
                                                fs_surface_rgb_index_file, 
                                           "correlation surface rgb index file");

        if (surface_rgb_index_fp == NULL)
        {
            kjb_fclose(illum_rgb_index_fp); 
            return ERROR;
        }

        num_illum = count_real_lines(illum_rgb_index_fp);

        if (num_illum != count_real_lines(surface_rgb_index_fp))
        {
            set_error("Mismatch in the number of lines in corr_illum_index_filei and corr_surface_index_file.");
            kjb_fclose(illum_rgb_index_fp); 
            kjb_fclose(surface_rgb_index_fp); 
            return ERROR;
        }
    }
    else 
    {
        ERE(verbose_pso(5, "from spectra.\n"));

        illum_spectra_version = get_illum_spectra(&illum_sp);

        if (illum_spectra_version == ERROR)
        {
            add_error("Failed to get illuminant spectra.");
            return ERROR;
        }

        num_illum = illum_sp->spectra_mp->num_rows;  /* Kobus */
    }
        
    num_surfaces = NOT_SET;
    num_colours  = fs_settable_num_bins * fs_settable_num_bins;
          
    if (get_target_correlation_matrix(&prob_uniform_data_cp,
                                      CMA_PROBABILITY,
                                      CMS_SPECTRA,
                                      CMD_UNIFORM,
                                      num_illum,
                                      num_surfaces,
                                      fs_settable_num_bins,
                                      NOT_SET, 
                                      CORR_RG_SPACE,
                                      0.0,
                                      DBL_NOT_SET)
        == ERROR)
    {
        add_error("Failed to allocate correlation matrix.");
        return ERROR;
    }
          
    if (illum_sp != NULL)
    {
        if (copy_spectra( &(prob_uniform_data_cp->illum_sp), illum_sp ) == ERROR)
        {
            add_error("Failed to copy illuminant spectra to uniform corr. matrix.");
            free_correlation_matrix(prob_uniform_data_cp);
            return ERROR;
        }
    }
        
    if (illum_rgb_index_fp != NULL) 
    {
        ERE(get_target_matrix(&(prob_uniform_data_cp->illum_chrom_mp), num_illum, 2));
    }
    else if (illum_sp != NULL)
    {
        ERE(get_RGB_data_from_spectra(&illum_rgb_mp, illum_sp, (Spectra*)NULL));
        ERE(get_divide_by_sum_projection_matrix(&(prob_uniform_data_cp->illum_chrom_mp),
                                                illum_rgb_mp, 
                                                SKIP_INVALID_CHROMATICITY,
                                                1000.0, DBL_NOT_SET, DBL_NOT_SET,
                                                (Matrix**)NULL));    
    }

    if (get_zero_matrix(&prob_colour_illum_mp, num_illum, num_colours) == ERROR)
    {
        add_error("Failed to allocate P(colour|illum) matrix.");
        free_correlation_matrix(prob_uniform_data_cp);
        return ERROR;
    }
        
    for (illum = 0; illum < num_illum; illum++)
    {
        IMPORT volatile int io_atn_flag;
        IMPORT volatile int halt_all_output;

        if (io_atn_flag)
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE;
            result = ERROR;
            break;
        }
        
        if (illum_rgb_index_fp != NULL) 
        {
            ERE(BUFF_GET_REAL_LINE(illum_rgb_index_fp, file_name)); 

            ERE(read_matrix_from_gamut_config_file(&illum_rgb_mp,
                                                   (const char*)NULL,
                                                   file_name,
                                                   "indexed surface rgb file",
                                                   (char*)NULL, 0)); 

            ERE(average_matrix_rows(&illum_rgb_vp, illum_rgb_mp)); 

            ERE(project_vector(&illum_chrom_vp, illum_rgb_vp, 
                               DIVIDE_BY_SUM)); 
            ERE(put_matrix_row(prob_uniform_data_cp->illum_chrom_mp, illum_chrom_vp, 
                               illum)); 
            
            ERE(BUFF_GET_REAL_LINE(surface_rgb_index_fp, file_name)); 

            ERE(read_matrix_from_gamut_config_file(&rgb_data_mp,
                                                   (const char*)NULL,
                                                   file_name,
                                                   "indexed surface rgb file",
                                                   (char*)NULL, 0)); 
        }
        else 
        {
            world_version = get_synthetic_RGB(&rgb_data_mp, illum);

            if (world_version == ERROR)
            {
                add_error("Failed to generate synthetic RGB data from spectra");
                result = ERROR;
                break;
            }
        }

        /* Last count of RGBs wins for the number of surfaces - hack ... */
        num_surfaces = rgb_data_mp->num_rows;
        prob_uniform_data_cp->num_surf = num_surfaces;
        
        if (get_divide_by_sum_projection_matrix(&chrom_mp, 
                                                rgb_data_mp, 
                                                SKIP_INVALID_CHROMATICITY,
                                                1000.0,
                                                DBL_NOT_SET,
                                                DBL_NOT_SET,
                                                (Matrix**)NULL) == ERROR )
        {
            add_error("Failed to get chromaticity matrix.");
            result = ERROR;
            break;
        }   
            
        ERE(verbose_pso(30, "Uniform probability correlation matrix - Illuminant %3d/%d\n",
                            illum + 1, num_illum));
              
        /* Insert a uniform probability binarized histogram into the
        // P(colour|illum) matrix for the current illuminant.
        */
        if ( (insert_uniform_histogram(&prob_colour_illum_mp,
                                       illum, chrom_mp,
                                       fs_settable_num_bins,
                                       MAX_RG_CHROM)) == ERROR )
        {
            add_error("Failed to make histogram data for illum %d", illum + 1);
            result = ERROR;
            break;
        }
            
        if (ow_zero_matrix(rgb_data_mp) == ERROR )
        {
            add_error("Failed to reset RGB matrix to zero.");
            result = ERROR;
            break;
        }
            
        if (ow_zero_matrix(chrom_mp) == ERROR )
        {
            add_error("Failed to set rg-chromaticity matrix to zero.");
            result = ERROR;
            break;
        }
            
    } /* for each illuminant  */

    if (result == NO_ERROR)
    {
        double sum_of_cols;
        double ave; 
        
        /* Compute the probability of each colour as the sum of
        // occurances for the colour under all illuminants. */
        ERE(sum_matrix_rows(&prob_colour_vp, prob_colour_illum_mp));
            
        sum_of_cols = sum_vector_elements(prob_colour_vp);
            
        if (sum_of_cols == 0.0)
        {
            set_error("Sum of probabilities for all colours is 0.");
            result = ERROR;                    
        }
        
        if (result == NO_ERROR)
        {
            /* Compute probability of an illuminant given a colour. */
            for (illum = 0; illum < num_illum; illum++)
            {
                for (colour = 0; colour < num_colours; colour++)
                {
                    double p_colour;       /* P(colour)       */
                    double p_colour_illum; /* P(colour|illum) */
                    double p_illum_colour; /* P(illum|colour) */
                    
                    p_colour = prob_colour_vp->elements[colour];
                    
                    p_colour_illum
                        = prob_colour_illum_mp->elements[illum][colour];
                    
                    if (p_colour > 0.0)
                    {
                        p_illum_colour = p_colour_illum / p_colour;
                    }
                    else /* Illuminant not possible given colour. */
                    /* Kobus: I would explain this further as no illuminant has
                    // this colour.
                    */
                    {
                        p_illum_colour = 0.0;
                    }
                    
                    prob_uniform_data_cp->cormat_mp->elements[illum][colour]
                        = p_illum_colour;
                }
            }
        }
              
        ave = average_matrix_elements(prob_uniform_data_cp->cormat_mp);

        result = ow_divide_matrix_by_scalar(prob_uniform_data_cp->cormat_mp,
                                            ave); 
    }
          
    if (result != ERROR)
    {
        result = copy_correlation_matrix(new_uniform_data_cp_ptr,
                                         prob_uniform_data_cp);
    }

    if (result == NO_ERROR)
    {
        *world_version_ptr           = world_version;
        *illum_spectra_version_ptr   = illum_spectra_version;

        ERE(verbose_pso(10, "Uniform probability data remade.\n"));
        
    }
    else
    {
        add_error("Failed to copy uniform probability correlation matrix");
    }

#ifdef TEST
    if (result != ERROR)
    {
        ASSERT( prob_uniform_data_cp != NULL );
        ASSERT( prob_uniform_data_cp->cormat_mp != NULL );
        ASSERT( (*new_uniform_data_cp_ptr) != NULL );
    }
#endif 
    
    free_correlation_matrix(prob_uniform_data_cp);
    free_spectra(illum_sp); 
    free_matrix(rgb_data_mp); 
    free_matrix(chrom_mp);
    free_matrix(prob_colour_illum_mp);
    free_vector(prob_colour_vp);
    free_vector(cur_prob_row_vp);
    free_matrix(illum_rgb_mp); 
    free_vector(illum_rgb_vp); 
    free_vector(illum_chrom_vp); 

    kjb_fclose(illum_rgb_index_fp); 
    kjb_fclose(surface_rgb_index_fp); 

    return result; 
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int insert_uniform_histogram
(
    Matrix** uniform_cormat_mpp,
    int      illum,
    Matrix*  chrom_mp,
    int      num_bins,
    double   max_projection_coordinate 
)
{
    int     i, row, col;
    Matrix* histogram_2D_mp   = NULL;
    Vector* histogram_1D_vp   = NULL;
    int     num_colours       = num_bins * num_bins;
    int     result = NO_ERROR;
    double    sum_of_votes;
    
    if ((chrom_mp == NULL) || (num_bins < 1))
    {
        SET_ARGUMENT_BUG();
        return ERROR;    
    }

    if (max_projection_coordinate == DBL_NOT_SET)
    {
        max_projection_coordinate = (double)1.0; /* Default for rg-chrom. */
    }

    /* Allocate 1-D and 2-D probability histograms. */ 
    ERE(get_zero_vector( &histogram_1D_vp, num_colours ));
    ERE(get_zero_matrix( &histogram_2D_mp, num_bins, num_bins ));

    /* Bin the chromaticity values for all colours under the current illum. */
    for (i = 0; i < chrom_mp->num_rows; i++)
    {
        /* g = G/(R+G+B) */
        row = bin_chromaticity(chrom_mp->elements[i][G_CHROM_INDEX],
                               num_bins, max_projection_coordinate); /* g */

        /* r = R/(R+G+B) */
        col = bin_chromaticity(chrom_mp->elements[i][R_CHROM_INDEX],
                               num_bins, max_projection_coordinate); /* r */

        histogram_2D_mp->elements[row][col] += (double)1.0;
    }

    sum_of_votes = sum_matrix_elements(histogram_2D_mp);
    
    if (sum_of_votes == 0.0)
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR;
    }
    else
    {
        /* Insert the histogram for the current illuminant into to the
        // colour voting matrix. */
        ERE(map_2D_to_1D_histogram( &histogram_1D_vp, histogram_2D_mp ));
        ERE(put_matrix_row( (*uniform_cormat_mpp),
                            histogram_1D_vp, illum) );
    }
    
    free_matrix(histogram_2D_mp);
    free_vector(histogram_1D_vp);

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* ===========================================================================
 *                 do_probability_algorithm
 *
 * Performs the probability matrix algorithm for either uniform or gaussian
 * probability data. Checks to see if the data has been converted to log
 * space already - if not, this function performs the log conversion.
 *
 */

/* Kobus: 08/07/99 | Added id_str and plot_enable args. */
static int do_probability_algorithm
(
    Matrix*      rgb_data_mp,
    Vector**     result_vpp,
    Vector**     mlm_result_vpp,
    Vector**     mmse_result_vpp,
    double*      cpu_time_ptr,
    double*      mlm_cpu_time_ptr,
    double*      mmse_cpu_time_ptr,
    Corr_matrix* source_prob_data_cp,
    int          plot_enable,
    char*        id_str 
)
{
    int j; 
    long base_cpu_time; 
    Matrix* gaussian_mask_mp = NULL; 

    /* Shorthand to log of probability data */
    Matrix*  source_log_data_mp;
    
    /* Input RGB data from "gamut"             */
    Matrix*  input_chrom_mp      = NULL;
    Matrix*  input_2D_hist_mp    = NULL;
    Vector*  input_1D_hist_vp    = NULL;

    /* Illuminant data from correlation matrix */
    Matrix*  illum_rgb_mp        = NULL;
    const Matrix*  illum_chrom_mp;

    Vector*  sum_of_log_probs_vp = NULL;
    Vector*  best_chrom_vp       = NULL;

    /* Actual illuminant data - for debugging */
    Vector* actual_rgb_vp        = NULL;
    Vector* actual_chrom_vp      = NULL;
    double    max_sum_of_log_probs;
    int     prob_num_bins, prob_num_illum;
    int     i, row, col;
    int     result = NO_ERROR;

    if (source_prob_data_cp == NULL)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (source_prob_data_cp->cormat_mp == NULL)
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    init_cpu_time(); 

    /*
    // Kobus--I have no idea whether this verification business is necessary,
    // actually here, and if so, if it works.
    */

    /* VERIFY THAT THE DATA HAS ALREADY BEEN CONVERTED TO LOG SPACE */

    source_log_data_mp = source_prob_data_cp->cormat_mp;

    prob_num_illum = source_prob_data_cp->num_illum;
    prob_num_bins  = source_prob_data_cp->num_bins;

    if ( source_log_data_mp->num_rows != prob_num_illum )
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    if ( source_log_data_mp->num_cols !=
         (prob_num_bins * prob_num_bins) )
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    
    ERE(get_zero_matrix(&input_2D_hist_mp, prob_num_bins, prob_num_bins));
    ERE(get_zero_vector(&input_1D_hist_vp, (prob_num_bins * prob_num_bins)));
    
    /* Convert input RGB data to chromaticity space. */
    ERE( get_divide_by_sum_projection_matrix(&input_chrom_mp,
                                             rgb_data_mp, 
                                             SKIP_INVALID_CHROMATICITY,
                                             1000.0, DBL_NOT_SET, DBL_NOT_SET,
                                             (Matrix**)NULL) );
    
    for (i=0; i<input_chrom_mp->num_rows; i++)
    {
        row = bin_chromaticity(input_chrom_mp->elements[i][G_CHROM_INDEX],
                               prob_num_bins, MAX_RG_CHROM);
        col = bin_chromaticity(input_chrom_mp->elements[i][R_CHROM_INDEX],
                               prob_num_bins, MAX_RG_CHROM);
        /* Kobus (01/08/99) */ 
#ifdef USE_ONLY_PRESENCE_INFO
        input_2D_hist_mp->elements[row][col] = (double)1.0;
#else 
        input_2D_hist_mp->elements[row][col] += (double)1.0;
#endif 
    }
    ERE(map_2D_to_1D_histogram(&input_1D_hist_vp, input_2D_hist_mp));
    
    /* Get the sum of the log probabilities for colours in input dataset. */
    ERE(multiply_matrix_and_vector(&sum_of_log_probs_vp, source_log_data_mp,
                                   input_1D_hist_vp));

    if (sum_of_log_probs_vp->length != source_log_data_mp->num_rows)
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    illum_chrom_mp = source_prob_data_cp->illum_chrom_mp; 

    /* Kobus: 07-07-99 */
    if ((is_interactive()) && plot_enable)
    {   
        plot_posterior(illum_chrom_mp, sum_of_log_probs_vp, id_str);
    }
        
    verbose_pso(500, "Illuminant rg-chromaticities:\n");
    for (i = 0; i < illum_chrom_mp->num_rows; i++)
    {
        verbose_pso(500, "     %3d: (%6.4f, %6.4f)", i + 1,
                    illum_chrom_mp->elements[i][RED_INDEX],
                    illum_chrom_mp->elements[i][GREEN_INDEX]);
        
        verbose_pso(500, " sum of log(P) = %12.10f\n",
                    sum_of_log_probs_vp->elements[i]);

    }
    verbose_pso(500,
                "---------------------------------------------------------\n");

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(get_initialized_matrix(&fs_posterior_dist_mp, POSTERIOR_RESOLUTION + 1, 
                               POSTERIOR_RESOLUTION + 1,
                               CORRELATION_MINUS_INFINITY)); 

    for (i = 0; i < illum_chrom_mp->num_rows; i++)
    {
        if (sum_of_log_probs_vp->elements[i] > 0.5 * CORRELATION_MINUS_INFINITY)
        {
            int ii = illum_chrom_mp->elements[ i ][ 0 ] * POSTERIOR_RESOLUTION * 2.0;
            int jj = illum_chrom_mp->elements[ i ][ 1 ] * POSTERIOR_RESOLUTION * 2.0;
            
            if (    (ii >= 0) && (jj >= 0)
                 && (ii <= POSTERIOR_RESOLUTION) 
                 && (jj <= POSTERIOR_RESOLUTION))
            {
                double log_prob = sum_of_log_probs_vp->elements[i];

                if (log_prob > fs_posterior_dist_mp->elements[ ii ][ jj ])
                {
                    fs_posterior_dist_mp->elements[ ii ][ jj ] = log_prob;
                }
            }
            else 
            {
                verbose_pso(1, "Skipping 2D C-by-C rg (%.3f, %.3f).\n",
                            (double)illum_chrom_mp->elements[ i ][ 0 ], 
                            (double)illum_chrom_mp->elements[ i ][ 1 ]);
            }
        }
    }

    base_cpu_time = get_cpu_time(); 

    init_cpu_time(); 

    /* Compute the result as the illuminant with the greatest sum(log(P)). */
    i = get_max_vector_element(sum_of_log_probs_vp, &max_sum_of_log_probs);

    dbe(max_sum_of_log_probs); 

    if (max_sum_of_log_probs <= 0.5 * CORRELATION_MINUS_INFINITY)
    {
        verbose_pso(1, "No solution found using log prob corr matrix (MAP).\n");
        result = NO_SOLUTION; 
    }
    else 
    {
        /* Convert the chromaticity of the best illuminant back to RGB. */
        ERE(get_matrix_row(&best_chrom_vp, illum_chrom_mp, i));
        ERE(back_project_vector(result_vpp, best_chrom_vp, DIVIDE_BY_SUM));
        /*
        ERE(ow_multiply_vector_by_scalar(*result_vpp, 255.0));
        */
    
        verbose_pso(50, "Best %3d: (%6.4f, %6.4f)", i + 1,
                    best_chrom_vp->elements[RED_INDEX],
                    best_chrom_vp->elements[GREEN_INDEX]);
        
        verbose_pso(50, " sum of log(P) = %12.10f\n",
                    sum_of_log_probs_vp->elements[i]);
        
        ERE(get_actual_illum(&actual_rgb_vp));
        ERE(get_projection_vector(&actual_chrom_vp, actual_rgb_vp,
                                  DIVIDE_BY_SUM, 1000.0));
    
        verbose_pso(50, "Actual  : (%6.4f, %6.4f)\n\n",
                    actual_chrom_vp->elements[RED_INDEX],
                    actual_chrom_vp->elements[GREEN_INDEX]);
    }

    *cpu_time_ptr = base_cpu_time + get_cpu_time();
    *cpu_time_ptr /= (double)MILLI_SECONDS_PER_SECOND; 	

    if (    (result != ERROR) && (result != NO_SOLUTION) 
         && (fs_posterior_dist_mp != NULL)
       )
    {
        double max_log_prob = max_matrix_element(fs_posterior_dist_mp); 
        double offset = CORRELATION_MIN_LOG - max_log_prob;
        double offset_2 = max_log_prob - CORRELATION_MAX_LOG;

        dbe(max_log_prob); 
        dbe(exp(max_log_prob)); 
        dbe(CORRELATION_MIN_LOG); 
        dbe(CORRELATION_MAX_LOG); 
        dbe(offset); 
        dbe(offset_2); 

        if (offset > 0.0)
        {
            verbose_pso(1, "Adding %.3e to posterior distribution.\n", offset); 

            ERE(ow_add_scalar_to_matrix(fs_posterior_dist_mp, offset)); 
        }

#define DEAL_WITH_OVERFLOW

#ifdef DEAL_WITH_OVERFLOW
        else if (offset_2 > 0.0)
        {
            verbose_pso(1, "Subtracting %.3e from posterior distribution.\n", 
                        offset_2); 

            ERE(ow_subtract_scalar_from_matrix(fs_posterior_dist_mp, offset_2)); 
        }

        for (i=0; i <= POSTERIOR_RESOLUTION; i++)
        {
            for (j=0; j <= POSTERIOR_RESOLUTION; j++)
            {
                double log_prob = fs_posterior_dist_mp->elements[ i ][ j ];

                fs_posterior_dist_mp->elements[ i ][ j ] = exp(log_prob);
            }
        }
#else 
        if (offset_2 > 0.0)
        {
            verbose_pso(1, "Unable to compute MMSE result due to overflow.\n");
            result = NO_SOLUTION;
        }
        else
        {
            for (i=0; i <= POSTERIOR_RESOLUTION; i++)
            {
                for (j=0; j <= POSTERIOR_RESOLUTION; j++)
                {
                    double log_prob = fs_posterior_dist_mp->elements[ i ][ j ];

                    fs_posterior_dist_mp->elements[ i ][ j ] = exp(log_prob);
                }
            }
        }
#endif 
    }

    if (    (result != ERROR) && (result != NO_SOLUTION) 
         && (fs_posterior_dist_mp != NULL)
       )
    {
        double i_sum = 0.0;
        double j_sum = 0.0;
        double posterior_sum = 0.0;

        init_cpu_time();

        for (i=0; i <= POSTERIOR_RESOLUTION; i++)
        {
            for (j=0; j <= POSTERIOR_RESOLUTION; j++)
            {
                double prob = fs_posterior_dist_mp->elements[ i ][ j ];

                i_sum += (i + 0.5) * prob; 
                j_sum += (j + 0.5) * prob; 
                posterior_sum += prob; 
            }
        }

        dbe(posterior_sum);
        dbe(i_sum);
        dbe(j_sum); 

        if (posterior_sum == 0.0)
        {
            verbose_pso(1,
                    "No solution found using log prob corr matrix (MMSE).\n");
            verbose_pso(1, "Max sum of log probs is %e.\n", 
                        (double)max_sum_of_log_probs); 
            result = NO_SOLUTION; 
        }
        else 
        {
            double r = (i_sum / posterior_sum) / (2.0 * POSTERIOR_RESOLUTION); 
            double g = (j_sum / posterior_sum) / (2.0 * POSTERIOR_RESOLUTION); 

            ERE(get_target_vector(mmse_result_vpp, 3)); 

            r = MIN_OF(1.0, r); 
            g = MIN_OF(1.0, g); 

            (*mmse_result_vpp)->elements[ 0 ] = r;
            (*mmse_result_vpp)->elements[ 1 ] = g;
            (*mmse_result_vpp)->elements[ 2 ] = 1.0 - r - g; 

        dbe(r);
        dbe(g);
        dbe((*mmse_result_vpp)->elements[ 2 ]);

        }

        *mmse_cpu_time_ptr = base_cpu_time + get_cpu_time();
        *mmse_cpu_time_ptr /= (double)MILLI_SECONDS_PER_SECOND; 
    }

    if (    (result != ERROR) && (result != NO_SOLUTION) 
         && (fs_posterior_dist_mp != NULL)
         && (fs_correlation_mlm_mask_sigma > 0.0) 
         && (fs_correlation_mlm_mask_size > 1)
       )
    {
        double max_prob  = 0.0;
        int    i_for_max = NOT_SET;
        int    j_for_max = NOT_SET;

        init_cpu_time();

        if (result != ERROR)
        {
            result = get_gaussian_mask(&gaussian_mask_mp, 
                                       fs_correlation_mlm_mask_size,
                                       (double)fs_correlation_mlm_mask_sigma); 
        }

        if (result != ERROR)
        {
            result = convolve(&fs_mlm_posterior_dist_mp, fs_posterior_dist_mp, 
                              gaussian_mask_mp);
        }

        for (i=0; i <= POSTERIOR_RESOLUTION; i++)
        {
            for (j=0; j <= POSTERIOR_RESOLUTION; j++)
            {
                if (fs_mlm_posterior_dist_mp->elements[ i ][ j ] > max_prob) 
                {
                    i_for_max = i;
                    j_for_max = j;
                    max_prob = fs_mlm_posterior_dist_mp->elements[ i ][ j ]; 
                }
            }
        }

        if (    (max_prob == 0.0)
             || (i_for_max == NOT_SET) || (j_for_max == NOT_SET)
           )
        {
            verbose_pso(1,
                    "No solution found using log prob corr matrix (MLM).\n");
            verbose_pso(1,
                    "Max prob is %e, i_for_max is %d, and j_for_max is %d.\n",
                    (double)max_prob, i_for_max, j_for_max);
            verbose_pso(1, "Max sum of log probs is %e.\n", 
                        (double)max_sum_of_log_probs); 
            result = NO_SOLUTION; 
        }
        else 
        {
            double r = ((double)i_for_max + 0.5) / (2.0 * POSTERIOR_RESOLUTION);
            double g = ((double)j_for_max + 0.5) / (2.0 * POSTERIOR_RESOLUTION); 

            ERE(get_target_vector(mlm_result_vpp, 3)); 

            r = MIN_OF(1.0, r); 
            g = MIN_OF(1.0, g); 

            (*mlm_result_vpp)->elements[ 0 ] = r;
            (*mlm_result_vpp)->elements[ 1 ] = g;
            (*mlm_result_vpp)->elements[ 2 ] = 1.0 - r - g; 
        }

        *mlm_cpu_time_ptr = base_cpu_time + get_cpu_time();
        *mlm_cpu_time_ptr /= (double)MILLI_SECONDS_PER_SECOND; 
    }

    free_matrix(input_chrom_mp);
    free_matrix(input_2D_hist_mp);
    free_vector(input_1D_hist_vp);

    free_matrix(illum_rgb_mp);

    free_vector(sum_of_log_probs_vp);
    free_vector(best_chrom_vp);
    
    free_vector(actual_rgb_vp);
    free_vector(actual_chrom_vp);

    free_matrix(gaussian_mask_mp); 
    
    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    int i;

    free_matrix(fs_posterior_dist_mp);
    free_matrix(fs_mlm_posterior_dist_mp);

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        free_correlation_matrix(fs_cached_01_data_cp[i]);
        free_correlation_matrix(fs_cached_log_uniform_gaussian_cp[i]);
    }
}
#endif 

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int bin_chromaticity
(
    double chrom_value,
    int    num_bins,
    double max_projection_coordinate 
)
{
    int    bin;

    bin = (chrom_value * num_bins) / max_projection_coordinate;

    if (bin >= num_bins) bin = num_bins - 1; 

    return bin;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ */

static int map_2D_to_1D_histogram
(
    Vector** hist_1D_vpp,
    Matrix*  hist_2D_mp 
)
{
    int x, y, index, num_bins;

    if (hist_2D_mp == NULL)
    {
        SET_ARGUMENT_BUG();
        add_error("map_2D_to_1D_histogram(): Input 2D histogram is empty");
        return ERROR;
    }
    
    /* Assume equal number of bins per histogram axis. */
    num_bins = hist_2D_mp->num_rows;
    ERE(get_zero_vector(hist_1D_vpp, num_bins * num_bins ));
    
    index = 0;
    for (y = 0; y < num_bins; y++)
    {
        for (x = 0; x < num_bins; x++)
        {
            (*hist_1D_vpp)->elements[index] =  hist_2D_mp->elements[y][x];
            index++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ */

static int map_1D_to_2D_histogram
(
    Matrix** hist_2D_mpp,
    Vector*  hist_1D_vp 
)
{
    int x, y, index, num_bins;

    if (hist_1D_vp == NULL)
    {
        SET_ARGUMENT_BUG();
        add_error("map_1D_to_2D_histogram(): Input 1D histogram is empty");
        return ERROR;
    }

    /* Assume equal number of histogram bins per axis. */
    num_bins = sqrt( (double)hist_1D_vp->length );
    
    if ( (num_bins * num_bins) != hist_1D_vp->length )
    {
        add_error("Input histogram vector does not represent a square matrix");
        return ERROR;
    }

    ERE(get_zero_matrix(hist_2D_mpp, num_bins, num_bins));
    
    index = 0;
    for (y = 0; y < num_bins; y++)
    {
        for (x = 0; x < num_bins; x++)
        {
            (*hist_2D_mpp)->elements[y][x] = hist_1D_vp->elements[index];
            index++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ */

static int get_gaussian_mask
(
    Matrix** mask_mpp,
    int      mask_size,
    double   sigma 
)
{
    int r, c;
    int    mask_centre;
    double dist_sqr; 
    double sigma_sqr; 
    double gauss;
    double   mask_sum;
    
    if ( (mask_size < 0 ) || IS_EVEN(mask_size) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    sigma_sqr = sigma * sigma;

    mask_centre = mask_size / 2;
    mask_sum  = (double)0.0;
    
    ERE(get_zero_matrix(mask_mpp, mask_size, mask_size));

    for (r = 0; r < mask_size; r++)
    {
        for (c = 0; c < mask_size; c++)
        {
            dist_sqr =(r - mask_centre)*(r - mask_centre) + 
                                            (c - mask_centre)*(c - mask_centre);

            gauss = exp( -0.5 * dist_sqr / sigma_sqr ); 
            (*mask_mpp)->elements[r][c] = gauss;
            mask_sum += gauss;
        }
    }
    ERE(ow_divide_matrix_by_scalar(*mask_mpp,  mask_sum));
   
    /* db_mat(*mask_mpp); */
    
    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int convolve
(
    Matrix** output_mpp,
    Matrix*  input_mp,
    Matrix*  mask_mp 
)
{
    int ir, ic; /* Current entry in input matrix.      */
    int nr, nc; /* Neighbouring entry in input matrix. */
    int mr, mc; /* Mask entry.                         */

    int mask_size, mask_centre;
    double convolved_val;
#ifdef NORMALIZE_CONVOLUTIONS   /* Not normally used (this was a bad idea) */
    double weight;
#endif 

    if (input_mp == NULL)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
    
    if (   (input_mp->num_rows < 0)
        || (input_mp->num_cols < 0)
        || (input_mp->elements == NULL)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
    
    if (mask_mp == NULL)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
    
    mask_size = mask_mp->num_rows;
    mask_centre = mask_size / 2;
    
    if(  (mask_size != mask_mp->num_cols)
       || IS_EVEN(mask_size)
      )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_zero_matrix(output_mpp, input_mp->num_rows, input_mp->num_cols));
    
    for (ir = 0; ir < input_mp->num_rows; ir++)
    {
        for (ic = 0; ic < input_mp->num_cols; ic++)
        {
            convolved_val = 0.0;
#ifdef NORMALIZE_CONVOLUTIONS   /* Not normally used (this was a bad idea) */
            weight = 0.0;
#endif 

            for (mr = 0; mr < mask_size; mr++)
            {
                for (mc = 0; mc < mask_size; mc++)
                {
                    nr = ir + mr - mask_centre;
                    nc = ic + mc - mask_centre;

                    if (   (nr >= 0) && (nr < input_mp->num_rows)
                        && (nc >= 0) && (nc < input_mp->num_cols))
                    {
                        convolved_val += input_mp->elements[nr][nc]
                                                    * mask_mp->elements[mr][mc];
#ifdef NORMALIZE_CONVOLUTIONS   /* Not normally used (this was a bad idea) */
                        weight += mask_mp->elements[mr][mc];
#endif 
                    }
                }
            }
            (*output_mpp)->elements[ir][ic] = convolved_val;
#ifdef NORMALIZE_CONVOLUTIONS   /* Not normally used (this was a bad idea) */
            (*output_mpp)->elements[ir][ic] /= weight;
#endif 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Kobus: 07-07-99 */
static void plot_posterior
(
    const Matrix* illum_chrom_mp,
    Vector*       sum_of_log_probs_vp,
    char*         id_str 
)
{
    int     plot_id     = plot_open3();
    int     result       = (plot_id == ERROR) ? ERROR : NO_ERROR;
    char    title[ 300 ];
    Matrix* mp           = NULL;
    Matrix* count_mp     = NULL;
    int     i;
    double    min_r = 1.0;
    double    max_r = 0.0;
    double    min_g = 1.0;
    double    max_g = 0.0; 
    double    max_prob = 0.0; 

    if (result != ERROR)
    {
        result = kjb_sprintf(title, sizeof(title),
                             "Posterior Probability Distribution: %s.", 
                             id_str);
    }

    if (result != ERROR)
    {
        result = plot_set_title(plot_id, title, 1, 1);
    }

    if (result != ERROR)
    {
        result = get_zero_matrix(&mp, 1 + POSTERIOR_PLOT_RG_RESOLUTION, 
                                 1 + POSTERIOR_PLOT_RG_RESOLUTION);
    }

    if (result != ERROR)
    {
        result = get_zero_matrix(&count_mp, 1 + POSTERIOR_PLOT_RG_RESOLUTION, 
                                 1 + POSTERIOR_PLOT_RG_RESOLUTION);
    }

    for (i = 0; i < illum_chrom_mp->num_rows; i++)
    {
        double prob = exp((double)sum_of_log_probs_vp->elements[i]);

        if (prob > max_prob)
        {
            max_prob = prob; 
        }
    }

    for (i = 0; i < illum_chrom_mp->num_rows; i++)
    {
        if (exp((double)sum_of_log_probs_vp->elements[i]) > 0.02 * max_prob)
        {
            double r = illum_chrom_mp->elements[ i ][ 0 ];
            double g = illum_chrom_mp->elements[ i ][ 1 ];

            if (r < min_r) min_r = r; 
            if (g < min_g) min_g = g; 
            if (r > max_r) max_r = r; 
            if (g > max_g) max_g = g; 
        }
    }

    if (result != ERROR) 
    {
        for (i = 0; i < illum_chrom_mp->num_rows; i++)
        {
            double r_range = max_r - min_r;
            double g_range = max_g - min_g;
            int ii = (illum_chrom_mp->elements[ i ][ 0 ] - min_r) / r_range * 
                                           POSTERIOR_PLOT_RG_RESOLUTION;
            int jj = (illum_chrom_mp->elements[ i ][ 1 ] - min_g) / g_range * 
                                           POSTERIOR_PLOT_RG_RESOLUTION;
            
            if (    (ii >= 0) 
                 && (jj >= 0)
                 && (ii <= POSTERIOR_PLOT_RG_RESOLUTION)
                 && (jj <= POSTERIOR_PLOT_RG_RESOLUTION)
               )  
            {
                mp->elements[ ii ][ jj ] += 
                         exp((double)sum_of_log_probs_vp->elements[i]);
                count_mp->elements[ ii ][ jj ] += 1.0;
            }
        }

        if (    (ow_min_thresh_matrix(count_mp, 1.0) == ERROR)
             || (ow_divide_matrices_ew(mp, count_mp) == ERROR)
             || (ow_scale_matrix_by_max_abs(mp) == ERROR)
           )
        {
            result = ERROR; 
        }
        else 
        {
            result = plot_matrix_values(plot_id, mp, 
                                        (double)min_r, (double)max_r, 
                                        (double)min_g, (double)max_g);
        }
    }

    free_matrix(mp); 
    free_matrix(count_mp); 

    if (result == ERROR)
    {
        add_error("Error occured while plotting 2D correlation posterior."); 
        kjb_print_error();
        plot_close(plot_id);
    }
}

