
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin 
|  contributed the interface with the correlation code). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "do_error_report.h"
#include "compute_output.h"
#include "luminance.h"

/*
// This is a big HACK. Use #define's to hide it. 
//
// The low level code treats something to print as an ERROR as in "verifying
// that there is nothing to print".  I used this backwards logic to make adding
// this capability much easier.
// 
// Sorry!
*/
#define HAVE_SOMETHING_TO_PRINT ERROR
#define DRY_RUN                 TRUE
#define WET_RUN                 FALSE

/* -------------------------------------------------------------------------- */

static int fs_report_actual                        = TRUE;
static int fs_report_world_best_2D                 = TRUE;
static int fs_report_world_best_3D                 = TRUE;
static int fs_report_world_best_2D_diag            = TRUE;
static int fs_report_world_best_3D_diag            = TRUE;
static int fs_report_scene_best_2D                 = FALSE;
static int fs_report_scene_best_3D                 = FALSE;
static int fs_report_scene_best_2D_diag            = FALSE;
static int fs_report_scene_best_3D_diag            = FALSE;

static int fs_report_nothing                       = TRUE;
static int fs_report_scale_by_max                       = TRUE;
static int fs_report_ave_illum                     = TRUE;
static int fs_report_grey_world                    = TRUE;
static int fs_report_db_grey_world                 = TRUE;

static int fs_report_2D_surface_hull_ave           = FALSE;
static int fs_report_2D_surface_max_vol            = FALSE;
static int fs_report_2D_surface_weighted_ave       = FALSE;

static int fs_report_2D_surface_illum_hull_ave     = TRUE;
static int fs_report_2D_surface_illum_icsa         = TRUE;
static int fs_report_2D_surface_illum_max_vol      = TRUE;
static int fs_report_2D_surface_illum_scca         = TRUE;
static int fs_report_2D_surface_illum_scia         = TRUE;
static int fs_report_2D_surface_illum_weighted_ave = TRUE;

static int fs_report_3D_surface_max_vol            = TRUE;

static int fs_report_3D_surface_hull_ave           = TRUE;
static int fs_report_3D_surface_weighted_ave       = TRUE;

static int fs_report_3D_surface_illum_max_vol      = TRUE;

static int fs_report_3D_surface_illum_hull_ave     = TRUE;
static int fs_report_3D_surface_illum_weighted_ave = TRUE;
static int fs_report_3D_surface_illum_icsa         = TRUE;
static int fs_report_3D_surface_illum_scia         = TRUE;
static int fs_report_3D_surface_illum_scca         = TRUE;

static int fs_report_3D_surface_lum_hull_ave       = TRUE;
static int fs_report_3D_surface_lum_max_vol        = TRUE;

static int fs_report_3D_sil_hull_ave               = TRUE;
static int fs_report_3D_sil_max_vol                = TRUE;
static int fs_report_3D_sil_scla                   = TRUE;
static int fs_report_3D_sil_lcsa                   = TRUE;
static int fs_report_3D_sil_scia                   = TRUE;
static int fs_report_3D_sil_scca                   = TRUE;

static int fs_report_3D_siv_max_vol                = TRUE;
static int fs_report_3D_siv_hull_ave               = TRUE;

static int fs_report_rg_neural_net                 = TRUE;

/* Lindsay: Colour by Correlation - Sept 30/98 */
static int fs_report_correlation_01                    = TRUE;
static int fs_report_correlation_uniform_inv_dist_fill = TRUE;
static int fs_report_correlation_uniform_gaussian_fill = TRUE;
static int fs_report_correlation_gaussian              = TRUE;
static int fs_report_correlation_uniform_fitted        = TRUE;
/* End Lindsay */

/* Lindsay: 3D Colour by Correlation - Jun 04/99 */
static int fs_report_corr_3D_uniform_gaussian          = TRUE;
static int fs_report_corr_3D_uniform_fitted            = TRUE;
/* End Lindsay */

/* Lindsay: Earth Mover's Distance - Oct 01/98 */
static int fs_report_earthmover_3D_all_data            = TRUE;
static int fs_report_earthmover_3D_fixed_bins          = TRUE;
static int fs_report_earthmover_3D_k_means             = TRUE;
static int fs_report_earthmover_3D_clarans             = TRUE;
/* End Lindsay: Oct 01/98 */

#ifdef SIMPLE_TEMPLATE
static int fs_report_simple_template_alg                = TRUE;
#endif

#ifdef TEMPLATE
static int fs_report_template_first_alg                = TRUE;
static int fs_report_template_second_alg               = TRUE;
#endif

#ifdef MULTI_TEMPLATE
static int fs_report_multi_template_alg                = TRUE;
#endif

static int fs_report_title                             = TRUE;
static int fs_report_separator                         = TRUE;

static int fs_report_input_data_error                  = TRUE;
static int fs_report_illum_angular_error               = TRUE;
static int fs_report_illum_sum_chrom_error             = TRUE;
static int fs_report_illum_lab_error                   = TRUE;
static int fs_report_illum_rgb_error                   = TRUE;
static int fs_report_illum_r_div_sum_error             = TRUE;
static int fs_report_illum_g_div_sum_error             = TRUE;
static int fs_report_illum_sum_lum_error               = TRUE;
static int fs_report_scene_blue_chrom_rms_error        = TRUE;
static int fs_report_scene_sum_chrom_rms_error         = TRUE;
static int fs_report_scene_rgb_rms_error               = TRUE;
static int fs_report_scene_lab_rms_error               = TRUE;
static int fs_report_world_blue_chrom_rms_error        = TRUE;
static int fs_report_world_sum_chrom_rms_error         = TRUE;
static int fs_report_world_rgb_rms_error               = TRUE;
static int fs_report_world_lab_rms_error               = TRUE;
static int fs_report_cpu                               = FALSE;

#ifdef TEST
static int fs_report_luminance_errors                  = TRUE;
static int fs_report_luminance_factors                 = TRUE;
#else
static int fs_report_luminance_errors                  = TRUE;
static int fs_report_luminance_factors                 = TRUE;
#endif

/* -------------------------------------------------------------------------- */

static int do_input_data_error_report
(
    Input_data_error* error_one_by_sharp_ptr,
    Input_data_error* error_two_by_sharp_ptr,
    int               dry_run 
);

static int do_error_report_by_algorithm
(
    Error_by_sharp* error_one_ptr,
    Error_by_sharp* error_two_ptr,
    int             dry_run 
);

static int do_best_fit_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);

static int do_standard_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);

static int do_2D_gamut_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

static int do_3D_gamut_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

static int do_3D_gamut_surface_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

static int do_3D_gamut_surface_lum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

static int do_3D_gamut_surface_illum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

static int do_3D_gamut_surface_illum_lum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
);

/* Lindsay: Added Feb 17/98 */
static int do_correlation_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);
/* End Lindsay: Feb 17/98 */              

#ifdef SIMPLE_TEMPLATE
static int do_simple_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);
#endif

#ifdef TEMPLATE
static int do_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);
#endif

#ifdef MULTI_TEMPLATE
static int do_multi_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
);
#endif

static int print_one_error_line
(
    const char* algorithm,
    int         sharp_method,
    double      error_one,
    double      error_two,
    int         dry_run 
);

static int print_one_relative_error_line
(
    const char* algorithm,
    int         sharp_method,
    double      error_one,
    double      error_two,
    int         dry_run 
);

static int print_head
(
    const char* title,
    const char* error_stat_one_str,
    const char* error_stat_two_str 
);

/* -------------------------------------------------------------------------- */

int set_report_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_boolean_value;
    int  do_all             = FALSE;
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "report-all")))
    {
        do_all = TRUE;
    }
    
    if (do_all || (match_pattern(lc_option, "report-title")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-title = %s\n",
                    fs_report_title ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Report titles %s used.\n", 
                    fs_report_title ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_title = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all 
         || (match_pattern(lc_option, "report-separator"))
         || (match_pattern(lc_option, "report-sep"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-separator = %s\n",
                    fs_report_separator ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Report separators %s used\n",
                    fs_report_separator ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_separator = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-actual")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-actual = %s\n",
                    fs_report_actual ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Actual illuminant results %s reported.\n",
                    fs_report_actual ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_actual = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-world-best-2d"))
               || (match_pattern(lc_option, "report-world-best-2d-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-best-2d = %s\n",
                    fs_report_world_best_2D ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D world fit results %s reported.\n",
                    fs_report_world_best_2D ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_best_2D = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-world-best-3d"))
               || (match_pattern(lc_option, "report-world-best-3d-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-best-3d = %s\n",
                    fs_report_world_best_3D ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D world fit results %s reported.\n",
                    fs_report_world_best_3D ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_best_3D = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-world-best-2d-diag"))
               || (match_pattern(lc_option, "report-world-best-2d-diag-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-best-2d-diag = %s\n",
                    fs_report_world_best_2D_diag ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D world diag fit results %s reported.\n",
                    fs_report_world_best_2D_diag ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_best_2D_diag = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-world-best-3d-diag"))
               || (match_pattern(lc_option, "report-world-best-3d-diag-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-best-3d-diag = %s\n",
                    fs_report_world_best_3D_diag ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D world diag fit results %s reported.\n",
                    fs_report_world_best_3D_diag ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_best_3D_diag = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-scene-best-2d"))
               || (match_pattern(lc_option, "report-scene-best-2d-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-best-2d = %s\n",
                    fs_report_scene_best_2D ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D scene fit results %s reported.\n",
                    fs_report_scene_best_2D ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_best_2D = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-scene-best-3d"))
               || (match_pattern(lc_option, "report-scene-best-3d-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-best-3d = %s\n",
                    fs_report_scene_best_3D ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D scene fit results %s reported.\n",
                    fs_report_scene_best_3D ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_best_3D = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-scene-best-2d-diag"))
               || (match_pattern(lc_option, "report-scene-best-2d-diag-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-best-2d-diag = %s\n",
                    fs_report_scene_best_2D_diag ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D scene diag fit results %s reported.\n",
                    fs_report_scene_best_2D_diag ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_best_2D_diag = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-scene-best-3d-diag"))
               || (match_pattern(lc_option, "report-scene-best-3d-diag-fit")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-best-3d-diag = %s\n",
                    fs_report_scene_best_3D_diag ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D scene diag fit results %s reported.\n",
                    fs_report_scene_best_3D_diag ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_best_3D_diag = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-nothing")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-nothing = %s\n",
                    fs_report_nothing ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The result of doing scale by max %s reported.\n",
                    fs_report_scale_by_max ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_nothing = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-ave-illum"))
               || (match_pattern(lc_option, "report-average-illumination")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-ave-illum = %s\n",
                    fs_report_ave_illum ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Average illumination results %s reported.\n",
                    fs_report_ave_illum ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_ave_illum = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-scale-by-max")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scale-by-max = %s\n",
                    fs_report_scale_by_max ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scale by max results %s reported.\n",
                    fs_report_scale_by_max ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scale_by_max = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-gw"))
               || (match_pattern(lc_option, "report-grey-world")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-gw = %s\n",
                    fs_report_grey_world ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Grey world results %s reported.\n",
                    fs_report_grey_world ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_grey_world = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-db-gw"))
               || (match_pattern(lc_option, "report-database-grey-world")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-db-gw = %s\n",
                    fs_report_db_grey_world ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Database grey world results %s reported.\n",
                    fs_report_db_grey_world ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_db_grey_world = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-s-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-s-mv = %s\n",
                    fs_report_2D_surface_max_vol ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface max vol results %s reported.\n",
                    fs_report_2D_surface_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-mv = %s\n", 
                    fs_report_2D_surface_illum_max_vol ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface illum max vol results %s reported.\n",
                    fs_report_2D_surface_illum_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-s-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-s-ha = %s\n",
                    fs_report_2D_surface_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface hull average results %s reported.\n",
                    fs_report_2D_surface_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-ha = %s\n",
                    fs_report_2D_surface_illum_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface illum hull average results %s reported.\n",
                    fs_report_2D_surface_illum_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-icsa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-icsa = %s\n",
                    fs_report_2D_surface_illum_icsa ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D illum constrained hull average results %s reported.\n",
                    fs_report_2D_surface_illum_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_icsa = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-scia")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-scia = %s\n",
                    fs_report_2D_surface_illum_scia ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface constrained illum average results %s reported.\n",
                    fs_report_2D_surface_illum_scia ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_scia = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-scca")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-scca = %s\n",
                    fs_report_2D_surface_illum_scca ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface constrained chromaticity average results %s reported.\n",
                    fs_report_2D_surface_illum_scca ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_scca = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-s-wa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-s-wa = %s\n",
                    fs_report_2D_surface_weighted_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface weighted average results %s reported.\n",
                    fs_report_2D_surface_weighted_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_weighted_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-2d-si-wa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-2d-si-wa = %s\n",
                    fs_report_2D_surface_illum_weighted_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D surface illum weighted average results %s reported.\n",
                    fs_report_2D_surface_illum_weighted_ave ?
                                                           "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_2D_surface_illum_weighted_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-3d-s-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-s-mv = %s\n",
                    fs_report_3D_surface_max_vol ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface max vol results %s reported.\n",
                    fs_report_3D_surface_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (do_all || (match_pattern(lc_option, "report-3d-s-wa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-s-wa = %s\n",
                    fs_report_3D_surface_weighted_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface weighted average results %s reported.\n",
                    fs_report_3D_surface_weighted_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_weighted_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-3d-s-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-s-ha = %s\n",
                    fs_report_3D_surface_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface hull average results %s reported.\n",
                    fs_report_3D_surface_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-3d-si-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-mv = %s\n",
                    fs_report_3D_surface_illum_max_vol ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum max vol results %s reported.\n",
                    fs_report_3D_surface_illum_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
     
    if (do_all || (match_pattern(lc_option, "report-3d-si-wa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-wa = %s\n",
                    fs_report_3D_surface_illum_weighted_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum weighted average results %s reported.\n",
                    fs_report_3D_surface_illum_weighted_ave ? 
                                                           "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_weighted_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
     
    if (do_all || (match_pattern(lc_option, "report-3d-si-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-ha = %s\n",
                    fs_report_3D_surface_illum_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum hull average results %s reported.\n",
                    fs_report_3D_surface_illum_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (do_all || (match_pattern(lc_option, "report-3d-si-icsa")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-icsa = %s\n",
                    fs_report_3D_surface_illum_icsa ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D illum constrained hull average results %s reported.\n",
                    fs_report_3D_surface_illum_icsa ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_icsa = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-3d-si-scia")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-scia = %s\n",
                    fs_report_3D_surface_illum_scia ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained illum average results %s reported.\n",
                    fs_report_3D_surface_illum_scia ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_scia = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (do_all || (match_pattern(lc_option, "report-3d-si-scca")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-scca = %s\n",
                    fs_report_3D_surface_illum_scca ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained chromaticity average results %s reported.\n",
                    fs_report_3D_surface_illum_scca ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_illum_scca = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (do_all || (match_pattern(lc_option, "report-3d-sl-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-sl-ha = %s\n",
                    fs_report_3D_surface_lum_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface luminance hull average results %s reported.\n",
                    fs_report_3D_surface_lum_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_lum_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
 
    if (do_all || (match_pattern(lc_option, "report-3d-sl-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-sl-ha = %s\n",
                    fs_report_3D_surface_lum_max_vol ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface luminance max volume results %s reported.\n",
                    fs_report_3D_surface_lum_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_surface_lum_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
 
    if (do_all || (match_pattern(lc_option, "report-3d-sil-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-sil-ha = %s\n",
                    fs_report_3D_sil_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum hull average results %s reported.\n",
                    fs_report_3D_sil_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-3d-sil-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-sil-ha = %s\n",
                    fs_report_3D_sil_max_vol ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum max volume results %s reported.\n",
                    fs_report_3D_sil_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-3d-sil-scla"))
         || (match_pattern(lc_option, "report-sil-scla"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-scla = %s\n",
                    fs_report_3D_sil_scla ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained illum+lum ave results %s reported.\n",
                    fs_report_3D_sil_scla ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_scla = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    do_all
         || (match_pattern(lc_option, "report-3d-sil-lcsa"))
         || (match_pattern(lc_option, "report-sil-lcsa"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-lcsa = %s\n",
                    fs_report_3D_sil_lcsa ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained illum+lum ave results %s reported.\n",
                    fs_report_3D_sil_lcsa ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_lcsa = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    do_all
         || (match_pattern(lc_option, "report-3d-sil-scia"))
         || (match_pattern(lc_option, "report-sil-scia"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-scia = %s\n",
                    fs_report_3D_sil_scia ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained illum+lum ave results %s reported.\n",
                    fs_report_3D_sil_scia ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_scia = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    do_all
         || (match_pattern(lc_option, "report-3d-sil-scca"))
         || (match_pattern(lc_option, "report-sil-scca"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-si-scca = %s\n",
                    fs_report_3D_sil_scca ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface constrained illum+lum ave results %s reported.\n",
                    fs_report_3D_sil_scca ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_sil_scca = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (do_all || (match_pattern(lc_option, "report-3d-siv-mv")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-siv-mv = %s\n",
                    fs_report_3D_siv_max_vol ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum var max vol results %s reported.\n",
                    fs_report_3D_siv_max_vol ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_siv_max_vol = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
     
    if (do_all || (match_pattern(lc_option, "report-3d-siv-ha")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-3d-siv-ha = %s\n",
                    fs_report_3D_siv_hull_ave ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D surface illum var hull average results %s reported.\n",
                    fs_report_3D_siv_hull_ave ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_3D_siv_hull_ave = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-rg-neural-net")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-rg-neural-net = %s\n",
                    fs_report_rg_neural_net ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("RG neural net results %s reported.\n",
                    fs_report_rg_neural_net ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_rg_neural_net = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    /* Lindsay: Colour by Correlation - Sept 30/98 */    
    /* 0/1 Correlation Matrix. */
    if (do_all || (match_pattern(lc_option, "report-corr-01"))
               || (match_pattern(lc_option, "report-correlation-01"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-01 = %s\n",
                    fs_report_correlation_01 ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation 0/1 results %s reported.\n",
                    fs_report_correlation_01 ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_correlation_01 = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* Uniform Probability Correlation Matrix - Inverse Distance Fill. */
    if (do_all || (match_pattern(lc_option,
                                 "report-corr-uniform-inv-dist-fill"))
               || (match_pattern(lc_option,
                                 "report-correlation-uniform-inv-dist-fill"))
               || (match_pattern(lc_option,
                                 "report-correlation-uniform-inverse-distance-fill"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-uniform-inverse-distance-fill = %s\n",
                    fs_report_correlation_uniform_inv_dist_fill ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation Uniform Probability (Inverse Distance Fill) results %s reported.\n",
                    fs_report_correlation_uniform_inv_dist_fill ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_correlation_uniform_inv_dist_fill = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* Uniform Probability Correlation Matrix - Gaussian Mask Fill. */
    if (do_all || (match_pattern(lc_option,
                                 "report-corr-uniform-gaussian-fill"))
               || (match_pattern(lc_option,
                                 "report-correlation-uniform-gaussian-fill"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-uniform-gaussian-fill = %s\n",
                    fs_report_correlation_uniform_gaussian_fill ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation Uniform Probability (Gaussian Mask Fill) results %s reported.\n",
                    fs_report_correlation_uniform_gaussian_fill ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_correlation_uniform_gaussian_fill = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* Gaussian Probability Correlation Matrix. */
    if (do_all || (match_pattern(lc_option, "report-corr-gaussian"))
               || (match_pattern(lc_option, "report-correlation-gaussian"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-gaussian = %s\n",
                    fs_report_correlation_gaussian ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation Gaussian Probability results %s reported.\n",
                    fs_report_correlation_gaussian ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_correlation_gaussian = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* Dummy algorithm (infrastructure) */
        if (do_all || (match_pattern(lc_option, "report-corr-fitted"))
                   || (match_pattern(lc_option, "report-correlation-fitted"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-fitted = %s\n",
                    fs_report_correlation_uniform_fitted ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation fitted results %s reported.\n",
                    fs_report_correlation_uniform_fitted ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_correlation_uniform_fitted = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    /* End Lindsay - Sept 30/98 */

    /* Lindsay: 3D Colour by Correlation - Jun 03/99 */

    /* 3D Uniform Correlation Matrix - Gaussian Mask Fill. */
    if (do_all || (match_pattern(lc_option,
                                 "report-corr-3d-uniform-gaussian"))
               || (match_pattern(lc_option,
                                 "report-correlation-3d-uniform-gaussian"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-3d-uniform-gaussian = %s\n",
                    fs_report_corr_3D_uniform_gaussian ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Uniform Correlation (Gaussian Fill) results %s reported.\n",
                    fs_report_corr_3D_uniform_gaussian ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_corr_3D_uniform_gaussian = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* 3D Uniform (Fitted) Probability Correlation Matrix  */
    if (do_all || (match_pattern(lc_option,
                                 "report-corr-3d-uniform-fitted"))
               || (match_pattern(lc_option,
                                 "report-correlation-3d-uniform-fitted"))
        )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-correlation-3d-uniform-fitted = %s\n",
                    fs_report_corr_3D_uniform_fitted ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Uniform Correlation (Fitted) results %s reported.\n",
                    fs_report_corr_3D_uniform_fitted ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_corr_3D_uniform_fitted = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    /* End Lindsay: Jun 03/99 */

    /* Lindsay: Oct 01/98 */
    /* 3D Earth Mover's Distance - Using All Data (No Clustering) */ 
    if (do_all || (match_pattern(lc_option, "report-emd-3d-all-data"))
               || (match_pattern(lc_option, "report-earthmover-3d-all-data"))
        )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-earthmover-3d-all-data = %s\n",
                    fs_report_earthmover_3D_all_data ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Earthmover (All Data) results %s reported.\n",
                    fs_report_earthmover_3D_all_data ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_earthmover_3D_all_data = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* 3D Earth Mover's Distance - Fixed Bin Clustering */
    if (do_all || (match_pattern(lc_option,
                                 "report-emd-3d-fixed-bins"))
               || (match_pattern(lc_option,
                                 "report-earthmover-3d-fixed-bins"))
        )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-earthmover-3d-fixed-bins = %s\n",
                    fs_report_earthmover_3D_fixed_bins ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Earthmover (Fixed Bin Clustering) results %s reported.\n",
                    fs_report_earthmover_3D_fixed_bins ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_earthmover_3D_fixed_bins = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* 3D Earth Mover's Distance - k-Means Clustering */
    if (do_all || (match_pattern(lc_option, "report-emd-3d-k-means"))
               || (match_pattern(lc_option, "report-earthmover-3d-k-means"))
        )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-earthmover-3d-k-means = %s\n",
                    fs_report_earthmover_3D_k_means ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Earthmover (k-Means Clustering) results %s reported.\n",
                    fs_report_earthmover_3D_k_means ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_earthmover_3D_k_means = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    /* 3D Earth Mover's Distance - CLARANS Clustering */
    if (do_all || (match_pattern(lc_option, "report-emd-3d-clarans"))
               || (match_pattern(lc_option, "report-earthmover-3d-clarans"))
        )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-earthmover-3d-clarans = %s\n",
                    fs_report_earthmover_3D_clarans ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Earthmover (CLARANS) results %s reported.\n",
                    fs_report_earthmover_3D_clarans ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_earthmover_3D_clarans = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    /* End Lindsay: Oct 01/98 */

#ifdef SIMPLE_TEMPLATE
    if (do_all || (match_pattern(lc_option, "report-simple-template-alg")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-simple-template-alg = %s\n",
                     fs_report_simple_template_alg? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Simple template alg results %s reported.\n",
                    fs_report_simple_template_alg ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_simple_template_alg = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
#endif 
    
#ifdef TEMPLATE
    if (do_all || (match_pattern(lc_option, "report-template-first-alg")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-template-first-alg = %s\n",
                     fs_report_template_first_alg? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Template first alg results %s reported.\n",
                    fs_report_template_first_alg ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_template_first_alg = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-template-second-alg")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-template-second-alg = %s\n",
                    fs_report_template_second_alg ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Template second alg results %s reported.\n",
                    fs_report_template_second_alg ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_template_second_alg = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
#endif 
    
#ifdef MULTI_TEMPLATE
    if (do_all || (match_pattern(lc_option, "report-multi-template-alg")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-multi-template-alg = %s\n",
                     fs_report_multi_template_alg? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Multi template alg results %s reported.\n",
                    fs_report_multi_template_alg ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_multi_template_alg = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
#endif 
    
    if (do_all || (match_pattern(lc_option, "report-input-data-error")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-input-data-error = %s\n",
                    fs_report_input_data_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Input data error estimates "));
            ERE(pso("required for intersections %s reported.\n",
                    fs_report_input_data_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_input_data_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-angular-error"))
               || (match_pattern(lc_option, "report-angular-error")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-angular-error = %s\n",
                    fs_report_illum_angular_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum angular error results %s reported.\n",
                    fs_report_illum_angular_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_angular_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-sum-chrom-error"))
               || (match_pattern(lc_option, "report-sum-chrom-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-sum-chrom-error = %s\n",
                    fs_report_illum_sum_chrom_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum sum chrom error results %s reported.\n",
                    fs_report_illum_sum_chrom_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_sum_chrom_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-lab-error"))
               || (match_pattern(lc_option, "report-lab-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-lab-error = %s\n",
                    fs_report_illum_lab_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum lab error results %s reported.\n",
                    fs_report_illum_lab_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_lab_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-rgb-error"))
               || (match_pattern(lc_option, "report-rgb-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-rgb-error = %s\n",
                    fs_report_illum_rgb_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum rgb error results %s reported.\n",
                    fs_report_illum_rgb_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_rgb_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-r-div-sum-error"))
               || (match_pattern(lc_option, "report-r-div-sum-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-r-div-sum-error = %s\n",
                    fs_report_illum_r_div_sum_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum R/(R+G+B) error results %s reported.\n",
                    fs_report_illum_r_div_sum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_r_div_sum_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-g-div-sum-error"))
               || (match_pattern(lc_option, "report-g-div-sum-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-g-div-sum-error = %s\n",
                    fs_report_illum_g_div_sum_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum G/(R+G+B) error results %s reported.\n",
                    fs_report_illum_g_div_sum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_g_div_sum_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-illum-sum-lum-error"))
               || (match_pattern(lc_option, "report-sum-lum-error"))) 
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-sum-lum-error = %s\n",
                    fs_report_illum_sum_lum_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum R+G+B error results %s reported.\n",
                    fs_report_illum_sum_lum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_illum_sum_lum_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-world-rgb-rms-error"))
         || (match_pattern(lc_option, "report-rgb-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-rgb-rms-error = %s\n",
                    fs_report_world_rgb_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World RGB rms error results %s reported.\n",
                    fs_report_world_rgb_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_rgb_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-blue-chrom-rms-error"))
         || (match_pattern(lc_option, "report-world-blue-chrom-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-blue-chrom-rms-error = %s\n",
                    fs_report_world_blue_chrom_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World blue chrom rms error results %s reported.\n",
                    fs_report_world_blue_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_blue_chrom_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-sum-chrom-rms-error"))
         || (match_pattern(lc_option, "report-world-sum-chrom-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-sum-chrom-rms-error = %s\n",
                    fs_report_world_sum_chrom_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World sum chrom rms error results %s reported.\n",
                    fs_report_world_sum_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_sum_chrom_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all 
         || (match_pattern(lc_option, "report-lab-rms-error"))
         || (match_pattern(lc_option, "report-world-lab-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-world-lab-rms-error = %s\n",
                    fs_report_world_lab_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World Lab rms error results %s reported.\n",
                    fs_report_world_lab_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_world_lab_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    do_all 
        || (match_pattern(lc_option, "report-rgb-rms-error"))
        || (match_pattern(lc_option, "report-scene-rgb-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-rgb-rms-error = %s\n",
                    fs_report_scene_rgb_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene RGB rms error results %s reported.\n",
                    fs_report_scene_rgb_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_rgb_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-blue-chrom-rms-error"))
         || (match_pattern(lc_option, "report-scene-blue-chrom-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-blue-chrom-rms-error = %s\n",
                    fs_report_scene_blue_chrom_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene blue chrom rms error results %s reported.\n",
                    fs_report_scene_blue_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_blue_chrom_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-sum-chrom-rms-error"))
         || (match_pattern(lc_option, "report-scene-sum-chrom-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-sum-chrom-rms-error = %s\n",
                    fs_report_scene_sum_chrom_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene sum chrom rms error results %s reported.\n",
                    fs_report_scene_sum_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_sum_chrom_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (    do_all
         || (match_pattern(lc_option, "report-lab-rms-error"))
         || (match_pattern(lc_option, "report-scene-lab-rms-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-scene-lab-rms-error = %s\n",
                    fs_report_scene_lab_rms_error ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene Lab rms error results %s reported.\n",
                    fs_report_scene_lab_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_scene_lab_rms_error = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    do_all 
        || (match_pattern(lc_option, "report-cpu"))
        || (match_pattern(lc_option, "report-cpu-time"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-cpu-time = %s\n",
                    fs_report_cpu ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cpu results %s reported.\n",
                    fs_report_cpu ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_cpu = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-luminance-errors")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-luminance-errors = %s\n",
                    fs_report_luminance_errors ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Luminance error results %s reported.\n",
                    fs_report_luminance_factors ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_luminance_errors = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    if (do_all || (match_pattern(lc_option, "report-luminance-factors")))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("report-luminance-factors = %s\n",
                    fs_report_luminance_factors ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Luminance factor results %s reported.\n",
                    fs_report_luminance_factors ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_report_luminance_factors = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }
    
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_error_report
(
    Colour_constancy_error* error_one_ptr,
    Colour_constancy_error* error_two_ptr,
    Input_data_error*       input_data_error_one_ptr,
    Input_data_error*       input_data_error_two_ptr,
    const char*             heading,
    const char*             error_stat_one_str,
    const char*             error_stat_two_str 
)
{
    int        dry_run_result; 

    if ((fs_report_title) && (heading != NULL))
    {
        ERE(pso(heading)); 
    }

    if (fs_report_input_data_error)
    {
        dry_run_result = do_input_data_error_report(input_data_error_one_ptr,
                                                    input_data_error_two_ptr,
                                                    DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Error Estimates Required for Intersections",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_input_data_error_report(input_data_error_one_ptr,
                                           input_data_error_two_ptr, WET_RUN));

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_angular_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->illum_angular),
                                         &(error_two_ptr->illum_angular),
                                         DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Angular Error Predicting White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_angular),
                                             &(error_two_ptr->illum_angular),
                                             WET_RUN));
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_sum_chrom_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                           &(error_one_ptr->illum_sum_chrom),
                                           &(error_two_ptr->illum_sum_chrom),
                                           DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head(
                        "Sum Chromaticity Error Predicting White Under Unknown",
                         error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_sum_chrom),
                                            &(error_two_ptr->illum_sum_chrom),
                                            WET_RUN));
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_rgb_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->illum_rgb),
                                         &(error_two_ptr->illum_rgb),
                                         DRY_RUN);
        
        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("RGB Error Predicting White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_rgb),
                                             &(error_two_ptr->illum_rgb),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_r_div_sum_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->illum_r_div_sum),
                                         &(error_two_ptr->illum_r_div_sum),
                                         DRY_RUN);
        
        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Error Predicting R/(R+G+B) of White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_r_div_sum),
                                             &(error_two_ptr->illum_r_div_sum),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_g_div_sum_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->illum_g_div_sum),
                                         &(error_two_ptr->illum_g_div_sum),
                                         DRY_RUN);
        
        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Error Predicting G/(R+G+B) of White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_g_div_sum),
                                             &(error_two_ptr->illum_g_div_sum),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_sum_lum_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->illum_sum_lum),
                                         &(error_two_ptr->illum_sum_lum),
                                         DRY_RUN);
        
        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Error Predicting R+G+B of White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_sum_lum),
                                             &(error_two_ptr->illum_sum_lum),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_illum_lab_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                  &(error_one_ptr->illum_lab),
                                                  &(error_two_ptr->illum_lab),
                                                  DRY_RUN);
        
        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("L*a*b Error Predicting White Under Unknown",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->illum_lab),
                                             &(error_two_ptr->illum_lab),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_world_sum_chrom_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->world_map_sum_chrom_rms),
                                         &(error_two_ptr->world_map_sum_chrom_rms),
                                         DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("World RMS Sum Chromaticity Mapping Error",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(
                                         &(error_one_ptr->world_map_sum_chrom_rms),
                                         &(error_two_ptr->world_map_sum_chrom_rms),
                                         WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_world_blue_chrom_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->world_map_blue_chrom_rms),
                                         &(error_two_ptr->world_map_blue_chrom_rms),
                                         DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("World RMS Blue Chromaticity Mapping Error",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(
                                         &(error_one_ptr->world_map_blue_chrom_rms),
                                         &(error_two_ptr->world_map_blue_chrom_rms),
                                         WET_RUN));
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_world_rgb_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                 &(error_one_ptr->world_map_3D_rms),
                                                 &(error_two_ptr->world_map_3D_rms),
                                                 DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("World RMS Mapping Error",
                               error_stat_one_str, error_stat_two_str));
                ERE(print_output_dependence_message()); 
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->world_map_3D_rms),
                                             &(error_two_ptr->world_map_3D_rms),
                                             WET_RUN));

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_world_lab_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                &(error_one_ptr->world_map_3D_lab),
                                                &(error_two_ptr->world_map_3D_lab),
                                                DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("World L*a*b Error", error_stat_one_str, 
                               error_stat_two_str));
                ERE(print_output_dependence_message()); 
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->world_map_3D_lab),
                                             &(error_two_ptr->world_map_3D_lab),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_scene_sum_chrom_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->scene_map_sum_chrom_rms),
                                         &(error_two_ptr->scene_map_sum_chrom_rms),
                                         DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Scene RMS Sum Chromaticity Mapping Error",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(
                                         &(error_one_ptr->scene_map_sum_chrom_rms),
                                         &(error_two_ptr->scene_map_sum_chrom_rms),
                                         WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_scene_blue_chrom_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                         &(error_one_ptr->scene_map_blue_chrom_rms),
                                         &(error_two_ptr->scene_map_blue_chrom_rms),
                                         DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Scene RMS Blue Chromaticity Mapping Error",
                               error_stat_one_str, error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(
                                         &(error_one_ptr->scene_map_blue_chrom_rms),
                                         &(error_two_ptr->scene_map_blue_chrom_rms),
                                         WET_RUN));
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_scene_rgb_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                 &(error_one_ptr->scene_map_3D_rms),
                                                 &(error_two_ptr->scene_map_3D_rms),
                                                 DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Scene RMS Mapping Error",
                               error_stat_one_str, error_stat_two_str));
                ERE(print_output_dependence_message()); 
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->scene_map_3D_rms),
                                             &(error_two_ptr->scene_map_3D_rms),
                                             WET_RUN));

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_scene_lab_rms_error)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                &(error_one_ptr->scene_map_3D_lab),
                                                &(error_two_ptr->scene_map_3D_lab),
                                                DRY_RUN);

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("Scene L*a*b Error", error_stat_one_str, 
                               error_stat_two_str));
                ERE(print_output_dependence_message()); 
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->scene_map_3D_lab),
                                             &(error_two_ptr->scene_map_3D_lab),
                                             WET_RUN));
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_cpu)
    {
        dry_run_result = do_error_report_by_algorithm(
                                                  &(error_one_ptr->cpu_time),
                                                  &(error_two_ptr->cpu_time),
                                                  DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title)
            {
                ERE(print_head("CPU Time (Seconds)", error_stat_one_str,
                               error_stat_two_str));
            }

            ERE(do_error_report_by_algorithm(&(error_one_ptr->cpu_time),
                                             &(error_two_ptr->cpu_time),
                                             WET_RUN)); 
            
            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    kjb_fflush(stdout); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_luminance_error_report(void)
{
    int        dry_run_result; 

    if (fs_report_luminance_errors)
    {
        dry_run_result = do_current_luminance_error_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                ERE(pso("        Error Predicting Luminance\n\n"));
            }

            ERE(do_current_luminance_error_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_luminance_factor_report(void)
{
    int        dry_run_result; 

    if (fs_report_luminance_factors)
    {
        dry_run_result = do_current_luminance_factor_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                 ERE(pso("        Luminance Factors\n\n"));
            }

            ERE(do_current_luminance_factor_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_cumulative_luminance_report(void)
{
    int        dry_run_result; 

    if (fs_report_luminance_errors)
    {
        dry_run_result = do_cumulative_luminance_error_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                ERE(pso("        Error Predicting Luminance\n\n"));
            }

            ERE(do_cumulative_luminance_error_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_luminance_factors)
    {
        dry_run_result = do_cumulative_luminance_factor_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                ERE(pso("        Relative "));
                ERE(pso("Standard Deviations of Luminance Factors\n\n"));
            }

            ERE(do_cumulative_luminance_factor_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_grand_total_luminance_report(void)
{
    int        dry_run_result; 

    if (fs_report_luminance_errors)
    {
        dry_run_result = do_grand_total_luminance_error_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                ERE(pso("        Error Predicting Luminance\n\n"));
            }

            ERE(do_grand_total_luminance_error_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    if (fs_report_luminance_factors)
    {
        dry_run_result = do_grand_total_luminance_factor_report(DRY_RUN); 

        if (dry_run_result == HAVE_SOMETHING_TO_PRINT) 
        {
            if (fs_report_title) 
            {
                ERE(pso("        Relative "));
                ERE(pso("Standard Deviations of Luminance Factors\n\n"));
            }

            ERE(do_grand_total_luminance_factor_report(WET_RUN)); 

            if (fs_report_title) 
            {
                ERE(pso("\n\n"));
            }
        }
    }

    return NO_ERROR; 
}
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_input_data_error_report
(
    Input_data_error* error_one_by_sharp_ptr,
    Input_data_error* error_two_by_sharp_ptr,
    int               dry_run 
)
{
    int    i, j; 
    Gamut_input_data_error* error_one_ptr; 
    Gamut_input_data_error* error_two_ptr; 
    char title[ 100 ];

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "2D Absolute Pixel Error [ %d ]",
                            i + 1));
            ERE(print_one_error_line(title,  j,
                                     error_one_ptr->absolute_pixel_error, 
                                     error_two_ptr->absolute_pixel_error,
                                     dry_run));
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "3D Absolute Pixel Error [ %d ]",
                            i + 1));
            ERE(print_one_error_line(title,  j,
                                     error_one_ptr->absolute_pixel_error, 
                                     error_two_ptr->absolute_pixel_error,
                                     dry_run));
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "2D Relative Pixel Error [ %d ]",
                            i + 1));
            ERE(print_one_relative_error_line(title,  j,
                                              error_one_ptr->relative_pixel_error, 
                                              error_two_ptr->relative_pixel_error,
                                              dry_run));
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "3D Relative Pixel Error [ %d ]",
                            i + 1));
            ERE(print_one_relative_error_line(title,  j,
                                              error_one_ptr->relative_pixel_error, 
                                              error_two_ptr->relative_pixel_error,
                                              dry_run));
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_2D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "2D Relative Diagonal Error [ %d ]",
                            i + 1));
            ERE(print_one_relative_error_line(title,  j,
                                              error_one_ptr->diag_error, 
                                              error_two_ptr->diag_error,
                                              dry_run));
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].gamut_3D[ i ]);

            ERE(kjb_sprintf(title,sizeof(title), 
                            "3D Relative Diagonal Error [ %d ]",
                            i + 1));
            ERE(print_one_relative_error_line(title,  j,
                                              error_one_ptr->diag_error, 
                                              error_two_ptr->diag_error,
                                              dry_run));
        }
    }
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_error_report_by_algorithm
(
    Error_by_sharp* error_one_ptr,
    Error_by_sharp* error_two_ptr,
    int             dry_run 
)
{
    int i; 

    if (dry_run) 
    {
        /*
        // Recall our hack: As soon as the dry run finds something to print,
        // then ERROR is returned. Returning NO_ERROR indicates nothing to
        // print!
        */
        ERE(do_best_fit_error_report(error_one_ptr, error_two_ptr, dry_run));
        ERE(do_standard_error_report(error_one_ptr, error_two_ptr, dry_run));

        for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
        {
            ERE(do_2D_gamut_error_report(error_one_ptr, error_two_ptr, 
                                         i, dry_run));
        }

        for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
        {
            ERE(do_3D_gamut_error_report(error_one_ptr, error_two_ptr, 
                                         i, dry_run));
        }

        /* Lindsay: Feb 17/98 */
        ERE(do_correlation_error_report(error_one_ptr, error_two_ptr,dry_run));

#ifdef SIMPLE_TEMPLATE
        ERE(do_simple_template_error_report(error_one_ptr, error_two_ptr, 
                                            dry_run));
#endif 

#ifdef TEMPLATE
        ERE(do_template_error_report(error_one_ptr, error_two_ptr, dry_run));
#endif 
    
#ifdef MULTI_TEMPLATE
        ERE(do_multi_template_error_report(error_one_ptr, error_two_ptr, 
                                           dry_run));
#endif 
    }
    else 
    {
        int dry_run_res; 

        dry_run_res = do_best_fit_error_report(error_one_ptr, error_two_ptr,
                                               DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_best_fit_error_report(error_one_ptr, error_two_ptr, 
                                         dry_run));
        }

        dry_run_res = do_standard_error_report(error_one_ptr, error_two_ptr,
                                               DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_standard_error_report(error_one_ptr, error_two_ptr, 
                                         dry_run));
        }

        for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
        {
            dry_run_res = do_2D_gamut_error_report(error_one_ptr, error_two_ptr,
                                                   i, DRY_RUN);

            if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
            {
                ERE(do_2D_gamut_error_report(error_one_ptr, error_two_ptr, 
                                             i, dry_run));
            }
        }

        /* Lindsay: Feb 17/98 */
        dry_run_res = do_correlation_error_report(error_one_ptr, error_two_ptr,
                                               DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_correlation_error_report(error_one_ptr, error_two_ptr, 
                                         dry_run));
        }
        /* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
        dry_run_res = do_simple_template_error_report(error_one_ptr, 
                                                     error_two_ptr,
                                                     DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_simple_template_error_report(error_one_ptr, error_two_ptr, 
                                               dry_run));
        }
#endif 

#ifdef TEMPLATE
        dry_run_res = do_template_error_report(error_one_ptr, error_two_ptr,
                                               DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_template_error_report(error_one_ptr, error_two_ptr, 
                                         dry_run));
        }
#endif 

#ifdef MULTI_TEMPLATE
        dry_run_res = do_multi_template_error_report(error_one_ptr, 
                                                     error_two_ptr,
                                                     DRY_RUN);

        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_multi_template_error_report(error_one_ptr, error_two_ptr, 
                                               dry_run));
        }
#endif 

        for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
        {
            dry_run_res = do_3D_gamut_error_report(error_one_ptr, error_two_ptr,
                                                   i, DRY_RUN);

            if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
            {
                ERE(do_3D_gamut_error_report(error_one_ptr, error_two_ptr, 
                                             i, dry_run));
            }
        }
    }

    return NO_ERROR;
}
        
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_best_fit_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    Best_fit_error* error_one_ptr;
    Best_fit_error* error_two_ptr;
    int             i;

    if (fs_report_world_best_3D)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("World best 3D fit", i,
                                     error_one_ptr->world_best_3D_fit,
                                     error_two_ptr->world_best_3D_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_world_best_2D)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("World best 2D fit", i,
                                     error_one_ptr->world_best_2D_fit,
                                     error_two_ptr->world_best_2D_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_world_best_3D_diag)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("World best 3D diagonal fit", i,
                                     error_one_ptr->world_best_3D_diag_fit,
                                     error_two_ptr->world_best_3D_diag_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_world_best_2D_diag)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("World best 2D diagonal fit", i,
                                     error_one_ptr->world_best_2D_diag_fit,
                                     error_two_ptr->world_best_2D_diag_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_scene_best_3D)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("Scene best 3D fit", i,
                                     error_one_ptr->scene_best_3D_fit,
                                     error_two_ptr->scene_best_3D_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_scene_best_2D)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("Scene best 2D fit", i,
                                     error_one_ptr->scene_best_2D_fit,
                                     error_two_ptr->scene_best_2D_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_scene_best_3D_diag)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("Scene best 3D diagonal fit", i,
                                     error_one_ptr->scene_best_3D_diag_fit,
                                     error_two_ptr->scene_best_3D_diag_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_scene_best_2D_diag)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("Scene best 2D diagonal fit", i,
                                     error_one_ptr->scene_best_2D_diag_fit,
                                     error_two_ptr->scene_best_2D_diag_fit,
                                     dry_run)); 
        }
    }

    if (fs_report_actual)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].best_fit);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].best_fit);

            ERE(print_one_error_line("Actual illuminant", i,
                                     error_one_ptr->actual,
                                     error_two_ptr->actual,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_standard_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    Standard_error* error_one_ptr;
    Standard_error* error_two_ptr;
    int             i;

    if (fs_report_nothing)
    {
        error_one_ptr = &(error_one_by_sharp_ptr->sharp[DONT_SHARPEN].standard);
        error_two_ptr = &(error_two_by_sharp_ptr->sharp[DONT_SHARPEN].standard);

        ERE(print_one_error_line("Nothing", DONT_SHARPEN, 
                                 error_one_ptr->nothing, 
                                 error_two_ptr->nothing,
                                 dry_run));
    }

    if (fs_report_ave_illum)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].standard);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].standard);

            ERE(print_one_error_line("Average illumination", i, 
                                     error_one_ptr->ave_illum, 
                                     error_two_ptr->ave_illum,
                                     dry_run));
        }
    }

    if (fs_report_grey_world)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].standard);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].standard);

            ERE(print_one_error_line("Grey world", i, 
                                     error_one_ptr->grey_world, 
                                     error_two_ptr->grey_world,
                                     dry_run));
        }
    }

    if (fs_report_db_grey_world)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].standard);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].standard);

            ERE(print_one_error_line("DB Grey world", i,
                                     error_one_ptr->db_grey_world,
                                     error_two_ptr->db_grey_world,
                                     dry_run)); 
        }
    }

    if (fs_report_scale_by_max)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].standard);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].standard);

            ERE(print_one_error_line("Scale by max", i,
                                     error_one_ptr->scale_by_max,
                                     error_two_ptr->scale_by_max,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_2D_gamut_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{
    Gamut_error* error_one_ptr;
    Gamut_error* error_two_ptr;
    int          i;
    char         title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_2D_surface_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Max vol with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->mv_surface,
                                     error_two_ptr->mv_surface,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_illum_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Max vol with surfaces and illum [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->mv_surface_illum,
                                     error_two_ptr->mv_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Hull ave with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->ha_surface,
                                     error_two_ptr->ha_surface,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_illum_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Hull ave with surfaces and illum [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title,
                                     i,
                                     error_one_ptr->ha_surface_illum,
                                     error_two_ptr->ha_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_illum_icsa)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Illum constrained hull average [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->icsa_surface_illum,
                                     error_two_ptr->icsa_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_illum_scia)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_SCA_WEIGHT_FUNCTIONS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "2D Surface constrained illum average [ %d ]",
                                canon_gamut + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scia_surface_illum[ j ],
                                         error_two_ptr->scia_surface_illum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_2D_surface_illum_scca)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_SCA_WEIGHT_FUNCTIONS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "2D Surface constrained chrom average [ %d ]",
                                canon_gamut + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scca_surface_illum[ j ],
                                         error_two_ptr->scca_surface_illum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_2D_surface_weighted_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Weighted ave with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title,
                                     i, error_one_ptr->wa_surface,
                                     error_two_ptr->wa_surface,
                                     dry_run)); 
        }
    }

    if (fs_report_2D_surface_illum_weighted_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_2D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "2D Weighted ave with S and I [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->wa_surface_illum,
                                     error_two_ptr->wa_surface_illum,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_gamut_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{

    if (dry_run) 
    {
        /*
        // Recall our hack: As soon as the dry run finds something to print,
        // then ERROR is returned. Returning NO_ERROR indicates nothing to
        // print!
        */
        ERE(do_3D_gamut_surface_error_report(error_one_by_sharp_ptr, 
                                             error_two_by_sharp_ptr, 
                                             canon_gamut, dry_run));
        ERE(do_3D_gamut_surface_lum_error_report(error_one_by_sharp_ptr, 
                                                 error_two_by_sharp_ptr, 
                                                 canon_gamut, dry_run));
        ERE(do_3D_gamut_surface_illum_error_report(error_one_by_sharp_ptr, 
                                                   error_two_by_sharp_ptr, 
                                                   canon_gamut, dry_run));
    }
    else 
    {
        int dry_run_res; 

        dry_run_res = do_3D_gamut_surface_error_report(error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr,
                                                       canon_gamut, DRY_RUN); 
        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_3D_gamut_surface_error_report(error_one_by_sharp_ptr, 
                                                 error_two_by_sharp_ptr, 
                                                 canon_gamut, dry_run));
        }

        dry_run_res = do_3D_gamut_surface_lum_error_report(
                                                       error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr,
                                                       canon_gamut, DRY_RUN); 
        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_3D_gamut_surface_lum_error_report(error_one_by_sharp_ptr, 
                                                     error_two_by_sharp_ptr, 
                                                     canon_gamut, dry_run));
        }

        dry_run_res = do_3D_gamut_surface_illum_error_report(
                                                       error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr,
                                                       canon_gamut, DRY_RUN); 
        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_3D_gamut_surface_illum_error_report(error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr, 
                                                       canon_gamut, dry_run));
        }

        dry_run_res = do_3D_gamut_surface_illum_lum_error_report(
                                                       error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr,
                                                       canon_gamut, DRY_RUN); 
        if (dry_run_res == HAVE_SOMETHING_TO_PRINT)
        {
            ERE(do_3D_gamut_surface_illum_lum_error_report(
                                                       error_one_by_sharp_ptr, 
                                                       error_two_by_sharp_ptr, 
                                                       canon_gamut, dry_run));
        }

    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_gamut_surface_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{
    Gamut_error* error_one_ptr;
    Gamut_error* error_two_ptr;
    int          i;
    char         title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_3D_surface_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Max vol with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->mv_surface,
                                     error_two_ptr->mv_surface,
                                     dry_run)); 
        }
    }

    if (fs_report_3D_surface_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Hull ave with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->ha_surface,
                                     error_two_ptr->ha_surface,
                                     dry_run)); 
        }
    }

    if (fs_report_3D_surface_weighted_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Weighted average with surfaces only [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title,i,
                                     error_one_ptr->wa_surface,
                                     error_two_ptr->wa_surface,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_gamut_surface_lum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{
    Gamut_error* error_one_ptr;
    Gamut_error* error_two_ptr;
    int          i;
    char         title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_3D_surface_lum_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D Ave with surfaces and lum [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->ha_surface_lum[ j ],
                                         error_two_ptr->ha_surface_lum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_3D_surface_lum_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D Max vol with surfaces and lum [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->mv_surface_lum[ j ],
                                         error_two_ptr->mv_surface_lum[ j ],
                                         dry_run)); 
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_gamut_surface_illum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{
    Gamut_error* error_one_ptr;
    Gamut_error* error_two_ptr;
    int          i;
    char         title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_3D_surface_illum_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Max vol with surfaces and illum [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i,
                                     error_one_ptr->mv_surface_illum,
                                     error_two_ptr->mv_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_3D_surface_illum_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Hull ave with surfaces and illum [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title,
                                     i, 
                                     error_one_ptr->ha_surface_illum,
                                     error_two_ptr->ha_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_3D_surface_illum_weighted_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Weighted ave with S and I [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i, 
                                     error_one_ptr->wa_surface_illum,
                                     error_two_ptr->wa_surface_illum,
                                     dry_run)); 
        }
    }

    if (fs_report_3D_surface_illum_icsa)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            ERE(kjb_sprintf(title, sizeof(title),
                            "3D Illum constrained hull average [ %d ]",
                            canon_gamut + 1));

            ERE(print_one_error_line(title, i, 
                        error_one_ptr->icsa_surface_illum,
                        error_two_ptr->icsa_surface_illum,
                        dry_run)); 
        }
    }

    if (fs_report_3D_surface_illum_scia)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_SCA_WEIGHT_FUNCTIONS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D Surface constrained illum ave [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scia_surface_illum[ j ],
                                         error_two_ptr->scia_surface_illum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_3D_surface_illum_scca)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_SCA_WEIGHT_FUNCTIONS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D Surface constrained chrom ave [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scca_surface_illum[ j ],
                                         error_two_ptr->scca_surface_illum[ j ],
                                         dry_run)); 
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_gamut_surface_illum_lum_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             canon_gamut,
    int             dry_run 
)
{
    Gamut_error* error_one_ptr;
    Gamut_error* error_two_ptr;
    int          i;
    char         title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_3D_sil_scla)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D Surface constrained illum and lum ave [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scla_surface_illum_lum[ j ],
                                         error_two_ptr->scla_surface_illum_lum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_3D_sil_lcsa)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D I and L constrained surface ave [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                  error_one_ptr->lcsa_surface_illum_lum[ j ],
                                  error_two_ptr->lcsa_surface_illum_lum[ j ],
                                  dry_run)); 
            }
        }
    }

    if (fs_report_3D_sil_hull_ave)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                               "3D Ave with surfaces, illum, and lum [ %d.%d ]",
                               canon_gamut+ 1, j + 1));

                ERE(print_one_error_line(title, i,
                                     error_one_ptr->ha_surface_illum_lum[ j ],
                                     error_two_ptr->ha_surface_illum_lum[ j ],
                                     dry_run)); 
            }
        }
    }

    if (fs_report_3D_sil_max_vol)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                           "3D Max vol with surfaces, illum, and lum [ %d.%d ]",
                            canon_gamut+ 1, j + 1));

                ERE(print_one_error_line(title, i,
                                    error_one_ptr->mv_surface_illum_lum[ j ],
                                    error_two_ptr->mv_surface_illum_lum[ j ],
                                    dry_run)); 
            }
        }
    }

    if (fs_report_3D_sil_scia)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D SIL scia [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scia_surface_illum_lum[ j ],
                                         error_two_ptr->scia_surface_illum_lum[ j ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_3D_sil_scca)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            int j;

            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].gamut_3D[ canon_gamut ]);

            for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
            {
                ERE(kjb_sprintf(title, sizeof(title),
                                "3D SIL scca [ %d.%d ]", 
                                canon_gamut + 1, j + 1));

                ERE(print_one_error_line(title, i,
                                         error_one_ptr->scca_surface_illum_lum[ j ],
                                         error_two_ptr->scca_surface_illum_lum[ j ],
                                         dry_run)); 
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay: Colour by Correlation - Sept 30/98 */
static int do_correlation_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    Correlation_error* error_one_ptr;
    Correlation_error* error_two_ptr;
    int                i, j;
    char               title[ 100 ];

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_correlation_01)
    {
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (0/1 matrix) [ %d ]", i+1));
                
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_01[ i ],
                                         error_two_ptr->correlation_01[ i ],
                                         dry_run));
            }
        }
    }

    if (fs_report_correlation_uniform_inv_dist_fill)
    {
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Inv. Distance, MAP) [ %d ]",
                                i+1));
                    
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MAP_inv_dist_fill[ i ],
                                         error_two_ptr->correlation_uniform_MAP_inv_dist_fill[ i ],
                                         dry_run)); 
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Inv. Distance, MLM) [ %d ]",
                                i+1));
                    
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MLM_inv_dist_fill[ i ],
                                         error_two_ptr->correlation_uniform_MLM_inv_dist_fill[ i ],
                                         dry_run)); 
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Inv. Distance, MMSE) [ %d ]",
                                i+1));
                    
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MMSE_inv_dist_fill[ i ],
                                         error_two_ptr->correlation_uniform_MMSE_inv_dist_fill[ i ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_correlation_uniform_gaussian_fill)
    {
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Gaussian Mask, MAP) [ %d ]",
                                i+1));
                
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MAP_gaussian_fill[ i ],
                                         error_two_ptr->correlation_uniform_MAP_gaussian_fill[ i ],
                                         dry_run)); 
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Gaussian Mask, MLM) [ %d ]",
                                i+1));
                
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MLM_gaussian_fill[ i ],
                                         error_two_ptr->correlation_uniform_MLM_gaussian_fill[ i ],
                                         dry_run)); 
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Gaussian Mask, MMSE) [ %d ]",
                                i+1));
                
                ERE(print_one_error_line(title, j,
                                         error_one_ptr->correlation_uniform_MMSE_gaussian_fill[ i ],
                                         error_two_ptr->correlation_uniform_MMSE_gaussian_fill[ i ],
                                         dry_run)); 
            }
        }
    }

    if (fs_report_correlation_uniform_fitted)
    {
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Fitted, MAP) [ %d ]", i+1));
            
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_uniform_MAP_fitted[ i ],
                                         error_two_ptr->correlation_uniform_MAP_fitted[ i ],
                                         dry_run));
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Fitted, MLM) [ %d ]", i+1));
            
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_uniform_MLM_fitted[ i ],
                                         error_two_ptr->correlation_uniform_MLM_fitted[ i ],
                                         dry_run));
            }
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Uniform - Fitted, MMSE) [ %d ]", i+1));
            
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_uniform_MMSE_fitted[ i ],
                                         error_two_ptr->correlation_uniform_MMSE_fitted[ i ],
                                         dry_run));
            }
        }
    }

    if (fs_report_correlation_gaussian)
    {
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Gaussian, MAP) [ %d ]", i+1));
                
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_gaussian_MAP[ i ],
                                         error_two_ptr->correlation_gaussian_MAP[ i ],
                                         dry_run)); 
            }

            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Gaussian, MLM) [ %d ]", i+1));
                
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_gaussian_MLM[ i ],
                                         error_two_ptr->correlation_gaussian_MLM[ i ],
                                         dry_run)); 
            }

            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].correlation);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].correlation);

                ERE(kjb_sprintf(title,sizeof(title),
                                "Correlation (Gaussian, MMSE) [ %d ]", i+1));
                
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->correlation_gaussian_MMSE[ i ],
                                         error_two_ptr->correlation_gaussian_MMSE[ i ],
                                         dry_run)); 
            }
        }
    }    

    return NO_ERROR;
}

/* End Lindsay: Sept 30/98 */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static int do_simple_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    int             i;
    Simple_template_error* error_one_ptr;
    Simple_template_error* error_two_ptr;

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_simple_template_alg)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].simple_template);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].simple_template);

            ERE(print_one_error_line("Simple template algorithm", i, 
                                     error_one_ptr->alg,
                                     error_two_ptr->alg,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static int do_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    int             i;
    Template_error* error_one_ptr;
    Template_error* error_two_ptr;

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_template_first_alg)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].template);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].template);

            ERE(print_one_error_line("Template first algorithm", i, 
                                     error_one_ptr->first_alg,
                                     error_two_ptr->first_alg,
                                     dry_run)); 
        }
    }

    if (fs_report_template_second_alg)
    {
        for (i=0; i<NUM_SHARP_METHODS; i++) 
        {
            error_one_ptr = &(error_one_by_sharp_ptr->sharp[ i ].template);
            error_two_ptr = &(error_two_by_sharp_ptr->sharp[ i ].template);

            ERE(print_one_error_line("Template second algorithm", i, 
                                     error_one_ptr->second_alg,
                                     error_two_ptr->second_alg,
                                     dry_run)); 
        }
    }

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE

static int do_multi_template_error_report
(
    Error_by_sharp* error_one_by_sharp_ptr,
    Error_by_sharp* error_two_by_sharp_ptr,
    int             dry_run 
)
{
    int                   i, j;
    Multi_template_error* error_one_ptr;
    Multi_template_error* error_two_ptr;

    if ((fs_report_separator) && (! dry_run))
    {
        ERE(rep_print(stdout, '-', 80));
        ERE(pso("\n")); 
    }

    if (fs_report_multi_template_alg)
    {
        for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
        {
            for (j=0; j<NUM_SHARP_METHODS; j++) 
            {
                char title[ 100 ];

                error_one_ptr = &(error_one_by_sharp_ptr->sharp[ j ].multi_template);
                error_two_ptr = &(error_two_by_sharp_ptr->sharp[ j ].multi_template);

                ERE(kjb_sprintf(title,sizeof(title), 
                                "Multi template algorithm [ %d ]",
                                i+1));
                ERE(print_one_error_line(title, j, 
                                         error_one_ptr->alg[ i ],
                                         error_two_ptr->alg[ i ],
                                         dry_run)); 
            }
        }
    }

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int print_one_error_line
(
    const char* algorithm,
    int         sharp_method,
    double      error_one,
    double      error_two,
    int         dry_run 
)
{

    if (error_one > DBL_HALF_MOST_NEGATIVE/2.0) 
    {
        char sharp_method_string[ 100 ];
        char heading[ 100 ];
        int len;
        int left_over;
        int room_left;
        int i; 

        /*
        // See coment above for more information about this hack. 
        */
        if (dry_run) return HAVE_SOMETHING_TO_PRINT; 

        BUFF_CPY(heading, algorithm);

        get_sharp_method_string(sharp_method, sharp_method_string, 
                                sizeof(sharp_method_string));

        BUFF_CAT(heading, " (");
        BUFF_CAT(heading, sharp_method_string);
        BUFF_CAT(heading, ") ");

        len = strlen(heading);
        left_over = 3 - len % 3;
        if (left_over == 3) left_over = 0;
        room_left = 60 - left_over - len; /* Divisible by 3 */

        ERE(pso(heading));
        ERE(print_blanks(stdout, left_over));

        for (i=0; i<room_left/3; i++)
        {
            ERE(pso(".  "));
        }

        if (IS_INTEGRAL_DBL(error_one)) 
        {
            ERE(pso("%4d     ", (int)error_one)); 
        }
        else 
        {
            ERE(pso("%9.4f", error_one));
        }

        if (error_two > DBL_HALF_MOST_NEGATIVE/2.0) 
        {
            if (IS_INTEGRAL_DBL(error_two))
            {
                ERE(pso(" %4d", (int)error_two)); 
            }
            else 
            {
                ERE(pso(" %9.4f", error_two));
            }
        }

        ERE(pso("\n"));
    }

    return NO_ERROR;
}
    
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int print_one_relative_error_line
(
    const char* algorithm,
    int         sharp_method,
    double      error_one,
    double      error_two,
    int         dry_run 
)
{

    if (error_one > DBL_HALF_MOST_NEGATIVE/2.0) 
    {
        char sharp_method_string[ 100 ];
        char heading[ 100 ];
        int len;
        int left_over;
        int room_left;
        int i; 

        /*
        // See coment above for more information about this hack. 
        */
        if (dry_run) return HAVE_SOMETHING_TO_PRINT; 

        BUFF_CPY(heading, algorithm);

        get_sharp_method_string(sharp_method, sharp_method_string, 
                                sizeof(sharp_method_string));

        BUFF_CAT(heading, " (");
        BUFF_CAT(heading, sharp_method_string);
        BUFF_CAT(heading, ") ");

        len = strlen(heading);
        left_over = 3 - len % 3;
        if (left_over == 3) left_over = 0;
        room_left = 60 - left_over - len; /* Divisible by 3 */

        ERE(pso(heading));
        ERE(print_blanks(stdout, left_over));

        for (i=0; i<room_left/3; i++)
        {
            ERE(pso(".  "));
        }

        if (IS_INTEGRAL_DBL(error_one))
        {
            ERE(pso("%4d     ", (int)error_one)); 
        }
        else 
        {
            ERE(pso("%8.2f%%", 100.0 * error_one));
        }

        if (error_two > DBL_HALF_MOST_NEGATIVE/2.0) 
        {
            if (IS_INTEGRAL_DBL(error_two))
            {
                ERE(pso(" %4d", (int)error_two)); 
            }
            else 
            {
                ERE(pso("%8.2f%%", 100.0 * error_two));
            }
        }

        ERE(pso("\n")); 
    }

    return NO_ERROR;
}
    
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int print_head
(
    const char* title,
    const char* error_stat_one_str,
    const char* error_stat_two_str 
)
{
    
    ERE(pso("     %-58s", title));
    
    if (error_stat_one_str != NULL)
    {
        ERE(pso("%-10s", error_stat_one_str));
    }
    else 
    {
        ERE(pso("          "));
        /*       0123456789      */
    }

    if (error_stat_two_str != NULL)
    {
        ERE(pso("%-10s", error_stat_two_str));
    }

    ERE(pso("\n\n")); 

    return NO_ERROR;
}

