
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "grey.h"
#include "illum_data.h"
#include "do_other_cc.h"

/* -------------------------------------------------------------------------- */

static int do_scale_by_max(Matrix* data_mp, Standard_result* result_ptr);
static int do_ave_illum(Standard_result* result_ptr);

/* -------------------------------------------------------------------------- */

int do_standard(Matrix* data_mp, Standard_result* standard_result_ptr)
{
    Vector* canon_illum_vp = NULL;

    ERE(do_grey_world(data_mp, standard_result_ptr));
    ERE(do_scale_by_max(data_mp, standard_result_ptr));
    ERE(do_ave_illum(standard_result_ptr));

    free_vector(canon_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_scale_by_max(Matrix* data_mp, Standard_result* result_ptr)
{
    Vector* max_vp = NULL;
    int     result; 

    result = get_max_matrix_col_elements(&max_vp, data_mp);

    if (result != ERROR)
    {
        result = get_map_to_canon_illum(&(result_ptr->scale_by_max_vp), max_vp);
    }

    free_vector(max_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_ave_illum(Standard_result* result_ptr)
{
    Vector* ave_illum_vp = NULL;
    Vector* canon_illum_vp = NULL;

    ERE(get_average_of_illum_gamut_data(&ave_illum_vp)); 
    ERE(ow_sharpen_vector_data(ave_illum_vp));
    ERE(get_sharp_canon_illum(&canon_illum_vp)); 

    ERE(divide_vectors(&(result_ptr->ave_illum_vp), canon_illum_vp, 
                       ave_illum_vp)); 

    free_vector(ave_illum_vp);
    free_vector(canon_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

