
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"     /* Only safe as first include in a ".c" file. */
#include "i/i_transform.h"

/* -------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                             ow_invert_image
 *                               
 *
 * Index: images, image inversion, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int ow_invert_image(KJB_image* ip)
{
    int   i, j, num_rows, num_cols;

    num_rows = ip->num_rows;
    num_cols = ip->num_cols; 

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            ip->pixels[i][j].r = 255.0 - ip->pixels[i][j].r;
            ip->pixels[i][j].g = 255.0 - ip->pixels[i][j].g;
            ip->pixels[i][j].b = 255.0 - ip->pixels[i][j].b;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             rotate_image_right
 *                               
 *
 * Index: images, image rotation, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int rotate_image_right(KJB_image** target_ipp, const KJB_image* ip)
{
    KJB_image* target_ip; 
    int   i, j, num_rows, num_cols;

    num_rows = ip->num_rows;
    num_cols = ip->num_cols; 

    ERE(get_target_image(target_ipp, ip->num_cols, ip->num_rows));
    target_ip = *target_ipp; 

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            target_ip->pixels[j][num_rows - 1 - i].r = ip->pixels[i][j].r;
            target_ip->pixels[j][num_rows - 1 - i].g = ip->pixels[i][j].g;
            target_ip->pixels[j][num_rows - 1 - i].b = ip->pixels[i][j].b;
            target_ip->pixels[j][num_rows - 1 - i].invalid = ip->pixels[i][j].invalid;
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             rotate_image_left
 *                               
 *
 * Index: images, image rotation, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int rotate_image_left(KJB_image** target_ipp, const KJB_image* ip)
{
    KJB_image* target_ip; 
    int   i, j, num_rows, num_cols;

    num_rows = ip->num_rows;
    num_cols = ip->num_cols; 

    ERE(get_target_image(target_ipp, ip->num_cols, ip->num_rows));
    target_ip = *target_ipp; 

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            target_ip->pixels[num_cols - 1 - j][i].r = ip->pixels[i][j].r;
            target_ip->pixels[num_cols - 1 - j][i].g = ip->pixels[i][j].g;
            target_ip->pixels[num_cols - 1 - j][i].b = ip->pixels[i][j].b;
            target_ip->pixels[num_cols - 1 - j][i].invalid = ip->pixels[i][j].invalid;
        }
    }

    return NO_ERROR; 
}

