
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"
#include "i/i_video.h"

/* -------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------- */

#define  MAX_NUM_CAPTURE_TRIES 5

/* -------------------------------------------------------------------------- */

static int fs_capture_frame_count = 1;

/* -------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------- */

/*
 * =============================================================================
 *                               set_video_options
 *
 *
 *
 * -----------------------------------------------------------------------------
 */

/*
// We have a set routine expecially for video so that this file is not
// automatically linked whenever kjb_set is used. Thus if these set commands are
// needed, then set_video_options has to be explicitly called.
*/

int set_video_options(const char* option, const char* value)
{ 
    char lc_option[ 100 ];
    int  temp_int;
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "capture-frame-count"))
         || (match_pattern(lc_option, "cfc"))
       )
    {
        if (value[ 0 ] == '\0')
        {
            ERE(pso("Capture frame count (cfc) is %d.\n", fs_capture_frame_count));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("capture-frame-count = %d\n", fs_capture_frame_count));
        }
        else
        {
            ERE(ss1pi(value, &temp_int));
            ERE(set_capture_frame_count(temp_int));     
        }
        result = NO_ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             set_capture_frame_count            
 *                               
 *
 *
 * -----------------------------------------------------------------------------
*/ 

/*
// We have to export this one because programs such as multi_capture use it. It
// is possible to get arround this, but since the video set routines are
// regarded as special (see comment for kjb_video_set()), it is easier to simply
// export this routine.
*/

int set_capture_frame_count(int new_value)
{
    if (new_value < 1)
    {
        set_error("Capture frame count (\"capture-frame-count\" must ");
        set_error("be greater than zero."); 
        return ERROR;
    }

    fs_capture_frame_count = new_value;
    
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_capture_frame_count(void)
{
    return fs_capture_frame_count;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int start_video_grabber(void)
{
    set_error("Video is not supported on this platform.");
    return ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int capture_image
(
    KJB_image**   dummy_ipp,
    Image_window* dummy_image_window_ptr 
)
{

    set_error("Video is not supported on this platform.");
    return ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int stop_video_grabber(void)
{
    set_error("Video is not supported on this platform.");
    return ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

