
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_ERROR_INCLUDED 
#define L_ERROR_INCLUDED

#include "l/l_def.h"

/*
// Lint complains about null clauses. Thus we give up and define two versions.
// One double one, and one for lint. The null clauses are not that critical,
// but they gaurd against a certain kind of error--unfortunately, I can't
// remember what that error is! Nonetheless, keep it for now. 
*/

#ifndef __lint

#define NOTE_ERROR() \
             {                                                         \
                  extern int kjb_debug_level;                          \
                                                                       \
                  if (kjb_debug_level > 0)                             \
                  {                                                    \
                      add_error("Error noted on line %d of %s.)",      \
                                __LINE__, __FILE__);                   \
                  }                                                    \
              }

#define ERE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ERE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return ERROR;                                        \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define ERN(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ERN on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return NULL;                                         \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define ER(x)         if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ER on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return;                                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define EPE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(EPE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define ETE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ETE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define EPETE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(EPETE on line %d of %s.)",           \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define BATCH_EPETE(x) \
                      if ((x) == ERROR)                                        \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(BATCH_EPETE on line %d of %s.)",     \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                                                                               \
                          if ( ! is_interactive())                             \
                          {                                                    \
                              kjb_exit(EXIT_FAILURE);                          \
                          }                                                    \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define NRE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NRE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return ERROR;                                        \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define NRN(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NRN on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return NULL;                                         \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#define NR(x)         if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NR on line %d of %s.)",              \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return;                                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define NPE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NPE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define NTE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NTE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#define NPETE(x)      if ((x) == NULL)                                         \
                      {                                                        \
                          extern int kjb_debug_level;                          \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NPETE on line %d of %s.)",           \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#else
/*
--------------------------------------------------------------------------------
|                       Definitions for "lint" only
|
|                                  ||   
|                                 \||/  
|                                  \/
*/

#define ERE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            return ERROR;                                        \
                      }

#define ERN(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            return NULL;                                         \
                      }

#define ER(x)         if ((x) == ERROR)                                        \
                      {                                                        \
                          return;                                              \
    /* LINT */        }

#define EPE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                      }

#define ETE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define EPETE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          kjb_print_error();                                   \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define BATCH_EPETE(x)    if ((x) == ERROR)                                    \
                          {                                                    \
                              kjb_print_error();                               \
    /* LINT */                kjb_exit(EXIT_FAILURE);                          \
                          }

#define NRE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          return ERROR;                                        \
    /* LINT */        }

#define NRN(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            return NULL;                                         \
                      }

#define NR(x)         if ((x) == NULL)                                         \
                      {                                                        \
                          return;                                              \
    /* LINT */        }

#define NPE(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                      }

#define NTE(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define NPETE(x)      if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }
/*
|                                  /\    
|                                 /||\  
|                                  ||
|
|                       Definitions for "lint" only
--------------------------------------------------------------------------------
*/

#endif 

void check_num_args(int, int, int, const char*);

#endif 

