
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_INT_LIST_INCLUDED
#define L_INT_LIST_INCLUDED

#include "l/l_def.h"

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                                   Int_vector 
 * 
 * Integer list type
 *
 * This type is the int_vector type for the KJB library
 *
 * See also:
 *    create_int_vector, free_int_vector, get_target_int_vector
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/

typedef struct Int_vector 
{   
    int  length;
    int* elements;
}   
Int_vector;

typedef struct Int_vector_vector
{   
    int          length;
    Int_vector** elements;
}   
Int_vector_vector;

typedef struct Indexed_int_vector_element
{
    int element;
    int index;
}
Indexed_int_vector_element;

typedef struct Indexed_int_vector 
{   
    int                         length;
    Indexed_int_vector_element* elements;
}   
Indexed_int_vector;

/* -------------------------------------------------------------------------- */

int get_zero_int_vector(Int_vector** mpp, int len);

int get_initialized_int_vector
(
    Int_vector** mpp,
    int          len,
    int          initial_value 
);

#ifdef TRACK_MEMORY_ALLOCATION
#   define create_int_vector(x)    debug_create_int_vector(x, __FILE__, __LINE__)

#   define get_target_int_vector(x,y)   \
                               debug_get_target_int_vector(x,y, __FILE__, __LINE__) 

    Int_vector* debug_create_int_vector
    (
        int         len,
        const char* file_name,
        int         line_number 
    );

    int debug_get_target_int_vector
    (
        Int_vector** target_mpp,
        int          length,
        const char*  file_name,
        int          line_number 
    );
#else 
    Int_vector* create_int_vector(int len);

    int get_target_int_vector(Int_vector** out_arg_mp, int length);

#endif

void                free_int_vector                   (Int_vector* vp);
Int_vector*         create_zero_int_vector            (int);
int                 ow_zero_int_vector                (Int_vector*);

int                 ow_set_int_vector                 
(
    Int_vector* ,
    int          
);
int                 ascend_sort_int_vector            (Int_vector* vp);
Indexed_int_vector* create_indexed_int_vector         (Int_vector* vp);

void                free_indexed_int_vector           
(
    Indexed_int_vector* vp 
);

int                 descend_sort_indexed_int_vector   
(
    Indexed_int_vector* vp 
);

Int_vector*         create_int_vector_copy            
(
    Int_vector* source_mp 
);

int                 copy_int_vector_vector            
(
    Int_vector_vector**      target_vvpp,
    const Int_vector_vector* source_vvp 
);

int                 copy_int_vector                   
(
    Int_vector**      target_mpp,
    const Int_vector* source_mp 
);

int                 copy_int_vector_section           
(
    Int_vector**      target_vpp,
    const Int_vector* source_vp,
    int               first_elem,
    int               length 
);

int                 read_int_vector                   
(
    Int_vector** vpp,
    const char*  file_name 
);

int                 fp_read_int_vector                
(
    Int_vector** vpp,
    FILE*        fp 
);

int                 write_col_int_vector              
(
    const Int_vector* vp,
    const char*       file_name 
);

int                 fp_write_col_int_vector           
(
    const Int_vector* vp,
    FILE*             fp 
);

int                 fp_write_col_int_vector_with_title
(
    const Int_vector* vp,
    FILE*             fp,
    const char*       title 
);

int                 write_row_int_vector              
(
    const Int_vector* vp,
    const char*       file_name 
);

int                 fp_write_row_int_vector           
(
    const Int_vector* vp,
    FILE*             fp 
);

int                 fp_write_row_int_vector_with_title
(
    const Int_vector* vp,
    FILE*             fp,
    const char*       title 
);

int                 get_target_int_vector_vector      
(
    Int_vector_vector** target_vvpp,
    int                 count 
);
Int_vector_vector*  create_int_vector_vector          (int count);

void                free_int_vector_vector            
(
    Int_vector_vector* vvp 
);

int                 concat_int_vectors                
(
    Int_vector**      vpp,
    int               num_vectors,
    const Int_vector* vectors[] 
);

int                 int_vector_linear_search          
(
    const Int_vector* vp,
    int               search_elem 
);

#endif

