
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"     /* Only safe as first include in a ".c" file. */ 

#include "l/l_sys_term.h" 
#include "l/l_string.h" 
#include "l/l_sys_str.h" 
#include "l/l_parse.h" 
#include "l/l_sys_io.h" 
#include "l/l_sys_scan.h" 
#include "l/l_sys_sig.h" 
#include "l/l_sys_time.h" 
#include "l/l_verbose.h" 

#ifdef UNIX
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <sys/param.h> 
#    include <pwd.h> 
#    include <fcntl.h> 
#    include <glob.h>
#endif 
#endif

#ifdef  SUN4 
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <sys/filio.h> 
#endif
#endif

#ifdef  HPUX
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <sys/ioctl.h> 
#endif 
#endif

#ifdef  LINUX
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <sys/ioctl.h> 
#    include <time.h>
#endif 
#endif

#ifdef  MAC_OSX
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <sys/ioctl.h> 
#endif 
#endif

/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#include <sys/stat.h> 
#endif 

#ifdef SUN4
    extern char* realpath(char*, char[ MAXPATHLEN ]);
#endif

#ifdef NeXT
    extern char* realpath(char*, char[ MAXPATHLEN ]);
#endif

#ifndef EXCHANGE_USR_AND_NET
#    define EXCHANGE_USR_AND_NET  FALSE
#endif 

#ifndef DISABLE_DIR_OPEN
#    define DISABLE_DIR_OPEN FALSE 
#endif 

static int   fs_uncompress_files     = TRUE;
static int   fs_exchange_usr_and_net = FALSE;
static int   fs_disable_dir_open     = FALSE;
static FILE* fs_stderr_shadow_fp     = NULL;
static FILE* fs_stdout_shadow_fp     = NULL;
static int   fs_enable_stdout        = TRUE;

#ifdef UNIX

/* 
 * Some extra stuff that we use which is defined in some sys/stats but not all
 * (NeXT for example). On those systems that I have deal with so far, the
 * building blocks have been defined defined, so we use them for now. 
*/

#ifndef S_IRWXU
#    define S_IRWXU         0000700        /* rwx, owner */
#endif

#ifndef S_ISSOCK
#    define S_ISSOCK(m)    (((m)&S_IFMT) == S_IFSOCK)
#endif

#ifndef S_ISFIFO
#    define S_ISFIFO(m)    (((m)&S_IFMT) == S_IFIFO)
#endif

#ifndef S_ISCHR
#    define S_ISCHR(mode)  (((mode) & (S_IFMT)) == (S_IFCHR))
#endif 

#ifndef S_ISDIR
#    define S_ISDIR(mode)  (((mode) & (S_IFMT)) == (S_IFDIR))
#endif 

#ifndef S_ISBLK
#    define S_ISBLK(mode)  (((mode) & (_S_IFMT)) == (_S_IFBLK))
#endif 

#ifndef S_ISREG
#    define S_ISREG(mode)  (((mode) & (_S_IFMT)) == (_S_IFREG))
#endif 

#endif    /* #ifdef UNIX  */

/* ------------------------------------------------------------------------- */

/*
// This must be at least the maximum number of file descriptors available.
*/
#define  MAX_NUM_NAMED_FILES      255

#define BUFF_EXPAND_PATH(x, y)    expand_path((x), (y), sizeof(x)) 
#define OPEN_FOR_WRITING(x)       (x != MODE_IS_READ)

/* ------------------------------------------------------------------------- */

typedef enum Mode
{
    MODE_IS_NOT_SET, 
    MODE_IS_READ, 
    MODE_IS_WRITE,
    MODE_IS_APPEND,
    MODE_IS_READ_PLUS, 
    MODE_IS_WRITE_PLUS, 
    MODE_IS_APPEND_PLUS
} 
Mode;

/* -------------------------------------------------------------------------- */

static FILE* kjb_fopen_2(const char* input_fd_name, const char* mode);

static FILE* kjb_fopen_3(const char* input_fd_name, const char* mode);

static int expand_path(char*, const char*, size_t);
static Path_type get_path_type_guts(struct stat* path_stats_ptr);

static int cache_file_name
(
    FILE*       fp,
    const char* expanded_path,
    const char* fd_name,
    const char* uncompressed_fd_name,
    const char* mode 
);

#ifdef TRACK_MEMORY_ALLOCATION
static void free_cached_file_names(void);
#endif

static int uncache_file_name(int file_des);
static int initialize_cache(void);
static Mode get_mode_for_cache(const char* mode);

static void get_mode_for_error_message
(
    char*  mode,
    char*  buff,
    size_t len 
);

static int kjb_isatty_guts(int file_des);

static int process_file_open_mode
(
    const char* mode,
    char*       system_dependent_mode,
    size_t      system_dependent_mode_max_len,
    char*       unix_mode,
    size_t      unix_mode_max_len 
);

static void close_debug_file(void);

/* ------------------------------------------------------------------------- */

static char* fs_cached_fnames                [ MAX_NUM_NAMED_FILES ];
static char* fs_cached_user_fnames           [ MAX_NUM_NAMED_FILES ];
static char* fs_cached_uncompressed_fnames   [ MAX_NUM_NAMED_FILES ];

/*
// The first element of the following does double duty as an invalidation
// indicator. 
*/
static int   fs_cached_file_reference_count  [ MAX_NUM_NAMED_FILES ] = { NOT_SET };

static FILE* fs_cached_fp                    [ MAX_NUM_NAMED_FILES ];
static Mode  fs_cached_mode                  [ MAX_NUM_NAMED_FILES ];
static int   fs_cached_isatty                [ MAX_NUM_NAMED_FILES ];

#ifdef OLD_SYSV_NON_BLOCKING
    /*
    // If we have to deal with systems where return from a non-blocking device
    // is indestinguishable from EOF, then we may want to resurect keeping
    // track of what we are blocking on. If the need arises, #define
    // OLD_SYSV_NON_BLOCKING in the header file (l_sys_io.h). This will implement
    // a is_blocking() routine for devices whose blocking is modified only
    // through set_blocking() and set_no_blocking(). 
    */
    static int fs_cached_is_blocking[ MAX_NUM_NAMED_FILES ];
#endif

/* ------------------------------------------------------------------------- */

/*
// Not sure if we want to have fs_debug_fp and pdo only for debugging. Leave 
// available to all code for now. 
*/
static FILE* fs_debug_fp = NULL;

/* -------------------------------------------------------------------------- */

int set_low_level_io_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_boolean_value;
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "exchange-usr-and-net"))
         || (match_pattern(lc_option, "exchange-net-and-usr"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("/usr/local-<x> and /net/local-<x> are treated as %s.\n",
                    fs_exchange_usr_and_net ? "equavalent" : "different" ));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_exchange_usr_and_net = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "uncompress-files"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Files %s (temporarily) uncompressed on open.\n",
                    fs_uncompress_files ? "are" : "are not" ));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_uncompress_files = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "disable-dir-open"))
         || (match_pattern(lc_option, "disable-directory-open"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Use of directories for file names %s an error.\n", 
                    fs_disable_dir_open ? "is" : "is not" ));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_disable_dir_open = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                                kjb_read_exact
 *
 * Reads a specifed number of bytes from a file descriptor. 
 *                                                            
 * This routine is very similar to kjb_read, execpt that exactly "len" bytes
 * must be successfully read for a successfull return. Otherwise, ERROR is
 * returned, with an approapriate error measure being set. 
 *                                                            
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_fread_exact returns the number of bytes read. On failure
 *    ERROR is returned and an error message is set. Other kjb_fread returns
 *    such as EOF will not occur. 
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_read_exact(int file_des, void* buff, size_t len)
{
    Size_result read_res;

    ERE(read_res = kjb_read(file_des, buff, len));

    if (read_res > 0) return read_res; 

    if (read_res == WOULD_BLOCK) 
    {
        set_error("Read from %D failed because it would block.", file_des);
    }
    else if (read_res == INTERRUPTED) 
    {
        set_error("Read from %D failed because it was interrupted.", file_des);
    }
    else if (read_res == EOF) 
    {
        set_error("Unexpected end of file reading %D.", file_des);
    }
    else 
    {  
        SET_CANT_HAPPEN_BUG();
    }

    return ERROR; 
} 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_read
 *
 * Reads up to specified number of bytes from a device
 *
 * Similer to system read except that error handling is simplified.   We
 * inherit the standard limitation of being only able to read INT_MAX bytes.
 * 
 * ==> Remove this limitation in the case of 16 bit OS's if we ever work 
 * ==> on them again. (Just need an extra read). 
 *  
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_read returns the number of bytes read. On end of file. EOF
 *    is returned. On failure ERROR is returned and an error message is set.
 *    If the I/O is non blocking, then WOULD_BLOCK may also be returned. 
 *    Finally, depending on the type of signal handling in place, INTERRUPTED
 *    may also be returned. 
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_read
(
    int des,           /* File descriptor */
    void* buff,        /* Buffer for read */
    size_t num_bytes   /* Number of bytes to read */ 
)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int errno;
#endif 
    Size_result result;

    /*
    // FIX
    //
    // This routine could be easily made to read up to 64k on 16 bit systems
    // with an extra read. When this is done, the following can be
    // conditionally on SIZE_T_IS_32_BITS, with INT32_MAX replacing INT_MAX. 
    */
    if (num_bytes > INT_MAX)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (num_bytes == 0) 
    {
        return 0;
    }

    result = read(des, (void*)buff, (size_t)num_bytes);

    if (result > 0) 
    {
        return result;
    }
    else if (result == 0) 
    {
        return EOF;
    }
    else if (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM)
    {
        return WOULD_BLOCK;
    }
    else if (errno == EINTR)
    {
        return INTERRUPTED;
    }
    else 
    {
        set_error("Read failed.%S");
        return ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void broken_pipe_trap_fn(TRAP_FN_DUMMY_ARGS)
{
    dbp("In broken_pipe_trap_fn.\n");

    TEST_PSO(("In broken_pipe_trap_fn.\n"));
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               safe_pipe_write
 *                               
 * Writes to pipe with trap for SIGPIPE
 *
 * This routine writes to a pipe, trapping SIGPIPE so that if raised, error is
 * returned. Otherwise, this routine is the same as kjb_write. This routine can
 * be safely called with a descriptor for any type of file, but if the type of
 * file is known not to be a pipe, then there is nothing to be gained over
 * kjb_write. 
 *
 * Returns: 
 *    On success safe_pipe_write returns the number of bytes written. On
 *    failure ERROR is returned and an error message is set.
 *
 * Index: output, I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int safe_pipe_write
(
     int des,           /* File descriptor */
     const void* buff,  /* Location of bytes to write.  */
     size_t num_bytes   /* Number of bytes to write. */    
) 
{
    Signal_info save_broken_pipe_vec;
    Signal_info new_broken_pipe_vec;
    int         result;
    Error_action save_error_action; 
    
    INIT_SIGNAL_INFO(new_broken_pipe_vec); 
    new_broken_pipe_vec.SIGNAL_HANDLER = broken_pipe_trap_fn; 
    kjb_sigvec(SIGPIPE, &new_broken_pipe_vec, &save_broken_pipe_vec); 
 
    verbose_pso(300, (const char*)buff); 

    result = kjb_write(des, buff, num_bytes);

    save_error_action = get_error_action();

    if (result == ERROR) 
    {
        set_error_action(FORCE_ADD_ERROR_ON_ERROR);
    }

    kjb_sigvec(SIGPIPE, &save_broken_pipe_vec, (Signal_info*)NULL);
    
    set_error_action(save_error_action); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_write
 *
 * Writes a specifed number of bytes to a device 
 *
 * Similer to system write except that error handling is simplified. We 
 * inherit the standard limitation of being only able to write INT_MAX bytes.
 *  
 * ==> Remove this limitation in the case of 16 bit OS's if we ever work 
 * ==> on them again. (Just need extra writes). 
 *  
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_write returns the number of bytes written.   On failure
 *    ERROR is returned and an error message is set.
 *
 * Index: output, I/O, paging
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_write
    (
         int des           /* File descriptor */,
         const void* buff  /* Location of bytes to write.  */,
         size_t num_bytes  /* Number of bytes to write. */    
    ) 
{
    Size_result result;

    /*
    // FIX
    //
    // This routine could be easily made to read up to 64k on 16 bit systems
    // with an extra read. When this is done, the following can be
    // conditionally on SIZE_T_IS_32_BITS, with INT32_MAX replacing INT_MAX. 
    */
    if (num_bytes > INT_MAX)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (num_bytes == 0) 
    {
        return 0;
    }

    result = write(des, buff, num_bytes);

    if (result == (Size_result)num_bytes) 
    {
        return result;
    }
    else 
    {
#ifdef TEST
        set_error("Write of %ld bytes returned %ld.%S.", (long)num_bytes, 
                  (long)result);
#else
        set_error("Write failed.%S");
#endif

        return ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                                kjb_fread_exact
 *
 * Reads a specifed number of bytes from a stream. 
 *                                                            
 * This routine is very similar to kjb_fread, execpt that exactly "len" bytes
 * must be successfully read for a successfull return. Otherwise, ERROR is
 * returned, with an approapriate error measure being set. 
 *                                                            
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_fread_exact returns the number of bytes read. On failure
 *    ERROR is returned and an error message is set. Other kjb_fread returns
 *    such as EOF will not occur. 
 *
 * Macros:
 *    FIELD_READ
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_fread_exact(FILE* fp, void* buff, size_t len)
{
    Size_result read_res;

    ERE(read_res = kjb_fread(fp, buff, len));

    if (read_res > 0) return read_res; 

    if (read_res == WOULD_BLOCK) 
    {
        set_error("Read from %F failed because it would block.", fp);
    }
    else if (read_res == INTERRUPTED) 
    {
        set_error("Read from %F failed because it was interrupted.", fp);
    }
    else if (read_res == EOF)
    {
        set_error("Unexpected end of file reading file %F.", fp);
    }
    else 
    {  
        SET_CANT_HAPPEN_BUG();
    }

    return ERROR; 
} 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                                kjb_fread
 *
 * Reads a specifed number of bytes from a stream. 
 *                                                            
 * This is essentially a front end for fread, but there are some differences.
 * First, the arguments are different. The file pointer is first, and the size
 * arguments are combined. Second, the error reporting is more consistent with
 * the rest of the library. 
 *                                                            
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_fread returns the number of bytes read. On end of file. EOF
 *    is returned. On failure ERROR is returned and an error message is set.
 *    If the I/O is non blocking, then WOULD_BLOCK may also be returned. 
 *    Finally, depending on the type of signal handling in place, INTERRUPTED
 *    may also be returned. 
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_fread(FILE* fp, void* buff, size_t len)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int          errno;
#endif 
    IMPORT volatile int num_term_lines;
    IMPORT volatile int pause_on_next;
    IMPORT volatile int pause_on_next;
    size_t              read_res;

    clearerr(fp); 

    if (kjb_isatty(fileno(fp))) 
    {
        pause_on_next = FALSE; 
        num_term_lines = 0; 
        num_term_chars = 0; 
    }

#ifdef MAX_IO_RESULT_IO_SIZE
    if (len > MAX_IO_RESULT_IO_SIZE)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if (len == 0) return 0; 

    read_res = fread((void*)buff, (size_t)1, (size_t)len, fp); 

    if (read_res > 0) 
    {
        return read_res; 
    }
    else if (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM) 
    {
        return WOULD_BLOCK;
    }
    else if (errno == EINTR) 
    {
        return INTERRUPTED;
    }
    else if (feof(fp)) 
    {
        return EOF; 
    }
    else 
    {  
        set_error("Read from %F failed.%S", fp);
        return ERROR; 
    }
} 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                             kjb_fwrite
 *                                                            
 * Writes a specified number of bytes to a stream
 *                                                          
 * This routine is similar to fwrite (but note the difference in arguments!). 
 * However, it pages the output if fp points to a terminal. In addition, the
 * error handling is made more consistant with the rest of the KJB library. 
 *                                                          
 * Index: output, I/O, paging
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    On success kjb_write returns the number of bytes written.   On failure
 *    ERROR is returned and an error message is set.
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_fwrite
    (
         FILE* fp          /* File pointer */, 
         const void* line  /* Pointer to bytes to write */, 
         size_t len        /* Number of bytes to write. */
    ) 
{
    IMPORT volatile int halt_all_output; 
    IMPORT volatile int halt_term_output; 
    size_t              fwrite_res;  
    int result = NO_ERROR; 

    if ((len == 0) || halt_all_output) return 0; 

    if (fp == stderr) 
    {
        ERE(kjb_fflush((FILE*)NULL));
    }

    if ((fp != stdout) || (fs_enable_stdout))
    {
        if (kjb_isatty(fileno(fp))) 
        {
            if (halt_term_output) return 0; 
            result = term_put_n_chars((const char *)line, len); 
        }
        else 
        {
            fwrite_res = fwrite((const void*)line,(size_t)1,
                                (size_t)len, fp); 

            if (fwrite_res == 0) 
            {
                set_error("Write to %F failed.%S", fp);
                result = ERROR; 
            }
            else 
            {
                result = fwrite_res; 
            }
        }
    }

    if ((fp == stdout) && (fs_stdout_shadow_fp != NULL))
    {
        fwrite_res = fwrite((const void*)line,(size_t)1,
                            (size_t)len, fs_stdout_shadow_fp); 

        if (fwrite_res == 0) 
        {
            if (result == ERROR) 
            {
                set_error("Write to %F also failed.%S", fs_stdout_shadow_fp);
                result = ERROR; 
            }
            else 
            {
                set_error("Write to %F failed.%S", fs_stdout_shadow_fp);
            }
        }
    }
   
    if ((fp == stderr) && (fs_stderr_shadow_fp != NULL))
    {
        fwrite_res = fwrite((const void*)line,(size_t)1,
                            (size_t)len, fs_stderr_shadow_fp); 

        if (fwrite_res == 0) 
        {
            if (result == ERROR) 
            {
                set_error("Write to %F also failed.%S", fs_stderr_shadow_fp);
                result = ERROR; 
            }
            else 
            {
                set_error("Write to %F failed.%S", fs_stderr_shadow_fp);
            }
        }
    }
   
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int kjb_safe_fflush
    (
         FILE* fp  /* File pointer, or NULL to flush all open files.*/
    )
{
    /*
    File static
        int fs_cached_file_reference_count[ MAX_NUM_NAMED_FILES ];
    */

    if (fs_cached_file_reference_count[ 0 ] != NOT_SET)
    {
        return kjb_fflush(fp);
    }
    else 
    {
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_fflush
 *                               
 * KJB library replacement for fflush. 
 *                               
 * This routine is an anolog of the system routine fflush. It implements
 * additional error reporting and also implements taking a NULL argument to
 * flush all open files. This is available on many systems, but here we
 * implement our own symantics. Specifically we flush stderr LAST. 
 * 
 * Warning: 
 *    It is critical than any file pointer passed to kjb_fflush was obtained
 *    from this library using kjb_fopen and friends. 
 *
 * Index: output, I/O, paging
 *
 * -----------------------------------------------------------------------------
*/ 

/*
// Even though many systems implement fflush(NULL), we want to impose our own
// symantics on this. This is because fflush(NULL) does not specify the order
// of the flushing, and we want stderr to be last (for sure!). 
*/
#define IMPLEMENT_FLUSH_ALL

int kjb_fflush
    (
         FILE* fp /* File pointer, or NULL to flush all open files. */
    )
{

    ERE(initialize_cache()); 

    if (fp == NULL)
    {
#ifdef IMPLEMENT_FLUSH_ALL
        int i;
        int error_flag = FALSE;

        /*
        // We do stderr last, stdout second last. Thus do the others first. 
        */

        for (i=3; i<MAX_NUM_NAMED_FILES; i++)
        {
            if (    (fs_cached_fp[ i ] != NULL)
                 && (OPEN_FOR_WRITING(fs_cached_mode[ i ]))
               )
            {
                if (fflush(fs_cached_fp[ i ]) == -1)
                {
                    if ( ! error_flag)
                    {
                        kjb_clear_error();
                        error_flag = TRUE;
                    }
                    add_error("Flush of %F failed.%S", fs_cached_fp[ i ]);
                }
            }
        }

        for (i=0; i<3; i++)
        {
            if (    (fs_cached_fp[ i ] != NULL)
                 && (OPEN_FOR_WRITING(fs_cached_mode[ i ]))
               )
            {
                if (fflush(fs_cached_fp[ i ]) == -1)
                {
                    if ( ! error_flag)
                    {
                        kjb_clear_error();
                        error_flag = TRUE;
                    }
                    add_error("Flush of %F failed.%S", fs_cached_fp[ i ]);
                }
            }
        }

        if (error_flag) return ERROR;
#else
        if (fflush(fp) == -1)
        {
            set_error("Attempt to flush all open files failed.%S");
            return ERROR;
        }
#endif
        return NO_ERROR;
    }
    else if (fflush(fp) == -1) 
    {
        set_error("Flush of %F failed.%S", fp);
        return ERROR;
    }
    else if (    (fp == stdout)
              && (fs_stdout_shadow_fp != NULL)
              && (fflush(fs_stdout_shadow_fp) == -1)
            )
    {
        set_error("Flush of %F failed.%S", fs_stdout_shadow_fp);
        return ERROR;
    }
    else if (    (fp == stderr)
              && (fs_stderr_shadow_fp != NULL)
              && (fflush(fs_stderr_shadow_fp) == -1)
            )
    {
        set_error("Flush of %F failed.%S", fs_stderr_shadow_fp);
        return ERROR;
    }
    else 
    {
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_ioctl
 * 
 * Front end for ioctl
 *
 * The function kjb_ioctl is very similar to the standard ioctl, except that
 * error reporting is made consistent with the other library routines, and the
 * error messages are more meaningfull. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    the result of the underlying ioctl call is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

#ifdef UNIX

int kjb_ioctl(int file_des, int request, void* arg)
{
    int result;

#ifdef LINUX_386
    /*
    // The red hat include file contradicts the manual. Hopefully the include
    // file is correct. Of course, currently int and long uint are the same
    // thing, so likely there is no double problem.
    */
    result = ioctl(file_des, (ulong)request, (caddr_t)arg);
#else 
    result = ioctl(file_des, request, (caddr_t)arg);
#endif 

    if (result == -1) 
    {
        set_error("Control of device failed.%S");
        return ERROR;
    }
    else 
    {
        return result;
    }
}

#else     /* Matches #ifdef UNIX .... */
/* -----------------------------------------------------------------------------
|                            Non UNIX code                     
|                                  ||   
|                                 \||/  
|                                  \/
*/

/*ARGSUSED*/
int kjb_ioctl(int file_des, int  request, void* arg) 
{

    set_bug("kjb_ioctl is not implemented on this system.");
    return ERROR;
}

#endif  /* #ifdef UNIX .... #else ....                            */
        /* (Ends handling of system dependent natures of "ioctl") */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef UNIX 

/*
 * =============================================================================
 *                              set_blocking
 *
 * Sets a device into blocking mode on subsequent reads.
 *
 * Note:
 *    Often better to use "poll". This routine is intended to be used where poll
 *    is not quite the right thing. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * See also:
 *    term_set_blocking, term_set_no_blocking
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
 */

int set_blocking(int des)
{
#ifdef OLD_SYSV_NON_BLOCKING
    extern int fs_cached_is_blocking[ MAX_NUM_NAMED_FILES ]; 
#endif
    int        ioctl_arg; 

    if ((des < 0) || (des >= MAX_NUM_NAMED_FILES))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

#ifdef OLD_SYSV_NON_BLOCKING
    ERE(initialize_cache()); 
#endif

#ifdef BSD
    ioctl_arg = 0;   
    ERE(kjb_ioctl(des, (int)FIONBIO, (void*)&ioctl_arg)); 
#else
    if ((ioctl_arg = fcntl(des, F_GETFL, 0)) < 0)
    {
        set_error("Fctl F_GETFL failed.%S");
        return ERROR;
    }

    ioctl_arg &= ~O_NONBLOCK;     /* Posix     */

    /* 
    // OLD SYSV way. Best avoided if possible. 
    // 
    //       ioctl_arg &= ~O_NDELAY; 
    */

    if (fcntl(des, F_SETFL, ioctl_arg) < 0)
    {
        set_error("Fctl F_SETFL failed.%S");
        return ERROR;
    }
#endif

#ifdef OLD_SYSV_NON_BLOCKING
    fs_cached_is_blocking[ des ] = TRUE; 
#endif

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef UNIX
/*
 * =============================================================================
 *                                set_no_blocking
 *
 * Sets a device into non blocking mode on subsequent reads.
 *
 * Note:
 *    Often better to use "poll". This routine is intended to be used where poll
 *    is not quite the right thing. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

int set_no_blocking(int des)
{
#ifdef OLD_SYSV_NON_BLOCKING
    extern int fs_cached_is_blocking[ MAX_NUM_NAMED_FILES ]; 
#endif
    int        ioctl_arg; 

    if ((des < 0) || (des >= MAX_NUM_NAMED_FILES))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

#ifdef OLD_SYSV_NON_BLOCKING
    ERE(initialize_cache()); 
#endif

#ifdef BSD
    ioctl_arg = 1;   
    ERE(kjb_ioctl(des, (int)FIONBIO, (void*) & ioctl_arg));
#else
    if ((ioctl_arg = fcntl(des, F_GETFL, 0)) < 0)
    {
        set_error("Fctl F_GETFL failed.%S");
        return ERROR;
    }

    /*
    // Using both breaks HPUX and who knows what else! 
    */
    ioctl_arg |= O_NONBLOCK;     /* Posix     */
#ifdef DONT_USE_EXCEPT_ON_SPECIFIC_SYSTEM
    ioctl_arg |= O_NDELAY;       /* NOT Posix */
#endif 

    if (fcntl(des, F_SETFL, ioctl_arg) < 0)
    {
        set_error("Fctl F_SETFL failed.%S");
        return ERROR;
    }
#endif

#ifdef OLD_SYSV_NON_BLOCKING
    fs_cached_is_blocking[ des ] = FALSE; 
#endif

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef OLD_SYSV_NON_BLOCKING

/*
 * =============================================================================
 *                                is_blocking
 *
 * Returns effect of last set_blocking or set_no_blocking
 *
 * This routine returns the effect of the last call to set_blocking or
 * set_no_blocking. It does not attempt to determine whether the device
 * actually will block on reads (although this is possible on some/all systems,
 * we wish to avoid this overhead, and the trouble required to code the
 * capability). If all setting and unsetting of blocking is done through
 * set_blocking and set_no_blocking, then the result can be trusted. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    is_blocking returns either TRUE or FALSE. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

int is_blocking(des)
    int des;
{
    extern int fs_cached_is_blocking[ MAX_NUM_NAMED_FILES ]; 

    if ((des < 0) || (des >= MAX_NUM_NAMED_FILES))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(initialize_cache()); 

    return (fs_cached_is_blocking[ des ]);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                                fget_line
 *                                                            
 * Reads a line from a stream into a null terminated string.
 *                                                          
 * This routine is similar to fgets(3C).  It reads a null terminated string
 * representing one line from the file/device represented by the parameter "fp"
 * into the buffer pointed to by the parameter "line". The new line character
 * is consumed but not returned. At most max_len characters are placed into
 * "line" intcluding the NULL. If there are more than max_len-1 characters on
 * the line, the extra ones are purged, and a truncated version of the line is
 * left in the buffer and TRUNCATED is returned. The maximum number of
 * characters that can be read is also limited by an internal buffer. If the
 * number of characters exceed the size of the internal buffer, then TRUNCATED
 * is also returned. Since fget_line is intended for reading text, extremely
 * long lines are not expected. The exact limit is given by GET_LINE_BUFF_SIZE
 * which is defined in l_sys_io.h. An internal buffer is needed to get the proper
 * behaviour on non-blocking reads. 
 *
 * If the device corresponding to fp has been set non-blocking then this
 * routine will return WOULD_BLOCK until an entire line can be returned.
 *
 * Similarly, if EOF is read before the new line is reached, then EOF is
 * returned until an entire line can be read. Thus if a separate process is
 * writing to a file, and this routine is used to read that file, then a loop
 * on EOF will read the file by lines, which can simplify parsing. 
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Note:
 *    The behaviour on non-blocking and EOF assumes that all calls in a
 *    sequence trying to read a given line are made with the same file pointer
 *    (there is only one buffer). This is only relavent for situations where
 *    one is actively waiting for additional input to arrive (due to some other
 *    process). 
 *
 * Note:
 *    This routine is not designed for performance. The assumption is that 
 *    line at a time reads correspond to small amounts of I/O. 
 *
 * Returns:
 *    On success fget_line returns the number of bytes placed in the buffer,
 *    excluding the NULL.  Since the new line chracter is consumed but not
 *    copied, zero is returned for null lines. Alternate return values are all
 *    negative. In the case of end of file, EOF returned. If the line was
 *    truncated, then TRUNCATED is returned. Depending on the signal traps and
 *    options in place, INTERRUPTED is an additional possible return value. In
 *    the case of non-blocking reads WOULD_BLOCK is returned unless a complete
 *    line can be returned. ERROR is returned for unexpected problems and an
 *    error message is set. 
 *
 * Macros:
 *    BUFF_FGET_LINE(FILE fp, char line[]) 
 *  
 *    The max_len parameter is set to sizeof(line) with the appropriate cast.
 *    Using sizeof to set the buffer size is recomended where applicable, as
 *    the code will not be broken if the buffer size changes. HOWEVER, neither
 *    this method, nor the macro, is  applicable if line is NOT a character
 *    array.  If line is declared by "char* line", then the size of line is the
 *    number of bytes in a character pointer (usually 4), which is NOT what is
 *    normally intended.  You have been WARNED! 
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

/*
// Previously fget_line was simply defined in terms of dget_line as below. 
// However, mixing stream and non-stream IO is bad practice at best, and 
// in this case broke some code when it was ported to HPUX. Nonetheless, 
// fget_line and dget_line are intended to be very similar, and thus
// changes should be synched. 
//
// Old (bad) code. 
// 
// long fget_line(fp, line, max_len)
//   FILE* fp; 
//   char* line;
//   long max_len; 
//    {
// 
// 
//     return dget_line(fileno(fp), line, max_len);
//    }
// 
*/

Size_result fget_line(FILE* fp, char* line, size_t max_len)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int    errno;
#endif 
    static char*  buff_pos                   = NULL;
    static char   buff[ GET_LINE_BUFF_SIZE ];
    static int    line_truncated_flag        = FALSE;
    static size_t i                          = 0;
    Size_result   read_res                   = 0;

#ifdef SIZE_T_IS_32_BITS
    if (max_len > INT32_MAX)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if (max_len == 0) return 0; 

    max_len--;  

    if (max_len > sizeof(buff))
    {
        max_len = sizeof(buff);
    }

    if (buff_pos == NULL)
    { 
        buff_pos = buff;
        i=0;
        line_truncated_flag = FALSE; 
    }

    while (i < max_len) 
    {
        read_res = fread(buff_pos, (size_t)1, (size_t)1, fp);

        if (read_res > 0)
        {
            if (*buff_pos == '\n')
            {
                break;
            }
            else
            {
                buff_pos++;
                i++; 
            }
        }
        else if (read_res == 0)     /* Case EOF */
        {
#ifdef OLD_SYSV_NON_BLOCKING
            if (is_blocking(fileno(fp)))
            {
                return WOULD_BLOCK;
            }
#endif 
            /*
            // CHANGED  01/07/99 
            //
            // WAS 
            //    break; 
            */

            /*
            // CHANGED  again 14/11/00 
            //
            // WAS
            //    clearerr(fp); 
            //    return EOF; 
            //
            // Basically the question is what to do with the last chunk of text
            // in the file which may not have a line terminator. My current
            // feeling is to return it as though there was a line feed, and so
            // EOF is only trigured if the EOF is reached with no other input
            // having been read. 
            */

            if (i == 0) 
            {
                clearerr(fp); 
                return EOF; 
            }
            else 
            {
                break; 
            }
        }
#ifdef UNIX
        else if ((read_res == EOF) && 
                 (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM)) 
        {
            return WOULD_BLOCK; 
        }
#endif
        else if ((read_res == EOF) && (errno == EINTR))
        {
            return INTERRUPTED;
        }
        else 
        {
            set_error("Read of %F failed.%S\n", fp);
            buff_pos = NULL;
            return ERROR;
        }
    }

    /* 
    // Despite first appearances, "i" could be greater than max_len if max_len
    // changed between non-blocking reads and we fell through. (Not likely, but
    // possible).
    */

    if (i >= max_len)    
    {
        char read_buff[ 1 ];

        read_res = fread(read_buff, (size_t)1, (size_t)1, fp);

        if ((read_res > 0) && (read_buff[ 0 ] != '\n')) 
        {
            line_truncated_flag = TRUE; 
        }

        while ((read_res > 0) && (read_buff[ 0 ] != '\n')) 
        {
            read_res = fread(read_buff, (size_t)1, (size_t)1, fp);
        }

        if (read_res >= 0) 
        {
            /*EMPTY*/
            ; /* Do nothing */ 
        }
        else if (read_res != EOF)
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }
        else if (errno == EINTR)
        {
            return INTERRUPTED;
        }
        else if (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM)
        {
            return WOULD_BLOCK;
        }
        else 
        {
            set_error("Read failed on %F.%S", fp);
            buff_pos = NULL;
            return ERROR;
        }
    } 

    /*
    // "i" is at most max_len which is one less than what max_len started as.
    */

    /*
    //  CHANGED, 14/11/00
    //
    //  WAS
    //      kjb_strncpy(line, buff, i + ROOM_FOR_NULL);
    //
    //  Changed because this fools the trap in kjb_strncpy which checks for bad
    //  buffer sizes. Also, what do we want to do when there are NULLs in the
    //  file? The previous would not copy them. The current will copy them,
    //  but normally they would not get seen. It is a small point because the
    //  routine is meant for ASCII, but the action on bad data must be
    //  considered. 
    */

    kjb_memcpy(line, buff, i);

    /*
    // HACK to deal with DOS format. 
    */
    if ((i > 0) && (line[ i - 1 ] == '\r'))
    {
        i--;
    }

    line[ i ] = '\0'; 

    /*
    // Prepare for next read. Next time all statics will be reset. 
    */
    buff_pos = NULL;

    if (line_truncated_flag)
    {
        return TRUNCATED; 
    }
    else if (i > 0)
    {
        return i;
    }
    else if (read_res == 0)
    {
        return EOF;
    }
    else 
    {
        return 0;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
 * =============================================================================
 *                            dget_line
 *                                                            
 * Gets a line into a null terimated string from a device.
 *                                                          
 * The routine dget_line reads a null terminated string representing one line
 * from the device designated by the descriptor "des" into the buffer pointed to
 * by the parameter "line". The new line character is consumed but not returned.
 * At most max_len characters are placed into "line" including the NULL. If
 * there are more than max_len-1 characters on the line, then the extra are
 * purged, , and a truncated version of the line is left in the buffer and
 * TRUNCATED is returned.   The maximum number of characters that can be read is
 * also limited by an internal buffer. If the number of characters exceed the
 * size of the internal buffer, then TRUNCATED is also returned. Since dget_line
 * is intended for reading text, extremely long lines are not expected. The
 * exact limit is given by GET_LINE_BUFF_SIZE which is defined in l_sys_io.h. An
 * internal buffer is needed to get the proper behaviour on non-blocking reads. 
 *
 * If the device corresponding to des has been set non-blocking then this
 * routine will return WOULD_BLOCK until an entire line can be returned.
 *
 * Similarly, if EOF is read before the new line is reached, then EOF is
 * returned until an entire line can be read. Thus if a separate process is
 * writing to a file, and this routine is used to read that file, then a loop
 * on EOF will read the file by lines, which can simplify parsing. 
 *
 * Note:
 *    The behaviour on non-blocking and EOF assumes that all calls in a
 *    sequence trying to read a given line are made with the same file pointer
 *    (there is only one buffer). This is only relavent for situations where
 *    one is actively waiting for additional input to arrive (due to some other
 *    process). 
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Note:
 *    This routine is not designed for performance. The assumption is that 
 *    line at a time reads correspond to small amounts of I/O. 
 *
 * Returns:
 *    On success dget_line returns the number of bytes placed in the buffer.
 *    Since the new line chracter is consumed but not copied, the return may be
 *    zero for null lines. Alternate return values are all negative. In the
 *    case of end of file, EOF returned). If the line was truncated, then
 *    TRUNCATED is returned. Depending on the signal traps and options in
 *    place, INTERRUPTED is an additional possible return value. In the case of
 *    non-blocking reads WOULD_BLOCK is returned unless a complete line can be
 *    returned. ERROR is returned for unexpected problems and an error message
 *    is set. 
 *
 * Macros:
 *    BUFF_DGET_LINE(FILE fp, char line[]) 
 *  
 *    The max_len parameter is set to sizeof(line) with the appropriate cast.
 *    Using sizeof to set the buffer size is recomended where applicable, as
 *    the code will not be broken if the buffer size changes. HOWEVER, neither
 *    this method, nor the macro, is  applicable if line is NOT a character
 *    array.  If line is declared by "char* line", then the size of line is the
 *    number of bytes in a character pointer (usually 4), which is NOT what is
 *    normally intended.  You have been WARNED! 
 *
 * Index: input, I/O
 *
 * -----------------------------------------------------------------------------
*/

/*
// Changes to dget_line should be synchronized with analogous changes in 
// fget_line (if applicable). 
*/ 

Size_result dget_line(int des, char* line, size_t max_len)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int    errno;
#endif 
    static char*  buff_pos                   = NULL;
    static char   buff[ GET_LINE_BUFF_SIZE ];
    static int    line_truncated_flag        = FALSE;
    static size_t i                          = 0;
    Size_result   read_res                   = 0;
#ifdef I_DONT_THINK_THIS_IS_A_GOOD_IDEA
    int inside_double_quote = FALSE; 
#endif 

#ifdef SIZE_T_IS_32_BITS
    if (max_len > INT32_MAX)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if ((des < 0) || (des >= MAX_NUM_NAMED_FILES))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (max_len == 0) 
    {
        return 0; 
    }

    max_len--;  

    if (max_len > sizeof(buff))
    {
        max_len = sizeof(buff);
    }

    if (buff_pos == NULL)
    { 
        buff_pos = buff;
        i=0;
        line_truncated_flag = FALSE; 
    }

    while (i < max_len) 
    {
        read_res = read(des, buff_pos, (size_t)1);

        if (read_res > 0)
        {
#ifdef I_DONT_THINK_THIS_IS_A_GOOD_IDEA
            if (*buff_pos == '"')
            {
                if (inside_double_quote)
                {
                    inside_double_quote = FALSE;
                }
                else 
                {
                    inside_double_quote = TRUE; 
                }
            }
            else
#endif 
            if (    (*buff_pos == '\n') 
#ifdef I_DONT_THINK_THIS_IS_A_GOOD_IDEA
                 && ( ! inside_double_quote )
#endif 
               )
            {
                break;
            }

            buff_pos++;
            i++; 
        }
        else if (read_res == 0)     /* Case EOF */
        {
#ifdef OLD_SYSV_NON_BLOCKING
            if (is_blocking(des))
            {
                return WOULD_BLOCK;
            }
#endif 
            /*
            // CHANGED  01/07/99 
            //
            break; 
            */
            return EOF; 
        }
#ifdef UNIX
        else if ((read_res == EOF) && 
                 (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM)) 
        {
            return WOULD_BLOCK; 
        }
#endif
        else if ((read_res == EOF) && (errno == EINTR))
        {
            return INTERRUPTED;
        }
        else 
        {
            set_error("Read of %D failed.%S\n", des);
            buff_pos = NULL;
            return ERROR;
        }
    }

    /* 
    // Despited first appearances, "i" could be greater than max_len if max_len
    // changed between non-blocking reads and we fell through. (Not likely, but
    // possible).
    */

    if (i >= max_len)    
    {
        char read_buff[ 1 ];

        read_res = read(des, read_buff, (size_t)1);

        if ((read_res > 0) && (read_buff[ 0 ] != '\n')) 
        {
            line_truncated_flag = TRUE; 
        }

        while ((read_res > 0) && (read_buff[ 0 ] != '\n')) 
        {
            read_res = read(des, read_buff, (size_t)1);
        }

        if (read_res -= 0) 
        {
            return EOF;
        }
        else if (read_res > 0) 
        {
            /*EMPTY*/
            ; /* Do nothing */ 
        }
        else if (read_res != EOF)
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }
        else if (errno == EINTR)
        {
            return INTERRUPTED;
        }
        else if (errno == NON_BLOCKING_READ_FAILURE_ERROR_NUM)
        {
            return WOULD_BLOCK;
        }
        else 
        {
            set_error("Read failed on %D.%S", des);
            buff_pos = NULL;
            return ERROR;
        }
    } 

    /*
    // "i" is at most max_len which is one less than what max_len started as.
    */

    kjb_strncpy(line, buff, i + ROOM_FOR_NULL);

    /*
    // Prepare for next read. Next time all statics will be reset. 
    */
    buff_pos = NULL;

    if (line_truncated_flag)
    {
        return TRUNCATED; 
    }
    else if (i > 0)
    {
        return i;
    }
    else if (read_res == 0)
    {
        return EOF;
    }
    else 
    {
        return 0;
    }
}

/* 
 * =============================================================================
 *                                   fput_line
 *                                                            
 * Writes a line to a file.
 *                                                            
 * Fput_line writes a string to a stream. A new line character is then appended.
 *                                                          
 * Returns:
 *    On success, the number of characters written is returned. On failure, 
 *    an error is set, and ERROR is returned.
 *                                                          
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Index: I/O, output
 * 
 * -----------------------------------------------------------------------------
*/

Size_result fput_line
(
    FILE* fp          /* Stream, assumed open for write or append */ , 
    const char* line  /* Null terminated string to output */
) 
{
    int res; 
    int res2; 

    ERE(res = kjb_fputs(fp, line)); 
    ERE(res2 = kjb_fputs(fp, "\n")); 

    return (res + res2);  
} 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_mkdir
 *                               
 * KJB library version of mkdir(2). 
 *                               
 * Makes the directory pointed to by dir_path. Unlike mkdir(2), there is no
 * mode argument. For now, the directory will be made with the usual permission
 * (RWX everybody) & ~ (umask). (The permission stuff has only been checked on
 * a few systems).
 *
 * Note:
 *    Unlike mkdir(2) and kjb_mkdir_2(), this routine succeeds if the directory
 *    already exists. Also unlike mkdir(2) on many systems, this routine
 *    recursivley makes subdirectories. See kjb_mkdir_2 for a simple wrapper for
 *    mkdir(2). 
 * 
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/ 

/*
// FIX
//
// Implement the permission semantics described above. 
*/

int kjb_mkdir(const char* dir_path)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int errno;
#endif 
    int        mkdir_res; 
    char       sub_dir_path[ MAX_FILE_NAME_SIZE ];
    char       sub_dir[ MAX_FILE_NAME_SIZE ];
    const char*      dir_path_pos = dir_path; 
    int count = 0; 

    sub_dir_path[ 0 ] = '\0';

    while (*dir_path_pos == DIR_CHAR)
    {
        sub_dir_path[ count ] = *dir_path_pos; 
        count++;
        dir_path_pos++;
    }

    sub_dir_path[ count ] = '\0';

    while (BUFF_GEN_GET_TOKEN(&dir_path_pos, sub_dir, DIR_STR) > 0)
    {
        BUFF_CAT(sub_dir_path, sub_dir); 

#ifdef UNIX
#ifdef _MODE_T 
        mkdir_res = mkdir(sub_dir_path, (mode_t)(S_IRWXU | S_IRWXG | S_IRWXO));
#else
        mkdir_res = mkdir(sub_dir_path, (S_IRWXU | S_IRWXG | S_IRWXO));
#endif 
#else
#ifdef MS_16_BIT_OS
        mkdir_res = mkdir(sub_dir_path);
#endif
#endif 
        if ((mkdir_res == -1) && (errno != EEXIST))
        {
            set_error("Unable to make directory %s.%S", dir_path);
            return ERROR; 
        }

        BUFF_CAT(sub_dir_path, DIR_STR); 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_mkdir_2
 *                               
 * Another KJB library version of mkdir(2). 
 *                               
 * Makes the directory pointed to by dir_path. Unlike mkdir(2), there is no
 * mode argument. For now, the directory will be made with the usual permission
 * (RWX everybody) & ~ (umask). (The permission stuff has only been checked on
 * a few systems).
 *
 * Note:
 *    See kjb_mkdir for a similar routine which succeeds even if the directory
 *    exists, and makes needed subdirectories. 
 * 
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/ 

/*
// FIX
//
// Implement the permission semantics described above. 
*/

int kjb_mkdir_2(const char* dir_path)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int errno;
#endif 
    int        mkdir_res; 

#ifdef UNIX
#ifdef _MODE_T 
    mkdir_res = mkdir(dir_path, (mode_t)(S_IRWXU | S_IRWXG | S_IRWXO));
#else
    mkdir_res = mkdir(dir_path, (S_IRWXU | S_IRWXG | S_IRWXO));
#endif 
#else
#ifdef MS_16_BIT_OS
    mkdir_res = mkdir(dir_path);
#endif
#endif

    if (mkdir_res == -1) 
    {
        set_error("Unable to make directory %s.%S", dir_path);
        return ERROR; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_unlink
 * 
 * KJB library version of unlnk(2)
 *
 * The main point of this routine is that error reporting is made consistent
 * with the rest of the library. In addition, if file_name is NULL, or if it is
 * a null string (IE, *file_name == '\0'), then this is NO-OP, returning
 * success.
 *
 * Returns:
 *    ERROR on failure (with error message set), and NO_ERROR on success. 
 * 
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/ 

int kjb_unlink(const char* file_name)
{
    Error_action save_error_action;
    char         real_path[ MAX_FILE_NAME_SIZE ];

    if ((file_name == NULL) ||  (*file_name == '\0')) return NO_ERROR;

    save_error_action = get_error_action();
    set_error_action(IGNORE_ERROR_ON_ERROR); 

    if (BUFF_REAL_PATH(file_name, real_path) == ERROR)
    {
        BUFF_CPY(real_path, file_name);
    }

    set_error_action(save_error_action); 

    if (unlink(real_path) == EOF) 
    {
        TEST_PSO(("Unable to remove %s.%S\n", real_path));
        set_error("Unable to remove %s.%S\n", real_path);
        return ERROR; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_rmdir
 *                               
 * KJB library version of rmdir(2)
 * 
 * The main point of this routine is that error reporting is made consistent
 * with the rest of the library. 
 *
 * Returns:
 *    ERROR on failure (with error message set), and NO_ERROR on success. 
 * 
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/ 

int kjb_rmdir(const char* directory_name)
{

    if (rmdir(directory_name) == EOF) 
    {
        set_error("Unable to remove directory %s.%S", directory_name);
        return ERROR; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_fopen
 *                               
 * A front end for fopen
 *
 * The routine kjb_fopen should be used to open files that are used with the KJB
 * library.  It implements additional features such as more comprehensive
 * diagnostics, and sytem independent determination of file name from the file
 * pointer. It also implements acceptance of the "b" (for binary) mode
 * character.  (This is STILL not universally available). Another feature is a
 * routine print_open_files (development version only!) to print all open files.
 * An additional feature is that "~", and "~users"" is exanded. Of course, this
 * means that you cannot open a file that starts with a double ~. 
 *
 * If the user settable option "exchange-usr-and-net" is set, then when an open
 * on a file beginining with /usr/ which fails, it is tried again with /net/
 * instead and vice versa. 
 *
 * If the user settable option "disable-dir-open" is set, then an open attempt
 * on a directory failes.  
 *
 * If the user settable option "uncompress-files" is set, and the filenanme has
 * suffix ".gz" or ".Z", then the file is uncompressed into tempory space, and
 * the temporary file is opened instead.  The temporary file is removed on file
 * close. 
 *
 * Files opened with kjb_fopen MUST be closed with kjb_fclose. 
 *
 * Warning:
 *     Many KJB library routines requiring a file pointer will assume that this
 *     pointer came from one of kjb_fopen, kjb_freopen, kjb_popen, or
 *     kjb_fdopen.  Furthermore the library assumes that all files opened with
 *     one of these routines will be closed with kjb_fclose.   Furthermore, the
 *     KJB library file IO routines assume that stdin, stdout, and stderr will
 *     not be closed with fclose (kjb_fclose is OK). In short, if one makes use
 *     of the IO portion of the library, then kjb_fopen and friends should be
 *     used. 
 *
 *     Using file pointers obtained from kjb_fopen with non KJB library
 *     routines is less critical, but can lead to minor problems such as
 *     incorrect paging. Of course, when there is no non-KJB library routine
 *     for the given task, then there is no problem. 
 *
 * Returns:
 *     A regular FILE* handle on success, and NULL on failure. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/ 

FILE* kjb_fopen(const char* input_fd_name, const char* mode)
{
    FILE* fp; 

    fp = kjb_fopen_2(input_fd_name, mode);

    /*
    // We put this here, after the error message has been set, so that the error
    // message is for the first attempt.
    // 
    // Currenly we only do the exchange if the user has entered a complete path
    // as this is the only practical case. 
    */
    if (    (fp == NULL)
         && (fs_exchange_usr_and_net)
         && (input_fd_name[ 0 ] == '/'))
    {
        char exchange_path[ MAX_FILE_NAME_SIZE ];

        exchange_path[ 0 ] = '\0';

        if (HEAD_CMP_EQ(input_fd_name, "/usr/local-"))
        {
            BUFF_CPY(exchange_path, "/net/local-");
        }
        else if (HEAD_CMP_EQ(input_fd_name, "/net/local-"))
        {
            BUFF_CPY(exchange_path, "/usr/local-");
        }

        if (exchange_path[ 0 ] != '\0')
        {
            const char* path_pos = input_fd_name + 11; 

            BUFF_CAT(exchange_path, path_pos);

            fp = kjb_fopen_2(exchange_path, mode);

            if (fp == NULL) 
            {
                add_error("Also tried %q.", exchange_path); 
            }
        }
    }

    return fp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static FILE* kjb_fopen_2(const char* input_fd_name, const char* mode)
{
    FILE* fp; 
    char  system_dependent_mode[ 100 ];
    char  unix_mode[ 100 ];
    char gz_fd_name[ MAX_FILE_NAME_SIZE ];
    char Z_fd_name[ MAX_FILE_NAME_SIZE ];
    char error[ 5000 ]; 

    gz_fd_name[ 0 ] = '\0'; 
    Z_fd_name[ 0 ] = '\0'; 

    ERN(process_file_open_mode(mode, 
                               system_dependent_mode,
                               sizeof(system_dependent_mode),
                               unix_mode,
                               sizeof(unix_mode)));

    fp = kjb_fopen_3(input_fd_name, mode);

    /* Pacify error checkers. */
    error[ 0 ] = '\0'; 
    kjb_get_error(error, sizeof(error));

    if (    (fp == NULL) 
         && (fs_uncompress_files)
         && (get_mode_for_cache(unix_mode) == MODE_IS_READ)
       )  
    {
        BUFF_CPY(gz_fd_name, input_fd_name);
        BUFF_CAT(gz_fd_name, ".gz");

        fp = kjb_fopen_3(gz_fd_name, mode);
    }

    if (    (fp == NULL) 
         && (fs_uncompress_files)
         && (get_mode_for_cache(unix_mode) == MODE_IS_READ)
       )  
    {
        BUFF_CPY(Z_fd_name, input_fd_name);
        BUFF_CAT(Z_fd_name, ".Z");

        fp = kjb_fopen_3(Z_fd_name, mode);
    }

    if (fp == NULL)
    {
        str_set_error(error);

        if (gz_fd_name[ 0 ] != '\0') 
        {
            add_error("Also tried %s.", gz_fd_name); 
        }

        if (Z_fd_name[ 0 ] != '\0') 
        {
            add_error("Also tried %s.", Z_fd_name); 
        }
    }

    return fp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static FILE* kjb_fopen_3(const char* input_fd_name, const char* mode)
{
    FILE* fp = NULL;
    char  system_dependent_mode[ 100 ];
    char  unix_mode[ 100 ];
    char  expanded_path[ MAX_FILE_NAME_SIZE ];
    char  fd_name[ MAX_FILE_NAME_SIZE ];
    size_t        expanded_path_len;
    char          uncompressed_temp_name[ MAX_FILE_NAME_SIZE ];
    int result = NO_ERROR; 

    uncompressed_temp_name[ 0 ] = '\0';

    ERN(process_file_open_mode(mode, 
                               system_dependent_mode,
                               sizeof(system_dependent_mode),
                               unix_mode,
                               sizeof(unix_mode)));

    ERN(BUFF_EXPAND_PATH(expanded_path, input_fd_name)); 

    /* If we are doing something other than read, then there is no special
    // treatement at this stage.
    */
    if (get_mode_for_cache(unix_mode) != MODE_IS_READ)
    {
        BUFF_CPY(fd_name, expanded_path); 
    }
    /* We are reading someithing which exists. We will decompress it if the
    // appropriate option is set, and it has a compression suffix. Otherwise,
    // we will just read it. 
    */ 
    else if (get_path_type(expanded_path) == PATH_IS_REGULAR_FILE)
    {
        expanded_path_len = strlen(expanded_path);

        if (    (fs_uncompress_files) 
             && (expanded_path_len > 3)
             && (    (    (expanded_path[ expanded_path_len - 3 ] == '.')
                       && (expanded_path[ expanded_path_len - 2 ] == 'g')
                       && (expanded_path[ expanded_path_len - 1 ] == 'z')
                     )
                  ||
                     (    (expanded_path[ expanded_path_len - 2 ] == '.')
                       && (expanded_path[ expanded_path_len - 1 ] == 'Z')
                     )
                )
           )
        {
            char        command[ 20 + 2 * MAX_FILE_NAME_SIZE ];
            const char* suffix = (expanded_path[ expanded_path_len - 1 ] == 'z') 
                                                                    ? ".gz" : ".Z"; 
            char        compressed_temp_name[ MAX_FILE_NAME_SIZE ];

            compressed_temp_name[ 0 ] = '\0'; 

            BUFF_GET_TEMP_FILE_NAME(uncompressed_temp_name); 
            BUFF_CPY(compressed_temp_name, uncompressed_temp_name); 
            BUFF_CAT(compressed_temp_name, suffix); 

            ERN(kjb_sprintf(command, sizeof(command), "/bin/cp %s %s", 
                            expanded_path, compressed_temp_name));
            ERN(kjb_system(command)); 

            result = kjb_sprintf(command, sizeof(command), "gunzip %s",
                                 compressed_temp_name);

            if (result != ERROR) 
            {
                verbose_pso(5, "Uncompressing a copy of %s as %s.\n", 
                            expanded_path, uncompressed_temp_name);
                result = kjb_system(command);
            }

            if (result != ERROR)
            {
                verbose_pso(5, "Uncompression seems to have succeeded.\n"); 
                BUFF_CPY(fd_name, uncompressed_temp_name); 
            }
            else 
            {
                (void)kjb_unlink(compressed_temp_name);  /* Just in case. */
            }

        }
        else
        {
            BUFF_CPY(fd_name, expanded_path); 
        }
    }
    else 
    {
        BUFF_CPY(fd_name, expanded_path); 
    }

    if (result != ERROR)
    {
        fp = fopen(fd_name, system_dependent_mode); 
    }

    if (fp == NULL) 
    {
        char mode_for_error_message[ 100 ]; 

        get_mode_for_error_message(unix_mode, mode_for_error_message, 
                                   sizeof(mode_for_error_message));

        set_error("Unable to open file %q for %s.%S",
                  fd_name, mode_for_error_message); 
    }

    if (fp != NULL)
    {
        if (fs_disable_dir_open && (fp_get_path_type(fp) == PATH_IS_DIRECTORY))
        {
            set_error("%q is a directory.", expanded_path);
            (void)fclose(fp);
            fp = NULL;
        }
        else if (cache_file_name(fp, expanded_path, input_fd_name,
                                 uncompressed_temp_name, 
                                 unix_mode)
                 == ERROR)
        {
            (void)fclose(fp);
            fp = NULL;
        }
    }

    if (    (fp == NULL)
         && (uncompressed_temp_name[ 0 ] != '\0')
       )
    {
        kjb_unlink(uncompressed_temp_name); 
    }

    return fp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_freopen
 *                               
 * A replacement for freopen
 *
 * See kjb_fopen for details
 *
 * Returns:
 *     A regular FILE* handle on success, and NULL on failure. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/ 

FILE* kjb_freopen(const char* fd_name, const char* mode, FILE* stream)
{
    FILE* fp; 
    char  system_dependent_mode[ 100 ]; 
    char  unix_mode[ 100 ]; 
    char  expanded_path[ MAX_FILE_NAME_SIZE ]; 

    ERN(process_file_open_mode(mode, 
                               system_dependent_mode,
                               sizeof(system_dependent_mode),
                               unix_mode, 
                               sizeof(unix_mode)));

    ERN(BUFF_EXPAND_PATH(expanded_path, fd_name)); 

    fp = freopen(expanded_path, system_dependent_mode, stream); 

    if (fp == NULL) 
    {
        char mode_for_error_message[ 100 ]; 

        get_mode_for_error_message(unix_mode, mode_for_error_message, 
                                   sizeof(mode_for_error_message));

        set_error("Unable to open file %q for %s.%S",
                  fd_name, mode_for_error_message); 
    }
    else 
    {
        if (cache_file_name(fp, expanded_path, fd_name, (char*)NULL, unix_mode) == ERROR) 
        {
            fclose(fp);
            fp = NULL;
        }
    }

    return fp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_fdopen
 *                               
 * Replacement for fdopen
 *
 * See kjb_fopen for details.
 *
 * Returns:
 *     A regular FILE* handle on success, and NULL on failure. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/ 

/*ARGSUSED*/ 
FILE* kjb_fdopen(FILE* old_fp, const char* mode)
{
    FILE* new_fp;
    char  system_dependent_mode[ 100 ];
    char  unix_mode[ 100 ];
    int   old_file_des;
    int   new_file_des;

    ERN(initialize_cache()); 

    old_file_des = fileno(old_fp); 

    if (    (old_file_des < MAX_NUM_NAMED_FILES)
         && (fs_cached_mode[ old_file_des ] == MODE_IS_NOT_SET)
       )
    {
        set_bug(
              "kjb_fdopen sent a file descriptor not opened with KJB library.");
        return NULL;
    }

    ERN(process_file_open_mode(mode, 
                               system_dependent_mode,
                               sizeof(system_dependent_mode),
                               unix_mode, 
                               sizeof(unix_mode)));

    new_file_des = dup(fileno(old_fp)); 

    if (new_file_des == EOF) 
    {
        /*
         * FIX
         *
         * Should check for failure due to mismatch of modes, and call 
         * set_bug (not set_error) in this case. 
        */

        set_error("Unable to duplicate a file descriptor for %F.%S", old_fp);
        return NULL;
    }

    new_fp = fdopen(new_file_des, system_dependent_mode); 

    if (new_fp == NULL) 
    {
        set_error("Unable to open a stream for a duplicate descriptor of %F.%S",
                  old_fp);
        return NULL;
    }

    if (    (new_file_des < MAX_NUM_NAMED_FILES) 
         && (old_file_des < MAX_NUM_NAMED_FILES)
       )
    {
        fs_cached_fnames[ new_file_des ] = 
                                kjb_strdup(fs_cached_fnames[ old_file_des ]);

        fs_cached_user_fnames[ new_file_des ] = 
                                kjb_strdup(fs_cached_user_fnames[ old_file_des ]);

        fs_cached_file_reference_count[ new_file_des ] = 1;
        fs_cached_fp[ new_file_des ] = new_fp;

        /*
        // New mode will be "compatible" with the old, but I don't think it is
        // gaurenteed to be identical;
        */

        fs_cached_mode[ new_file_des ] = get_mode_for_cache(unix_mode);
    }

    return new_fp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_fclose
 *                               
 * Replacement for fclose
 *
 * This function should be used to close files opened with a KJB library routine
 * such as kjb_fopen. File pointers obtained elsewhere should NOT be sent to 
 * kjb_fclose. If kjb_fclose is sent a null file pointer, then this routine is a
 * successful no=op. 
 *
 * Returns:
 *    ERROR on failure (with error message set), and NO_ERROR on success. 
 * 
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int kjb_fclose(FILE* fp)
{
    int result; 
    int file_des;

    if (fp == NULL)
    {
        return NO_ERROR;
    }

    result = NO_ERROR;

    file_des = fileno(fp); 

    /*
    // If we need to close one of stdin, stdout, stderr, then we should make 
    // routines called kjb_fclose_stdin, etc, so the following error check can
    // remain.
    */
    if (file_des < 3)
    {
        set_bug("Attempt to close file with descriptor %d (%D).", file_des);
        return ERROR;
    }

    if (fclose(fp) == EOF) 
    {
        set_error("Close of file %F failed.%S", fp); 
        result = ERROR;
    }

    if (uncache_file_name(file_des) == ERROR)
    {
        result = ERROR;
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

/*
// Currently not used. 
*/

/* =============================================================================
 *                                kjb_popen
 *                               
 * A front end for popen
 *
 * The routine kjb_popen should be used in place of popen for streams that are
 * that are used with the KJB library.   kjb_popen implements additional
 * features such as more comprehensive diagnostics.
 *
 * The current implementation uses popen, so it shares the problem that the
 * open can succeed even if the execution of the command fails. 
 *
 * Streams opened with kjb_popen MUST be closed with kjb_pclose. 
 *
 * Warning:
 *     Many KJB library routines requiring a file pointer will assume that this
 *     pointer came from one of kjb_fopen, kjb_freopen, kjb_popen, or
 *     kjb_fdopen, Furthermore the library assumes that all files opened with
 *     one of these routines will be closed with kjb_fclose. Furthermore, the
 *     KJB library file IO routines assume that stdin, stdout, and stderr will
 *     not be closed with fclose (kjb_fclose is OK). In short, if one makes use
 *     of the IO portion of the library, then kjb_fopen and friends should be
 *     used. 
 *
 *     Using file pointers obtained from kjb_fopen with non KJB library
 *     routines is less critical, but can lead to minor problems such as
 *     incorrect paging. Of course, when there is no non-KJB library routine
 *     for the given task, then there is no problem. 
 *
 * -----------------------------------------------------------------------------
*/ 

FILE* kjb_popen(const char* cmd, const char* mode)
{
    FILE* fp; 
    char  system_dependent_mode[ 100 ]; 
    char  unix_mode[ 100 ]; 
    char  name_to_cache[ 1000 ]; 

    ERN(process_file_open_mode(mode, 
                               system_dependent_mode,
                               sizeof(system_dependent_mode),
                               unix_mode, 
                               sizeof(unix_mode)));

    BUFF_CPY(name_to_cache, "pipe to ");
    BUFF_CAT(name_to_cache, cmd);

    fp = popen(cmd, system_dependent_mode); 

    if (fp == NULL) 
    {
        char mode_for_error_message[ 100 ]; 

        get_mode_for_error_message(unix_mode, mode_for_error_message, 
                                   sizeof(mode_for_error_message));

        set_error("Unable to open pipe to %q with mode %q.%S",
                  cmd, mode_for_error_message); 
    }
    else 
    {
        if (cache_file_name(fp, (char*)NULL, name_to_cache, (char*)NULL, unix_mode) == ERROR) 
        {
            (void)fclose(fp);
            fp = NULL;
        }
    }

    return fp; 
}

#endif   /* TEST */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

/*
// Currently only used in test programs. 
*/

/* =============================================================================
 *                                kjb_pclose
 *                               
 * Replacement for pclose
 *
 * This function should be used to close streams opened with kjb_popen File
 * pointers obtained elsewhere should NOT be sent to kjb_pclose. 
 *
 * -----------------------------------------------------------------------------
*/ 

int kjb_pclose(FILE* fp)
{
    int result; 
    int file_des;

    result = NO_ERROR;

    file_des = fileno(fp); 

    if (pclose(fp) == EOF) 
    {
        set_error("Close of file %F failed.%S", fp); 
        result = ERROR;
    }

    if (uncache_file_name(file_des) == ERROR)
    {
        result = ERROR;
    }

    return result; 
}

#endif   /* TEST */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                      uncache_file_name
 * -----------------------------------------------------------------------------
*/ 

static int uncache_file_name(int file_des)
{

    if ((file_des < 0) || (file_des >= MAX_NUM_NAMED_FILES))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (fs_cached_file_reference_count[ 0 ] == NOT_SET)
    {
        set_bug("uncache_file_name called when cache was invalid.");
        return ERROR;
    }
    else if (fs_cached_mode[ file_des ] == MODE_IS_NOT_SET)
    {
        set_bug("uncache_file_name recieved fp not opened by KJB library");
        return ERROR; 
    }
    else
    {
        fs_cached_file_reference_count[ file_des ] --;
    }

    if (fs_cached_file_reference_count[ file_des ] == 0) 
    {
        kjb_free(fs_cached_fnames[ file_des ] );
        fs_cached_fnames[ file_des ] = NULL; 

        kjb_free(fs_cached_user_fnames[ file_des ] );
        fs_cached_user_fnames[ file_des ]=NULL; 

        if (fs_cached_uncompressed_fnames[ file_des ] != NULL)
        {
            EPE(kjb_unlink(fs_cached_uncompressed_fnames[ file_des ]));
            kjb_free(fs_cached_uncompressed_fnames[ file_des ] );
            fs_cached_uncompressed_fnames[ file_des ]=NULL; 
        }

        fs_cached_mode[ file_des ] = MODE_IS_NOT_SET;
        fs_cached_fp[ file_des ]   = NULL; 

        fs_cached_isatty[ file_des ] = NOT_SET;

#ifdef OLD_SYSV_NON_BLOCKING
        fs_cached_is_blocking[ file_des ] = FALSE;
#endif
    }
#ifdef TEST 
    else if (fs_cached_file_reference_count[ file_des ] > 0) 
    {
        kjb_fprintf(stderr, "Reference count for %d reduced to %d.\n", file_des,
                    fs_cached_file_reference_count[ file_des ]);
    }
    else 
    {
        kjb_fprintf(stderr, "Reference count for %d reduced to %d.\n", file_des,
                    fs_cached_file_reference_count[ file_des ]);

        kjb_fprintf(stderr, 
                    "Presumably an attempt to close the same stream twice.\n");

        kjb_abort();
    }
#endif

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                      get_mode_for_error_message
 * -----------------------------------------------------------------------------
*/ 

/*
// We assume that we are being sent a unix style mode. EG, the "b" designators
// have been removed where appropriate.
*/

static void get_mode_for_error_message(char* mode, char* buff, size_t len)
{

    if (STRCMP_EQ(mode, "r")) 
    {
        kjb_strncpy(buff, "reading", len);
    }
    else if (STRCMP_EQ(mode, "w")) 
    {
        kjb_strncpy(buff, "writing", len);
    }
    else if (STRCMP_EQ(mode, "w+")) 
    {
        kjb_strncpy(buff, "seek/write", len);
    }
    else if (STRCMP_EQ(mode, "r+")) 
    {
        kjb_strncpy(buff, "read/write", len);
    }
    else if (STRCMP_EQ(mode, "a+")) 
    {
        kjb_strncpy(buff, "read/expansion", len);
    }
    else if (STRCMP_EQ(mode, "a")) 
    {
        kjb_strncpy(buff, "expansion", len);
    }
    else 
    {
        kjb_strncpy(buff, mode, len);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                         cache_file_name
 * -----------------------------------------------------------------------------
*/ 

static int cache_file_name(FILE* fp, const char* expanded_path, const char* user_file_name, const char* uncompressed_fd_name, const char* mode)
{
    int          file_des;
#ifdef VMS
    char         buff[ 200 ];
    char*        buff_ptr;
#else
    char         real_path [ MAXPATHLEN ];
#endif

    if (fp == NULL) 
    {
        set_bug("Null fp passed to cache_file_name."); 
        return ERROR;
    } 

    file_des = fileno(fp); 

    if (file_des >= MAX_NUM_NAMED_FILES)
    {
        set_error("Number of open files cannot exceed %d.", 
                  MAX_NUM_NAMED_FILES);
        return ERROR;
    }

    ERE(initialize_cache()); 

#ifdef OLD_SYSV_NON_BLOCKING
    fs_cached_is_blocking[ file_des ] = TRUE;
#endif

    /*
     * It is possible to have more than one stream attached to 
     * an file descriptor, and in this case the reference
     * count will be greater than 0. This does not
     * happen in any code as of 10/02/96. 
    */

    if (fs_cached_file_reference_count[file_des] == 0) 
    {
        /*
         * If this is a re-open, then we are in some sense, 
         * "re-naming" the stream. The system has done the close
         * on the previous stream for us, and thus the cleanup code
         * in kjb_fclose has not been executed. If the strings are not
         * null for any other reason, then there is a mistake. (?)
        */
        if (fs_cached_fnames[ file_des ] != NULL) 
        {
            kjb_free( fs_cached_fnames[ file_des ] );
        }

        if (fs_cached_user_fnames[ file_des ] != NULL) 
        {
            kjb_free( fs_cached_user_fnames[ file_des ] );
        }

        if (kjb_isatty(fileno(fp))) 
        {
            fs_cached_fnames[ file_des ] = kjb_strdup("the computer terminal"); 
            fs_cached_user_fnames[ file_des ] = 
                kjb_strdup("the computer terminal");
            fs_cached_isatty[ file_des ] = TRUE; 
            fs_cached_uncompressed_fnames[ file_des ] = NULL; 
        }
        else 
        {
            fs_cached_isatty[ file_des ] = FALSE; 
#ifdef VMS 
            buff_ptr = getname(file_des, buff); 
            fs_cached_fnames[file_des ] = kjb_strdup(buff); 
#else 
            if (expanded_path != NULL)
            {
                ERE(BUFF_REAL_PATH(expanded_path, real_path)); 
                fs_cached_fnames[file_des ] = kjb_strdup(real_path); 
            }
            else 
            {
                fs_cached_fnames[file_des ] = kjb_strdup(user_file_name);
            }
#endif 
            fs_cached_user_fnames[file_des ] = kjb_strdup(user_file_name);

            if (    (uncompressed_fd_name != NULL)
                 && (uncompressed_fd_name[ 0 ] != '\0')
               )
            {
                fs_cached_uncompressed_fnames[ file_des ] = 
                                       kjb_strdup(uncompressed_fd_name); 
            }
        }

        fs_cached_fp[ file_des ] = fp;
        fs_cached_mode[ file_des ] = get_mode_for_cache(mode);

    }
#ifdef TEST
    else 
    {
        kjb_fprintf(stderr, 
                    "Duplicate descriptor (%d) noted in cache_file_name.\n", 
                    file_des);
        kjb_fprintf(stderr, "This occured while attempting to cache %s\n",
                    user_file_name);
    }
#endif

    fs_cached_file_reference_count[ file_des ] ++; 

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                        initialize_cache
 * -----------------------------------------------------------------------------
*/ 

/*
// Important note: We assume that if the functions in this file are being 
// used at all, then they are not being mixed with the standard ones in 
// devious ways. Specifically, we assume that stdin, stdout, and stderr, 
// have NOT been closed with fclose (kjb_fclose is OK).
*/

static int initialize_cache(void)
{
    static int   cache_is_ready = FALSE; 
    int          i; 

    if ( ! cache_is_ready)
    {
        fs_cached_file_reference_count[ 0 ] = 1;
        fs_cached_mode[ 0 ] = MODE_IS_READ;
        fs_cached_fp[ 0 ] = stdin; 
#ifdef OLD_SYSV_NON_BLOCKING
        fs_cached_is_blocking[ 0 ] = TRUE;
#endif

        /*
        // All allocated storage is tested for NULL together once all three
        // default file descriptors have been dealt with. 
        */

        if (kjb_isatty_guts(fileno(stdin)))
        {
            fs_cached_isatty[ 0 ]      = TRUE;
            fs_cached_fnames[ 0 ]      = kjb_strdup("the computer terminal");
            fs_cached_user_fnames[ 0 ] = kjb_strdup("the computer terminal");
        }
        else
        {
            fs_cached_isatty[ 0 ]      = FALSE;
            fs_cached_fnames[ 0 ]      = kjb_strdup("standard input");
            fs_cached_user_fnames[ 0 ] = kjb_strdup("standard input");
        }

        fs_cached_uncompressed_fnames[ 0 ] = '\0';

        fs_cached_file_reference_count[ 1 ] = 1;
        fs_cached_mode[ 1 ]                 = MODE_IS_WRITE;
        fs_cached_fp[ 1 ]                   = stdout; 
#ifdef OLD_SYSV_NON_BLOCKING
        fs_cached_is_blocking[ 1 ]          = TRUE;
#endif

        if (kjb_isatty_guts(fileno(stdout)))
        {
            fs_cached_isatty[ 1 ]      = TRUE;
            fs_cached_fnames[ 1 ]      = kjb_strdup("the computer terminal");
            fs_cached_user_fnames[ 1 ] = kjb_strdup("the computer terminal");
        }
        else
        {
            fs_cached_isatty[ 1 ]      = FALSE;
            fs_cached_fnames[ 1 ]      = kjb_strdup("standard output");
            fs_cached_user_fnames[ 1 ] = kjb_strdup("standard output");
        }

        fs_cached_uncompressed_fnames[ 1 ] = NULL;

        fs_cached_file_reference_count[ 2 ] = 1;
        fs_cached_mode[ 2 ]                 = MODE_IS_WRITE;
        fs_cached_fp[ 2 ]                   = stderr; 

#ifdef OLD_SYSV_NON_BLOCKING
        fs_cached_is_blocking[ 2 ]          = TRUE;
#endif

        if (kjb_isatty_guts(fileno(stderr)))
        {
            fs_cached_isatty[ 2 ]      = TRUE; 
            fs_cached_fnames[ 2 ]      = kjb_strdup("the computer terminal");
            fs_cached_user_fnames[ 2 ] = kjb_strdup("the computer terminal");
        }
        else
        {
            fs_cached_isatty[ 2 ]      = FALSE; 
            fs_cached_fnames[ 2 ]      = kjb_strdup("standard error");
            fs_cached_user_fnames[ 2 ] = kjb_strdup("standard error");
        }

        /*
        // This checks that we got memory when calling kjb_strdup.
        */
        if (    (fs_cached_fnames[ 0 ]      == NULL)
             || (fs_cached_user_fnames[ 0 ] == NULL)
             || (fs_cached_fnames[ 1 ]      == NULL)
             || (fs_cached_user_fnames[ 1 ] == NULL)
             || (fs_cached_fnames[ 2 ]      == NULL) 
             || (fs_cached_user_fnames[ 2 ] == NULL)
           )
        {
            return ERROR; 
        }

        for(i = 3; i<MAX_NUM_NAMED_FILES; i++) 
        {
            fs_cached_fnames[ i ]               = NULL;
            fs_cached_user_fnames[ i ]          = NULL;
            fs_cached_file_reference_count[ i ] = 0;
            fs_cached_mode[ i ]                 = MODE_IS_NOT_SET;
            fs_cached_fp[ i ]                   = NULL; 
            fs_cached_isatty[ i ]               = NOT_SET;
#ifdef OLD_SYSV_NON_BLOCKING
            fs_cached_is_blocking[ i ]          = TRUE;
#endif
            fs_cached_uncompressed_fnames[ i ] = NULL;
        }

        cache_is_ready = TRUE;

#ifdef TRACK_MEMORY_ALLOCATION
        add_cleanup_function(free_cached_file_names);
#endif
        return NO_ERROR; 
    }
    else if (fs_cached_file_reference_count[ 0 ] == NOT_SET) 
    {
        set_bug("File name cache has been invalidated.");
        return ERROR;
    }
    else 
    {
        return NO_ERROR; 
    } 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                         get_mode_for_cache
 * -----------------------------------------------------------------------------
*/ 

/*
// We assume that we are being sent a unix style mode. EG, the "b" designators
// have been removed where appropriate.
*/

static Mode get_mode_for_cache(const char* mode)
{

    if (STRCMP_EQ(mode, "r")) 
    {
        return MODE_IS_READ;
    }
    else if (STRCMP_EQ(mode, "w")) 
    {
        return MODE_IS_WRITE;
    }
    else if (STRCMP_EQ(mode, "w+")) 
    {
        return MODE_IS_WRITE_PLUS;
    }
    else if (STRCMP_EQ(mode, "r+")) 
    {
        return MODE_IS_READ_PLUS;
    }
    else if (STRCMP_EQ(mode, "a+")) 
    {
        return MODE_IS_APPEND_PLUS; 
    }
    else if (STRCMP_EQ(mode, "a")) 
    {
        return MODE_IS_APPEND; 
    }
    else 
    {
        return MODE_IS_NOT_SET;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                kjb_realpath
 *                               
 *
 *
 * -----------------------------------------------------------------------------
*/ 

/* ALL */   int kjb_realpath(const char* fd_name, char* path_buff, size_t max_len) 
/* ALL */                            
/* ALL */                              
/* ALL */                             
/* ALL */   {
#ifdef UNIX
/* UNIX */      char* path_res; 
/* UNIX */      char  resolved_path[ MAXPATHLEN ]; 
/* UNIX */      char  expanded_name[ MAX_FILE_NAME_SIZE ]; 
#ifdef SGI_NEXT
/* SGI_NEXT */  char  absolute_path[ MAX_FILE_NAME_SIZE ]; 
#endif
#endif

#ifdef MAX_SIZE_T_BUFF_SIZE
/* ALL */       if (max_len > MAX_SIZE_T_BUFF_SIZE)
/* ALL */       {
/* ALL */           SET_ARGUMENT_BUG();
/* ALL */           return ERROR;
/* ALL */       }
#endif 
/* ALL */       if ((max_len == 0) || (fd_name == NULL) || (path_buff == NULL))
/* ALL */       {
/* ALL */           SET_ARGUMENT_BUG();
/* ALL */           return ERROR;
/* ALL */       }
/* ALL */        
/* ALL */       *path_buff = '\0'; 
/* ALL */        
#ifdef UNIX
/*    
 * On some systems, EG AIX, realpath does not touch the result if there is
 * failure. However, since we want to use the result on other systems
 * even in the case of failure, we want to make sure that we don't have
 * garbage.  
 *
 * It should be noted that we are assuming that realpath either puts
 * something useful in resolved_path, or does not touch  it. A little
 * dangerous perhaps (FIX ?) 
 */

/* UNIX */      resolved_path[ 0 ] = '\0';
/* UNIX */
/* UNIX */      ERE(BUFF_EXPAND_PATH(expanded_name, fd_name));
/* UNIX */ 
#ifdef SGI_NEXT
/* Sgi's realpath is slow, but use it for now.                 */
/* Also, if requires READ access to all directories. Oh well.  */

/* SGI_NEXT */             
/* SGI_NEXT */  if (*expanded_name == '/') 
/* SGI_NEXT */  {
/* SGI_NEXT */      path_res = realpath(expanded_name, resolved_path);
/* SGI_NEXT */  }
/* SGI_NEXT */  else 
/* SGI_NEXT */  {
/* SGI_NEXT */      NRE(getwd(absolute_path)); 
/* SGI_NEXT */      BUFF_CAT(absolute_path, "/"); 
/* SGI_NEXT */      BUFF_CAT(absolute_path, expanded_name); 
/* SGI_NEXT */      path_res = realpath(absolute_path, resolved_path);
/* SGI_NEXT */  }
#else
/* UNIX */      path_res = realpath(expanded_name, resolved_path);
#endif
/* UNIX */       
/* UNIX */      /* We copy the resolved path on BOTH success and failure */ 
/* UNIX */      /* as even on failure it contains valuable information.  */ 
/* UNIX */      /* The system version of "realpath" contains the absolute */ 
/* UNIX */      /* path of the unresolved componant. Normally this is the */ 
/* UNIX */      /* path which x would have if x was created exactly now, */
/* UNIX */      /* assuming that x could be created now. */
/* UNIX */       
/* UNIX */      if (max_len < (strlen(resolved_path) + 1)) 
/* UNIX */      {
/* UNIX */          SET_BUFFER_OVERFLOW_BUG();
/* UNIX */          return ERROR;
/* UNIX */      }
/* UNIX */      else 
/* UNIX */      {
/* UNIX */          kjb_strncpy(path_buff, resolved_path, max_len);
/* UNIX */      }
/* UNIX */
/* UNIX */      if (path_res != NULL) 
/* UNIX */      {
/* UNIX */          return NO_ERROR; 
/* UNIX */      } 
/* UNIX */      else 
/* UNIX */      {
/* UNIX */          set_error("Unable to find %q.%S", fd_name); 
/* UNIX */          return ERROR;
/* UNIX */      }
#else  
/* default */   if (max_len < strlen(fd_name)+1) 
/* default */   {
/* default */       SET_BUFFER_OVERFLOW_BUG();
/* default */       return ERROR;
/* default */   }
/* default */   else 
/* default */   {
/* default */       kjb_strncpy(path_buff, fd_name, max_len);    
/* default */       return NO_ERROR;  
/* default */   }
#endif 
/* ALL */   }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                        expand_path
 * -----------------------------------------------------------------------------
*/ 

/* ALL */   static int expand_path(char* exanded_path, const char* path, size_t max_len) 
/* ALL */                                  
/* ALL */                          
/* ALL */                             
/* ALL */   {
#ifdef UNIX
/* UNIX */      const char*    path_pos;
/* UNIX */      struct passwd* pw;
/* UNIX */      char           logname_buff[ 200 ];
#endif
/* ALL */       size_t         exanded_path_len; 
/* ALL */
/* ALL */
#ifdef MAX_SIZE_T_BUFF_SIZE
/* ALL */       if (max_len > MAX_SIZE_T_BUFF_SIZE)
/* ALL */       {
/* ALL */           SET_ARGUMENT_BUG();
/* ALL */           return ERROR;
/* ALL */       }
#endif 
/* ALL */       if ((max_len == 0) || (path == NULL) || (exanded_path == NULL)) 
/* ALL */       {
/* ALL */           SET_ARGUMENT_BUG();
/* ALL */           return ERROR;
/* ALL */       }
/* ALL */
#ifdef UNIX
/* UNIX */      if (*path == '~') 
/* UNIX */      {
/* UNIX */          path_pos = path; 
/* UNIX */          path_pos++;
/* UNIX */
/* UNIX */          if ((*path_pos == '/') || (*path_pos == '\0')) 
/* UNIX */          {
/* UNIX */              ERE(BUFF_GET_USER_ID(logname_buff));
/* UNIX */          }
/* UNIX */          else 
/* UNIX */          {
/* UNIX */              BUFF_GEN_GET_TOKEN(&path_pos, logname_buff, "/");
/* UNIX */          }
/* UNIX */             
/* UNIX */          pw = getpwnam(logname_buff);
/* UNIX */  
/* UNIX */          if (pw == NULL) 
/* UNIX */          {
/* UNIX */              set_error("No user %q", logname_buff);
/* UNIX */              return ERROR;
/* UNIX */          }
/* UNIX */       
/* UNIX */          exanded_path_len = strlen(pw->pw_dir) +
/* UNIX */                                        strlen(path_pos) + 1;
/* UNIX */          if (max_len < exanded_path_len) 
/* UNIX */          {
/* UNIX */              SET_BUFFER_OVERFLOW_BUG();
/* UNIX */              return ERROR;
/* UNIX */          }
/* UNIX */       
/* UNIX */          kjb_strncpy(exanded_path, pw->pw_dir, max_len); 
/* UNIX */          kjb_strncat(exanded_path, path_pos, max_len);
/* UNIX */      }
/* UNIX */      else 
/* UNIX */      {
/* UNIX */          exanded_path_len = strlen(path) + 1;
/* UNIX */        
/* UNIX */          if (max_len < exanded_path_len) 
/* UNIX */          {
/* UNIX */              SET_BUFFER_OVERFLOW_BUG();
/* UNIX */              return ERROR;
/* UNIX */          }
/* UNIX */        
/* UNIX */          kjb_strncpy(exanded_path, path, max_len); 
/* UNIX */      }
/* UNIX */
/* UNIX */      return NO_ERROR; 
#else 
/* default */   exanded_path_len = strlen(path) + 1;
/* default */       
/* default */   if (max_len < exanded_path_len) 
/* default */   {
/* default */       SET_BUFFER_OVERFLOW_BUG();
/* default */       return ERROR;
/* default */   }
/* default */   else 
/* default */   {
/* default */       kjb_strncpy(exanded_path, path, max_len);    
/* default */       return NO_ERROR;  
/* default */   }
#endif 
/* ALL */   }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                get_fd_name
 * 
 * Gets the name of a file  or device associated with a stream
 *
 * The name of the file or device associated with descripter des is copied into
 * the buffer fd_name.  Max_len gives the size of the buffer which will not be
 * overflowed.   The name returned is the "real" name of the file, and thus
 * "~", "../", etc are expanded. Also, if the file was uncompressed behind the
 * scenes, then the uncompressed version is provided. The routine
 * get_user_fd_name can be used to find the name of the file as passed to
 * kjb_fopen. 
 *
 * Note:
 *   The file must have been opened with kjb_fopen.
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

int get_fd_name(int des, char* fd_name, size_t max_len)
{
    int result;

#ifdef MAX_SIZE_T_BUFF_SIZE
    if (max_len > MAX_SIZE_T_BUFF_SIZE)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if (max_len == 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(initialize_cache()); 

    result = NO_ERROR;

    /*
    // Have to be able to handle descriptors that did not come from kjb_fopen 
    // and friends, as it could have come from opening a pipe or other similar
    // source, and we do not have replacements for these opens. 
    */

    if (des >= MAX_NUM_NAMED_FILES)
    {
        kjb_strncpy(fd_name, "<unknown>",  max_len);
    }
    else if (fs_cached_uncompressed_fnames[ des ] != NULL)
    {
        kjb_strncpy(fd_name, fs_cached_uncompressed_fnames[ des ], max_len);
    }
    else if (fs_cached_fnames[ des ] != NULL)
    {
        kjb_strncpy(fd_name, fs_cached_fnames[ des ], max_len);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG(); 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                get_user_fd_name
 *
 * Gets the name of a file associated with a stream
 *
 * The name of the file or devicde associated with desciptor des is copied into
 * the buffer fd_name.  Max_len gives the size of the buffer which will not be
 * overflowed.  The name returned is the name as specified to kjb_fopen and
 * thus "~", "../", etc are not expanded. Get_fd_name can be used to find the
 * "double" file name. 
 *
 * Note:
 *   The file must have been opened with kjb_fopen.
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

int get_user_fd_name(int des, char* fd_name, size_t max_len)
{
    int result;

#ifdef MAX_SIZE_T_BUFF_SIZE
    if (max_len > MAX_SIZE_T_BUFF_SIZE)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if (max_len == 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(initialize_cache()); 

    result = NO_ERROR;

    /*
    // Have to be able to handle descriptors that did not come from kjb_fopen 
    // and friends, as it could have come from opening a pipe or other similar
    // source, and we do not have replacements for these opens. 
    */

    if ((des < MAX_NUM_NAMED_FILES) && (fs_cached_fnames[ des ] != NULL))
    {
        kjb_strncpy(fd_name, fs_cached_user_fnames[ des ], max_len);
    }
    else 
    {
        kjb_strncpy(fd_name, "<unknown>",  max_len);
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_fseek
 *
 * Moves the file pointer
 *
 * kjb_fseek is very similar to the standard fseek, except that error reporting
 * is made consistent with the other library routines, and the messages are 
 * more meaningfull. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

int kjb_fseek(FILE* fp, long int offset, int mode)
{
    int seek_res; 

    seek_res = fseek(fp, offset, mode);

    if (seek_res == 0) 
    {
        return NO_ERROR; 
    }
    else 
    {
        set_error("File positioning failed.%S");
        return ERROR; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_ftell
 *
 * Returns the file offset
 *
 * kjb_ftell is very similar to the standard ftell, except that error reporting
 * is made consistent with the other library routines, and the error messages
 * are more meaningfull. 
 *
 * Returns:
 *    On failure an error message is set, and ERROR is returned. Otherwise 
 *    NO_ERROR is retuned. 
 *
 * Index: I/O
 *
 * -----------------------------------------------------------------------------
*/

long kjb_ftell(FILE* fp)
{
    struct stat file_stats; 
    long        result; 

    if (kjb_isatty(fileno(fp))) 
    {
        set_error("Attempt to find file position of a terminal.");
        return ERROR;
    }

    if (fstat((int)fileno(fp), &file_stats) == -1) 
    {
#ifdef TEST
        set_error("Fstat system called failed for %F.%S", fp); 
#else
        set_error("Can't position %F.%S", fp); 
#endif
        return ERROR;
    }

#ifdef UNIX
    if (S_ISSOCK(file_stats.st_mode)) 
    {
        set_error("Attempt to find file position of a socket.");
        return ERROR;
    }
    if (S_ISFIFO(file_stats.st_mode)) 
    {
        set_error("Attempt to find file position of a pipe.");
        return ERROR;
    }
#else
#ifdef MS_16_BIT_OS
    if (file_stats.st_mode && S_IFCHR) 
    {
        set_error("Attempt to find file position of a device.");
        return ERROR;
    }
#endif
#endif

    result = ftell(fp);

    if (result == -1) 
    {
        set_error("Can't position %F.%S", fp);    
        return ERROR;
    }
    else 
    {
        return result;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_fputs
 *
 * Writes a null terminated string to a stream
 *
 * This is the KJB library version of fputs.   This routine works together with
 * the rest of the KJB library to implement paging and flushing of all files
 * when stderr is written to. 
 *
 * Warning:
 *    The that the arguments are REVERSED from fputs. In the KJB library 
 *    routines we also put the file pointer first. 
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    ERROR on failure (with error message set), and NO_ERROR on success. 
 * 
 * Index: output, I/O, paging
 *
 * -----------------------------------------------------------------------------
*/

Size_result kjb_fputs
(
     FILE* fp        /* File pointer */  , 
     const char* str /* String to write to file */ 
) 
{
    IMPORT volatile int halt_all_output;
    IMPORT volatile int halt_term_output;
    Size_result         fputs_res = NO_ERROR;

    if (halt_all_output) return NO_ERROR;

    if (fp == stderr) 
    {
        ERE(kjb_fflush((FILE*)NULL));
    }

    if ((fp != stdout) || (fs_enable_stdout))
    {
        if (kjb_isatty(fileno(fp))) 
        {
            if ( ! halt_term_output) 
            {
                fputs_res = term_puts(str);
            }
        }
        else 
        {
            fputs_res = fputs(str, fp); 

            if (fputs_res == EOF) 
            {
                set_error("Write to %F failed.%S\n", fp);
                fputs_res = ERROR; 
            }
        }
    }

    if ((fp == stdout) && (fs_stdout_shadow_fp != NULL))
    {
        if (fputs(str, fs_stdout_shadow_fp) == EOF)
        {
            if (fputs_res == ERROR)
            {
                add_error("Write to %F also failed.%S\n", fs_stdout_shadow_fp);
            }
            else 
            {
                set_error("Write to %F failed.%S\n", fs_stdout_shadow_fp);
                fputs_res = ERROR;
            }
        }
    }

    if ((fp == stderr) && (fs_stderr_shadow_fp != NULL))
    {
        if (fputs(str, fs_stderr_shadow_fp) == EOF)
        {
            if (fputs_res == ERROR)
            {
                add_error("Write to %F also failed.%S\n", fs_stderr_shadow_fp);
            }
            else 
            {
                set_error("Write to %F failed.%S\n", fs_stderr_shadow_fp);
                fputs_res = ERROR;
            }
        }
    }

    return fputs_res; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_fgetc
 *
 * Reads a character from a stream
 *
 * kjb_getc is very similar to the standard getc, except that error reporting
 * is made consistent with the other library routines, and the error messages
 * are more meaningfull. 
 *
 * Returns:
 *     On success, the character read is returned. On end of file, EOF is
 *     returned. If the system does not restart the read after an interupt, and
 *     the read was interupted, then INTERRUPTED is returned.   On more serious
 *     failures an error message is set, and ERROR is returned. 
 *
 * Index: input, I/O
 *
 * ==> This routine is not used much, as term_getc is usually what is meant.
 * -----------------------------------------------------------------------------
*/

int kjb_fgetc(FILE* fp)
{
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
    IMPORT int errno;
#endif 
    int        save_errno;
    int        c;

    if (kjb_isatty(fileno(fp))) 
    {
        return term_getc();
    }

    /*
    // FIX (someday). 
    //
    // This use of errno is a bit of a HACK. 
    */

    /* If we don't have error, we want to restore errno. */
    save_errno = errno; 
    errno = 0;
    c = fgetc(fp); 

    if  (c == EOF) 
    {
        if (errno == EINTR) 
        {
            return INTERRUPTED; 
        }
        else if (errno != 0) 
        {
            set_error("Read from file failed.%S");
            return ERROR;
        }
        else 
        {
            errno = save_errno; 
            return EOF;
        }
    }
    else 
    {
        errno = save_errno; 
        return c; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_fputc
 *
 *
 * -----------------------------------------------------------------------------
*/

int kjb_fputc(FILE* fp, int c)
{
    IMPORT volatile int halt_all_output; 
    IMPORT volatile int halt_term_output; 
    int                 res; 
    char                char_buff[ 1 ]; 

    if (halt_all_output) return NO_ERROR;

    if (fp == stderr) 
    {
        ERE(kjb_fflush((FILE*)NULL));
    }

    if (kjb_isatty(fileno(fp))) 
    {
        if (halt_term_output) return NO_ERROR;

        char_buff[ 0 ] = c; 
        ERE(term_put_n_chars(char_buff, 1)); 
        return NO_ERROR; 
    }
    else 
    {
        res = fputc(c, fp); 

        if (res == EOF) return ERROR; 
        else return NO_ERROR; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              pso  
 *
 * Writes a formatted string to stdout.
 *
 * The formating options are similar to the standard ones, except that the
 * standard ones are not very standard! Thus the KJB library  routine provides
 * some semblence of platform independent IO. Also, there are additional
 * formatting options available (see kjb_fprintf for details). 
 *
 * This routine works together with the rest of the KJB library to implement
 * paging and flushing of all files when stderr is written to. 
 *
 * Macros:
 *    This routine is similar is spirit to the system macro "printf", and 
 *    hence an alternate name "kjb_printf" is provided. 
 * 
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    ERROR on failure and the number of characters printed on success. 
 * 
 * See also:
 *    kjb_fprintf
 * 
 * Index: output, I/O, paging
 * 
 * -----------------------------------------------------------------------------
*/

/*PRINTFLIKE1*/
Size_result pso(const char* format_str, ...)
{
    int     result;
    va_list ap;

    va_start(ap, format_str); 

    result = kjb_vfprintf(stdout, format_str, ap); 

    va_end(ap);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              p_stderr
 *
 * Writes a formatted string to stderr.
 *
 * The formating options are similar to the standard ones, except that the
 * standard ones are not very standard! Thus the KJB library  routine provides
 * some semblence of platform independent IO. Also, there are additional
 * formatting options available (see kjb_fprintf for details). 
 *
 * This routine works together with the rest of the KJB library to implement
 * paging and flushing of all files when stderr is written to. 
 *
 * Notes:
 *    This routine used to called "pse" in analogy with "pso", but it was 
 *    found to be too easy to type the wrong one. Hence the less usual case 
 *    was changed to a more obtuse name. 
 *
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *
 * Returns:
 *    ERROR on failure and the number of characters printed on success. 
 * 
 * See also:
 *    kjb_fprintf
 * 
 * Index: output, I/O, paging
 * 
 * -----------------------------------------------------------------------------
*/

/*PRINTFLIKE1*/
Size_result p_stderr(const char* format_str, ...)
{
    int     result;
    va_list ap;

    va_start(ap, format_str); 

    result = kjb_vfprintf(stderr, format_str, ap); 

    va_end(ap);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_fprintf
 *
 * Writes a formatted string to a stream.
 *
 * The formating options are similar to the standard ones, except that the
 * standard ones are not very standard! Thus the KJB library  routine provides
 * some semblence of platform independent IO. There are additional formatting
 * options available. %F takes a file pointer, and prints the file name as
 * specified by the user (or internal to the code). %D does the same with a file
 * descriptor. %P is similar to %F, except that the complete path is provided
 * instead. The option %t (truncate) is similar to %s, exept that if a field
 * width is given, either by a numeric constant or a *, then if that field width
 * would be overrun, then we form a string of the form "<str> ...".  "t" is
 * short for trucated). The option %q is similar to %t, except that quotes are
 * added arround the string. Finally, the option %R converts an integer to the
 * return enum symbols defined as part of Return_status, or simply the integer
 * itself, if the integer is not one of those values. For example, if the
 * integers has the value ERROR, then "ERROR" is printed. This options is
 * obviously mostly for debugging. 
 *
 * This routine works together with the rest of the KJB library to implement
 * paging and flushing of all files when stderr is written to. 
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    ERROR on failure and the number of characters printed on success. 
 * 
 * See also:
 *    pso
 * 
 * Index: output, I/O, paging
 * 
 * -----------------------------------------------------------------------------
*/

/*PRINTFLIKE2*/
Size_result kjb_fprintf(FILE* fp, const char* format_str, ...)
{
    int     result;
    va_list ap;

    va_start(ap, format_str); 

    result = kjb_vfprintf(fp, format_str, ap); 

    va_end(ap);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                pdo
 *
 * Writes formatted string file "debug-<pid>.output". 
 *
 * This routine writes formatted string file "debug-<pid>.output". It is
 * ocasionally useful when the amount of debugging output is large.  The details
 * of its use can be deduced from pso(3-KJB) or p_stderr(3-KJB). 
 *
 * Note:
 *    The return type Size_result is an integer, except when size_t is 16 bits
 *    (IE, DOS and friends), in which case it is a long. 
 *                                                            
 * See also:
 *    pso, kjb_fprintf 
 *
 * Index: debugging
 *  
 * -----------------------------------------------------------------------------
*/

Size_result pdo(const char* format_str, ...)
{
    va_list      ap;
    char         debug_file_name[ MAX_FILE_NAME_SIZE ];
    int          result; 
    /* File static 
         FILE* fs_debug_fp;
    */

    if (fs_debug_fp == NULL)
    {
        ERE(kjb_sprintf(debug_file_name, sizeof(debug_file_name), 
                        "debug-%ld.output", (long)MY_PID)); 
        NRE(fs_debug_fp = kjb_fopen(debug_file_name, "w"));
        add_cleanup_function(close_debug_file); 
    }

    va_start(ap, format_str); 

    result =  kjb_vfprintf(fs_debug_fp, format_str, ap); 

    va_end(ap); 

    if (result != ERROR)
    {
        result = kjb_fflush(fs_debug_fp);
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                         close_debug_file
 * -----------------------------------------------------------------------------
*/ 

static void close_debug_file(void)
{

    EPE(kjb_fclose(fs_debug_fp));  /* Print error, as this is a void cleanup fn. */
    fs_debug_fp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                kjb_vfprintf
 *
 *
 * -----------------------------------------------------------------------------
*/

/*
// FIX   We should looking at sprintf's return.
*/

Size_result kjb_vfprintf(FILE* fp, const char* format_str, va_list ap)
{
    IMPORT volatile int halt_all_output; 
    IMPORT volatile int halt_term_output; 
    char                new_format_str[ MAX_FORMAT_STRING_SIZE ]; 
    const char*         format_str_pos; 
    char*               new_format_str_pos; 
    int                 format_item; 
    char                type_char; 
    int                 num_precision_args; 
    int                 precision_arg1; 
    int                 precision_arg2; 
    int                 short_flag; 
    int                 long_flag; 
#ifdef LONG_DOUBLE_OK
    int                 long_double_flag; 
#endif
    long                res                                      = 0; 
    long                byte_count                               = 0; 
    size_t              char_count; 
    char                output_buff[ MAX_FORMAT_STRING_SIZE ]; 
    char                trunc_buff[ 1000 ]; 
    int                 comma_flag; 
    int                 justify_flag; 
    int                 field_width_flag; 
    int                 precision_flag; 
    int                 field_width_arg_flag; 
    int                 field_width; 
    size_t              field_size; 
    int                 temp_field_width; 

    if (halt_all_output || (halt_term_output && kjb_isatty(fileno(fp))))
    {
        return NO_ERROR;
    }

    /*
    // Moved to kjb_fputs and kjb_fwrite, which are the only ways this routine
    // outputs?
    //
    //  if (fp == stderr) 
    //  {
    //      ERE(kjb_fflush((FILE*)NULL));
    //  }
    */

    if (*format_str == '\0') return NO_ERROR; 

    while (*format_str != '\0') 
    {
        format_item = FALSE; 

        char_count = 0; 
        format_str_pos = format_str; 

        while ((!format_item) &&
               (*format_str_pos != '\0'))
        { 
            if (*format_str_pos == '%') 
            {
                format_item = TRUE;
            }
            else 
            {
                format_str_pos++;
                char_count++; 
            }
        }

        if (char_count != 0) 
        {
            ERE(res = kjb_fwrite(fp, (const void*)format_str, char_count)); 
            byte_count += res; 
            format_str += char_count; 
        }

        if (format_item)  /* Can't be at string end if true */ 
        {
            short_flag = FALSE;
            long_flag = FALSE;
#ifdef LONG_DOUBLE_OK
            long_double_flag = FALSE; 
#endif
            num_precision_args = 0; 
            justify_flag = field_width_flag = precision_flag = FALSE;
            comma_flag = field_width_arg_flag = FALSE;

            new_format_str_pos = new_format_str;
            *new_format_str_pos = *format_str; 
            format_str++; 
            new_format_str_pos++; 

            /* 
             * FIX: Currently we accept stuff like "-+  -  +". Not serious,
             *      as this stuff just gets passed down to sprintf anyway. 
             *      BUT we DONT look at sprintf's return! 
            */
            while (    (*format_str == '-') 
                    || (*format_str == ' ') 
                    || (*format_str == '0') 
                    || (*format_str == '#') 
                    || (*format_str == '+') 
                    || (*format_str == ',') 
                  )
            {
                if (*format_str == '-') 
                {
                    justify_flag = TRUE;
                }

                if (*format_str == ',') 
                {
                    comma_flag = TRUE;
                }
                else 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                }

                format_str++;
            }

            if (*format_str == '*') 
            {
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;

                num_precision_args++;
                field_width_flag = TRUE; 
                field_width_arg_flag = TRUE; 
            }
            else 
            {
                if (isdigit((int)(*format_str))) 
                {
                    field_width_flag = TRUE; 
                }

                while (isdigit((int)(*format_str))) 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                    format_str++;
                }
            } 

            if (*format_str == '.') 
            {
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;

                if (*format_str == '*') 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                    format_str++;

                    num_precision_args++;
                    precision_flag = TRUE;
                }
                else 
                {
                    if (isdigit((int)(*format_str))) 
                    {
                        precision_flag = TRUE; 
                    }

                    while (isdigit((int)(*format_str))) 
                    {
                        *new_format_str_pos = *format_str; 
                        new_format_str_pos++;
                        format_str++;
                    }
                }
            }

            if (*format_str == 'l') 
            {
                long_flag = TRUE; 
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;
            }
            else if (*format_str == 'h') 
            {
                short_flag = TRUE; 
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;
            }
            else if (*format_str == 'L') 
            {
#ifdef LONG_DOUBLE_OK
                long_double_flag = TRUE; 
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
#endif
                format_str++;
            }

            if (*format_str == '\0') 
            {
                SET_FORMAT_STRING_BUG();
                return ERROR;
            }

            type_char = *format_str; 
            format_str++;

            if ((type_char == 'd') || (type_char == 'i'))
            {
                *new_format_str_pos = type_char;
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff, 
                                new_format_str,
                                va_arg(ap, int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                va_arg(ap, int)); 
                    }
                } 
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, int)); 
                    }
                } 
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, int)); 
                    }
                } 
                else 
                {
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                /*
                // FIX 
                //
                // Currently, commas will not be inserted into anything other
                // than non-justified positive numbers. Note that commas do not
                // properly work with precision arguments!
                //
                // Note: This code must be synched with unsigned case, and
                //       vsprintf. 
                */
                if (comma_flag && all_digits(output_buff))
                {
                    char buff_copy[ 100 ];
                    char* buff_copy_pos = buff_copy;
                    int  len = strlen(output_buff);
                    int first_chunk = (len - 1) % 3 + 1;
                    int i; 

                    i = 0;

                    while (i<first_chunk)
                    {
                        *buff_copy_pos = output_buff[ i ];
                        buff_copy_pos++;
                        i++;
                    } 

                    while (i<len)
                    {
                        *buff_copy_pos++ = ',';
                        *buff_copy_pos++ = output_buff[ i++ ];
                        *buff_copy_pos++ = output_buff[ i++ ];
                        *buff_copy_pos++ = output_buff[ i++ ];
                    }

                    *buff_copy_pos = '\0';

                    BUFF_CPY(output_buff, buff_copy);
                }
                    
                ERE(res = kjb_fputs(fp, output_buff)); 
            } 
            else if (    (type_char == 'o') || (type_char == 'u')
                      || (type_char == 'x') || (type_char == 'X') 
                    ) 
            {
                *new_format_str_pos = type_char;
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff, 
                                new_format_str,
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else 
                {
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                /*
                // FIX 
                //
                // Currently, commas will not be inserted into anything other
                // than non-justified positive numbers. Note that commas do not
                // properly work with precision arguments!
                */
                if ((type_char == 'u') && comma_flag && all_digits(output_buff))
                {
                    char buff_copy[ 100 ];
                    char* buff_copy_pos = buff_copy;
                    int  len = strlen(output_buff);
                    int first_chunk = (len - 1) % 3 + 1;
                    int i; 

                    i = 0;

                    while (i<first_chunk)
                    {
                        *buff_copy_pos = output_buff[ i ];
                        buff_copy_pos++;
                        i++;
                    } 

                    while (i<len)
                    {
                        *buff_copy_pos++ = ',';
                        *buff_copy_pos++ = output_buff[ i++ ];
                        *buff_copy_pos++ = output_buff[ i++ ];
                        *buff_copy_pos++ = output_buff[ i++ ];
                    }

                    *buff_copy_pos = '\0';

                    BUFF_CPY(output_buff, buff_copy);
                }
                    
                ERE(res = kjb_fputs(fp, output_buff)); 
            } 
            else if (    (type_char == 'e') || (type_char == 'f')
                      || (type_char == 'g') || (type_char == 'E') 
                      || (type_char == 'G')
                    )
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(output_buff, new_format_str,
                                va_arg(ap, long_double));
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(output_buff,
                            new_format_str, 
                            va_arg(ap, double));
#endif
                }
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(output_buff, new_format_str, precision_arg1, 
                                va_arg(ap, long_double)); 
                    }
                    else 
                    {
                        sprintf(output_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(output_buff,
                            new_format_str,
                            precision_arg1, 
                            va_arg(ap, double));
#endif
                }
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(output_buff, new_format_str,
                                precision_arg1, precision_arg2, 
                                va_arg(ap, long_double)); 
                    }
                    else 
                    {
                        sprintf(output_buff, 
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(output_buff, 
                            new_format_str,
                            precision_arg1, 
                            precision_arg2, 
                            va_arg(ap, double));
#endif
                }
                else 
                {
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                ERE(res = kjb_fputs(fp, output_buff)); 
            }
            else if (type_char == 's') 
            {
                if ((justify_flag) || (precision_flag) || (field_width_flag)) 
                {
                    *new_format_str_pos = type_char; 
                    new_format_str_pos++;
                    *new_format_str_pos = '\0';

                    /*
                    // FIX 
                    //
                    // Should check length of va_arg(ap, char*) and precision
                    // args where applicable in order to check for buffer
                    // over-run. 
                    */
                    if (num_precision_args == 0) 
                    {
                        sprintf(output_buff, new_format_str, 
                                va_arg(ap, char*));
                    }
                    else if (num_precision_args == 1) 
                    {
                        precision_arg1 = va_arg(ap, int); 

                        sprintf(output_buff, new_format_str, 
                                precision_arg1, va_arg(ap, char*));
                    } 
                    else if (num_precision_args == 2) 
                    {
                        precision_arg1 = va_arg(ap, int); 
                        precision_arg2 = va_arg(ap, int); 

                        sprintf(output_buff, new_format_str, 
                                precision_arg1, precision_arg2, 
                                va_arg(ap, char*));
                    }

                    ERE(res = kjb_fputs(fp, output_buff)); 
                }
                else 
                {
                    ERE(res = kjb_fputs(fp, va_arg(ap, char *)));
                }
            }
            else if (type_char == 'D') 
            {
                BUFF_GET_USER_FD_NAME(va_arg(ap, int), output_buff);
                ERE(res = kjb_fputs(fp, output_buff));
            }
            else if (type_char == 'R') 
            {
                get_status_string(va_arg(ap, int), output_buff, 
                                  sizeof(output_buff));
                ERE(res = kjb_fputs(fp, output_buff));
            }
            else if (type_char == 'F') 
            {
                BUFF_GET_USER_FD_NAME(fileno(va_arg(ap, FILE *)), output_buff);
                ERE(res = kjb_fputs(fp, output_buff));
            }
            else if (type_char == 'P') 
            {
                BUFF_GET_FD_NAME(fileno(va_arg(ap, FILE *)), output_buff);
                ERE(res = kjb_fputs(fp, output_buff));
            }
            else if (type_char == 'S') 
            {
#ifdef HAVE_STRERROR
                const char* sys_mess = strerror(errno);

                if (sys_mess != NULL) 
                {
                    ERE(res = kjb_fputs(fp, "\nSystem error message is: ")); 
                    byte_count += res;

                    ERE(res = kjb_fputs(fp, sys_mess));
                    byte_count += res;
                }
#else 
                const char*  sys_mess; 
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
                IMPORT int   errno;
#endif 
                IMPORT int   sys_nerr;
                IMPORT SYS_ERRLIST_TYPE sys_errlist[];

                if (    (errno != NO_SYSTEM_ERROR) 
                     && (errno < sys_nerr)
                   )
                {
                    ERE(res = kjb_fputs(fp, "\nSystem error message is: ")); 
                    byte_count += res;

                    sys_mess = sys_errlist[errno];

                    ERE(res = kjb_fputs(fp, sys_mess));
                    byte_count += res;
                }
#endif 
            }
            else if ((type_char == 'q') || (type_char == 't')) 
            {
                field_width = sizeof(trunc_buff); 

                if (num_precision_args == 1)
                {
                    temp_field_width = va_arg(ap, int); 

                    if (temp_field_width < field_width)
                    {
                        field_width = temp_field_width;
                    }
                }
                else if (field_width_flag)
                {
                    *new_format_str_pos = '\0';     

                    if (ss1pi(new_format_str + 1, &temp_field_width) == ERROR)
                    {
                        SET_FORMAT_STRING_BUG();
                        return ERROR; 
                    }

                    ASSERT(temp_field_width >= 0);

                    if (temp_field_width < field_width)
                    {
                        field_width = temp_field_width;
                    }
                }

                if (type_char == 't') 
                {
                    if (field_width < 5) 
                    {
                        field_size = 5;
                    }
                    else field_size = field_width;

                    str_trunc_cpy(trunc_buff, va_arg(ap, char *), field_size); 
                }
                else 
                {
                    if (field_width < 7) 
                    {
                        field_size = 7;
                    }
                    else field_size = field_width;

                    trunc_quote_cpy(trunc_buff, va_arg(ap, char *), field_size);
                }

                ERE(res = kjb_fputs(fp, trunc_buff)); 
            }
            else if (type_char == 'c') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (field_width_arg_flag) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    sprintf(output_buff, new_format_str,
                            precision_arg1, va_arg(ap, int));
                }
                else 
                {
                    sprintf(output_buff, new_format_str,
                            va_arg(ap, int));
                }

                ERE(res = kjb_fputs(fp, output_buff)); 
            }
            else if (type_char == 'p') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';
                sprintf(output_buff, new_format_str, va_arg(ap, void*));
                ERE(res = kjb_fputs(fp, output_buff)); 
            }
            else if (type_char == 'n') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';
                sprintf(output_buff, new_format_str, va_arg(ap, int*));
                ERE(res = kjb_fputs(fp, output_buff)); 
            }
            else if (type_char == '%') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                sprintf(output_buff, new_format_str); 
                ERE(res = kjb_fputs(fp, output_buff)); 
            }
            else 
            {
                SET_FORMAT_STRING_BUG();
                return ERROR; 
            }

            byte_count += res; 
        }
    }

    return byte_count; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
* =============================================================================
 *                              is_file
 *                                
 * Determins if the argument is a valid file
 *
 * This routines determines whether the arguement is a valid file.  It is the
 * intention that in the case of a symbolic link, this routine determines
 * whether what is pointed to by the link exists, and if so, is a file.  (This
 * works on linux). 
 *
 * Returns:
 *     TRUE if "path" is a file, FALSE otherwise.
 *
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/

int is_file(const char* path)
{

    return (get_path_type(path) == PATH_IS_REGULAR_FILE);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
* =============================================================================
 *                              is_directory
 *                                
 * Determins if the argument is a valid directory
 *
 * This routines determines whether the arguement is a valid directory. It is
 * the intention that in the case of a symbolic link, this routine determines
 * whether what is pointed to by the link exists, and if so, is a directory.
 * (This works on linux). 
 *
 * Returns:
 *     TRUE if "path" is a directory, FALSE otherwise.
 *
 * Index: directorys, standard library
 *
 * -----------------------------------------------------------------------------
*/

int is_directory(const char* path)
{
    /*
    char real_path[ MAX_FILE_NAME_SIZE ];

    dbs(path);
    BUFF_REAL_PATH(path, real_path);
    dbs(real_path); 

    dbi(get_path_type(path));
    dbi(get_path_type(real_path));
    */

    return (get_path_type(path) == PATH_IS_DIRECTORY);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
* =============================================================================
 *                              fp_get_path_type
 *                                
 * Determins type of path
 *
 * This routines determines the path type (file, directory, ...). It also is a
 * reasonable way to find out if a path exists at all. 
 *
 * Returns:
 *    One of { PATH_DOES_NOT_EXIST,  PATH_IS_REGULAR_FILE, PATH_IS_DIRECTORY,
 *             PATH_IS_PIPE, PATH_IS_SOCKET, PATH_IS_BLOCK_SPECIAL, 
 *             PATH_IS_CHARACTER_SPECIAL,  PATH_IS_UNCLASSIFIED, PATH_ERROR }
 *
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/

Path_type fp_get_path_type(FILE* fp)
{
    struct stat path_stats;

    if (fstat(fileno(fp), &path_stats) == -1) 
    {
#ifdef TEST
        set_error("Fstat system called failed for %F.%S", fp); 
#else
        set_error("Can't determine the size of %F.%S", fp); 
#endif
        return PATH_ERROR;
    }
    else 
    {
        return get_path_type_guts(&path_stats); 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                get_path_type
 *                                
 * Determins type of path
 *
 * This routines determines the path type (file, directory, ...). It also is a
 * reasonable way to find out if a path exists at all. 
 *
 * Returns:
 *    One of { PATH_DOES_NOT_EXIST,  PATH_IS_REGULAR_FILE, PATH_IS_DIRECTORY,
 *             PATH_IS_PIPE, PATH_IS_SOCKET, PATH_IS_BLOCK_SPECIAL, 
 *             PATH_IS_CHARACTER_SPECIAL,  PATH_IS_UNCLASSIFIED, PATH_ERROR }
 *
 * Index: files, standard library
 *
 * -----------------------------------------------------------------------------
*/

Path_type get_path_type(const char* path)
{
    struct stat path_stats;
    char        real_path[ MAX_FILE_NAME_SIZE ];

    if (BUFF_REAL_PATH(path, real_path) == ERROR)
    {
        BUFF_CPY(real_path, path);
    }

    if (stat(real_path, &path_stats) == -1) 
    {
#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
        IMPORT int  errno;
#endif 

        if (errno == ENOENT)
        {
            return PATH_DOES_NOT_EXIST; 
        }
        else 
        {
            set_error("Problem accessing %s.%S", path);
            return PATH_ERROR;
        }
    }
    else 
    {
        return get_path_type_guts(&path_stats); 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 * STATIC                         get_path_type_guts 
 * -----------------------------------------------------------------------------
*/ 

static Path_type get_path_type_guts(struct stat *path_stats_ptr)
{

    if (S_ISREG(path_stats_ptr->st_mode))
    {
        return PATH_IS_REGULAR_FILE; 
    }
    else if (S_ISDIR(path_stats_ptr->st_mode))
    {
        return PATH_IS_DIRECTORY; 
    }
    else if (S_ISFIFO(path_stats_ptr->st_mode))
    {
        return PATH_IS_PIPE; 
    }
    else if (S_ISSOCK(path_stats_ptr->st_mode))
    {
        return PATH_IS_SOCKET; 
    }
    else if (S_ISBLK(path_stats_ptr->st_mode))
    {
        return PATH_IS_BLOCK_SPECIAL; 
    }
    else if (S_ISCHR(path_stats_ptr->st_mode))
    {
        return PATH_IS_CHARACTER_SPECIAL; 
    }
    else 
    {
        return PATH_IS_UNCLASSIFIED; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                get_file_size
 *
 * Determins the size of a file in bytes
 *
 * See also:
 *     fp_get_byte_size 
 *
 * Returns:
 *     NO_ERROR on success, ERROR on failure, with an appropriate error message
 *     being set. 
 *
 * Index: files, standard library
 * 
 * -----------------------------------------------------------------------------
*/

int get_file_size(const char* file_name, off_t *size_ptr)
{
    struct stat file_stats;
    char        real_path[ MAX_FILE_NAME_SIZE ];

    if (BUFF_REAL_PATH(file_name, real_path) == ERROR)
    {
        BUFF_CPY(real_path, file_name);
    }

    if (stat(real_path, &file_stats) == -1) 
    {
#ifdef TEST
        set_error("Stat system called failed for %s.%S", file_name); 
#else
        set_error("Can't determine the size of %s.%S", file_name); 
#endif
        return ERROR;
    }

#ifdef UNIX
    if (S_ISSOCK(file_stats.st_mode)) 
    {
        set_error("Attempt to determine file size of a socket.");
        return ERROR;
    }
    else if (S_ISFIFO(file_stats.st_mode)) 
    {
        set_error("Attempt to determine file size of a pipe.");
        return ERROR;
    }
#else
#ifdef MS_16_BIT_OS
    if (file_stats.st_mode && S_IFCHR) 
    {
        set_error("Attempt to determine file size of a device.");
        return ERROR;
    }
#endif
#endif

    *size_ptr = file_stats.st_size;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                fp_get_byte_size
 *
 * Determines the size of a file in bytes
 *
 * See also:
 *     get_file_size 
 *
 * Returns:
 *     NO_ERROR on success, ERROR on failure, with an appropriate error message
 *     being set. 
 *
 * Index: files, standard library
 * 
 * -----------------------------------------------------------------------------
*/

int fp_get_byte_size(FILE* fp, off_t *size_ptr)
{
    struct stat file_stats;

    if (kjb_isatty(fileno(fp))) 
    {
        set_error("Attempt to determine file size of a terminal.");
        return ERROR;
    }

    if (fstat(fileno(fp), &file_stats) == -1) 
    {
#ifdef TEST
        set_error("Fstat system called failed for %F.%S", fp); 
#else
        set_error("Can't determine the size of %F.%S", fp); 
#endif
        return ERROR;
    }

#ifdef UNIX
    if (S_ISSOCK(file_stats.st_mode)) 
    {
        set_error("Attempt to determine file size of a socket.");
        return ERROR;
    }
    else if (S_ISFIFO(file_stats.st_mode)) 
    {
        set_error("Attempt to determine file size of a pipe.");
        return ERROR;
    }
#else
#ifdef MS_16_BIT_OS
    if (file_stats.st_mode && S_IFCHR) 
    {
        set_error("Attempt to determine file size of a device.");
        return ERROR;
    }
#endif
#endif

    *size_ptr = file_stats.st_size;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                get_file_age
 *
 * Determines the number of milliseconds since a file has been modified
 *
 * Returns:
 *     NO_ERROR on success, ERROR on failure, with an appropriate error message
 *     being set. 
 *
 * Index: files, standard library
 * 
 * -----------------------------------------------------------------------------
*/

int get_file_age(const char* file_name, time_t *age_ptr)
{
    struct stat file_stats; 
    time_t current_time;

    if (stat(file_name, &file_stats) == -1) 
    {
        set_error("Can't get modification time of %s.%S", file_name); 
        return ERROR;
    }

    if (time(&current_time) == TIME_ERROR) 
    {
        set_error("Attempt to get current time failed.");
        return ERROR;
    }

    *age_ptr =  current_time - file_stats.st_mtime;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

/*
 * =============================================================================
 *                                print_open_files
 *
 * Prints the names of all files that are currently open
 *
 * Notes:
 *    This routine is only available in the development version of the library.
 *
 *    Only files opened with kjb_fopen will be listed.
 *
 * Index: debugging
 *
 * -----------------------------------------------------------------------------
*/

int print_open_files(FILE* fp)
{
    int          i; 

    ERE(initialize_cache()); 

    ERE(kjb_fprintf(fp, "\nOpen files for process %ld:\n", (long)MY_PID));

    for (i=0; i<MAX_NUM_NAMED_FILES; i++) 
    {
        if (fs_cached_file_reference_count[ i ] > 0) 
        {
            ERE(kjb_fprintf(fp, "    %-3d %s (%d ref).\n", i, 
                            fs_cached_user_fnames[ i ],
                            fs_cached_file_reference_count[ i ]));
        }
    }

    ERE(kjb_fputs(fp, "\n")); 

    return kjb_fflush(fp); 
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

/*
 * =============================================================================
 * STATIC                         free_cached_file_names
 * -----------------------------------------------------------------------------
*/

static void free_cached_file_names(void)
{
    int          i; 
#ifdef TEST
    int          num_open_files = 0;
#endif 

    /*
    // Mimic print_open_files here, since it is problematic calling it at the
    // end of the program, while things are being dismantled--especially since
    // paging is not a bad feature of print_open_files. Also, we want a slightly
    // different behaviour. We only want to hear about open files if there are
    // more than the standard ones. 
    */
#ifdef TEST
    for (i=0; i<MAX_NUM_NAMED_FILES; i++) 
    {
        if (fs_cached_user_fnames[ i ] != NULL) num_open_files++;
    }

    if (num_open_files > 3)
    {
        printf("\nOpen files for process %ld:\n", (long)MY_PID);
    }
#endif 

    for (i=0; i<MAX_NUM_NAMED_FILES; i++) 
    {
        kjb_free(fs_cached_fnames[ i ]);
        fs_cached_fnames[ i ] = NULL;

        if (fs_cached_user_fnames[ i ] != NULL)
        {
#ifdef TEST
            if (num_open_files > 3)
            {
                printf("    %-3d %s (%d ref).\n", i, fs_cached_user_fnames[ i ],
                       fs_cached_file_reference_count[ i ]);
            }
#endif 
            kjb_free(fs_cached_user_fnames[ i ]);
            fs_cached_user_fnames[ i ] = NULL;
        }

        kjb_free(fs_cached_uncompressed_fnames[ i ]);
        fs_cached_uncompressed_fnames[ i ] = NULL;

        fs_cached_file_reference_count[ i ] = NOT_SET;   /* Invalidation */
    }

#ifdef TEST
    if (num_open_files > 5)
    {
        puts("\n"); 
        fflush(stdout); 
    }
#endif 

}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                         process_file_open_mode
 * -----------------------------------------------------------------------------
*/

static int process_file_open_mode
(
    const char* mode,
    char*       system_dependent_mode,
    size_t      system_dependent_mode_max_len,
    char*       unix_mode,
    size_t      unix_mode_max_len
)
{

#ifdef MAX_SIZE_T_BUFF_SIZE
    if (system_dependent_mode_max_len > MAX_SIZE_T_BUFF_SIZE)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }
#endif 

    if (system_dependent_mode_max_len == 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    --system_dependent_mode_max_len;  /* Account for room needed by null. */ 

    while (    (system_dependent_mode_max_len > 0) 
            && (unix_mode_max_len > 0) 
            && (*mode != '\0')
          ) 
    {
#ifdef MS_16_BIT_OS

        /* Straight copy.  */

        *system_dependent_mode = *mode; 
        system_dependent_mode++;
        system_dependent_mode_max_len--;

#else   /* Case not MS_16_BIT_OS follows */

        /* Copy, but skip 'b's.   */

        if (*mode != 'b')
        {
            *system_dependent_mode = *mode; 
            system_dependent_mode++;
            system_dependent_mode_max_len--;
        }
#endif

        /* Copy, but skip 'b's.   */

        if (*mode != 'b')
        {
            *unix_mode = *mode; 
            unix_mode++;
            unix_mode_max_len--;
        }

        mode++; 
    }

    *system_dependent_mode = '\0'; 
    *unix_mode = '\0'; 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */
/*
 * =============================================================================
 *                                kjb_isatty
 * -----------------------------------------------------------------------------
*/

int kjb_isatty(int file_des)
{
    int result;

    if (file_des < 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(initialize_cache()); 

    if (file_des < MAX_NUM_NAMED_FILES)
    {
        if (IS_SET(fs_cached_isatty[ file_des ]))
        {
            return fs_cached_isatty[ file_des ];
        }
    }

    result = kjb_isatty_guts(file_des);

    if (file_des < MAX_NUM_NAMED_FILES)
    {
        fs_cached_isatty[ file_des ] = result;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                         kjb_isatty_guts
 * -----------------------------------------------------------------------------
*/

#ifdef SUN 

static int kjb_isatty_guts(file_des)
    int file_des;
{
#ifdef SUN5
    extern char* ctermid_r(char *); 
    extern char* ttyname_r(int, char *, int); 
#endif
    static char  control_ttyname_buff[ L_ctermid ] = { '\0' }; 
    static int   first_time                        = TRUE; 
    char         ttyname_buff[ MAXPATHLEN ]; 
    char*        ttyname_ptr; 
    char*        tty_name_pos; 
    char         path_piece_one[ MAXPATHLEN ]; 
    char         path_piece_two[ MAXPATHLEN ]; 
#ifdef SUN5
    char         path_piece_three[ MAXPATHLEN ]; 
#endif

    if (first_time)
    {
#ifdef SUN5
        if (ctermid_r(control_ttyname_buff) == NULL)
#else 
            if (ctermid(control_ttyname_buff) == NULL)
#endif
            {
                control_ttyname_buff[ 0 ] = '\0';
            }
    }

#ifdef SUN5
    ttyname_ptr = ttyname_r(file_des, ttyname_buff, sizeof(ttyname_buff));
#else
    ttyname_ptr = ttyname(file_des);
#endif

    if (ttyname_ptr == NULL)
    {
        return FALSE;
    }

#ifndef SUN5
    BUFF_CPY(ttyname_buff, ttyname_ptr);
#endif

    if (STRCMP_EQ(control_ttyname_buff, ttyname_buff))
    {
        return TRUE;
    }

    /*
    // Find out if it is a pseudo-terminal. If so, then call it a terminal
    // (most terminals -- IE networked ones -- are pseudo-terminals). 
    */

    tty_name_pos = ttyname_buff;

    if (*tty_name_pos != '/') 
    {
        return FALSE;
    }

    tty_name_pos++; 

    BUFF_GEN_GET_TOKEN(&tty_name_pos, path_piece_one, "/");
    BUFF_GEN_GET_TOKEN(&tty_name_pos, path_piece_two, "/");

#ifdef SUN5

    BUFF_GEN_GET_TOKEN(&tty_name_pos, path_piece_three, "/");

    if (    (    (STRCMP_EQ(path_piece_one, "dev"))
              && (STRCMP_EQ(path_piece_two, "pts"))
              && (path_piece_three[ 0 ] != '\0')
              && (all_digits(path_piece_three))
            )
         ||
            (    (STRCMP_EQ(path_piece_one, "dev"))
              && (STRCMP_EQ(path_piece_two, "console"))
            )
         ||
            (    (STRCMP_EQ(path_piece_one, "devices"))
              && (STRCMP_EQ(path_piece_one, "pseudo"))
              && (HEAD_CMP_EQ(path_piece_three, "pts"))
            )
       )
    {
        return TRUE;
    }
#else
#ifdef SUN4
    if (    (STRCMP_EQ(path_piece_one, "dev"))
         && 
            (    (HEAD_CMP_EQ(path_piece_two, "ttyp"))
              || (STRCMP_EQ(path_piece_two, "console"))
            )
       )
    {
        return TRUE;
    }
#endif
#endif

    return FALSE;
}

#else  /* All others except for SUN -- to be updated as needed.  */

/*
// Go with the system supplied version until alternates are desired. 
// 
// Next controlling terminal can be found out by using ttyslot() and the file
// /etc/ttys. Pseudo-tty's is done much like SUN4 (as it is a BSD system).
//
// MS_16_BIT_OS with Microsoft compiler version 5 had isatty( <printer> )
// returning TRUE. I don't know if this will be the case with a late model
// Borland Compiler.
*/

static int kjb_isatty_guts(int file_des)
{

    return isatty(file_des);

}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                              print_underlined                          
 *                               
 * Prints a string as underlined
 *
 * The method to print underlined text is the same as that used by nroff to
 * format text which is underlined both on the printed page, and on the 
 * screen. (Some tools, on some systems, convert this to bold). 
 *
 * Note:
 *    This routine returns a "Size_result", which is currently typedef'd to
 *    long. This type is used whenever a routine wants to return "size_t"
 *    characters on success, and a negative value on failure. The main
 *    technical point in choosing this type is that it should be able to
 *    represent all 16 bit size_t values, and some negative values, and thus
 *    should be at least 32 bits. Currently long is used, even when equivalent
 *    to int, to aid in cross-platform developement. 
 *  
 * Returns:
 *    The number of characters output, which is three for every character in
 *    the input string sucessfully dealt with. This is a bit confusing, since
 *    the return is not general the length of buff, but it is the more "proper"
 *    return. The output should be properly paged if the output is the
 *    terminal. (I have not tested this, however, as the usual use of this
 *    routine is to build man pages). 
 *
 * ==> FIX  (test the routine more thouroughly, including the point raised 
 * ==>       above). 
 *
 * -----------------------------------------------------------------------------
*/ 

Size_result print_underlined(FILE* fp, const char* buff)
{
    IMPORT volatile int halt_all_output;
    IMPORT volatile int halt_term_output;
    int                 result           = NO_ERROR;
    int                 count            = 0;
    int                 cache_is_ttty    = kjb_isatty(fileno(fp));

    kjb_clear_error();

    /*
    // For each character in buff, we output "_", "" (ctl-H), and the
    // charcter.  The main complicating factors are paging and error handling. 
    // We use the standard library fputc for the cases where we do not want the
    // characters counted for paging, specifically "_", and "". 
    */
    while (    (*buff) && ( !halt_all_output )
            && !(halt_term_output && cache_is_ttty)
          )
    {
        if (fputc('_', fp) == -1)
        {
            result = ERROR;
            break;
        }

        count++;

        if (fputc('', fp) == -1)
        {
            result = ERROR;
            break;
        }

        count++;

        if (cache_is_ttty)
        {
            if (term_put_n_chars(buff, 1) != 1)
            {
                result = ERROR;
            }
        }
        else
        {
            if (fputc(*buff, fp) == -1)
            {
                result = ERROR;
                break;
            }
        }

        buff++; 
        count++;
    }
    
    if (result == ERROR)
    {
        insert_error("Underlined output to %F failed.%S.", fp);
        return result;
    }
    else
    {
        return count;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int start_stdout_shadow(const char* file_name)
{

    stop_stdout_shadow(); 
    NRE(fs_stdout_shadow_fp = kjb_fopen(file_name, "a"));
    ERE(add_cleanup_function(stop_stdout_shadow));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void stop_stdout_shadow(void)
{
    
    (void)kjb_fclose(fs_stdout_shadow_fp);
    fs_stdout_shadow_fp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int start_stderr_shadow(const char* file_name)
{

    stop_stderr_shadow(); 
    NRE(fs_stderr_shadow_fp = kjb_fopen(file_name, "a"));
    ERE(add_cleanup_function(stop_stderr_shadow));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void stop_stderr_shadow(void)
{

    (void)kjb_fclose(fs_stderr_shadow_fp);
    fs_stderr_shadow_fp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void enable_stdout(void)
{
    fs_enable_stdout = TRUE;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void disable_stdout(void)
{
    fs_enable_stdout = FALSE;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int kjb_glob(Word_list** dir_files_pp, const char* pattern, 
             int filter_fn(const char* path))
{
    glob_t pglob; 
    int result = NO_ERROR;
    int flags = GLOB_ERR;
    int glob_res;

    if (pattern == NULL)
    {
        return get_target_word_list(dir_files_pp, 0);
    }

    glob_res = glob(pattern, flags, NULL, &pglob);

#ifdef MAC_OSX
    if (glob_res == 0)
    {
        ERE(get_target_word_list(dir_files_pp, 0));
    }
#else
    if (glob_res == GLOB_NOMATCH)
    {
        ERE(get_target_word_list(dir_files_pp, 0));
    }
#endif 
    else if (glob_res != 0)
    {
        set_error("Unable to execute matching on %s.%S",
                  pattern);
        result = ERROR;
    }
    else
    {
        int    word_count = 0;
        char** words;
        uint   i;
        int    filter_res = NO_ERROR;

        result = get_target_word_list(dir_files_pp, 
                                      (int)pglob.gl_pathc);

        if (result != ERROR)
        {
            words = (*dir_files_pp)->words;

            for (i = 0; i < pglob.gl_pathc; i++)
            {
                if (filter_fn == NULL)
                {
                    filter_res = TRUE;
                }
                else 
                {
                    filter_res = filter_fn(pglob.gl_pathv[ i ]);
                }

                if (filter_res == ERROR)
                {
                    NOTE_ERROR();
                    result = ERROR;
                    break;
                }
                else if (filter_res)
                {
                    if ((words[ word_count ] = kjb_strdup(pglob.gl_pathv[ i ])) == NULL)
                    {
                        NOTE_ERROR();
                        result = ERROR;
                        break;
                    }

                    word_count++;
                }
            }

            (*dir_files_pp)->num_words = word_count;
        }
    }

    globfree(&pglob); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int kjb_simple_glob(Word_list** dir_files_pp, Word_list** star_str_pp, 
                    const char* beg_pattern, const char* end_pattern,
                    int filter_fn(const char* path))
{
    char glob_pattern[ 10000 ]; 
    int num_words, i;
    char** words;
    int beg_pattern_len = (beg_pattern == NULL) ? 0  : strlen(beg_pattern); 
    int end_pattern_len = (end_pattern == NULL) ? 0  : strlen(end_pattern); 
    int word_len; 
    char* word_ptr; 
    
    glob_pattern[ 0 ] = '\0';

    if (beg_pattern != NULL)
    {
        BUFF_CAT(glob_pattern, beg_pattern);
    }

    BUFF_CAT(glob_pattern, "*");

    if (end_pattern != NULL)
    {
        BUFF_CAT(glob_pattern, end_pattern);
    }

    ERE(kjb_glob(dir_files_pp, glob_pattern, filter_fn));

    num_words = (*dir_files_pp)->num_words;
    words = (*dir_files_pp)->words;

    ERE(get_target_word_list(star_str_pp, num_words));
    
    for (i = 0; i < num_words; i++)
    {
        word_ptr = words[ i ]; 

        word_ptr += beg_pattern_len;

        NRE((*star_str_pp)->words[ i ] = kjb_strdup(word_ptr));

        word_len = strlen((*star_str_pp)->words[ i ]);
        (*star_str_pp)->words[ i ][ word_len - end_pattern_len ] = '\0';
    }

    return NO_ERROR;
}

