
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"     /* Only safe as first include in a ".c" file. */ 
#include "l/l_parse.h"
#include "l/l_sys_mal.h"
#include "l/l_string.h"
 
#ifdef SUN5
#    ifndef __GNUC__ 
#        define MALLINFO
#    endif
#endif 

#ifdef SUN4
#    define MALLINFO
#endif 

/*
// Works if we add "-lmalloc", but then we get preemption messages.
// Not very important either way. 
//
#ifdef SGI 
#    define MALLINFO
#endif 
//
*/

#ifdef AIX
#    define MALLINFO
#endif 

#ifdef MALLINFO
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#    include <malloc.h> 
#endif 
#endif

#ifdef NO_MALLINFO
#    undef MALLINFO
#endif 

#ifdef MS_16_BIT_OS
/*Some versions of makedepend cannot be trusted to honor the "-Y" option. */
#ifndef MAKE_DEPEND
#include <alloc.h>
#endif 
#endif

#ifdef __C2MAN__
#ifndef TRACK_MEMORY_ALLOCATION_OR_C2MAN
#define TRACK_MEMORY_ALLOCATION_OR_C2MAN
#endif
#endif

#ifdef TRACK_MEMORY_ALLOCATION
#ifndef TRACK_MEMORY_ALLOCATION_OR_C2MAN
#define TRACK_MEMORY_ALLOCATION_OR_C2MAN
#endif
#endif

#ifdef TRACK_MEMORY_ALLOCATION_OR_C2MAN  /* ----------------------------------*/

#   include "l/l_io.h"
#   include "l/l_sort.h"

#ifdef TEST
#   include "l/l_sys_scan.h"
#   include "l/l_sys_rand.h"
    static double fs_simulated_heap_failure_frequency = DBL_NOT_SET;
#endif 

#   define MAX_NUM_POINTERS             2000000
#   define MAX_NUM_POINTERS_2            200000
#   define MAX_ALLOCATION_FILE_NAME_LEN      24
#   define MALLOC_OVER_RUN_SIZE              48

    typedef struct Sorted_pointers
    {
        int ptr;
        int index;
    }
    Sorted_pointers;

    static int fs_heap_checking_enable = TRUE;
    static int fs_initialization_checking_enable = TRUE;
    static int fs_num_pointers = 0;
    static void*       fs_pointers_to_allocated[ MAX_NUM_POINTERS ];
    static Malloc_size fs_amount_allocated[ MAX_NUM_POINTERS ];
    static char        fs_allocation_file_name[ MAX_NUM_POINTERS ][ MAX_ALLOCATION_FILE_NAME_LEN ];
    static int         fs_allocation_line_number[ MAX_NUM_POINTERS ];
    static void*       fs_pointer_to_watch     = NULL;
    static int         fs_fast_lookup_is_valid = FALSE;
    static Sorted_pointers fs_sorted_pointers[ MAX_NUM_POINTERS ];

    static char  fs_start_heap_check_2_file_name[ MAX_FILE_NAME_SIZE ]; 
    static int   fs_start_heap_check_2_line_number; 
    static int fs_heap_checking_enable_2 = FALSE;
    static int fs_heap_checking_skip_2 = FALSE;
    static int fs_num_pointers_2 = 0;
    static void*       fs_pointers_to_allocated_2[ MAX_NUM_POINTERS_2 ];
    static Malloc_size fs_amount_allocated_2[ MAX_NUM_POINTERS ];
    static char        fs_allocation_file_name_2[ MAX_NUM_POINTERS_2 ][ MAX_ALLOCATION_FILE_NAME_LEN ];
    static int         fs_allocation_line_number_2[ MAX_NUM_POINTERS_2 ];
    static int         fs_fast_lookup_is_valid_2 = FALSE;
    static Sorted_pointers fs_sorted_pointers_2[ MAX_NUM_POINTERS ];
#else 
    static int             fs_heap_checking_enable = FALSE; 
    static int             fs_initialization_checking_enable = FALSE; 

#endif    /* --- #ifdef TRACK_MEMORY_ALLOCATION ----------------------------- */

#ifdef MALLINFO
    static int fs_mallinfo_call_ok = FALSE;
#endif

/* -------------------------------------------------------------------------- */

#ifdef TRACK_MEMORY_ALLOCATION
static void save_pointer_info
(
    void*       ptr,
    Malloc_size num_bytes,
    const char* file_name,
    int         line_number 
);

static void save_pointer_info_2
(
    void*       ptr,
    Malloc_size num_bytes,
    const char* file_name,
    int         line_number 
);

#endif 

/* ------------------------------------------------------------------------- */

int set_heap_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option); 

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "heap-checking"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {

            ERE(pso("Heap checking is %s.\n", 
                    fs_heap_checking_enable ? "enabled" : "disabled" )); 
        }
        else
        {
            int temp_boolean_value;

            ERE(temp_boolean_value = get_boolean_value(value));

            if (temp_boolean_value) 
            {
                set_error("Heap checking can only be disabled.");
                return ERROR;
            }

            fs_heap_checking_enable = temp_boolean_value; 
            
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
          || match_pattern(lc_option, "initialization-checking")
          || match_pattern(lc_option, "init-checking")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {

            ERE(pso("Initialization checking is %s.\n", 
                    fs_initialization_checking_enable ? "enabled" : "disabled" )); 
        }
        else
        {
            int temp_boolean_value;

            ERE(temp_boolean_value = get_boolean_value(value));

            if (temp_boolean_value) 
            {
                set_error("Initialization checking can only be disabled.");
                return ERROR;
            }

            fs_initialization_checking_enable = temp_boolean_value; 
            
        }
        result = NO_ERROR; 
    }

#ifdef TRACK_MEMORY_ALLOCATION
#ifdef TEST
    if (    (lc_option[ 0 ] == '\0') 
         || match_pattern(lc_option, "heap-failure-simulation-frequency")
         || match_pattern(lc_option, "heap-failure-frequency")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_simulated_heap_failure_frequency >= 0.0) 
            {
                ERE(pso("heap-failure-simulation-frequency = %.5f\n",
                        fs_simulated_heap_failure_frequency));
            }
            else 
            {
                ERE(pso("heap-failure-simulation-frequency = off\n"));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_simulated_heap_failure_frequency >= 0.0) 
            {
                ERE(pso("Storage allocation failure is simulated with "));
                ERE(pso("a frequency of %.5f\n",
                        fs_simulated_heap_failure_frequency));
            }
            else 
            {
                ERE(pso("Storage allocation failure is not simulated.\n"));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_simulated_heap_failure_frequency = DBL_NOT_SET; 
        }
        else
        {
            double temp_double;

            ERE(ss1d(value, &temp_double));
            fs_simulated_heap_failure_frequency = temp_double; 
            
        }
        result = NO_ERROR; 
    }
#endif 
#endif 

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

int disable_heap_checking_for_fork(void)
{
    disable_heap_check_2(); 

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

void skip_heap_check_2(void)
{
    fs_heap_checking_skip_2 = TRUE;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

void continue_heap_check_2(void)
{
    fs_heap_checking_skip_2 = FALSE;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             memory_used
 *                               
 * Returns the amount of heap storage used
 *
 * This routine returns the amount of heap storage used provided that this
 * facility is available on the current platform. If this facility is not
 * available, then ERROR is returned, with an error message being set.
 * 
 * Note that the amount of storage used is always greater than the number of
 * bytes allocated, sometimes by a substantial amount. This is because memory is
 * not returned to the system when it is freed; freed memory is simply available
 * for re-use by the allocating process. 
 *
 * Returns:
 *    The amount of heap storage used in bytes, if available, otherwise ERROR
 *    is returned, with an error message being set.
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
*/

#ifdef MALLINFO

/* MALLINFO */ long memory_used()
/* MALLINFO */ {
/* MALLINFO */     struct mallinfo malloc_info;
/* MALLINFO */   
/* MALLINFO */   
/* MALLINFO */     if (fs_mallinfo_call_ok == FALSE) 
/* MALLINFO */     {
/* MALLINFO */         set_error("Heap storage information is not available.");
/* MALLINFO */         return ERROR;
/* MALLINFO */     }
/* MALLINFO */     else 
/* MALLINFO */     {
/* MALLINFO */         malloc_info = mallinfo();
/* MALLINFO */         return malloc_info.arena;
/* MALLINFO */     }
/* MALLINFO */ }

#else 

/* default  */ long memory_used(void)
/* default  */ {
/* default  */
/* default  */  
/* default  */     set_error("Heap storage information is not available.");
/* default  */ 
/* default  */     return ERROR;
/* default  */ }

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                 kjb_malloc
 *                               
 * Allocates memory from the heap. 
 *
 * This routine is a replacement for malloc. It provides some error checking in
 * the case of development code (compiled with -DTEST), some platform
 * independence with respect to malloc's behaviour, and error reporting
 * consistent with the KJB library. 
 *
 * If TRACK_MEMORY_ALLOCATION is defined (on unix, this is normally defined if
 * TEST is #defined) then this routine is #define'd to be debug_kjb_malloc,
 * which is the version available in the development library. In development
 * code, memory is tracked so that memory leaks can be found more easily.
 * Furthermore, all memory free'd is checked that it was allocated by a KJB
 * library routine, and that it has not already been freed (this is a common
 * source of nasty bugs). Finally, memory is checked for overuns. Most of these
 * problems are spotted when the code calls kjb_free. If a problem is spotted,
 * then you are notified, and asked if you want to abort. If you then abort,
 * then you simply need to go up a few stack frames with the debugger to find ut
 * where you were free'ing something that you should not be, or what was freed
 * that was over-run. The memory leak check is simply a report supplied at
 * program exit of the memory not freed, and where it was allocated. Since it is
 * often more useful that the allocation locatation is specified with respect
 * to higher level operations (such as creating a matrix), some routines such as
 * create_matrix will set it up so that the location reported is the matrix
 * creation call, not the call to kjb_malloc. 
 *
 * The routine kjb_free should be used to free memory allocated by this routine.
 * kjb_free is normally accessed through the macro kjb_free. 
 *
 * Macros:
 *    Normally kjb_malloc is accessed through macros to provide the casts. 
 *    The available macros are:
 *
 *        KJB_MALLOC(x), TYPE_MALLOC(t), N_TYPE_MALLOC(t,n), VOID_MALLOC(x),
 *        UCHAR_MALLOC(x), BYTE_MALLOC(x), STR_MALLOC(x), SHORT_MALLOC(x),
 *        INT_MALLOC(x), LONG_MALLOC(x), FLT_MALLOC(x), DBL_MALLOC(x),
 *        INT16_MALLOC(x), UINT16_MALLOC(x), INT32_MALLOC(x), UINT32_MALLOC(x),
 *        DBL_MALLOC(x)
 *
 *    The use of these is relatively obvious, except possibly N_TYPE_MALLOC,
 *    which takes a type name as the first argument, and the number that you
 *    want as the second. 
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the memory, just like malloc.
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

#ifdef TRACK_MEMORY_ALLOCATION
void* debug_kjb_malloc(Malloc_size num_bytes, 
                       const char* file_name, int line_number)
{
    void* malloc_res;
    long  memory_used_result;
    uint  i;

    if (num_bytes == 0) num_bytes = 1;

    if (    (fs_heap_checking_enable)
         || ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
       )
    {
        num_bytes += MALLOC_OVER_RUN_SIZE;
    }

#ifdef MAX_MALLOC_SIZE
    if (num_bytes > MAX_MALLOC_SIZE)
    {
        set_error("Memory allocation of %lu request exceeds %lu bytes.",
                  (ulong)num_bytes, (ulong)MAX_MALLOC_SIZE);
        return NULL;
    }
#endif 

#ifdef TEST
    if (    (fs_simulated_heap_failure_frequency >= 0.0)
         && (fs_simulated_heap_failure_frequency > kjb_rand_2())
       )
    {
        set_error("Memory allocation failed due to simulated failure.");
        return NULL;
    }
#endif 

#ifdef MS_16_BIT_OS
    malloc_res = (void*)farmalloc((ulong)num_bytes); 
#else
    malloc_res = (void*)malloc(num_bytes); 
#endif 

    if (malloc_res == NULL) 
    {
        memory_used_result = memory_used();

        set_error("Memory allocation failed trying to allocate %,lu bytes.",
                  (ulong)num_bytes);

        if (memory_used_result != ERROR)
        {
            add_error("%,ld bytes already allocated.", memory_used_result);
        }
    }
#ifdef TEST
    else if ((int)malloc_res < 256) 
    {
        memory_used_result = memory_used();

        set_error("Memory allocation failed trying to allocate %,lu bytes.", 
                   (ulong)num_bytes); 

        add_error("Unusual return value (%d) from malloc.", 
                  (int)malloc_res); 

        if (memory_used_result != ERROR)
        {
            add_error("%,ld bytes already allocated.", memory_used_result); 
        }

        malloc_res = NULL; 
    }
#endif
#ifdef MALLINFO
    else 
    {   /* Success */
        fs_mallinfo_call_ok = TRUE;
    }
#endif

    if (    (malloc_res != NULL) 
         && (    (fs_heap_checking_enable) 
              || ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
            )
       )
    {
        num_bytes -= MALLOC_OVER_RUN_SIZE;

        for (i=0; i<MALLOC_OVER_RUN_SIZE; i++)
        {
            *((unsigned char*)malloc_res + num_bytes + i) = 0xff;
        }

        if (fs_heap_checking_enable)
        {
            save_pointer_info(malloc_res, num_bytes, file_name, line_number);
        }

        if ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
        {
            save_pointer_info_2(malloc_res, num_bytes, file_name, line_number);
        }
    }

    if (    (malloc_res != NULL)
#ifdef DISABLE_CONDITIONAL_INIT
         && (fs_initialization_checking_enable) 
#endif 
       )
    {
        for (i=0; i<num_bytes; i++)
        {
            *((unsigned char*)malloc_res + i) = 0xff;
        }
    }

    return malloc_res; 
}
        /* ==>                                 /\                             */
        /* ==>    Development code above       ||                             */
#else   /* ------------------------------------------------------------------ */
        /* ==>    Production code below        ||                             */
        /* ==>                                 \/                             */

void* kjb_malloc(Malloc_size num_bytes)
{
    void* malloc_res; 
    long memory_used_result;

    if (num_bytes == 0) num_bytes = 1;

#ifdef MS_16_BIT_OS
    malloc_res = (void*) farmalloc((ulong)num_bytes); 
#else
    malloc_res = (void*) malloc(num_bytes); 
#endif 

    if (malloc_res == NULL) 
    {
        memory_used_result = memory_used();

        set_error("Memory allocation failed trying to allocate %,lu bytes.",
                  (ulong)num_bytes);

        if (memory_used_result != ERROR)
        {
            add_error("%,ld bytes already allocated.", memory_used_result);
        }
    }

#ifdef MALLINFO
    else 
    {      /* Success */
        fs_mallinfo_call_ok = TRUE;
    }
#endif

    return malloc_res; 
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                 kjb_calloc
 *                               
 * Allocates memory from the heap. 
 *
 * This routine is a replacement for calloc. It is similar to kjb_malloc, in
 * anology with calloc. (See kjb_malloc for more information). 
 *
 * The routine kjb_free should be used to free memory allocated by this routine.
 * kjb_free is normally accessed through the macro kjb_free. 
 *
 * Macros:
 *    Normally kjb_calloc is accessed through the macro KJB_CALLOC(x). 
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  On
 *    success it returns a pointer to the memory which has been zero'd, just
 *    like calloc.
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

#ifdef TRACK_MEMORY_ALLOCATION
void* debug_kjb_calloc(Malloc_size num_items, Malloc_size item_size, 
                       const char* file_name, int line_number)
{
    void* calloc_res;
    long  memory_used_result;
    int   num_overun_items = 0;
    uint  i;

    if (num_items == 0) num_items = 1;
    if (item_size == 0) item_size = 1;

    if (    (fs_heap_checking_enable) 
         || ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
       )
    {
        num_overun_items = ((MALLOC_OVER_RUN_SIZE-1)/item_size)+1;
        ASSERT(num_overun_items*item_size >= MALLOC_OVER_RUN_SIZE); 
        num_items += num_overun_items; 
    }

#ifdef MALLOC_SIZE_IS_32_BITS
    if (    (num_items > INT32_MAX) || (item_size > INT32_MAX)
         || (num_items * item_size > INT32_MAX)
       )
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }
#endif

#ifdef MAX_MALLOC_SIZE
    if (num_bytes > MAX_MALLOC_SIZE)
    {
        set_error("Memory allocation of %,lu request exceeds %,lu.",
                  (ulong)num_bytes, (ulong)MAX_MALLOC_SIZE);
        return NULL;
    }
#endif 

#ifdef TEST
    if (    (fs_simulated_heap_failure_frequency >= 0.0)
         && (fs_simulated_heap_failure_frequency > kjb_rand_2())
       )
    {
        set_error("Memory allocation failed due to simulated failure.");
        return NULL;
    }
#endif 

#ifdef MS_16_BIT_OS
    calloc_res = (void*)farcalloc((ulong)num_items, (ulong)item_size); 
#else
    calloc_res = (void*)calloc(num_items, item_size); 
#endif 

    if (calloc_res == NULL) 
    {
        memory_used_result = memory_used();

        set_error("Memory allocation failed trying to allocate %,lu bytes.",
                  (ulong)(num_items * item_size));

        if (memory_used_result != ERROR)
        {
            add_error("%,ld bytes already allocated.", memory_used_result);
        }
    }
#ifdef UNIX
#ifdef TEST
    else if ((int)calloc_res < 256) 
    {
        set_bug("Unusual return value (%d) from calloc.", (int)calloc_res); 
        calloc_res = NULL; 
    }
#endif    /* TEST */
#endif

#ifdef MALLINFO
    else 
    {      /* Success */
        fs_mallinfo_call_ok = TRUE;
    }
#endif

    if (    (calloc_res != NULL)
         && (    (fs_heap_checking_enable) 
              || ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2)) 
            )
       )
    {
        num_items -= num_overun_items;

        for (i=0; i<MALLOC_OVER_RUN_SIZE; i++)
        {
            *((unsigned char*)calloc_res + item_size * num_items + i) = 0xff;
        }

        if (fs_heap_checking_enable)
        {
            save_pointer_info(calloc_res, item_size * num_items, file_name, 
                              line_number);
        }

        if ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
        {
            save_pointer_info_2(calloc_res, item_size * num_items, file_name, 
                                line_number);
        }
    }
    
    return calloc_res; 
}
        /* ==>                                 /\                             */
        /* ==>    Development code above       ||                             */
#else   /* ------------------------------------------------------------------ */
        /* ==>    Production code below        ||                             */
        /* ==>                                 \/                             */

void* kjb_calloc(Malloc_size num_items, Malloc_size item_size)
{
    void* calloc_res; 
    long memory_used_result;

    if (num_items == 0) num_items = 1;
    if (item_size == 0) item_size = 1;

#ifdef MS_16_BIT_OS
    calloc_res = (void*) farcalloc((ulong)num_items, (ulong)item_size); 
#else
    calloc_res = (void*) calloc(num_items, item_size); 
#endif 

    if (calloc_res == NULL) 
    {
        memory_used_result = memory_used();

        set_error("Memory allocation failed trying to allocate %,lu bytes.",
                  (ulong)(num_items * item_size));

        if (memory_used_result != ERROR)
        {
            add_error("%,ld bytes already allocated.", memory_used_result);
        }
    }

#ifdef MALLINFO
    else 
    {   /* Success */
        fs_mallinfo_call_ok = TRUE;
    }
#endif

    return calloc_res; 
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION  ...   #else  ....    */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void save_pointer_info(void* ptr, Malloc_size num_bytes, 
                              const char* file_name, int line_number)
{
    int i;

    fs_fast_lookup_is_valid = FALSE; 

    if (fs_num_pointers == 0)
    {
        /* Double call is OK. */
        arrange_cleanup();

        for (i=0; i<MAX_NUM_POINTERS; i++)
        {
            fs_pointers_to_allocated[ i ] = NULL;
        }
    }
    else if (fs_num_pointers == MAX_NUM_POINTERS) 
    {

        fs_num_pointers = 0;

        for (i=0; i<MAX_NUM_POINTERS; i++)
        {
            if (fs_pointers_to_allocated[ i ] != NULL)
            {
                fs_pointers_to_allocated[ fs_num_pointers ] = 
                                                    fs_pointers_to_allocated[ i ];

                fs_amount_allocated[ fs_num_pointers ] = fs_amount_allocated[ i ];

                BUFF_CPY(fs_allocation_file_name[ fs_num_pointers ], 
                         fs_allocation_file_name[ i ]); 

                fs_allocation_line_number[ fs_num_pointers ] = 
                                                   fs_allocation_line_number[ i ];

                fs_num_pointers++;
            }
        }

#ifdef LOOKS_UNECESSARY
        for (i = fs_num_pointers; i<MAX_NUM_POINTERS; i++)
        {
            fs_pointers_to_allocated[ i ] = NULL;
        }
#endif 

    }

    if (fs_num_pointers < MAX_NUM_POINTERS) 
    {
        /*
        // Who knows, file_name could be const? 
        */
        char file_name_copy[ 200 ];

        fs_pointers_to_allocated[ fs_num_pointers ]  = (void*)ptr;
        fs_amount_allocated[ fs_num_pointers ]       = num_bytes; 
        fs_allocation_line_number[ fs_num_pointers ] = line_number; 

        BUFF_CPY(file_name_copy, file_name);

        BUFF_GEN_GET_LAST_TOKEN(file_name_copy, 
                                fs_allocation_file_name[ fs_num_pointers ], "/");
        fs_num_pointers++;
    }
    else if (fs_num_pointers == MAX_NUM_POINTERS) 
    {
        /*
        // Too dangerous to use KJB IO at this point.
        */ 
        fprintf(stderr, "Max num \"checked\" pointers exceeded.\n");
        fprintf(stderr, "Pointer checking is now disabled.\n");

        /*
        //  Make it so that pointer checking is disabled.
        */
        fs_num_pointers++;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void save_pointer_info_2(void* ptr, Malloc_size num_bytes, 
                                const char* file_name, int line_number)
{
    int i;

    if ((! fs_heap_checking_enable_2) || (fs_heap_checking_skip_2))
    {
        SET_CANT_HAPPEN_BUG();
        return;
    }

    fs_fast_lookup_is_valid_2 = FALSE; 

    if (fs_num_pointers_2 == 0)
    {
        /* Unlike save_pointer_info, not need to arrange cleanup. */
        /*
        arrange_cleanup();
        */

        for (i=0; i<MAX_NUM_POINTERS_2; i++)
        {
            fs_pointers_to_allocated_2[ i ] = NULL;
        }
    }
    else if (fs_num_pointers_2 == MAX_NUM_POINTERS_2) 
    {
/* #ifdef DEF_OUT */
        /*
        // Too dangerous to use KJB IO at this point.
        */ 
        fprintf(stderr, 
                "\nCompacting \"checked\" pointers (2) because table is full ...");
/* #endif  */

        fs_num_pointers_2 = 0;

        for (i=0; i<MAX_NUM_POINTERS_2; i++)
        {
            if (fs_pointers_to_allocated_2[ i ] != NULL)
            {
                fs_pointers_to_allocated_2[ fs_num_pointers_2 ] = 
                                                    fs_pointers_to_allocated_2[ i ];

                fs_amount_allocated_2[ fs_num_pointers_2 ] = fs_amount_allocated_2[ i ];

                BUFF_CPY(fs_allocation_file_name_2[ fs_num_pointers_2 ], 
                         fs_allocation_file_name_2[ i ]); 

                fs_allocation_line_number_2[ fs_num_pointers_2 ] = 
                                                   fs_allocation_line_number_2[ i ];

                fs_num_pointers_2++;
            }
        }

#ifdef LOOKS_UNECESSARY
        for (i = fs_num_pointers_2; i<MAX_NUM_POINTERS_2; i++)
        {
            fs_pointers_to_allocate_2d[ i ] = NULL;
        }
#endif 

/* #ifdef DEF_OUT */
        fprintf(stderr, " done\n\n");
/* #endif  */
    }

    if (fs_num_pointers_2 < MAX_NUM_POINTERS_2) 
    {
        /*
        // Who knows, file_name could be const? 
        */
        char file_name_copy[ 200 ];

        fs_pointers_to_allocated_2[ fs_num_pointers_2 ]  = (void*)ptr;
        fs_amount_allocated_2[ fs_num_pointers_2 ]       = num_bytes; 
        fs_allocation_line_number_2[ fs_num_pointers_2 ] = line_number; 

        BUFF_CPY(file_name_copy, file_name);

        BUFF_GEN_GET_LAST_TOKEN(file_name_copy, 
                                fs_allocation_file_name_2[ fs_num_pointers_2 ], "/");
        fs_num_pointers_2++;
    }
    else if (fs_num_pointers_2 == MAX_NUM_POINTERS_2) 
    {
        /*
        // Too dangerous to use KJB IO at this point.
        */ 
        fprintf(stderr, "Max num \"checked\" pointers exceeded.\n");
        fprintf(stderr, "Pointer checking is now disabled.\n");

        /*
        //  Make it so that pointer checking is disabled.
        */
        fs_num_pointers_2++;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void null_and_float_align(char** ptr_ptr)
{

    while (! FLT_ALIGNED((kjb_uint32)(*ptr_ptr))) 
    {
        **ptr_ptr = '\0'; 
        (*ptr_ptr)++; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void float_align(char** ptr_ptr)
{

    while (! FLT_ALIGNED((kjb_uint32)(*ptr_ptr))) 
    {
        (*ptr_ptr)++; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int print_allocated_memory(FILE* fp)
{
#ifdef TRACK_MEMORY_ALLOCATION
    long  total_bytes, i;
    int   minimal_report = FALSE; 
#endif
    long  heap_memory; 

    /*
    // Special case, normally used at program termination. 
    */
    if (fp == NULL) 
    {
#ifdef TRACK_MEMORY_ALLOCATION
        minimal_report = TRUE;
#endif
        fp = stderr;
    }

#ifdef TRACK_MEMORY_ALLOCATION
    total_bytes = 0;

    /*
    // If fs_num_pointers is greater than MAX_NUM_POINTERS the
    // table is full and heap checking is disabled.
    */
    if (fs_num_pointers <= MAX_NUM_POINTERS)
    {
        for (i=0; i<fs_num_pointers; i++) 
        {
            if (fs_pointers_to_allocated[ i ] != NULL)
            {
                total_bytes += fs_amount_allocated[ i ];
            }
        }
    }

    if (    ! fs_heap_checking_enable 
         || ((total_bytes == 0) && minimal_report)
       )
    {
        return NO_ERROR; 
    }

    /*
    // Use fprint as opposed to kjb_printf to avoid allocating any more 
    // memory or doing anything that may invoke add_cleanup_fn.
    */

    fprintf(fp, "\nResidual memory allocation for process %ld: \n", 
            (long)MY_PID);

    /*
    // If fs_num_pointers is greater than MAX_NUM_POINTERS the
    // table is full and heap checking is disabled.
    */
    if (fs_num_pointers <= MAX_NUM_POINTERS)
    {
        for (i=0; i<fs_num_pointers; i++) 
        {
            if (fs_pointers_to_allocated[ i ] != NULL)
            {
                fprintf(fp, "      %-7lx    %-8lu   %*s %5d\n",
                        (ulong)fs_pointers_to_allocated[i], 
                        (ulong)fs_amount_allocated[ i ], 
                        MAX_ALLOCATION_FILE_NAME_LEN,
                        fs_allocation_file_name[ i ], 
                        fs_allocation_line_number[ i ]);
            }
        }
        fprintf(fp, "    Sum of allocated: %ld\n", total_bytes);
    }

#endif

    if ((heap_memory = memory_used()) != ERROR)
    {
#ifdef TRACK_MEMORY_ALLOCATION
        fprintf(fp, "    ");
#endif
        fprintf(fp, "Heap memory used: %ld\n\n", heap_memory);
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

int start_heap_check_2(const char* file_name, int line_number)
{

    if (fs_heap_checking_enable_2)
    {
        set_error("Heap check 2 is already started.");
        return ERROR; 
    }

    fs_heap_checking_enable_2 = TRUE;
    fs_heap_checking_skip_2 = FALSE; 
    BUFF_CPY(fs_start_heap_check_2_file_name, file_name);
    fs_start_heap_check_2_line_number = line_number;
    fs_num_pointers_2 = 0;

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

void disable_heap_check_2(void)
{

    fs_heap_checking_enable_2 = FALSE;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

int finish_heap_check_2(const char* file_name, int line_number)
{
    long  total_bytes = 0, i;

    if ( ! fs_heap_checking_enable_2)
    {
        set_error("Cannot complete memory check because it is disabled.");
        return ERROR; 
    }

    if (fs_num_pointers_2 > MAX_NUM_POINTERS) 
    {
        set_error("Cannot complete memory check because the table got full.");
        return ERROR; 
    }

    for (i=0; i<fs_num_pointers_2; i++) 
    {
        if (fs_pointers_to_allocated_2[ i ] != NULL)
        {
            total_bytes += fs_amount_allocated_2[ i ];
        }
    }

    if (total_bytes > 0)
    {
        /*
        // Use fprint as opposed to kjb_printf to avoid allocating any more 
        // memory or doing anything that may invoke add_cleanup_fn.
        */

        fprintf(stderr, "\nLeaked\n");  
        fprintf(stderr, "   %s:%d\n", fs_start_heap_check_2_file_name, 
                fs_start_heap_check_2_line_number);
        fprintf(stderr, "   %s:%d\n\n", 
                file_name, line_number);

        /*
        // If fs_num_pointers is greater than MAX_NUM_POINTERS the
        // table is full and heap checking is disabled.
        */
        if (fs_num_pointers_2 <= MAX_NUM_POINTERS_2)
        {
            for (i=0; i<fs_num_pointers_2; i++) 
            {
                if (fs_pointers_to_allocated_2[ i ] != NULL)
                {
                    fprintf(stderr, "      %-7lx    %-8lu   %*s %5d\n",
                            (ulong)fs_pointers_to_allocated_2[i], 
                            (ulong)fs_amount_allocated_2[ i ], 
                            MAX_ALLOCATION_FILE_NAME_LEN,
                            fs_allocation_file_name_2[ i ], 
                            fs_allocation_line_number_2[ i ]);
                }
            }

            fprintf(stderr, "\nSum of allocated: %ld\n\n", total_bytes);
        }
    }

    /* 
    // Reset the table.
    */
    fs_heap_checking_enable_2 = FALSE;
    BUFF_CPY(fs_start_heap_check_2_file_name, "NOT SET"); 
    fs_start_heap_check_2_line_number = NOT_SET;

    return NO_ERROR; 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_allocated_memory(void)
{
#ifdef TRACK_MEMORY_ALLOCATION
    long total_bytes = 0;
    long i;

    /*
    // If fs_num_pointers is greater than MAX_NUM_POINTERS the
    // table is full and heap checking is disabled.
    */
    if (    ( ! fs_heap_checking_enable) 
         || (fs_num_pointers > MAX_NUM_POINTERS)
       )
    {
        return ERROR; 
    }
    else 
    {
        for (i=0; i<fs_num_pointers; i++) 
        {
            if (fs_pointers_to_allocated[ i ] != NULL)
            {
                total_bytes += fs_amount_allocated[ i ];
            }
        }

        return total_bytes; 
    }

#else

    return ERROR; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST 

int debug_print_heap_memory_used(char* file, int line)
{
    long heap_memory; 

    if ((heap_memory = memory_used()) != ERROR)
    {
        test_pso("Heap memory used is %,ld at %s:%d\n", heap_memory, 
                 file, line);
    }

    return NO_ERROR; 
}

#else 

/* 
// Nothing here! In the non-test case, "print_heap_memory_used" is #define'd as
// NULL. 
*/

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  watch_for_free
 *
 * (Debugging) Watches for free of a pointer
 *
 * In development code, this routine checks for an inadvertant free of the
 * pointer argument. The watch is terminated by using a NULL argument. If the
 * pointer is freed before that time, then you are prompted for an abort. To
 * state the obvious, the watch should be terminated before a non-inadvertant
 * free is called. The typical use is :
 *
 * |     allocation of "ptr"
 * |     watch_for_free(ptr);
 * |     code which is suspected of freeing ptr when it should not
 * |     watch_for_free(NULL);
 * |     free of "ptr"
 *
 *
 * In production code, watch_for_free() is defined as a NULL macro, so it
 * disapears. However, normally watch_for_free() is simply inserted for a run,
 * and then removed. 
 *
 * Unfortunately, only one pointer can be watched at present. 
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

#ifdef TRACK_MEMORY_ALLOCATION_OR_C2MAN

void watch_for_free(void* ptr)
{
    fs_pointer_to_watch = ptr;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  check_initialization
 *
 * (Debugging) Checks that memory was initialized
 *
 * In development code, this routine checks for memory which should have been
 * initialized, but was not. It can only be used for memory where consecutive
 * 0xff for block_size bytes is invalid data (as in the case of floats with
 * block_size 4 and and doubles with block_size 8 but not ints and char). If the
 * initialization checking option is not set, then this routine silently does
 * nothing 
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

#ifdef TRACK_MEMORY_ALLOCATION_OR_C2MAN

void check_initialization(const void* ptr,
                          size_t count, /* Number of items of size block_size to check. */
                          size_t block_size /* Size of item in bytes. */
                         )
{
    const uchar* uchar_ptr = (const uchar*)ptr; 
    size_t i, j; 
    int    ok; 

    if ( ! fs_initialization_checking_enable) return; 

    for (i=0; i<count; i++)
    {
        ok = FALSE; 

        for (j=0; j<block_size; j++)
        {
            if (*uchar_ptr != 0xff) 
            {
                ok = TRUE; 
            }
            uchar_ptr++; 
        }

        if (!ok)
        {
            set_bug("Uninitialized memory detected.");
            return; 
        }
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  optimize_free
 *
 * Speeds up frees while heap checking
 *
 * If this routine is called before a code segment which does lots of frees,
 * then heap checking should go faster. 
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

int optimize_free(void)
{
#ifdef TRACK_MEMORY_ALLOCATION
    int i;
    int save_num_pointers = fs_num_pointers;

    if ( ! fs_heap_checking_enable) return NO_ERROR;

    /*
    // If fs_num_pointers is greater than MAX_NUM_POINTERS the
    // table is full and heap checking is disabled.
    */
    if (fs_num_pointers > MAX_NUM_POINTERS)
    {
        return ERROR; 
    }

    if (fs_fast_lookup_is_valid) return NO_ERROR;

    fs_num_pointers = 0;

    for (i=0; i<save_num_pointers; i++)
    {
        if (fs_pointers_to_allocated[ i ] != NULL)
        {
            fs_pointers_to_allocated[ fs_num_pointers ] = fs_pointers_to_allocated[ i ];
            fs_amount_allocated[ fs_num_pointers ] = fs_amount_allocated[ i ];

            BUFF_CPY(fs_allocation_file_name[ fs_num_pointers ], 
                     fs_allocation_file_name[ i ]); 

            fs_allocation_line_number[ fs_num_pointers ] = 
                                               fs_allocation_line_number[ i ];

            fs_num_pointers++;
        }
    }

    for (i = 0; i < fs_num_pointers; i++)
    {
        /* CONSTCOND */
        if (sizeof(void*) != sizeof(int))
        {
            /*
            // If the above assumption is false (and it will become false on
            // many machines soon), then we may want to set up hash function
            // which ignores the last couple of bytes, and other which have
            // proven to be not useful. 
            */
            fprintf(stderr, 
                    "\nsizeof(void*) != sizeof(int) in optimize_free. ");
            fprintf(stderr, "Likely some coding adjustments are in order.\n\n");
        }

        /* Clearly a bit dodgy! */
        fs_sorted_pointers[ i ].ptr = (int)fs_pointers_to_allocated[ i ];
        fs_sorted_pointers[ i ].index = i; 
    }

    ERE(int_sort(fs_sorted_pointers, 
                 fs_num_pointers,
                 sizeof(fs_sorted_pointers[ 0 ]),
                 (size_t)0, 
                 USE_CURRENT_ATN_HANDLING));

    fs_fast_lookup_is_valid = TRUE;

    return NO_ERROR;

#else

    return ERROR;

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  optimize_free_2
 *
 * Speeds up frees while heap checking
 *
 * This is like optimize_free, but is used in conjunction with
 * start_heap_checking_2 and finish_heap_checking_2. 
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

int optimize_free_2(void)
{
#ifdef TRACK_MEMORY_ALLOCATION
    int i;
    int save_num_pointers = fs_num_pointers_2;

    if ( ! fs_heap_checking_enable_2) return NO_ERROR;

    /*
    // If fs_num_pointers_2 is greater than MAX_NUM_POINTERS_2 the
    // table is full and heap checking is disabled.
    */
    if (fs_num_pointers_2 > MAX_NUM_POINTERS_2)
    {
        return ERROR; 
    }

    if (fs_fast_lookup_is_valid_2) return NO_ERROR;

    fs_num_pointers_2 = 0;

    for (i=0; i<save_num_pointers; i++)
    {
        if (fs_pointers_to_allocated_2[ i ] != NULL)
        {
            fs_pointers_to_allocated_2[ fs_num_pointers_2 ] = fs_pointers_to_allocated_2[ i ];
            fs_amount_allocated_2[ fs_num_pointers_2 ] = fs_amount_allocated_2[ i ];

            BUFF_CPY(fs_allocation_file_name_2[ fs_num_pointers_2 ], 
                     fs_allocation_file_name_2[ i ]); 

            fs_allocation_line_number_2[ fs_num_pointers_2 ] = 
                                               fs_allocation_line_number_2[ i ];

            fs_num_pointers_2++;
        }
    }

    for (i = 0; i < fs_num_pointers_2; i++)
    {
        /* CONSTCOND */
        if (sizeof(void*) != sizeof(int))
        {
            /*
            // If the above assumption is false (and it will become false on
            // many machines soon), then we may want to set up hash function
            // which ignores the last couple of bytes, and other which have
            // proven to be not useful. 
            */
            fprintf(stderr, 
                    "\nsizeof(void*) != sizeof(int) in optimize_free. ");
            fprintf(stderr, "Likely some coding adjustments are in order.\n\n");
        }

        /* Clearly a bit dodgy! */
        fs_sorted_pointers_2[ i ].ptr = (int)fs_pointers_to_allocated_2[ i ];
        fs_sorted_pointers_2[ i ].index = i; 
    }

    ERE(int_sort(fs_sorted_pointers_2, 
                 fs_num_pointers_2,
                 sizeof(fs_sorted_pointers_2[ 0 ]),
                 (size_t)0, 
                 USE_CURRENT_ATN_HANDLING));

    fs_fast_lookup_is_valid_2 = TRUE;

    return NO_ERROR;

#else

    return ERROR;

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  kjb_free
 *
 * KJB library replacement for free
 *
 * This routine frees memory allocated by kjb_malloc or kjb_calloc, and in
 * conjunction with those routines provides error checking in the case of
 * development code as described in kjb_malloc. One important additional
 * difference between kjb_free and "free", is that it is always safe to send
 * kjb_free a NULL pointer. (In fact, it should be save to send all KJB library
 * destruction code a NULL pointer.) It is OK to send free a NULL pointer in
 * conformaing ANSI C, but many non-ANSI C environments break if you send free a
 * NULL. Using kjb_free is one way to be sure. 
 *
 * Index: memory allocation, debugging
 *
 * -----------------------------------------------------------------------------
 */

void kjb_free(void* ptr)
{
#ifdef TRACK_MEMORY_ALLOCATION
    static int prompt_for_abort = TRUE;
    int        ptr_index;
    int        i;
#endif 

    if (ptr == NULL) return; 

#ifdef TRACK_MEMORY_ALLOCATION
    if (    (    (fs_heap_checking_enable) 
              || ((fs_heap_checking_enable_2) && (! fs_heap_checking_skip_2))
            )
         && (ptr == fs_pointer_to_watch)
       )
    {
        if (prompt_for_abort)
        {
            IMPORT volatile int halt_all_output;
            IMPORT volatile int halt_term_output;

            halt_all_output = FALSE;
            halt_term_output = FALSE;

            fprintf(stderr, 
                    "Freeing of pointer being watched for free.\n");
        }

        if ((is_interactive()) && (prompt_for_abort)) 
        {
            static int guard = FALSE;

            /*
               Since yes_no_query allocates storage, make sure we do not
               loop.
            */

            if (guard)
            {
                fprintf(stderr, "Abort called (regardless of response) ");
                fprintf(stderr, "due to mutex failure.\n");
                kjb_abort();
            }
                    
            guard = TRUE;

            if (yes_no_query("Abort (y/n) ") ) 
            {
                kjb_abort();
            }
            else if ( ! (yes_no_query(
                            "Prompt for abort on future violations (y/n) ")
                        )
                    )
            {
                prompt_for_abort = FALSE;
            }

            guard = FALSE;
        }
    }

    /*
    // If fs_num_pointers is greater than MAX_NUM_POINTERS the
    // table is full and heap checking is disabled.
    */
    if (fs_heap_checking_enable && (fs_num_pointers <= MAX_NUM_POINTERS))
    {
        int pointer_overun = FALSE;

        if (fs_fast_lookup_is_valid)
        {
            int ptr_sorted_index = int_binary_search(fs_sorted_pointers, 
                                                     fs_num_pointers, 
                                                 sizeof(fs_sorted_pointers[ 0 ]),
                                                     offsetof(Sorted_pointers,
                                                              ptr),
                                                     (int)ptr);

            if (ptr_sorted_index == NOT_FOUND)
            {
                ptr_index = NOT_SET;
            }
            else 
            {
                ptr_index = fs_sorted_pointers[ ptr_sorted_index ].index;
                
                if (fs_pointers_to_allocated[ ptr_index ] == NULL)
                {
                    ptr_index = NOT_SET;
                }
                else 
                {
                    ASSERT(fs_pointers_to_allocated[ ptr_index ] == ptr); 
                }
            }
        }
        else
        {
            ptr_index = NOT_SET;

            /*
            // We are more likely to have allocated it recently, so search
            // backwards.
            */
            for (i = fs_num_pointers - 1; i >= 0; i--) 
            {
                if (ptr == fs_pointers_to_allocated[ i ]) 
                {
                    ptr_index = i;
                    break;
                }
             }
        }

        if (ptr_index != NOT_SET) 
        {
            unsigned char* temp_ptr = (unsigned char*)ptr + 
                                                  fs_amount_allocated[ ptr_index ];

            for (i=0; i<MALLOC_OVER_RUN_SIZE; i++)
            {
                if (*temp_ptr != 0xff)
                {
                    pointer_overun = TRUE;
                    break;
                }
                temp_ptr++;
            }
        }

        if ((ptr_index == NOT_SET) || pointer_overun)
        {
            if (prompt_for_abort)
            {
                IMPORT volatile int halt_all_output;
                IMPORT volatile int halt_term_output;

                halt_all_output = FALSE;
                halt_term_output = FALSE;

                fprintf(stderr, 
                        "Heap checker reporting on process %ld.\n",
                        (long)MY_PID);

                if (ptr_index == NOT_SET) 
                {
                    fprintf(stderr, 
                            "Freeing a ptr not allocated / already been freed.\n"); 
                }
                else
                {
                    fprintf(stderr, 
                            "Freeing storage which was overun.\n");
                    fprintf(stderr, "Allocated on line %d of file %s.\n",
                            fs_allocation_line_number[ ptr_index ],
                            fs_allocation_file_name[ ptr_index ]);
                    fprintf(stderr, "Allocation size is %u\n", 
                            fs_amount_allocated[ ptr_index ]);
                    fprintf(stderr, "Pointer in question is %p.\n", ptr);
                    fprintf(stderr, "Location of over-run is %p.\n", 
                                    (void*)((unsigned char*)ptr + 
                                              fs_amount_allocated[ ptr_index ]));
                }
            }

            if ((is_interactive()) && (prompt_for_abort)) 
            {
                static int guard = FALSE;

                /*
                   Since yes_no_query allocates storage, make sure we do not
                   loop.
                */

                if (guard)
                {
                    p_stderr("Abort called (regardless of response) ");
                    p_stderr("due to mutex failure.\n");
                    kjb_abort();
                }
                        
                guard = TRUE;

                if (yes_no_query("Abort (y/n) ") ) 
                {
                    kjb_abort();
                }
                else if ( ! (yes_no_query(
                                "Prompt for abort on future violations (y/n) ")
                            )
                        )
                {
                    prompt_for_abort = FALSE;
                }

                guard = FALSE;
            }

            return;
        }
        else 
        {
            fs_pointers_to_allocated[ ptr_index ] = NULL;

            if (! fs_fast_lookup_is_valid)
            {
                while (    (fs_num_pointers > 0)
                        && (fs_pointers_to_allocated[ fs_num_pointers - 1 ] == NULL)
                      )
                {
                    fs_num_pointers--;
                }
            }
        }
    }

    /*
    // If fs_num_pointers_2 is greater than MAX_NUM_POINTERS_2 the
    // table is full and heap checking is disabled.
    */
    if (    (fs_heap_checking_enable_2) 
         && (fs_num_pointers_2 <= MAX_NUM_POINTERS_2)
         && ( ! fs_heap_checking_skip_2)
       )
    {
        int pointer_overun = FALSE;

        if (fs_fast_lookup_is_valid_2)
        {
            int ptr_sorted_index = int_binary_search(fs_sorted_pointers_2, 
                                                     fs_num_pointers_2, 
                                                 sizeof(fs_sorted_pointers_2[ 0 ]),
                                                     offsetof(Sorted_pointers,
                                                              ptr),
                                                     (int)ptr);

            if (ptr_sorted_index == NOT_FOUND)
            {
                ptr_index = NOT_SET;
            }
            else 
            {
                ptr_index = fs_sorted_pointers_2[ ptr_sorted_index ].index;
                
                if (fs_pointers_to_allocated_2[ ptr_index ] == NULL)
                {
                    ptr_index = NOT_SET;
                }
                else 
                {
                    ASSERT(fs_pointers_to_allocated_2[ ptr_index ] == ptr); 
                }
            }
        }
        else
        {
            ptr_index = NOT_SET;

            /*
            // We are more likely to have allocated it recently, so search
            // backwards.
            */
            for (i = fs_num_pointers_2 - 1; i >= 0; i--) 
            {
                if (ptr == fs_pointers_to_allocated_2[ i ]) 
                {
                    ptr_index = i;
                    break;
                }
             }
        }

        if (ptr_index != NOT_SET) 
        {
            unsigned char* temp_ptr = (unsigned char*)ptr + 
                                                  fs_amount_allocated_2[ ptr_index ];

            for (i=0; i<MALLOC_OVER_RUN_SIZE; i++)
            {
                if (*temp_ptr != 0xff)
                {
                    pointer_overun = TRUE;
                    break;
                }
                temp_ptr++;
            }
        }

        if ((ptr_index == NOT_SET) || pointer_overun)
        {
            if (prompt_for_abort)
            {
                IMPORT volatile int halt_all_output;
                IMPORT volatile int halt_term_output;

                halt_all_output = FALSE;
                halt_term_output = FALSE;

                fprintf(stderr, 
                        "Heap check 2 reporting on process %ld.\n",
                        (long)MY_PID);

                if (ptr_index == NOT_SET) 
                {
                    fprintf(stderr, 
                            "Freeing a ptr which was not allocated or has already been freed.\n");
                }
                else
                {
                    fprintf(stderr, 
                            "Freeing storage which was overun.\n");
                    fprintf(stderr, "Allocated on line %d of file %s.\n",
                            fs_allocation_line_number_2[ ptr_index ],
                            fs_allocation_file_name_2[ ptr_index ]);
                    fprintf(stderr, "Allocation size is %u\n", 
                            fs_amount_allocated_2[ ptr_index ]);
                    fprintf(stderr, "Pointer in question is %p.\n", ptr);
                    fprintf(stderr, "Location of over-run is %p.\n", 
                                    (void*)((unsigned char*)ptr + 
                                              fs_amount_allocated_2[ ptr_index ]));
                }
            }

            if ((is_interactive()) && (prompt_for_abort)) 
            {
                static int guard = FALSE;

                /*
                   Since yes_no_query allocates storage, make sure we do not
                   loop.
                */

                if (guard)
                {
                    fprintf(stderr, "Abort called (regardless of response) ");
                    fprintf(stderr, "due to mutex failure.\n");
                    kjb_abort();
                }
                        
                guard = TRUE;

                if (yes_no_query("Abort (y/n) ") ) 
                {
                    kjb_abort();
                }
                else if ( ! (yes_no_query(
                                "Prompt for abort on future violations (y/n) ")
                            )
                        )
                {
                    prompt_for_abort = FALSE;
                }

                guard = FALSE;
            }

            return;
        }
        else 
        {
            fs_pointers_to_allocated_2[ ptr_index ] = NULL;

            if (! fs_fast_lookup_is_valid_2)
            {
                while (    (fs_num_pointers_2 > 0)
                        && (fs_pointers_to_allocated_2[ fs_num_pointers_2 - 1 ] == NULL)
                      )
                {
                    fs_num_pointers_2--;
                }
            }
        }
    }

#endif

#ifdef __STDC__
    free(ptr);
#else
    free((char*)ptr);
#endif

}

