
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_SYS_MAL_INCLUDED 
#define L_SYS_MAL_INCLUDED 

#include "l/l_def.h"

#ifdef __C2MAN__
#ifndef TRACK_MEMORY_ALLOCATION_OR_C2MAN
#define TRACK_MEMORY_ALLOCATION_OR_C2MAN
#endif
#endif

#ifdef TRACK_MEMORY_ALLOCATION
#ifndef TRACK_MEMORY_ALLOCATION_OR_C2MAN
#define TRACK_MEMORY_ALLOCATION_OR_C2MAN
#endif
#endif

#ifdef TRACK_MEMORY_ALLOCATION
#    define kjb_malloc(x)   \
                       debug_kjb_malloc((Malloc_size)(x), __FILE__ , __LINE__ )

#    define kjb_calloc(x, y)  \
                debug_kjb_calloc((Malloc_size)(x), (Malloc_size)(y), __FILE__ ,\
                                 __LINE__ )

/* 
 * The following are used only for writing debug versions of constructor
 * functions. Normally we use the non-debug versions regardless of whether
 * TRACK_MEMORY_ALLOCATION is defined. 
 *
 * The macro variables x and y invariably are file_name and line_number (perhaps
 * I should call them that?). We could save some typing by having the MACROS
 * insert "file_name" and "line_number", but I think this is being too clever.
*/

#define DEBUG_KJB_MALLOC(x, y, z)       debug_kjb_malloc((Malloc_size)(x), y, z)

#define DEBUG_KJB_CALLOC(u, v, y, z) \
                    debug_kjb_calloc((Malloc_size)(u), (Malloc_size)(v), y, z)

#define DEBUG_TYPE_MALLOC(t, y, z)      (t *)debug_kjb_malloc(sizeof(t), y, z)

#define DEBUG_N_TYPE_MALLOC(t, n, y, z) \
              (t *)debug_kjb_malloc((sizeof(t))*((kjb_int32)(n)), y, z)

#define DEBUG_UCHAR_MALLOC(x, y, z)  \
            (uchar*)debug_kjb_malloc((Malloc_size)(x), y, z)

#define DEBUG_BYTE_MALLOC(x, y, z) \
            (Byte*)debug_kjb_malloc((Malloc_size)(x), y, z)

#define DEBUG_STR_MALLOC(x, y, z) \
            (char*)debug_kjb_malloc((Malloc_size)(x), y, z)

#define DEBUG_SHORT_MALLOC(x, y, z) \
          (short*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(short)), y, z)

#define DEBUG_INT_MALLOC(x, y, z)    \
          (int*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(int)), y, z)

#define DEBUG_UINT_MALLOC(x, y, z)    \
          (uint*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(uint)), y, z)

#define DEBUG_LONG_MALLOC(x, y, z)   \
          (long*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(Malloc_size)), \
                                  y, z)

#define FLT_DEBUG_MALLOC(x, y, z)  \
          (float*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(float)), y, z)

#define DBL_DEBUG_MALLOC(x, y, z) \
          (double*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(double)), y, z)

#define DEBUG_INT16_MALLOC(x, y, z)  \
          (kjb_int16*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_int16)), y, z)

#define DEBUG_UINT16_MALLOC(x, y, z) \
          (kjb_uint16*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_uint16)), y, z)

#define DEBUG_INT32_MALLOC(x, y, z)  \
          (kjb_int32*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(Malloc_size)), \
                                   y, z)
#define DEBUG_UINT32_MALLOC(x, y, z) \
          (kjb_uint32*)debug_kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_uint32)), y, z)

#define START_HEAP_CHECK_2()  start_heap_check_2(__FILE__, __LINE__)
#define FINISH_HEAP_CHECK_2() finish_heap_check_2(__FILE__, __LINE__)

#define SKIP_HEAP_CHECK_2     skip_heap_check_2
#define CONTINUE_HEAP_CHECK_2 continue_heap_check_2
#define DISABLE_HEAP_CHECK_2  disable_heap_check_2
#define DISABLE_HEAP_CHECK_FOR_FORK  disable_heap_checking_for_fork

#else 

#define START_HEAP_CHECK_2()  
#define FINISH_HEAP_CHECK_2() 

#define SKIP_HEAP_CHECK_2() 
#define CONTINUE_HEAP_CHECK_2()
#define DISABLE_HEAP_CHECK_2() 
#define DISABLE_HEAP_CHECK_FOR_FORK() 

#endif    /*  TRACK_MEMORY_ALLOCATION  */

#define KJB_MALLOC(x)       kjb_malloc((Malloc_size)(x))
#define KJB_CALLOC(x,y)     kjb_calloc((Malloc_size)(x), (Malloc_size)(y))

#define TYPE_MALLOC(t)      (t *)kjb_malloc(sizeof(t))
#define N_TYPE_MALLOC(t, n) (t *)kjb_malloc((sizeof(t))*((kjb_int32)(n)))

#define UCHAR_MALLOC(x)     (uchar*)kjb_malloc((Malloc_size)(x))
#define BYTE_MALLOC(x)      (uchar*)kjb_malloc((Malloc_size)(x))
#define STR_MALLOC(x)       (char*)kjb_malloc((Malloc_size)(x))

#define SHORT_MALLOC(x) \
              (short*)kjb_malloc(((Malloc_size)(x))*(sizeof(short)))

#define INT_MALLOC(x)    \
              (int*)kjb_malloc(((Malloc_size)(x))*(sizeof(int)))

#define UINT_MALLOC(x)    \
              (uint*)kjb_malloc(((Malloc_size)(x))*(sizeof(uint)))

#define LONG_MALLOC(x)   \
              (long*)kjb_malloc(((Malloc_size)(x))*(sizeof(Malloc_size)))

#define FLT_MALLOC(x)  \
              (float*)kjb_malloc(((Malloc_size)(x))*(sizeof(float)))

#define DBL_MALLOC(x) \
              (double*)kjb_malloc(((Malloc_size)(x))*(sizeof(double)))

#define INT16_MALLOC(x)  \
              (kjb_int16*)kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_int16)))
#define UINT16_MALLOC(x) \
              (kjb_uint16*)kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_uint16)))
#define INT32_MALLOC(x)  \
              (kjb_int32*)kjb_malloc(((Malloc_size)(x))*(sizeof(Malloc_size)))
#define UINT32_MALLOC(x) \
              (kjb_uint32*)kjb_malloc(((Malloc_size)(x))*(sizeof(kjb_uint32)))

int set_heap_options(const char* option, const char* value);

long memory_used(void);

#ifdef TRACK_MEMORY_ALLOCATION
    void* debug_kjb_malloc(Malloc_size, const char*, int);
    void* debug_kjb_calloc
    (
        Malloc_size ,
        Malloc_size ,
        const char* ,
        int          
    );
#else
    void* kjb_calloc(Malloc_size, Malloc_size);
    void* kjb_malloc(Malloc_size);
#endif

void float_align(char**);
void null_and_float_align(char**);

#ifdef TRACK_MEMORY_ALLOCATION_OR_C2MAN
    void watch_for_free(void* ptr);
    void check_initialization
    (
        const void* ptr,
        size_t      count,
        size_t      block_size 
    );
#else 
#   define watch_for_free(x)
#   define check_initialization(x, y, z)
#endif 

int optimize_free(void);
void kjb_free(void*);
int print_allocated_memory(FILE*);
int get_allocated_memory(void);

#ifdef TEST
#define print_heap_memory_used()  \
        debug_print_heap_memory_used(__FILE__, __LINE__)

int debug_print_heap_memory_used(char* file, int line);
#else 
#define print_heap_memory_used() 
#endif 

int optimize_free_2(void);

#ifdef TRACK_MEMORY_ALLOCATION

int  start_heap_check_2            
(
    const char* file_name,
    int         line_number 
);

int  finish_heap_check_2           
(
    const char* file_name,
    int         line_number 
);
void skip_heap_check_2             (void);
void continue_heap_check_2         (void);
void disable_heap_check_2          (void);
int  disable_heap_checking_for_fork(void);

#endif 

#endif 

