
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifdef MAKE_DEPEND
#    ifndef L_SYS_STANDARD_INCLUDED
#        define L_SYS_STANDARD_INCLUDED
#    endif
#endif 

#ifndef L_SYS_STANDARD_INCLUDED 
#define L_SYS_STANDARD_INCLUDED 

/* The file l/l_sys_sys.h canonicalizes some definitions. It does not include
// any system includes, so we can still set up for including system includes
// after this point.
*/
#include "l/l_sys_sys.h"  

#ifdef __GNUC__

/*
// Gcc does some fancy optimizations inline on some string functions like
// "strcpy". At a glance, however, it does not necessarily look like an
// improvment in the case of strcpy (bench marking would be interesting), and
// causes the compiler to emit warnings. Since do not normally spend a lot of
// CPU doing string stuff, we disable this behaviour (only active when
// optimizing) with the following define. 
*/
#define __NO_STRING_INLINES

#ifdef __C2MAN__

#define _ANSI_STDARG_H_
#define  __restrict
#define _G_va_list void*
#define __gnuc_va_list void*
/*
#define __builtin_va_list void*
#define _STDARG_H
*/

#endif 
#endif 

/*
// --- HPUX --- HPUX --- HPUX --- HPUX --- HPUX --- HPUX --- HPUX --- HPUX --- 
// 
// FIX -- Try and do without this hack ASAP.
// 
// Prepare gcc include of non-gcc header in the case of HPUX. What happens is
// that gcc uses its own type.h which does not define blkcnt_t, off32_t, etc.
// I'm not sure if this is a problem with how we installed gcc, or if they just
// blew it. Nonetheless, there is no <sys/stat.h> in the gcc include directory,
// and so the <sys/stat.h> directory below uses /usr/include/sys/stat.h. The
// gcc include dir is:
// /opt/gcc/lib/gcc-lib/hppa1.1-hp-hpux10.10/2.7.2.1/include Perhaps the
// problem is that we are using hp-10.20, and the gcc we have is for 10.10.
// Another way to fix this problem is to use -I/usr/include but this means all
// users of this code will have to remember to do this, so I don't think it is
// a good idea. 
*/
#ifdef HPUX
#ifdef __GNUC__
#  include <sys/sigevent.h>

#  include <inttypes.h>

#  ifndef _BLKCNT_T
#    define _BLKCNT_T
#      ifdef _FILE64
         typedef int64_t blkcnt_t;        /* 64-bit # of blocks */
#      else
         typedef int32_t blkcnt_t;        /* 32-bit # of blocks */
#    endif
#  endif /* _BLKCNT_T */

#  ifndef _OFF32_T
#    define _OFF32_T
        typedef int32_t off32_t;	/* 32bit offsets and sizes */
#  endif /* _OFF32_T */
#endif
#endif

#ifdef SUN4
#ifdef __GNUC__
#define __USE_FIXED_PROTOTYPES__
#endif
#endif

#ifdef STANDARD_VAR_ARGS
#    include <stdarg.h>
#else
#    include <varargs.h>
#endif

#include <sys/types.h> 

#ifdef SUN5
#if __STDC__ - 0 == 1
/*
// Maximally conforming ANSI does not have long long, and soe 
// <sys/types.h> defines longlong_t and u_longlong_t in order to support ANSI
// level 1 as well as 0. However, int64_t and uint64_t do not get defined. This
// is a problem, because some include files do not prepare for this case. As of
// 13/07/99, the only problematic file required is <sys/kstat.h> called from
// <sys/vm.h> called from <sys/swap.h> called from Mike's videograb code
// (actually, code that Parallax suppled.) Likely, once that stuff is turfed,
// the following lines can also be turfed. 
*/
typedef longlong_t int64_t;
typedef u_longlong_t uint64_t;
#endif 
#endif 

#include <ctype.h> 
#include <stdio.h> 

#include <string.h> 
#include <math.h> 
#include <stdlib.h> 
#include <errno.h> 
#include <limits.h> 

#ifdef __STDC__
#    include <float.h> 
#endif 

/*  ---------------------------------- Unix --------------------------------- */
#ifdef UNIX_SYSTEM

#ifdef NEXT
#    include <libc.h>
#else
#    include <unistd.h>
#endif

/*  ---------------------------------- Unix - MAC_OSX  ------------------ */

#ifdef  MAC_OSX
#    include <sys/param.h>
#endif 

/*  ---------------------------------- Unix - LINUX_386  ------------------ */

#ifdef  LINUX_386
#    include <stdint.h>
#    include <stddef.h> 
#endif 

/*  ---------------------------------- Unix - DEBIAN ----------------------- */

#ifdef  DEBIAN_386
#    include <ioctls.h>
#    include <ioctl-types.h>
#endif 

/*  ---------------------------------- Unix - Solaris ----------------------- */

#ifdef  SUN5
#    include <stddef.h>  
#    include <ieeefp.h>
#endif 

/*  ---------------------------------- Unix - HPUX -------------------------- */

#ifdef  HPUX
#endif 

/*  ---------------------------------- Unix - Sun --------------------------- */

#ifdef  SUN4
#endif 

/*  ---------------------------------- Unix - NeXT -------------------------- */
#ifdef  NEXT
#endif 

/*  ---------------------------------- Unix - Sgi  -------------------------- */

#ifdef  SGI
#endif 

/*  ---------------------------------- Unix - Sysv -------------------------- */

#ifdef SYSV_SYSTEM
#endif 

/*  ------------------------------------------------------------------------- */
#endif                             /* End of unix */
/*  ------------------------------------------------------------------------- */

/*  -------------------------------- MS_16_BIT_OS --------------------------- */
#ifdef  MS_16_BIT_OS 

#include <sys\types.h> 
#include <sys\stat.h> 
#include <io.h> 

#endif   /* End of MS_16_BIT_OS */
/*  ------------------------------------------------------------------------- */

/*  -------------------------------------- VMS ------------------------------ */
#ifdef VMS 

#include <unixlib.h>
#include <stat.h> 
#include <iodef.h> 
#include <descrip.h> 

#endif  /* End of VMS */
/*  --------------------------------------------------------------------- */

/* We want to make it so that everybody defines the __sparcv8 symbol the same
// way to avoid messages about it being different. 
*/
#ifdef FAST_INCLUDE_PP
#    ifdef __sparcv8
#        undef __sparcv8
#        define __sparcv8 1
#    endif
#endif 

#endif    /* Include this file. */

