
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef M_MAT_BASIC_INCLUDED 
#define M_MAT_BASIC_INCLUDED 

#include "m/m_gen.h"

Matrix* create_matrix_copy(const Matrix* source_mp);

#ifdef TRACK_MEMORY_ALLOCATION

#   define copy_matrix(x,y)  debug_copy_matrix(x,y, __FILE__, __LINE__)
 
    int debug_copy_matrix
    (
        Matrix**      target_mpp,
        const Matrix* source_mp,
        const char*   file_name,
        int           line_number 
    );
#else 
    int copy_matrix(Matrix** target_mpp, const Matrix* source_mp);
#endif 

int copy_matrix_block
(
    Matrix**      target_mpp,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           num_rows,
    int           num_cols 
);

int ow_copy_matrix_block
(
    Matrix*       target_mp,
    int           target_row_offset,
    int           target_col_offset,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           num_rows,
    int           num_cols 
);

int copy_matrix_block_2
(
    Matrix**      target_mpp,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           source_row_step,
    int           source_col_step,
    int           num_rows,
    int           num_cols 
);

int ow_copy_matrix_block_2
(
    Matrix*       target_mp,
    int           target_row_offset,
    int           target_col_offset,
    int           target_row_step,
    int           target_col_step,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           source_row_step,
    int           source_col_step,
    int           num_rows,
    int           num_cols 
);

int ow_copy_matrix
(
    Matrix*       target_mp,
    int           target_row_offset,
    int           target_col_offset,
    const Matrix* source_mp 
);

int copy_matrix_data
(
    Matrix**      target_mpp,
    const Matrix* mp,
    const int*    invalid_data 
);

int copy_int_matrix_to_matrix
(
    Matrix**          target_mpp,
    const Int_matrix* source_mp 
);

int copy_matrix_to_int_matrix
(
    Int_matrix**  target_mpp,
    const Matrix* source_mp 
);

int copy_matrix_vector(Matrix_vector**, const Matrix_vector*);

int concat_matrix_vectors
(
    Matrix_vector**      mvpp,
    int                  num_vectors,
    const Matrix_vector* matrix_vectors[] 
);

int get_random_matrix_row(Vector** vpp, const Matrix* mp);

int mp_row_get_indexed_vector
(
    Indexed_vector** vpp,
    Matrix*          init_mp,
    int              row 
);

int get_matrix_row(Vector** vpp, const Matrix* mp, int row_num);

int remove_matrix_row(Vector** vpp, Matrix* mp, int row_num);

int put_matrix_row(Matrix* mp, const Vector* vp, int row_num);

int copy_matrix_row
(
    Matrix*       target_mp,
    int           target_row_num,
    const Matrix* source_mp,
    int           source_row_num 
);

int get_random_matrix_col(Vector** vpp, const Matrix* mp);

int get_matrix_col(Vector** vpp, const Matrix* mp, int col_num);

int remove_matrix_col(Vector** vpp, Matrix* mp, int col_num);

int put_matrix_col(Matrix* mp, const Vector* vp, int col_num);

int swap_matrix_rows(Matrix*, int, int);
int swap_matrix_cols(Matrix*, int, int);

int select_matrix_data
(
    Matrix**      selected_mpp,
    const Matrix* data_mp,
    int           num_data_points,
    const int*    data_point_nums,
    const int*    invalid_data 
);

int vector_is_matrix_row(const Matrix* mp, const Vector* vp);

#ifdef TEST

#define verify_matrix_vector(x) debug_verify_matrix_vector(x,__FILE__,__LINE__)
#define verify_matrix(x)        debug_verify_matrix(x, __FILE__, __LINE__)

void debug_verify_matrix_vector
(
    const Matrix_vector* mvp,
    const char*          file_name,
    int                  line_number 
);
void debug_verify_matrix
(
    const Matrix* mp,
    const char*   file_name,
    int           line_number 
);

#else 

#define verify_matrix_vector(x)
#define verify_matrix(x)

#endif 

#endif 

