
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef M_STAT_INCLUDED 
#define M_STAT_INCLUDED 

#include "m/m_gen.h"

/* -------------------------------------------------------------------------- */

typedef struct Stat 
{   
    int n; 
    double mean;
    double stdev;
    double min;
    double max;
}   
Stat;  

/* -------------------------------------------------------------------------- */

int add_data_point(double);
Stat* get_data_stats(void);
int clear_data_stats(void);
void cleanup_data_stats(void);

int get_correlation_coefficient
(
    Vector* first_vp,
    Vector* second_vp,
    double* result_ptr 
);

int gauss_convolve_matrix
(
    Matrix**      out_mpp,
    const Matrix* in_mp,
    double        sigma 
);

int get_log_gaussian_density
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       log_prob_ptr 
);

int get_gaussian_density
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       prob_ptr 
);

int get_malhalanobis_distance
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       dist_ptr 
);

int get_malhalanobis_distance_sqrd
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       dist_sqrd_ptr 
);

int get_log_gaussian_density_2
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       log_prob_ptr 
);

int get_gaussian_density_2
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       prob_ptr 
);

int get_malhalanobis_distance_2
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       dist_ptr 
);

int get_malhalanobis_distance_sqrd_2
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       dist_sqrd_ptr 
);

int get_log_gaussian_density_3
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       log_prob_ptr 
);

int get_gaussian_density_3
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       prob_ptr 
);

int get_malhalanobis_distance_3
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       dist_ptr 
);

int get_malhalanobis_distance_sqrd_3
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double        max_norm_var,
    double*       dist_sqrd_ptr 
);

int compute_mean_and_stdev
(
    double* mean_ptr,
    double* stdev_ptr,
    int     num_data_points,
    double  data_sum,
    double  data_sqrd_sum 
);

int compute_weighted_mean_and_stdev
(
    double* mean_ptr,
    double* stdev_ptr,
    double  total_weight,
    double  data_sum,
    double  data_sqrd_sum 
);

#endif 

