
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef M_VECTOR_INCLUDED
#define M_VECTOR_INCLUDED

#include "l/l_gen.h"

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                                   Vector 
 * 
 * Vector type
 *
 * This type is the vector type for the KJB library and is used by many
 * routines. 
 *
 * See also:
 *    create_vector, free_vector, get_target_vector
 *
 * Acknowlegement:
 *    This type grew out of one written by Ian Harder. However, it is important
 *    to note that this type is somewhat different than Ian's Matrix type. 
 *
 * Index: vectors
 *
 * -----------------------------------------------------------------------------
*/

typedef struct Vector 
{   
    int   length;
    double* elements;
}   
Vector;

typedef struct Vector_vector
{   
    int      length;
    Vector** elements;
}   
Vector_vector;

typedef struct V_v_v
{   
    int             length;
    Vector_vector** elements;
}   
V_v_v;

typedef struct V_v_v_v
{   
    int     length;
    V_v_v** elements;
}   
V_v_v_v;

typedef struct Indexed_vector_element
{
    double element;
    int  index;
}
Indexed_vector_element;

typedef struct Indexed_vector 
{   
    int                     length;
    Indexed_vector_element* elements;
}   
Indexed_vector;

/* -------------------------------------------------------------------------- */

int get_unit_vector(Vector** vpp, int len);
int get_zero_vector(Vector** vpp, int len);

int get_initialized_vector(Vector** vpp, int len, double initial_value);

#ifdef TRACK_MEMORY_ALLOCATION

#   define create_vector(x)    debug_create_vector(x, __FILE__, __LINE__)

#   define get_target_vector(x,y)   \
                               debug_get_target_vector(x,y, __FILE__, __LINE__) 
#   define allocate_2D_vp_array(x, y) \
                 debug_allocate_2D_vp_array((x), (y), __FILE__, __LINE__)

    Vector* debug_create_vector
    (
        int         len,
        const char* file_name,
        int         line_number 
    );

    int debug_get_target_vector
    (
        Vector**    target_vpp,
        int         length,
        const char* file_name,
        int         line_number 
    );

    Vector*** debug_allocate_2D_vp_array(int, int, const char*, int);
#else 
    Vector* create_vector(int len);
    int get_target_vector(Vector** out_arg_vp, int length);
    Vector*** allocate_2D_vp_array(int, int);
#endif

void free_vector(Vector* vp);
Vector* create_zero_vector(int);
int ow_zero_vector(Vector*);
int ow_set_vector(Vector*, double);
Vector* create_random_vector(int length);
Vector* create_random_vector_2(int length);

int get_random_vector(Vector** target_vpp, int length);

int get_random_vector_2(Vector** target_vpp, int length);

double* get_1D_dp_array_from_vector(Vector* vp);

Vector* put_1D_dp_array_into_vector(int length, double* data_ptr);

int ascend_sort_vector(Vector* vp);

int vp_get_indexed_vector(Indexed_vector** vpp, const Vector* init_vp);

int get_random_indexed_vector(Indexed_vector** vpp, int length);

int get_zero_indexed_vector(Indexed_vector** vpp, int length);

int get_target_indexed_vector(Indexed_vector** vpp, int length);

Indexed_vector* create_indexed_vector(int len);
Indexed_vector* create_zero_indexed_vector(int len);
Indexed_vector* vp_create_indexed_vector(const Vector* vp);
void free_indexed_vector(Indexed_vector* vp);
int ascend_sort_indexed_vector(Indexed_vector* vp);
int descend_sort_indexed_vector(Indexed_vector* vp);

int get_target_vector_vector(Vector_vector** target_vvpp, int count);

Vector_vector* create_vector_vector(int count);
void free_vector_vector(Vector_vector* vvp);

void free_2D_vp_array_and_vectors
(
    Vector*** array,
    int       num_rows,
    int       num_cols 
);

int get_target_v3(V_v_v** target_vvvpp, int count);
V_v_v* create_v3(int length);
void free_v3(V_v_v* vvvp);
int get_target_v4(V_v_v_v** target_vvvvpp, int count);
V_v_v_v* create_v4(int length);
void free_v4(V_v_v_v* vvvvp);

#endif

