
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "n/n_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                          get_1D_dp_array_from_matrix
 *                               
 *                               
 *
 * Index: matrices, fortran
 *
 * -----------------------------------------------------------------------------
*/

double* get_1D_dp_array_from_matrix(const Matrix* mp)
{
    int  num_rows, num_cols; 
    long num_elements; 
    double* data_array;
    double* data_array_pos; 
    double*   matrix_row_pos; 
    int i, j; 

    num_rows = mp->num_rows;
    num_cols = mp->num_cols;

    num_elements = num_rows * num_cols;

    NRN(data_array = DBL_MALLOC(num_elements));

    data_array_pos = data_array;

    for (i=0; i<num_rows; i++)
    {
        matrix_row_pos = mp->elements[ i ];

        for(j=0; j<num_cols; j++)
        {
            *data_array_pos++ = *matrix_row_pos++;
        }
    }

    return data_array; 
}
            
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          put_1D_dp_array_into_matrix
 *                               
 *
 * Index: matrices, fortran
 *
 * -----------------------------------------------------------------------------
*/ 

Matrix* put_1D_dp_array_into_matrix
(
    int           num_rows,
    int           num_cols,
    const double* data_ptr 
)
{
    Matrix*       mp;
    int           i;
    int           j;
    const double* data_pos;
    double*         matrix_row_pos;

    NRN(mp = create_matrix(num_rows, num_cols)); 

    data_pos = data_ptr;

    for (i=0; i<num_rows; i++)
    {
        matrix_row_pos = mp->elements[ i ];

        for(j=0; j<num_cols; j++)
        {
            *matrix_row_pos++ = *data_pos++;
        }
    }

    return mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       get_fortran_1D_dp_array_from_matrix
 *                               
 *
 * Index: matrices, fortran
 *
 * -----------------------------------------------------------------------------
*/ 

double* get_fortran_1D_dp_array_from_matrix(const Matrix* mp)
{
    int  num_rows, num_cols; 
    long num_elements; 
    double* data_array;
    double*   matrix_row_pos; 
    int i, j; 

    num_rows = mp->num_rows;
    num_cols = mp->num_cols;

    num_elements = num_rows * num_cols;

    NRN(data_array = DBL_MALLOC(num_elements));

    for (i=0; i<num_rows; i++)
    {
        matrix_row_pos = mp->elements[ i ];

        for(j=0; j<num_cols; j++)
        {
            data_array[ j * num_rows + i ] = *matrix_row_pos++;
        }
    }

    return data_array; 
}
            
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                      put_fortran_1D_dp_array_into_matrix
 *                               
 *
 * Index: matrices, fortran
 *
 * -----------------------------------------------------------------------------
*/ 

Matrix* put_fortran_1D_dp_array_into_matrix
(
    int           num_rows,
    int           num_cols,
    const double* data_ptr 
)
{
    Matrix* mp;
    int     i;
    int     j;
    double*   matrix_row_pos;

    NRN(mp = create_matrix(num_rows, num_cols)); 

    for (i=0; i<num_rows; i++)
    {
        matrix_row_pos = mp->elements[ i ];

        for(j=0; j<num_cols; j++)
        {
            *matrix_row_pos++ = data_ptr[ j * num_rows + i ];
        }
    }

    return mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

