
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "n/n_gen_f.h"        /*  Only safe if first #include in a ".c" file  */
#include "n/n_slatec.h"
#include "n/n_quadratic.h"

/* -------------------------------------------------------------------------- */

int constrained_least_squares_solve
(
    Vector** result_vpp,
    Matrix*  mp,
    Vector*  target_vp,
    Matrix*  le_constraint_mp,
    Vector*  le_constraint_col_vp,
    Matrix*  eq_constraint_mp,
    Vector*  eq_constraint_col_vp,
    Vector*  lb_row_arg_vp,
    Vector*  ub_row_arg_vp 
)
{
    int     num_vars  = mp->num_cols;
    Vector* lb_row_vp = NULL;
    Vector* ub_row_vp = NULL;
    int     result = NO_ERROR;

    ERE(get_target_vector(result_vpp, num_vars));

    if (lb_row_arg_vp == NULL)
    {
        result = get_zero_vector(&lb_row_vp, num_vars);
        lb_row_arg_vp = lb_row_vp; 
    }

    if ((result != ERROR) && (ub_row_arg_vp == NULL))
    {
       result = get_initialized_vector(&ub_row_vp, num_vars, (DBL_MAX));
       ub_row_arg_vp = ub_row_vp;
    }

    if (result != ERROR)
    {
        result = do_quadratic(mp, target_vp, 
                              le_constraint_mp, le_constraint_col_vp, 
                              eq_constraint_mp, eq_constraint_col_vp, 
                              lb_row_arg_vp, ub_row_arg_vp, *result_vpp);
    }

    free_vector(lb_row_vp);
    free_vector(ub_row_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_quadratic
(
    Matrix* mp,
    Vector* target_vp,
    Matrix* le_constraint_mp,
    Vector* le_constraint_col_vp,
    Matrix* eq_constraint_mp,
    Vector* eq_constraint_col_vp,
    Vector* lb_row_vp,
    Vector* ub_row_vp,
    Vector* result_vp 
)
{
    int res;

    /*
    res = do_dlsei_quadratic(mp,target_vp,le_constraint_mp, 
                             le_constraint_col_vp,
                             eq_constraint_mp, eq_constraint_col_vp, 
                             lb_row_vp, ub_row_vp, result_vp);
    */

    res = do_dbocls_quadratic(mp, target_vp,
                              le_constraint_mp, le_constraint_col_vp,
                              eq_constraint_mp, eq_constraint_col_vp, 
                              lb_row_vp, ub_row_vp, result_vp);

    return res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

