
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

/*
#define NO_DBOCLS_EQUALITY_SLOP
*/
#define DBOCLS_EQUALITY_SLOP    (10000.0 * DBL_EPSILON)

/*   
   Copyright (c) 1994, by Kobus Barnard (author). 
   
   Personal and educational use of this code is granted, provided 
   that this header is kept intact, and that the authorship is not 
   misrepresented. Commercial use is not permitted.
*/

/*
// This file implements wrappers to the SLATEC library. Code should go into
// this file if and only if it is involved with SLATEC. Thus this file forces
// loading of SLATEC. Other modules may implicitly force the loading of this 
// file (and thus SLATEC), but then they should not contain modules which 
// may be used independently of SLATEC.
// 
// The code that follows is a bit ugly with #ifdef's because we want code that
// calls these routines to compile and link even if the SLATEC library is not
// available. Since we cannot know (at the library level) if the routine(s)
// used from this module are essential for the higher level code, it is better
// if the higher level code can pretend that the routine is available. Also, we
// do not want to force the higher level code to do the #ifdef'ing.
*/

#include "n/n_gen_f.h"        /*  Only safe if first #include in a ".c" file  */
#include "n/n_slatec.h"

/* -------------------------------------------------------------------------- */

/*
// Un-comment to force NO SLATEC
// 
#undef HAVE_SLATEC
*/

/* -------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------- */

#ifdef HAVE_SLATEC

#define MAX_CONSTRAINT_TOLERANCE  0.0001

#ifdef HPUX 
#    define dlsei_ dlsei
#    define dbocls_ dbocls
#endif 

/* -----------------------------------------------------------------------------
|                              HAVE_SLATEC
|                                  ||   
|                                 \||/  
|                                  \/
*/

/*
// Prototypes to turn of warnings. Independent prototypes would be nice ..
*/

void dlsei_
(
    double*    W,
    kjb_int32* MDW,
    kjb_int32* ME,
    kjb_int32* MA,
    kjb_int32* MG,
    kjb_int32* N,
    double*    PRGOPT,
    double*    X,
    double*    RNORME,
    double*    RNORML,
    kjb_int32* MODE,
    double*    WS,
    kjb_int32* IP 
);

void dbocls_
(
    double*    W,
    kjb_int32* MDW,
    kjb_int32* MCON,
    kjb_int32* MROWS,
    kjb_int32* NCOLS,
    double*    BL,
    double*    BU,
    kjb_int32* IND,
    kjb_int32* IOPT,
    double*    X,
    double*    RNORMC,
    double*    RNORM,
    kjb_int32* MODE,
    double*    RW,
    kjb_int32* IW 
);
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_SLATEC        
----------------------------------------------------------------------------- */
#else 
/* -----------------------------------------------------------------------------
|                              no SLATEC 
|                                  ||   
|                                 \||/  
|                                  \/
*/
    static void set_dont_have_slatec_error(void);
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no SLATEC        
----------------------------------------------------------------------------- */
#endif

#ifdef HAVE_SLATEC
/* -----------------------------------------------------------------------------
|                              HAVE_SLATEC
|                                  ||   
|                                 \||/  
|                                  \/
*/

int do_dlsei_quadratic
(
    Matrix* mp,
    Vector* target_vp,
    Matrix* le_constraint_mp,
    Vector* le_constraint_col_vp,
    Matrix* eq_constraint_mp,
    Vector* eq_constraint_col_vp,
    Vector* lb_row_vp,
    Vector* ub_row_vp,
    Vector* result_vp 
)
{
    Matrix* bound_constraint_mp   = NULL;
    int     le_m;
    int     i;
    int     j;
    int     count;
    int     num_bound_constraints;
    int     result;

    /* Variable names as given in SLATEC docs. */
    kjb_int32   N;
    double* WS;
    kjb_int32*  IP;
    kjb_int32   dim_IP;
    kjb_int32   dim_WS;
    kjb_int32   MDW;
    kjb_int32   ME;
    kjb_int32   MA;
    kjb_int32   MG;
    kjb_int32   MODE;
    double* W;
    double* X;
    double  RNORML;
    double  RNORME;
    double  PRGOPT[ 1 ];

    PRGOPT[ 0 ] = 1.0;

    N = NOT_SET;

    if (mp != NULL) 
    {
        N = mp->num_cols;
        MA = mp->num_rows;
    }
    else 
    {
        MA = 0;
    }

    if (le_constraint_mp != NULL) 
    {
        if (N == NOT_SET) 
        {
            N = le_constraint_mp->num_cols;
        }
        else 
        {
            if (N != le_constraint_mp->num_cols) 
            {
                SET_ARGUMENT_BUG();
                return ERROR; 
            }
        }
        
        le_m = le_constraint_mp->num_rows;

        if (le_constraint_col_vp == NULL) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }

        if (le_m != le_constraint_col_vp->length) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }
    }
    else 
    {
        le_m = 0;
    }

    if (eq_constraint_mp != NULL) 
    {
        if (N == NOT_SET) 
        {
            N = eq_constraint_mp->num_cols;
        }
        else 
        {
            if (N != eq_constraint_mp->num_cols) 
            {
                SET_ARGUMENT_BUG();
                return ERROR; 
            }
        }
        
        ME = eq_constraint_mp->num_rows;

        if (eq_constraint_col_vp == NULL) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }

        if (ME != eq_constraint_col_vp->length) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }
    }
    else 
    {
        ME = 0;
    }

    if (N == NOT_SET) 
    {
        set_bug("All matrices passed to do_dlsie_quadratic are NULL.");
    }

    num_bound_constraints = 0;

    if (lb_row_vp != NULL)
    {
        for (i=0; i<N; i++) 
        {
            if ((lb_row_vp->elements)[i] > DBL_HALF_MOST_NEGATIVE / 3.0) 
            {
                num_bound_constraints++;
            }
        }
    }

    if (ub_row_vp != NULL)
    {
        for (i=0; i<N; i++) 
        {
            if ((ub_row_vp->elements)[i] < DBL_HALF_MOST_POSITIVE / 3.0) 
            {
                num_bound_constraints++;
            }
        }
    }

    if (num_bound_constraints > 0) 
    {
        kjb_int32 bound_inequality_count = 0;

        NRE(bound_constraint_mp = create_zero_matrix(num_bound_constraints, 
                                                     N + 1));

        if (lb_row_vp != NULL)
        {
            for (i=0; i<N; i++) 
            {
                if ((lb_row_vp->elements)[i]> DBL_HALF_MOST_NEGATIVE/3.0)
                {
                    (bound_constraint_mp->elements)[bound_inequality_count][i] =
                                                                       1.0;
                    (bound_constraint_mp->elements)[bound_inequality_count][N] =
                                                      (lb_row_vp->elements)[i];
                    bound_inequality_count++;
                }
            }
        }

        if (ub_row_vp != NULL)
        {
            for (i=0; i<N; i++) 
            {
                if ((ub_row_vp->elements)[i]< DBL_HALF_MOST_POSITIVE/3.0)
                {
                    (bound_constraint_mp->elements)[bound_inequality_count][i] =
                                                                      -1.0;
                    (bound_constraint_mp->elements)[bound_inequality_count][N] =
                                                      -(ub_row_vp->elements)[i];
                    bound_inequality_count++;
                }
            }
        }
    }

    MG = le_m + num_bound_constraints; 
    MDW = MA + ME + MG;

    NRE(X = DBL_MALLOC(N)); 
    dim_WS = 2*(ME+N)+(MAX_OF(MA+MG,N))+(MG+2)*(N+7);
    NRE(WS = DBL_MALLOC(dim_WS));
    dim_IP = MG+2*N+2;
    NRE(IP = INT32_MALLOC(dim_IP));
    IP[ 0 ] = dim_WS;
    IP[ 1 ] = dim_IP; 
    
    NRE(W = DBL_MALLOC( MDW * (N + 1)));

    for (i=0; i<MDW * (N+1); i++) 
    {
        W[i] = DBL_NOT_SET;
    }

    /* F77 style arrays */

    count = 0;

    for (i=0; i<ME; i++) 
    {
        for (j=0; j<N; j++) 
        {
            W[ j*MDW + count ] = (eq_constraint_mp->elements)[i][j];
        }

        W[ N*MDW + count ] = (eq_constraint_col_vp->elements)[i];
        count++;
    }

    for (i=0; i<MA; i++) 
    {
        for (j=0; j<N; j++) 
        {
            W[ j*MDW + count ] = (mp->elements)[i][j];
        }
        W[ N*MDW + count ] = (target_vp->elements)[i];
        count++;
    }

    for (i=0; i<le_m; i++) 
    {
        for (j=0; j<N; j++) 
        {
            W[ j*MDW + count ] = -(le_constraint_mp->elements)[i][j]; 
        }

        W[ N*MDW + count ] = -(le_constraint_col_vp->elements)[i];
        count++;
    }
            
    for (i=0; i<num_bound_constraints; i++) 
    {
        for (j=0; j<N+1; j++) 
        {
            W[ j*MDW + count] = (bound_constraint_mp->elements)[i][j];
        }
        count++;
    }

    dlsei_(W,&MDW, &ME, &MA, &MG, &N, PRGOPT, X, &RNORME, &RNORML, &MODE, WS, 
           IP);

    if (MODE == 0) 
    {
        Vector* est_vp  = NULL;
        Vector* diff_vp = NULL;
        double  rnorm;

        for (i=0; i<result_vp->length; i++) 
        {
            (result_vp->elements)[i] = X[i];
        }

        if (multiply_matrix_and_vector(&est_vp, mp, result_vp) == ERROR) 
        {
            result = ERROR;
        }
        else if (subtract_vectors(&diff_vp, est_vp, target_vp) == ERROR)
        {
            result = ERROR;
        }
        else
        {
            rnorm = vector_magnitude(diff_vp);

            if (    (rnorm > 1000 * DBL_EPSILON)
                 && (fabs(rnorm - RNORML) / rnorm > 0.01)
               )
            {
                set_high_light(stdout); 
                pso("\n"); 
                rep_print(stdout, '-', 80);
                pso("Warning:\n\n");
                pso("Externally computed residual norm differs ");
                pso("from dlsei() RNORML by more than 1%%.\n");
                pso("Computed residual norm is %e.\n", rnorm);
                pso("dlsei() RNORML is %e.\n", RNORML);
                rep_print(stdout, '-', 80);
                pso("\n\n");
                unset_high_light(stdout); 
            }

            result = check_quadratic_constraints(result_vp, le_constraint_mp,
                                                 le_constraint_col_vp, 
                                                 lb_row_vp, ub_row_vp);
        }
        
        free_vector(est_vp);
        free_vector(diff_vp);
    }
    else if (MODE == 1) 
    {
        set_error("Equality constraints are contradictory.");
        result = ERROR; 
    }
    else if (MODE == 2) 
    {
        set_error("Inequality constraints are contradictory.");
        result = ERROR; 
    }
    else if (MODE == 3) 
    {
        set_error(
             "Both equality and inequality constraints are contradictory.");
        result = ERROR; 
    }
    else 
    {
        set_bug("Unexpected MODE returned from dlsei.");
        result = ERROR; 
    }

    free_matrix(bound_constraint_mp); 

    kjb_free(W);
    kjb_free(X);
    kjb_free(WS);
    kjb_free(IP);

    return result; 
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_SLATEC        
----------------------------------------------------------------------------- */
#else 
/* -----------------------------------------------------------------------------
|                              no SLATEC 
|                                  ||   
|                                 \||/  
|                                  \/
*/
/*ARGSUSED*/
int do_dlsei_quadratic
(
    Matrix* dummy_mp,
    Vector* dummy_target_vp,
    Matrix* dummy_le_constraint_mp,
    Vector* dummy_le_constraint_col_vp,
    Matrix* dummy_eq_constraint_mp,
    Vector* dummy_eq_constraint_col_vp,
    Vector* dummy_lb_row_vp,
    Vector* dummy_ub_row_vp,
    Vector* dummy_result_vp 
)
{

    set_dont_have_slatec_error();
    return ERROR;
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no SLATEC        
----------------------------------------------------------------------------- */
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef HAVE_SLATEC
/* -----------------------------------------------------------------------------
|                              HAVE_SLATEC
|                                  ||   
|                                 \||/  
|                                  \/
*/

int do_dbocls_quadratic
(
    Matrix* mp,
    Vector* target_vp,
    Matrix* le_constraint_mp,
    Vector* le_constraint_col_vp,
    Matrix* eq_constraint_mp,
    Vector* eq_constraint_col_vp,
    Vector* lb_row_vp,
    Vector* ub_row_vp,
    Vector* result_vp 
)
{
    Matrix* constraint_mp   = NULL;
    Vector* input_lb_row_vp;
    Vector* input_ub_row_vp;
    int     eq_n;
    int     eq_m;
    int     le_n;
    int     le_m;
    int     n, m, i, j;
    int     result          = NO_ERROR;
    /* Variable names as given in SLATEC docs. */
    kjb_int32   I;
    double* RW;
    kjb_int32*  IW;
    kjb_int32   MDW;
    kjb_int32   MCON;
    kjb_int32   MROWS;
    kjb_int32   NCOLS;
    kjb_int32   MODE;
    double* W;
    double* X;
    double  RNORMC;
    double  RNORM;
    double* BL;
    double* BU;
    kjb_int32   IOPT[ 17 + 9   ];   /* 17 is base, 9 is extra for OPTION 2. */
    kjb_int32*  IND;

    if (mp->num_cols <= 0) 
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    if (le_constraint_mp == NULL) 
    {
        le_n = le_m = 0;
    }
    else 
    {
        le_m = le_constraint_mp->num_rows;
        le_n = le_constraint_mp->num_cols; 

        if ((le_constraint_col_vp == NULL) || 
            (le_constraint_col_vp->length != le_m)) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }
    }

    if (eq_constraint_mp == NULL) 
    {
        eq_n = eq_m = 0;
    }
    else 
    {
        eq_m = eq_constraint_mp->num_rows;
        eq_n = eq_constraint_mp->num_cols; 

        if ((eq_constraint_col_vp == NULL) || 
            (eq_constraint_col_vp->length != eq_m)) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }
    }

    if ((le_n != 0) && (eq_n != 0)) 
    {
        if (le_n != eq_n) 
        {
            SET_ARGUMENT_BUG();
            return ERROR; 
        }
        n = le_n; 
    }
    else if (le_n != 0) 
    {
        n = le_n; 
    }
    else if (eq_n != 0) 
    {
        n = eq_n;
    }
    else 
    {
        n = mp->num_cols; 
    }

    if (n != mp->num_cols) 
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    if (result_vp == NULL) 
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    m = le_m + eq_m; 

    if (m > 0) 
    {
        NRE(constraint_mp = create_zero_matrix(m, n));
    }

    I = 0;

    for (i=0; i<le_m; i++) 
    {
        for (j=0; j<n; j++) 
        {
            (constraint_mp->elements)[I][j]= (le_constraint_mp->elements)[i][j];
        }
        I++;
    }

    for (i=0; i<eq_m; i++) 
    {
        for (j=0; j<n; j++) 
        {
            (constraint_mp->elements)[I][j]= (eq_constraint_mp->elements)[i][j];
        }
        I++;
    }

    NRE(input_lb_row_vp = create_vector(n + eq_m + le_m));
    NRE(input_ub_row_vp = create_vector(n + eq_m + le_m));

    for (j=0; j<n; j++) 
    {
        if (lb_row_vp != NULL)
        {
            (input_lb_row_vp->elements)[j] = (lb_row_vp->elements)[j];
        }
        else 
        {
            (input_lb_row_vp->elements)[j] = DBL_MOST_NEGATIVE;
        }

        if (ub_row_vp != NULL)
        {
            (input_ub_row_vp->elements)[j] = (ub_row_vp->elements)[j];
        }
        else
        {
            (input_ub_row_vp->elements)[j] = DBL_MOST_POSITIVE;
        }
    }

    /* Constraints */
    for (j=0; j<le_m; j++) 
    {
        (input_lb_row_vp->elements)[n + j] = DBL_MOST_NEGATIVE; 
        
        (input_ub_row_vp->elements)[n + j]=(le_constraint_col_vp->elements)[j];
    }
    
    for (j=0; j<eq_m; j++) 
    {
#ifdef NO_DBOCLS_EQUALITY_SLOP
        (input_lb_row_vp->elements)[n + le_m + j] =
                                              eq_constraint_col_vp->elements[j];

        (input_ub_row_vp->elements)[n + le_m + j] =
                                              eq_constraint_col_vp->elements[j];
#else 
#ifdef DBOCLS_EQUALITY_SLOP
        (input_lb_row_vp->elements)[n + le_m + j] =
                              SUB_RELATIVE_DBL(eq_constraint_col_vp->elements[j],
                                           DBOCLS_EQUALITY_SLOP);

        (input_ub_row_vp->elements)[n + le_m + j] =
                              ADD_RELATIVE_DBL(eq_constraint_col_vp->elements[j],
                                           DBOCLS_EQUALITY_SLOP);
#else
        (input_lb_row_vp->elements)[n + le_m + j] =
                              SUB_EPSILON_DBL(eq_constraint_col_vp->elements[j]);

        (input_ub_row_vp->elements)[n + le_m + j] =
                              ADD_DBL_EPSILON(eq_constraint_col_vp->elements[j]);
#endif 
#endif 
    }

    MROWS = mp->num_rows;
    NCOLS = n;
    MCON = le_m + eq_m;
    MDW = MCON + MAX_OF(MROWS, NCOLS);

    NRE(X = DBL_MALLOC(2*(NCOLS+MCON)+2));   /* No extra due to options */
    NRE(RW = DBL_MALLOC(6*NCOLS+5*MCON));
    NRE(IW = INT32_MALLOC(2*(NCOLS+MCON)));
    
    NRE(IND = INT32_MALLOC(MCON + NCOLS));

    for (i=0; i<MCON + NCOLS; i++) 
    {
        if (    ((input_lb_row_vp->elements)[i] > DBL_HALF_MOST_NEGATIVE / 3.0)
             && ((input_ub_row_vp->elements)[i] < DBL_HALF_MOST_POSITIVE / 3.0)
           )
        {
            IND[ i ] = 3;
        }
        else if ((input_lb_row_vp->elements)[i] > DBL_HALF_MOST_NEGATIVE / 3.0)
        {
            IND[ i ] = 1;
        }
        else if ((input_ub_row_vp->elements)[i] < DBL_HALF_MOST_POSITIVE / 3.0)
        {
            IND[ i ] = 2;
        }
        else 
        {
            IND[ i ] = 4;
        }
    }   

    NRE(BU = DBL_MALLOC(input_ub_row_vp->length));
    NRE(BL = DBL_MALLOC(input_lb_row_vp->length));

    ASSERT(input_ub_row_vp->length == input_lb_row_vp->length);

    for (i=0; i<input_ub_row_vp->length; i++) 
    {
        BU[i] = (input_ub_row_vp->elements)[i];
        BL[i] = (input_lb_row_vp->elements)[i];
    }

    NRE(W = DBL_MALLOC( MDW * (NCOLS + MCON + 1) ));

    for (i=0; i<MDW * (NCOLS + MCON + 1); i++) 
    {
        W[i] = 0.0;
    }

    /* F77 style arrays */

    for (i=0; i<mp->num_rows; i++) 
    {
        for (j=0; j<mp->num_cols; j++) 
        {
            W[ j*MDW + i + MCON  ] = (mp->elements)[i][j];
        }

        W[ NCOLS*MDW + i + MCON ] = (target_vp->elements)[i];
    }

    /*
    // If constraint_mp is NULL, then m is zero.
    */
    for (i=0; i<m; i++) 
    {
        for (j=0; j<n; j++) 
        {
            W[ j*MDW + i ] = (constraint_mp->elements)[i][j]; 
        }
    }
            
    /* 
    // The first option, flagged by the 2, says to check the array sizes. The
    // sizes follow for the next 7 values. 
    */
    IOPT[ 0 ] = 2;
    IOPT[ 1 ] = MDW;
    IOPT[ 2 ] = NCOLS + MCON + 1 ;
    IOPT[ 3 ] = input_lb_row_vp->length;
    IOPT[ 4 ] = 2*(NCOLS+MCON)+2;
    IOPT[ 5 ] = 6*NCOLS+5*MCON;
    IOPT[ 6 ] = 2*(NCOLS+MCON);
    IOPT[ 7 ] = 17 + 9;
    IOPT[ 8 ] = 99;

/* ========================================================================== */

    dbocls_(W, &MDW, &MCON, &MROWS, &NCOLS, BL, BU, 
            IND, IOPT, X, &RNORMC, &RNORM, &MODE, RW, IW); 

    if (MODE == -22) 
    {
        set_high_light(stdout); 
        pso("\n"); 
        rep_print(stdout, '-', 80);
        pso("Warning:\n\n");
        pso("Quadratic programming problem reached max number of ");
        pso("allowed iterations.\n");
        pso("Solution may not be accurate.");
        rep_print(stdout, '-', 80);
        pso("\n\n");
        unset_high_light(stdout); 

        result = NO_ERROR;
    }
    else if (MODE < 0) 
    {
        set_error("Routine dbocls() failed (MODE=%ld).", (long)MODE);
        result = ERROR;
    }

    if (result != ERROR) 
    {
        Vector* est_vp  = NULL;
        Vector* diff_vp = NULL;
        double  rnorm;

        for (i=0; i<result_vp->length; i++) 
        {
            (result_vp->elements)[ i ] = X[i];
        }

        if (multiply_matrix_and_vector(&est_vp, mp, result_vp) == ERROR)
        {
            result = ERROR;
        }
        else if (subtract_vectors(&diff_vp, est_vp, target_vp) == ERROR)
        {
            result = ERROR;
        }
        else
        {
            rnorm = vector_magnitude(diff_vp);

            if (    (rnorm > 1000 * DBL_EPSILON)
                 && (fabs(rnorm - RNORM) / rnorm > 0.01)
               )
            {
                set_high_light(stdout); 
                pso("\n"); 
                rep_print(stdout, '-', 80);
                pso("Warning:\n\n");
                pso("Externally computed residual norm differs ");
                pso("from dbocls() RNORM by more than 1%%.\n");
                pso("Computed residual norm is %e.\n", rnorm);
                pso("dbocls() RNORM is %e.\n", RNORM);
                rep_print(stdout, '-', 80);
                pso("\n\n");
                unset_high_light(stdout); 
            }
    
            result = check_quadratic_constraints(result_vp, le_constraint_mp,
                                                 le_constraint_col_vp, 
                                                 lb_row_vp, ub_row_vp);
        }

        free_vector(est_vp);
        free_vector(diff_vp);
    }

#ifdef TEST
    if ((result != ERROR) && (RNORMC > MAX_CONSTRAINT_TOLERANCE))
    {
        set_high_light(stdout); 
        pso("\n"); 
        rep_print(stdout, '-', 80);
        pso("Warning:\n\n");
        pso("Routine dbocls() reports RNORMC=%e.\n", RNORMC);
        pso("This is larger than the (arbitrary) limit of %e.\n\n", 
            MAX_CONSTRAINT_TOLERANCE); 

        pso("Normally this would indicate infeasibility. However the value\n");
        pso("of RNORMC can be bogus, at least for reasonable tolerances.\n\n");
        pso("This problem seems to occur more on HPUX than SUN5.\n");
        pso("I don't know what is going on. Sorry!\n\n");

        pso("The result is likely approximately correct, as it passed some ");
        pso("sanity checks.\n"); 
        rep_print(stdout, '-', 80);
        pso("\n\n");
        unset_high_light(stdout); 
    }
#endif 

    kjb_free(W);
    kjb_free(BL);
    kjb_free(BU);
    kjb_free(IND);
    kjb_free(X);
    kjb_free(RW);
    kjb_free(IW); 

    free_vector(input_lb_row_vp);
    free_vector(input_ub_row_vp); 
    free_matrix(constraint_mp); 

    return result; 
    
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_SLATEC        
----------------------------------------------------------------------------- */
#else 
/* -----------------------------------------------------------------------------
|                              no SLATEC 
|                                  ||   
|                                 \||/  
|                                  \/
*/
/*ARGSUSED*/
int do_dbocls_quadratic
(
    Matrix* dummy_mp,
    Vector* dummy_target_vp,
    Matrix* dummy_le_constraint_mp,
    Vector* dummy_le_constraint_col_vp,
    Matrix* dummy_eq_constraint_mp,
    Vector* dummy_eq_constraint_col_vp,
    Vector* dummy_lb_row_vp,
    Vector* dummy_ub_row_vp,
    Vector* dummy_result_vp 
)
{

    set_dont_have_slatec_error();
    return ERROR;
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no SLATEC        
----------------------------------------------------------------------------- */
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifndef HAVE_SLATEC
/* -----------------------------------------------------------------------------
|                              no SLATEC 
|                                  ||   
|                                 \||/  
|                                  \/
*/
static void set_dont_have_slatec_error(void)
{
    set_error("Operation failed because the program was built without ");
    cat_error("the SLATEC interface.");
    add_error("(Appropriate re-compiling is needed to fix this.)");
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no SLATEC        
----------------------------------------------------------------------------- */
#endif 

