
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "stat.h"
#include "do_error_report.h"
#include "report.h"

/* -------------------------------------------------------------------------- */

int process_report_command(char* arg)
{ 
    static const char *key_words[ ] = { "total" } ;
    int   key_words_found_list[ sizeof(key_words)/sizeof(char*) ];
    int   num_key_words = sizeof(key_words)/sizeof(char*);
    int   i, num_found_key_words;
    char  report_request[ 1000 ];
    char* lc_report_request_pos; 
    char  lc_report_request[ 1000 ];
    char  heading[ 1000 ]; 
    int   do_grand_total = FALSE; 
    char  lc_report_request_item[ 100 ]; 

    trim_beg(&arg);

    if (*arg == '%') 
    {
        report_request[ 0 ] = '\0';
    }
    else
    {
        BUFF_GEN_GET_TOKEN(&arg, report_request, "%");
        trim_end(report_request);
    }

    if (*arg == '%') arg++;
    trim_beg(&arg); 
    BUFF_CPY(heading, arg);

    ERE(num_found_key_words = ic_parse_key_words(report_request, num_key_words,
                                                 key_words,key_words_found_list,
                                                 (int*)NULL));

    for (i=0; i<num_found_key_words; i++)
    {
        char lc_key_word[ 100 ]; 

        EXTENDED_LC_BUFF_CPY(lc_key_word, key_words[key_words_found_list[i]]);

        if (STRCMP_EQ(lc_key_word, "total"))
        {
            do_grand_total = TRUE;
        }
    }

    EXTENDED_LC_BUFF_CPY(lc_report_request, report_request);
    lc_report_request_pos = lc_report_request;
    trim_beg(&lc_report_request_pos);
    BUFF_GET_TOKEN(&lc_report_request_pos, lc_report_request_item);

    do
    {
        int valid_request = FALSE; 

        if (    (lc_report_request_item[ 0 ] == '\0')
             || (STRCMP_EQ(lc_report_request_item, "error"))
            )
        {
            if (do_grand_total)
            {
                ERE(do_grand_total_error_report(heading));
            }
            else 
            {
                ERE(do_cumulative_error_report(heading));
            }
            valid_request = TRUE; 
        }

        if (    (lc_report_request_item[ 0 ] == '\0')
             || (STRCMP_EQ(lc_report_request_item, "luminance"))
            )
        {
            if (do_grand_total)
            {
                ERE(do_grand_total_luminance_report());
            }
            else 
            {
                ERE(do_cumulative_luminance_report());
            }
            valid_request = TRUE; 
        }

        if ( ! valid_request )
        {
            set_error("%q is an invalid report request.\n", report_request);
            return ERROR; 
        }
    }
    while (BUFF_GET_TOKEN_OK(&lc_report_request_pos, lc_report_request_item));

    return NO_ERROR; 
}

