
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#ifdef SIMPLE_TEMPLATE /* Controlled in def.h */

#include "illum_data.h"
#include "spectra.h"
#include "simple_template.h"

/* -------------------------------------------------------------------------- */

static int do_simple_template_algorithm
(
    const Matrix* data_mp,
    Vector**      result_vpp 
);

/* -------------------------------------------------------------------------- */

int do_simple_template
(
    const Matrix*    data_mp,
    Simple_template_result* simple_template_result_ptr,
    Simple_template_error*  cpu_time_ptr
)
{
    long    cpu_time_in_ms;
    Vector* illum_rgb_vp   = NULL;
    Vector* rgb_map_vp     = NULL;
    int     result;

    init_cpu_time();

    result = do_simple_template_algorithm(data_mp, &illum_rgb_vp); 

    if ((result != ERROR) && (result != NOT_AVAILABLE)) 
    {
        cpu_time_in_ms = get_cpu_time(); 

        cpu_time_ptr->alg = ((double)cpu_time_in_ms) / 
                                             ((double)MILLI_SECONDS_PER_SECOND) ;

        result = get_map_to_canon_illum(&rgb_map_vp, illum_rgb_vp);
    }

    if ((result != ERROR) && (result != NOT_AVAILABLE))
    {
        result = get_projection_vector(&(simple_template_result_ptr->alg_vp), 
                                       rgb_map_vp, DIVIDE_BY_BLUE, MAX_X_DIV_B);
    }

    if (result == ERROR) 
    {
        free_vector(rgb_map_vp);
        free_vector(illum_rgb_vp);
        return ERROR;
    }

    free_vector(rgb_map_vp);
    free_vector(illum_rgb_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine should return one of NO_ERROR, ERROR, and NOT_AVAILABLE. It is
// somewhat arbitrary where one draws the line between NOT_AVAILABLE and ERROR. 
*/
static int do_simple_template_algorithm
(
    const Matrix* data_mp,
    Vector**      result_vpp 
)
{

    /* 
    // Invent a grey world type algorithm for testing. We can't just call the
    // double grey world, as this code is tied together into a module. 
    */
    ERE(average_matrix_rows(result_vpp, data_mp)); 

    /*
    // Make it a little bit unlike the grey world, again for testing purposes
    // (Make it redder by 10%).
    */
    ((*result_vpp)->elements)[ 0 ] *= 1.1;

    return NO_ERROR; 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#endif  /* SIMPLE_TEMPLATE */

