
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef TYPE_INCLUDED
#define TYPE_INCLUDED

#include "l/l_incl.h"
#include "m/m_incl.h"
#include "def.h"

/*
// Warning : We must have DONT_SHARPEN first, followed by all "CORRECTION"
//           methods, followed by FIXED_SHARP, followed by the other sharp
//           methods, followed by NUM_SHARP_METHODS which must be last.
//           Otherwise, code in do_cc.c will have to be changed. Also, if
//           DONT_SHARPEN is not first (i.e., it is not zero), then tests for
//           sharpening against sharpening method will be broken (I'm not sure
//           if there are any such tests in the code at present.)
//
//           (I know this is a big HACK -- sorry.) 
*/
typedef enum Sharp_method 
{
    DONT_SHARPEN = 0, 
    FIXED_SHARP_CORRECTION, 
    SYNTHETIC_DB_SHARP_CORRECTION,
    SYNTHETIC_MIP_DB_SHARP_CORRECTION,
    WORLD_DB_SHARP_CORRECTION,
    SCENE_DB_SHARP_CORRECTION,
    ITERATIVE_SHARP_CORRECTION,
    FIXED_SHARP, 
    SYNTHETIC_DB_SHARP,
    SYNTHETIC_MIP_DB_SHARP,
    WORLD_DB_SHARP,
    SCENE_DB_SHARP,
    ITERATIVE_SHARP,
    NUM_SHARP_METHODS
} 
Sharp_method;

typedef struct Colour_constancy_methods 
{   
    int hull_intersection;

    int do_2D;
    int do_3D;
    int do_luminance;
    int do_luminance_factors;

    int do_standard; 
    int do_gamut; 
    int do_correlation;    /* Lindsay: Feb 17/98 */
    
#ifdef SIMPLE_TEMPLATE
    /*
    // Only one flag for each GROUP of algorithms.
    */
    int do_simple_template; 
#endif 

#ifdef TEMPLATE
    /*
    // Only one flag for each GROUP of algorithms.
    */
    int do_template; 
#endif 

#ifdef MULTI_TEMPLATE
    /*
    // Only one flag for each GROUP of algorithms.
    */
    int do_multi_template; 
#endif 

    int reset_specular; 

    int sharp[ NUM_SHARP_METHODS ];

    int plot_enable;
    int report_result;  
    int skip_on_error; 
    int print_algorithm_error; 
    int debug;
}    
Colour_constancy_methods;

typedef enum Negative_data_action
{
    USE_NEGATIVE_DATA,
    SKIP_NEGATIVE_DATA,
    SKIP_NEGATIVE_BLUE_DATA,
    ZERO_NEGATIVE_DATA,
    NEGATIVE_DATA_IS_ERROR
}
Negative_data_action;

typedef struct Constraints 
{   
    Queue_element* surface_hull_list_head;
    Queue_element* illum_hull_list_head;   /* Illum hull approx in map space. */
    Hull*          illum_hp;               /* RG Illum */
    Vector*        canon_illum_vp;
    double         sum_luminance[ MAX_NUM_LUMINANCE_METHODS ];
}    
Constraints;

typedef struct Hull_result 
{   
    Hull* surface_hp;
    Hull* surface_illum_hp;
    Hull* illum_constrained_surface_hp; 
    Hull* luminance_surface_hp[ MAX_NUM_LUMINANCE_METHODS ]; 
    Hull* luminance_surface_illum_hp[ MAX_NUM_LUMINANCE_METHODS ]; 
    Hull* luminance_constrained_surface_hp[ MAX_NUM_LUMINANCE_METHODS ]; 
    Hull* surface_illum_var_hp;
    Matrix* scia_points_mp;
    Matrix* scca_points_mp;
}    
Hull_result; 

typedef struct Best_fit_result 
{   
    Matrix* world_best_2D_fit_mp;
    Matrix* world_best_3D_fit_mp;
    Vector* world_best_3D_diag_fit_vp;
    Vector* world_best_2D_diag_fit_vp;
    Matrix* scene_best_2D_fit_mp;
    Matrix* scene_best_3D_fit_mp;
    Vector* scene_best_3D_diag_fit_vp;
    Vector* scene_best_2D_diag_fit_vp;
    Vector* actual_vp;
}    
Best_fit_result;

typedef struct Standard_result 
{   
    Vector* grey_world_vp;
    Vector* db_grey_vp;
    Vector* scale_by_max_vp;
    Vector* ave_illum_vp;
}    
Standard_result;

typedef struct Gamut_result 
{   
    Vector* mv_surface_vp;
    Vector* ha_surface_vp;
    Vector* wa_surface_vp;
    /* With surface and illum info */
    Vector* mv_surface_illum_vp;
    Vector* ha_surface_illum_vp;
    Vector* wa_surface_illum_vp;
    Vector* icsa_surface_illum_vp;
    Vector* scia_surface_illum_vp[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    Vector* scca_surface_illum_vp[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    /* With surface and lum info */
    Vector* ha_surface_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* mv_surface_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    /* With surface, illum, and lum info */
    Vector* lcsa_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* scla_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* ha_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* mv_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* scia_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
    Vector* scca_surface_illum_lum_vp[ MAX_NUM_LUMINANCE_METHODS ];
}    
Gamut_result; 

/* Lindsay: Feb 17/98 */
typedef struct Correlation_result
{   
    Vector *correlation_01_vp[ NUM_CORMATS_PER_ALG ]; 
    Vector *correlation_gaussian_MAP_vp[ NUM_CORMATS_PER_ALG ]; 
    Vector *correlation_uniform_MAP_inv_dist_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MAP_gaussian_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MAP_fitted_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_gaussian_MLM_vp[ NUM_CORMATS_PER_ALG ]; 
    Vector *correlation_uniform_MLM_inv_dist_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MLM_gaussian_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MLM_fitted_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_gaussian_MMSE_vp[ NUM_CORMATS_PER_ALG ]; 
    Vector *correlation_uniform_MMSE_inv_dist_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MMSE_gaussian_fill_vp[ NUM_CORMATS_PER_ALG ];
    Vector *correlation_uniform_MMSE_fitted_vp[ NUM_CORMATS_PER_ALG ];
}   
Correlation_result;
/* End Lindsay: Feb 17/98 */

#ifdef TEMPLATE
typedef struct Template_result
{   
    Vector* first_alg_vp;
    Vector* second_alg_vp;
}   
Template_result;
#endif 

#ifdef MULTI_TEMPLATE
typedef struct Multi_template_result
{   
    Vector* alg_vp[ MAX_NUM_MULTI_TEMPLATE_ALGS ];
}   
Multi_template_result;
#endif 

#ifdef SIMPLE_TEMPLATE
typedef struct Simple_template_result
{   
    Vector* alg_vp;
}   
Simple_template_result;
#endif 

typedef struct Result_by_algorithm 
{   
    Best_fit_result       best_fit; 
    Standard_result       standard;
    Gamut_result          gamut_3D[ MAX_NUM_CANON_GAMUTS ];
    Gamut_result          gamut_2D[ MAX_NUM_CANON_GAMUTS ];
    Correlation_result    correlation;    /* Lindsay: Feb 17/98 */
#ifdef SIMPLE_TEMPLATE
    Simple_template_result simple_template;
#endif 
#ifdef TEMPLATE
    Template_result       template;
#endif 
#ifdef MULTI_TEMPLATE
    Multi_template_result multi_template;
#endif 
}    
Result_by_algorithm; 

typedef struct Colour_constancy_result 
{   
    Result_by_algorithm sharp[ NUM_SHARP_METHODS ];
}    
Colour_constancy_result;

typedef struct Best_fit_error 
{   
    double world_best_3D_fit;
    double world_best_2D_fit;
    double world_best_3D_diag_fit;
    double world_best_2D_diag_fit;
    double scene_best_3D_fit;
    double scene_best_2D_fit;
    double scene_best_3D_diag_fit;
    double scene_best_2D_diag_fit;
    double actual; 
}    
Best_fit_error;

typedef struct Standard_error 
{   
    double nothing;
    double grey_world; 
    double db_grey_world; 
    double cluster_grey_world; 
    double cluster_db_grey_world; 
    double scale_by_max; 
    double ave_illum; 
}    
Standard_error;

typedef struct Gamut_error 
{   
    /* With surface and lum info */
    double mv_surface;
    double ha_surface;
    double wa_surface;
    /* With surface and illum info */
    double mv_surface_illum;
    double ha_surface_illum;
    double icsa_surface_illum;
    double scia_surface_illum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    double scca_surface_illum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    double wa_surface_illum;
    /* With surface and lum info */
    double ha_surface_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double mv_surface_lum[ MAX_NUM_LUMINANCE_METHODS ];
    /* With surface, illum, and lum info */
    double ha_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double mv_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double lcsa_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double scla_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double scca_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
    double scia_surface_illum_lum[ MAX_NUM_LUMINANCE_METHODS ];
}    
Gamut_error;

/* Lindsay: Feb 17/98 */
typedef struct Correlation_error 
{
    double correlation_01[ NUM_CORMATS_PER_ALG ]; 
    double correlation_gaussian_MAP[ NUM_CORMATS_PER_ALG ]; 
    double correlation_uniform_MAP_inv_dist_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MAP_gaussian_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MAP_fitted[ NUM_CORMATS_PER_ALG ];
    double correlation_gaussian_MLM[ NUM_CORMATS_PER_ALG ]; 
    double correlation_uniform_MLM_inv_dist_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MLM_gaussian_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MLM_fitted[ NUM_CORMATS_PER_ALG ];
    double correlation_gaussian_MMSE[ NUM_CORMATS_PER_ALG ]; 
    double correlation_uniform_MMSE_inv_dist_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MMSE_gaussian_fill[ NUM_CORMATS_PER_ALG ];
    double correlation_uniform_MMSE_fitted[ NUM_CORMATS_PER_ALG ];
}    
Correlation_error;
/* End Lindsay */

#ifdef TEMPLATE
typedef struct Template_error 
{   
    double first_alg; 
    double second_alg; 
}    
Template_error;
#endif 

#ifdef MULTI_TEMPLATE
typedef struct Multi_template_error 
{   
    double alg[ MAX_NUM_MULTI_TEMPLATE_ALGS ]; 
}    
Multi_template_error;
#endif 

#ifdef SIMPLE_TEMPLATE
typedef struct Simple_template_error 
{   
    double alg; 
}    
Simple_template_error;
#endif 

typedef struct Error_by_algorithm
{   
    Best_fit_error       best_fit;
    Standard_error       standard;
    Gamut_error          gamut_2D[ MAX_NUM_CANON_GAMUTS ];
    Gamut_error          gamut_3D[ MAX_NUM_CANON_GAMUTS ];
    Correlation_error    correlation;    /* Lindsay: Feb 17/98 */
#ifdef SIMPLE_TEMPLATE
    Simple_template_error simple_template;
#endif 
#ifdef TEMPLATE
    Template_error       template;
#endif 
#ifdef MULTI_TEMPLATE
    Multi_template_error multi_template;
#endif 
}   
Error_by_algorithm; 
        
typedef struct Error_by_sharp 
{   
    /*
    // A bit of a hack here, as the space allocated for fixed sharp methods
    // never gets used! Going for readability here. 
    */
    Error_by_algorithm sharp[ NUM_SHARP_METHODS ];
}    
Error_by_sharp;

typedef struct Colour_constancy_error 
{   
    Error_by_sharp illum_angular;
    Error_by_sharp illum_sum_chrom;
    Error_by_sharp illum_lab;
    Error_by_sharp illum_rgb;
    Error_by_sharp illum_r_div_sum;
    Error_by_sharp illum_g_div_sum;
    Error_by_sharp illum_sum_lum;
    Error_by_sharp scene_map_blue_chrom_rms;
    Error_by_sharp scene_map_sum_chrom_rms;
    Error_by_sharp scene_map_3D_rms;
    Error_by_sharp scene_map_3D_lab;
    Error_by_sharp world_map_blue_chrom_rms;
    Error_by_sharp world_map_sum_chrom_rms;
    Error_by_sharp world_map_3D_rms;
    Error_by_sharp world_map_3D_lab;
    Error_by_sharp cpu_time;
}   
Colour_constancy_error;

typedef struct Gamut_input_data_error
{   
    double relative_pixel_error;
    double absolute_pixel_error;
    double diag_error;
}   
Gamut_input_data_error;

typedef struct Input_data_error_by_algorithm
{   
    Gamut_input_data_error gamut_2D[ MAX_NUM_CANON_GAMUTS ];
    Gamut_input_data_error gamut_3D[ MAX_NUM_CANON_GAMUTS ];
}   
Input_data_error_by_algorithm;

typedef struct Input_data_error
{   
    Input_data_error_by_algorithm sharp[ NUM_SHARP_METHODS ];
}   
Input_data_error;

#endif

