
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "gamut_alg_input.h"

/* -------------------------------------------------------------------------- */

EXTERN_C int set_gamut_alg_input_options(const char* option, const char* value)
{
    double temp_double;
    double temp_real;
    int  temp_boolean_value;
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "negative-3d-data-action"))
       )
    {
        IMPORT Negative_data_action negative_3D_data_action;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Input data points with negative R/G/B "));

            if (negative_3D_data_action == USE_NEGATIVE_DATA)
            {
                ERE(pso("are used with 3D algorithms.\n"));
            }
            if (negative_3D_data_action == SKIP_NEGATIVE_DATA)
            {
                ERE(pso("are skipped with 3D algorithms.\n"));
            }
            else if (negative_3D_data_action == ZERO_NEGATIVE_DATA)
            {
                ERE(pso("are set to zero with 3D algorithms.\n"));
            }
            else if (negative_3D_data_action == NEGATIVE_DATA_IS_ERROR)
            {
                ERE(pso("cause error with 3D algorithms.\n"));
            }
        }
        else
        {
            char lc_value[ 100 ];

            EXTENDED_LC_BUFF_CPY(lc_value, value);

            if (STRCMP_EQ(lc_value, "use"))
            {
                negative_3D_data_action = USE_NEGATIVE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "skip"))
            {
                negative_3D_data_action = SKIP_NEGATIVE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "zero"))
            {
                negative_3D_data_action = ZERO_NEGATIVE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "error"))
            {
                negative_3D_data_action = NEGATIVE_DATA_IS_ERROR;
            }
            else
            {
                set_error("%q is and invalid negative data action.",value);
                return ERROR;
            }
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "udh")))
    {
        IMPORT int use_data_hull;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (use_data_hull)
            {
                ERE(pso("Hull of data is used instead of individual points "));
                ERE(pso("(udh=t).\n"));
            }
            else
            {
                ERE(pso("Data points are used instead of hull (udh=f).\n"));
            }
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            use_data_hull = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "ape"))
         || (match_pattern(lc_option, "min-absolute-pixel-error"))
         || (match_pattern(lc_option, "absolute-pixel-error"))
       )
    {
        IMPORT double min_absolute_pixel_error;
#ifdef USE_ONLY_ONE_OF_RE_AND_AE
        IMPORT double min_relative_pixel_error;
#endif

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (min_absolute_pixel_error < 0.0)
            {
                ERE(pso("absolute-pixel-error = off\n"));
            }
            else
            {
                ERE(pso("absolute-pixel-error = %.1f.\n",
                        min_absolute_pixel_error));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (min_absolute_pixel_error < 0.0)
            {
                ERE(pso("Minimum absolute pixel error is not modeled.\n"));
            }
            else
            {
                ERE(pso("Minimum absolute pixel error (ape) is %.1f.\n",
                        min_absolute_pixel_error));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                 min_absolute_pixel_error = DBL_HALF_MOST_NEGATIVE;
            }
            else
            {
                ERE(ss1d(value, &temp_double));

                if (temp_double < 0.0)
                {
                    set_error("Minimum absolute pixel error must be non-negative.");
                    return ERROR;
                }

#ifdef USE_ONLY_ONE_OF_RE_AND_AE
                min_relative_pixel_error = DBL_HALF_MOST_NEGATIVE;
#endif
                min_absolute_pixel_error = temp_double;
            }
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "rpe"))
         || (match_pattern(lc_option, "min-relative-pixel-error"))
         || (match_pattern(lc_option, "relative-pixel-error"))
       )
    {
        IMPORT double min_relative_pixel_error;
#ifdef USE_ONLY_ONE_OF_RE_AND_AE
        IMPORT double min_absolute_pixel_error;
#endif

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (min_relative_pixel_error < 0.0)
            {
                ERE(pso("relative-pixel-error = off\n"));
            }
            else
            {
                ERE(pso("relative-pixel-error = %.1f%%.\n",
                        100.0 * min_relative_pixel_error));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (min_relative_pixel_error < 0.0)
            {
                ERE(pso("Minimum relative pixel error is not modeled.\n"));
            }
            else
            {
                ERE(pso("Minimum relative pixel error (rpe) is %.1f%%.\n",
                        100.0 * min_relative_pixel_error));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                 min_relative_pixel_error = DBL_HALF_MOST_NEGATIVE;
            }
            else
            {
                ERE(ss1d(value, &temp_double));

                if (temp_double < 0.0)
                {
                    set_error("Minimum relative pixel error must be non-negative.");
                    return ERROR;
                }

#ifdef USE_ONLY_ONE_OF_RE_AND_AE
                min_absolute_pixel_error = DBL_HALF_MOST_NEGATIVE;
#endif
                min_relative_pixel_error = temp_double;
            }
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "rse"))
         || (match_pattern(lc_option, "relative-sensor-error"))
       )
    {
        IMPORT double relative_sensor_error;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("relative-sensor-error = %.1f%%.\n",
                    (100.0 * relative_sensor_error)));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Relative sensor error (rse) is %.1f%%.\n",
                    (100.0 * relative_sensor_error)));
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < -DBL_EPSILON)
            {
                set_error("Relative sensor error must be non-negative.");
                return ERROR;
            }

            if (temp_real < 0.0) temp_real = 0.0;
            relative_sensor_error = temp_real;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "rde"))
         || (match_pattern(lc_option, "relative-diag-error"))
         || (match_pattern(lc_option, "relative-diagonal-error"))
       )
    {
        IMPORT double min_relative_diag_error;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("relative-diagonal-error = %.1f%%.\n",
                    (100.0 * min_relative_diag_error)));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Relative diag error (rde) is %.1f%%.\n",
                    (100.0 * min_relative_diag_error)));
        }
        else
        {
            ERE(ss1d(value, &temp_double));

            if (temp_double < -DBL_EPSILON)
            {
                set_error("Relative diag error must be non-negative.");
                return ERROR;
            }

            if (temp_double < 0.0) temp_double = 0.0;
            min_relative_diag_error = temp_double;
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int process_negative_data
(
    Matrix**             result_mpp,
    const Matrix*        data_mp,
    Negative_data_action negative_data_action,
    const char*          data_id_str
)
{

    if (negative_data_action == USE_NEGATIVE_DATA)
    {
        return copy_matrix(result_mpp, data_mp);
    }
    else if (negative_data_action == NEGATIVE_DATA_IS_ERROR)
    {
        if (min_matrix_element(data_mp) < 0.0)
        {
            set_error("Negative %s value found.", data_id_str);
            return ERROR;
        }
        else
        {
            return copy_matrix(result_mpp, data_mp);
        }
    }
    else if (negative_data_action == ZERO_NEGATIVE_DATA)
    {
        return min_thresh_matrix(result_mpp, data_mp, 0.0);
    }
    else if (negative_data_action == SKIP_NEGATIVE_DATA)
    {
        int     i;
        int     count     = 0;
        Matrix* result_mp ;

        ERE(get_target_matrix(result_mpp, data_mp->num_rows,
                              data_mp->num_cols));

        result_mp = *result_mpp;

        for (i=0; i<data_mp->num_rows; i++)
        {
            if (    ((data_mp->elements)[ i ][ 0 ] >= 0.1)
                 && ((data_mp->elements)[ i ][ 1 ] >= 0.1)
                 && ((data_mp->elements)[ i ][ 2 ] >= 0.1)
               )
            {
                (result_mp->elements)[ count ][0] = (data_mp->elements)[ i ][0];
                (result_mp->elements)[ count ][1] = (data_mp->elements)[ i ][1];
                (result_mp->elements)[ count ][2] = (data_mp->elements)[ i ][2];

                count++;
            }
            else
            {
                verbose_pso(20,
                            "Skipping %s point %d due to small component.\n",
                            data_id_str, i + 1);
            }
        }

        if (count == 0)
        {
            set_error("No %s points with positive values > 0.1.", data_id_str);
            return ERROR;
        }
        else if (count < data_mp->num_rows)
        {
            verbose_pso(4, "Skipping %d %s points to small componants.\n",
                        data_mp->num_rows - count, data_id_str);
            result_mp->num_rows = count;
        }

        return NO_ERROR;
    }
    else if (negative_data_action == SKIP_NEGATIVE_BLUE_DATA)
    {
        int     i;
        int     count     = 0;
        Matrix* result_mp ;

        ERE(get_target_matrix(result_mpp, data_mp->num_rows,
                              data_mp->num_cols));

        result_mp = *result_mpp;

        for (i=0; i<data_mp->num_rows; i++)
        {
            if ((data_mp->elements)[ i ][ 2 ] >= 0.0)
            {
                (result_mp->elements)[ count ][0] = (data_mp->elements)[ i ][0];
                (result_mp->elements)[ count ][1] = (data_mp->elements)[ i ][1];
                (result_mp->elements)[ count ][2] = (data_mp->elements)[ i ][2];

                count++;
            }
            else
            {
                verbose_pso(20,
                        "Skipping %s point %d due to negative blue component.\n",
                         data_id_str, i + 1);
            }
        }

        if (count == 0)
        {
            set_error("No %s points with only positive blue values.",
                      data_id_str);
            return ERROR;
        }
        else if (count < data_mp->num_rows)
        {
            verbose_pso(2,
                        "Skipping %d %s points to negative blue componants.\n",
                        data_mp->num_rows - count, data_id_str);
            result_mp->num_rows = count;
        }

        return NO_ERROR;
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

