/* =============================================================================
 *
 * NOTE: This is a header file.
 *
 * Author: Andrew Emmott (aemmott)
 *
 * -------------------------------------------------------------------------- */

#include <g/g_point_projection.h>

#ifdef __cplusplus
extern "C" {
#endif

/* =============================================================================
 *                                 points_from_plucker
 *
 * TODO Short Description
 *
 * TODO Description
 *
 * Returns:
 *    TODO Returns
 *
 * See also:
 *    TODO See also
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: TODO Index
 * -------------------------------------------------------------------------- */
int points_from_plucker
(
    Vector **point_one,
    Vector **point_two,
    const Vector *plucker_line
)
{
    Vector *_vp1 = NULL;
    Vector *_vp2 = NULL;

    /*Error Checking*/
    /*The line must be a valid plucker representation.*/
    if (is_plucker_line(plucker_line) == 0)
    {
        G_POINT_PROJECTION_PLUCKER_ERROR;
        return ERROR;
    }
    /*Allocation errors.*/
    if (get_target_vector(point_one, 4) == ERROR || get_target_vector(point_two, 4) == ERROR)
    {
        G_POINT_PROJECTION_ALLOC_ERROR;
        return ERROR;
    }
    /*End Error Checking*/

    /*The strategy is to observe that at least one of the axis planes is intersected.
      It can be assumed that one of the coordinates is 1 and -1 respectively, and the others
      can be reverse engineered.*/
    _vp1 = *point_one;
    _vp1->elements[3] = 1.0;
    _vp2 = *point_two;
    _vp2->elements[3] = 1.0;
    /*X-axis plane is intersected.*/
    if (fabs(plucker_line->elements[2]) > DBL_EPSILON)
    {
        _vp1->elements[0] = 1;
        _vp2->elements[0] = -1;
        _vp1->elements[1] = (plucker_line->elements[0]-plucker_line->elements[4])/plucker_line->elements[2];
        _vp2->elements[1] = (plucker_line->elements[0]+plucker_line->elements[4])/plucker_line->elements[2];
        _vp1->elements[2] = (plucker_line->elements[1]+plucker_line->elements[5])/plucker_line->elements[2];
        _vp2->elements[2] = (plucker_line->elements[1]-plucker_line->elements[5])/plucker_line->elements[2];
    }
    /*Y-axis plane is intersected.*/
    else if (fabs(plucker_line->elements[4]) > DBL_EPSILON)
    {
        _vp1->elements[1] = -1;
        _vp2->elements[1] = 1;
        _vp1->elements[0] = (plucker_line->elements[0]+plucker_line->elements[2])/plucker_line->elements[4];
        _vp2->elements[0] = (plucker_line->elements[0]-plucker_line->elements[2])/plucker_line->elements[4];
        _vp1->elements[2] = (plucker_line->elements[5]-plucker_line->elements[3])/plucker_line->elements[4];
        _vp2->elements[2]= (plucker_line->elements[5]+plucker_line->elements[3])/-plucker_line->elements[4];
    }
    /*Z=axis plane is intersected.*/
    else /*Implicit: if (plucker_line->elements[5] != 0)*/
    {
        _vp1->elements[2] = 1;
        _vp2->elements[2] = -1;
        _vp1->elements[0]=(plucker_line->elements[2]-plucker_line->elements[1])/plucker_line->elements[5];
        _vp2->elements[0]=(plucker_line->elements[2]+plucker_line->elements[1])/-plucker_line->elements[5];
        _vp1->elements[1]=(plucker_line->elements[4]+plucker_line->elements[3])/-plucker_line->elements[5];
        _vp2->elements[1]=(plucker_line->elements[4]-plucker_line->elements[3])/plucker_line->elements[5];
    }

    return NO_ERROR;
}

/* =============================================================================
 *                                 distance_between_lines_as_points
 *
 * TODO Short Description
 *
 * TODO Description
 *
 * Returns:
 *    TODO Returns
 *
 * See also:
 *    TODO See also
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: TODO Index
 * -------------------------------------------------------------------------- */
int distance_between_lines_as_points
(
    double *distance,
    int *is_parallel,
    const Vector *line_one_point_one,
    const Vector *line_one_point_two,
    const Vector *line_two_point_one,
    const Vector *line_two_point_two
)
{
    double ux;
    double uy;
    double uz;
    double vx;
    double vy;
    double vz;
    double wx;
    double wy;
    double wz;
    double a;
    double b;
    double c;
    double d;
    double e;
    double denom;
    double Sc;
    double Tc;
    double distx;
    double disty;
    double distz;
    
    /*Error Checking*/
    if (line_one_point_one->length != 4 || line_one_point_two->length != 4 || line_two_point_one->length != 4 || line_two_point_two->length != 4)
    {
        G_POINT_PROJECTION_SIZE_ERROR;
        return ERROR;
    }
    if (line_one_point_one->elements[3] != 1.0 || line_one_point_two->elements[3] != 1.0 ||
        line_two_point_one->elements[3] != 1.0 || line_two_point_two->elements[3] != 1.0)
    {
        G_POINT_PROJECTION_NORM_ERROR;
        return ERROR;
    }
    /*End Error Checking*/

    /*Calculating faux-vectors in stack memory: u, v and w*/
    ux = line_one_point_two->elements[0] - line_one_point_one->elements[0];
    uy = line_one_point_two->elements[1] - line_one_point_one->elements[1];
    uz = line_one_point_two->elements[2] - line_one_point_one->elements[2];
    vx = line_two_point_two->elements[0] - line_two_point_one->elements[0];
    vy = line_two_point_two->elements[1] - line_two_point_one->elements[1];
    vz = line_two_point_two->elements[2] - line_two_point_one->elements[2];
    wx = line_one_point_one->elements[0] - line_two_point_one->elements[0];
    wy = line_one_point_one->elements[1] - line_two_point_one->elements[1];
    wz = line_one_point_one->elements[2] - line_two_point_one->elements[2];

    /*The needed dot products.*/
    a = ux*ux + uy*uy + uz*uz;
    b = ux*vx + uy*vy + uz*vz;
    c = vx*vx + vy*vy + vz*vz;
    d = ux*wx + uy*wy + uz*wz;
    e = vx*wx + vy*wy + vz*wz;

    /*The denominator of the Sc and Tc scalars.*/
    denom = a*c - b*b;

    /*If denominator is zero.*/
    if (fabs(denom) < DBL_EPSILON)
    {
        *is_parallel = 1;
        /*Implied: Sc = 0*/
        Tc = d/b;
        distx = wx + Tc*vx;
        disty = wy + Tc*vy;
        distz = wz + Tc*vz;
    } else { /*Implied: Denominator is non-zero.*/
        *is_parallel = 0;
        Sc = (b*e - c*d)/denom;
        Tc = (a*e - b*d)/denom;
        distx = wx + Sc*ux + Tc*vx;
        disty = wy + Sc*uy + Tc*vy;
        distz = wz + Sc*uz + Tc*vz;
    }

    /*The magnitude of the distance vector.*/
    *distance = sqrt(distx*distx + disty*disty + distz*distz);

    return NO_ERROR;
}

/* =============================================================================
 *                                 plane_from_lines_as_points
 *
 * TODO Short Description
 *
 * TODO Description
 *
 * Returns:
 *    TODO Returns
 *
 * See also:
 *    TODO See also
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: TODO Index
 * -------------------------------------------------------------------------- */
int plane_from_lines_as_points
(
    Vector **plane,
    const Vector *line_one_point_one,
    const Vector *line_one_point_two,
    const Vector *line_two_point_one,
    const Vector *line_two_point_two
)
{
    double distance;
    int is_parallel;
    Vector *line1 = NULL;
    Vector *line2 = NULL;
    Vector *_vp;

    /*Error Checking*/
    /*Distance must be checked*/
    distance_between_lines_as_points(&distance, &is_parallel, line_one_point_one, line_one_point_two, line_two_point_one, line_two_point_two);
    if (distance > DBL_EPSILON)
    {
        if (is_parallel)
        {
            subtract_vectors(&line1, line_two_point_one, line_one_point_one);
            subtract_vectors(&line2, line_two_point_two, line_one_point_two);
        }
        else
        {
            G_POINT_PROJECTION_NOT_COPLANAR;
            return ERROR;
        }
    }
    else
    {
        subtract_vectors(&line1, line_one_point_two, line_one_point_one);
        subtract_vectors(&line2, line_two_point_two, line_two_point_one);
    }
    /*Allocation Errors*/
    if (get_target_vector(plane, 4) == ERROR)
    {
        G_POINT_PROJECTION_ALLOC_ERROR;
        return ERROR;
    }
    /*End Error Checking*/

    _vp = *plane;
    /*Cross-Product of the two lines to get the normal.*/
    _vp->elements[0] = line1->elements[1]*line1->elements[2] - line1->elements[2]*line1->elements[1];
    _vp->elements[1] = line1->elements[0]*line1->elements[2] - line1->elements[2]*line1->elements[0];
    _vp->elements[2] = line1->elements[0]*line1->elements[1] - line1->elements[1]*line1->elements[0];
    _vp->elements[3] = -(_vp->elements[0]*line_one_point_one->elements[0]+_vp->elements[1]*line_one_point_one->elements[1]+_vp->elements[2]*line_one_point_one->elements[2]);

    return NO_ERROR;
}

/* =============================================================================
 *                                 plane_from_plucker_lines
 *
 * TODO Short Description
 *
 * TODO Description
 *
 * Returns:
 *    TODO Returns
 *
 * See also:
 *    TODO See also
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: TODO Index
 * -------------------------------------------------------------------------- */
int plane_from_plucker_lines
(
    Vector **plane,
    const Vector *plucker_one,
    const Vector *plucker_two
)
{
    Vector *line_one_point_one;
    Vector *line_one_point_two;
    Vector *line_two_point_one;
    Vector *line_two_point_two;

    /*No explicit error checking but ERE follows*/
    ERE(points_from_plucker(&line_one_point_one, &line_one_point_two, plucker_one));
    ERE(points_from_plucker(&line_two_point_one, &line_two_point_two, plucker_two));
    ERE(plane_from_lines_as_points(plane, line_one_point_one, line_one_point_two, line_two_point_one, line_two_point_two));
    /*End Error checking*/
    free(line_one_point_one);
    free(line_one_point_two);
    free(line_two_point_one);
    free(line_two_point_two);

    return NO_ERROR;
}
/* =============================================================================
 *                                 point_projection_matrix_from_line_projection_matrix
 *
 * TODO Short Description
 *
 * TODO Description
 *
 * Returns:
 *    TODO Returns
 *
 * See also:
 *    TODO See also
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: TODO Index
 * -------------------------------------------------------------------------- */
int point_projection_matrix_from_line_projection_matrix
(
    Matrix **point_projection_matrix,
    const Matrix *line_projection_matrix
)
{
    Vector *plucker_one;
    Vector *plucker_two;
    Vector *plucker_three;
    Vector *plane;
    
    /*Error Checking*/
    /*Ensure line projection matrix is the right size.*/
    if (line_projection_matrix->num_rows != 3 || line_projection_matrix->num_cols != 6)
    {
        G_POINT_PROJECTION_MAT_SIZE_ERROR;
        return ERROR;
    }
    /*Allocation Errors*/
    if (get_target_matrix(point_projection_matrix, 3, 4) == ERROR)
    {
        G_POINT_PROJECTION_MAT_ALLOC_ERROR;
        return ERROR;
    }
    /*End explicit error checking, but ERE follows*/
    ERE(get_matrix_row(&plucker_one, line_projection_matrix, 0));
    ERE(get_matrix_row(&plucker_two, line_projection_matrix, 1));
    ERE(get_matrix_row(&plucker_three, line_projection_matrix, 2));

    ERE(plane_from_plucker_lines(&plane, plucker_two, plucker_three));
    ERE(put_matrix_row(*point_projection_matrix, plane, 0));

    ERE(plane_from_plucker_lines(&plane, plucker_one, plucker_three));
    ERE(put_matrix_row(*point_projection_matrix, plane, 1));

    ERE(plane_from_plucker_lines(&plane, plucker_one, plucker_two));
    ERE(put_matrix_row(*point_projection_matrix, plane, 2));
    /*End Error checking*/

    return NO_ERROR;
}

#ifdef __cplusplus
}
#endif
