
/* $Id: l_debug.h 6352 2010-07-11 20:13:21Z kobus $ */

/* =========================================================================== *
|
|  Copyright (c) 1994-2008 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowledged in publications, and relevant papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarantee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_DEBUG_INCLUDED
#define L_DEBUG_INCLUDED

#include "l/l_def.h"

#ifdef __cplusplus
extern "C" {
#ifdef COMPILING_CPLUSPLUS_SOURCE
namespace kjb_c {
#endif
#endif

#ifdef TEST
#    define TEST_PSO(x) test_pso x
#else
#    define TEST_PSO(x)
#endif

#define TEST_P_STDERR  TEST_PSE

#ifdef TEST
#    define TEST_PSE(x) test_pse x
#else
#    define TEST_PSE(x)
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

/* ============================================================================
 *
 * (MACRO) Prints a warning message, but only on the first time it is executed.
 *
 * This macro alerts the user that the library code currently being executing
 * is especially untested, and not to be regarded as reliable.  However, if
 * the flow of control passes again through this macro, the message is not
 * repeated.  Thus one may invoke this macro inside a loop without causing
 * pages of identical warning messages.
 *
 * Since this macro embeds a static variable in the context of wherever it
 * gets executed, you should be wary of invoking it inside of another macro.
 *
 * Related:
 *   test_pse
 *
 * Author:
 *   Kobus Barnard
 *
 * Documentor:  Andrew Predoehl
 *
 * Index:  debugging, Macros, warning messages, TEST
 *
 * ----------------------------------------------------------------------------
 */
#ifndef __C2MAN__  /* I'm not using #else, following the example of 
                    * l/l_error.h line 63, which says that using #else would
                    * confuse ctags.
                    */

#endif /* ! C2MAN */

int  set_debug_options  (const char* option, const char* value);
int  kjb_set_debug_level(int new_level);
int  kjb_get_debug_level(void);
void hex_print          (FILE*, void*, size_t);
void get_status_string  (int status_num, char* buff, size_t buff_size);

#ifdef STANDARD_VAR_ARGS
#ifndef SGI /* SGI pretends it is lint and gives spurious messages. */
    /*PRINTFLIKE1*/
#endif
    void test_pse(const char*, ...);
    void test_pso(const char*, ...);
#else
    void test_pse(void);
    void test_pso(void);
#endif

#ifdef __cplusplus
#ifdef COMPILING_CPLUSPLUS_SOURCE
}
#endif
}
#endif

#endif

