
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                     /* Otherwise use gen.h.                       */

#include "version.h"

/* -------------------------------------------------------------------------- */

/*
// We put print_version in a separate file so that we can force compiling it
// with every compile with minimal overhead. We wish to force the compile of
// this routine so that the compilation time is correct.
*/

int print_version(void)
{
    char compile_time_str[ 10 ];
    const char* str_pos; 
#ifdef WITH_LIBS_HAVE_FLAGS
    const char *kjb_have_flags_pos;
    char kjb_have_define[ 100 ];
    int  first_have_string = TRUE;
#endif

    BUFF_CPY(compile_time_str, __TIME__);
    compile_time_str[ 5 ] = '\0';

    ERE(pso("\nGamut version %d.%d.%d%s (%s, %s).\n",
            MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL, DEVELOPMENT_STAGE,
            __DATE__, compile_time_str));

    if (DEVELOPMENT_MESSAGE[ 0 ] != '\0')
    {
        ERE(pso("\n  [  %s  ]\n", DEVELOPMENT_MESSAGE));
    }

    ERE(pso("\n"));

#ifdef PROGRAMMER
    ERE(pso("Compiled by:    %s \n", PROGRAMMER));
#endif

#ifdef CC_FLAGS_USED
    ERE(pso("Compile flags: %s\n", CC_FLAGS_USED));
#endif

#ifdef LIBS_USED
    str_pos = LIBS_USED;
    const_trim_beg(&str_pos); 
    ERE(pso("Libs used:     %s\n", str_pos));
#endif

#ifdef WITH_LIBS_HAVE_FLAGS
    /*
    ERE(pso("Have flags:    %s\n", WITH_LIBS_HAVE_FLAGS));
    */

    ERE(pso("Have flags:    "));

    kjb_have_flags_pos = WITH_LIBS_HAVE_FLAGS;

    const_trim_beg(&kjb_have_flags_pos);

    while (BUFF_CONST_GET_TOKEN_OK(&kjb_have_flags_pos, kjb_have_define))
    {
        const char* kjb_have_define_pos = kjb_have_define;

        if (first_have_string)
        {
            first_have_string = FALSE;
        }
        else
        {
             ERE(pso(" "));
        }

        ERE(pso(kjb_have_define_pos));
    }

    ERE(pso("\n"));
#endif

    ERE(pso("\nWritten by Kobus Barnard with help from Lindsay Martin.\n\n"));

    if ( ! is_interactive() )
    {
        char host_str[ 100 ];
        char cur_time_str[ 100 ];

        if (    (BUFF_GET_ENV("HOST", host_str) != ERROR)
             && (BUFF_GET_TIME(cur_time_str) != ERROR)
           )
        {
            pso("\n  [ Process %ld on %s (started at %s) ]\n\n",
                (long)MY_PID, host_str, cur_time_str);
        }
    }

#ifdef MAC_OSX
    kjb_puts("\n");
    kjb_puts("The default stacksize on the mac (8K) is a bit small for this program.\n");
    kjb_puts("If it crashes, try inscreasing it with the shell command 'unlimit 64000'.\n");
#endif 

    kjb_puts("\n");

    ERE(kjb_fflush(stdout));

    return NO_ERROR;
}

