
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#define EMERGENCY_PATCH_FOR_THESIS

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "illum_data.h"
#include "compute_output.h"

#include "compute_error.h"

/* ------------------------------------------------------------------------- */

static int fs_compute_illum_angular_error        = TRUE;
static int fs_compute_illum_sum_chrom_error      = TRUE;
static int fs_compute_illum_lab_error            = TRUE;
static int fs_compute_illum_rgb_error            = TRUE;
static int fs_compute_illum_r_div_sum_error      = FALSE;
static int fs_compute_illum_g_div_sum_error      = FALSE;
static int fs_compute_illum_sum_lum_error        = FALSE;
static int fs_compute_scene_blue_chrom_rms_error = FALSE;
static int fs_compute_scene_sum_chrom_rms_error  = FALSE;
static int fs_compute_scene_rgb_rms_error        = FALSE;
static int fs_compute_scene_lab_rms_error        = FALSE;
static int fs_compute_world_blue_chrom_rms_error = FALSE;
static int fs_compute_world_sum_chrom_rms_error  = FALSE;
static int fs_compute_world_rgb_rms_error        = FALSE;
static int fs_compute_world_lab_rms_error        = FALSE;
static int fs_actual_illum_lab_white             = FALSE;
static int fs_normalize_3D_illum_estimate        = FALSE;

/* -------------------------------------------------------------------------- */

static int do_errors
(
    const Vector*        actual_illum_vp,
    const Vector*        canon_illum_vp,
    const Matrix*        data_mp,
    const Matrix*        sharp_data_mp,
    const Matrix*        canon_data_mp,
    Result_by_algorithm* cc_result_ptr,
    Error_by_algorithm*  error_ptr,
    int                  (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);

static int do_best_fit_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Best_fit_result* best_fit_ptr,
    Best_fit_error*  error_ptr,
    int              (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);

static int do_standard_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Standard_result* standard_result_ptr,
    Standard_error*  error_ptr,
    int              (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);

static int do_gamut_errors
(
    const Vector* actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    Gamut_result* gamut_result_ptr,
    Gamut_error*  error_ptr,
    int           (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);

/* Lindsay: Feb 17/98 */
static int do_correlation_errors
(
    const Vector*       actual_illum_vp,
    const Vector*       canon_illum_vp,
    const Matrix*       data_mp,
    const Matrix*       sharp_data_mp,
    const Matrix*       canon_data_mp,
    Correlation_result* correlation_result_ptr,
    Correlation_error*  error_ptr,
    int                 (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);
/* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
static int do_simple_template_errors
(
    const Vector*          actual_illum_vp,
    const Vector*          canon_illum_vp,
    const Matrix*          data_mp,
    const Matrix*          sharp_data_mp,
    const Matrix*          canon_data_mp,
    Simple_template_result* simple_template_result_ptr,
    Simple_template_error*  error_ptr,
    int                    (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);
#endif

#ifdef TEMPLATE
static int do_template_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Template_result* template_result_ptr,
    Template_error*  error_ptr,
    int              (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);
#endif

#ifdef MULTI_TEMPLATE
static int do_multi_template_errors
(
    const Vector*          actual_illum_vp,
    const Vector*          canon_illum_vp,
    const Matrix*          data_mp,
    const Matrix*          sharp_data_mp,
    const Matrix*          canon_data_mp,
    Multi_template_result* multi_template_result_ptr,
    Multi_template_error*  error_ptr,
    int                    (*error_fn)(const Vector*, const Vector*, const Matrix*, const Matrix*, const Matrix*, const Vector*, const Matrix*, double*)
);
#endif

static int get_illum_angular_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_sum_chrom_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_lab_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_rgb_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_r_div_sum_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_g_div_sum_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_illum_sum_lum_error
(
    const Vector* actual_illum_vp,
    const Vector* derived_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
);

static int get_sum_chrom_map_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* sum_chrom_canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_sum_chrom_map_error_guts
(
    const Matrix* sharp_data_mp,
    const Matrix* sum_chrom_canon_data_mp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_blue_chrom_map_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* projected_canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_blue_chrom_map_error_guts
(
    const Matrix* sharp_data_mp,
    const Matrix* projected_canon_data_mp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_map_rms_error
(
    const Vector* actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_map_lab_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
);

static int get_delta_lab
(
    const Vector* actual_vp,
    const Vector* est_vp,
    double*       error_ptr
);

/* -------------------------------------------------------------------------- */

EXTERN_C int set_compute_error_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_boolean_value;
    int  do_all             = FALSE;
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "compute-all")))
    {
        do_all = TRUE;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-angular-error"))
               || (match_pattern(lc_option, "compute-angular-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-angular-error = %s\n",
                    fs_compute_illum_angular_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum angular error results %s computed.\n",
                    fs_compute_illum_angular_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_angular_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-sum-chrom-error"))
               || (match_pattern(lc_option, "compute-sum-chrom-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-sum-chrom-error = %s\n",
                    fs_compute_illum_sum_chrom_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum sum chrom error results %s computed.\n",
                    fs_compute_illum_sum_chrom_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_sum_chrom_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-lab-error"))
               || (match_pattern(lc_option, "compute-lab-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-lab-error = %s\n",
                    fs_compute_illum_lab_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum L*a*b error results %s computed.\n",
                    fs_compute_illum_lab_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_lab_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-rgb-error"))
               || (match_pattern(lc_option, "compute-rgb-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-rgb-error = %s\n",
                    fs_compute_illum_rgb_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum RGB error results %s computed.\n",
                    fs_compute_illum_rgb_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_rgb_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-r-error"))
               || (match_pattern(lc_option, "compute-illum-r-div-sum-error"))
               || (match_pattern(lc_option, "compute-r-error"))
               || (match_pattern(lc_option, "compute-r-div-sum-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-r-div-sum-error = %s\n",
                    fs_compute_illum_r_div_sum_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum R/(R+G+B) error results %s computed.\n",
                    fs_compute_illum_r_div_sum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_r_div_sum_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-g-error"))
               || (match_pattern(lc_option, "compute-illum-g-div-sum-error"))
               || (match_pattern(lc_option, "compute-g-error"))
               || (match_pattern(lc_option, "compute-g-div-sum-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-g-div-sum-error = %s\n",
                    fs_compute_illum_g_div_sum_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum G/(R+G+B) error results %s computed.\n",
                    fs_compute_illum_g_div_sum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_g_div_sum_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (do_all || (match_pattern(lc_option, "compute-illum-sum-lum-error"))
               || (match_pattern(lc_option, "compute-sum-lum-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-illum-sum-lum-error = %s\n",
                    fs_compute_illum_sum_lum_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum R+G+B error results %s computed.\n",
                    fs_compute_illum_sum_lum_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_illum_sum_lum_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-rgb-rms-error")) /* Does both */
         || (match_pattern(lc_option, "compute-scene-rgb-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scene-rgb-rms-error = %s\n",
                    fs_compute_scene_rgb_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene RGB rms error results %s computed.\n",
                    fs_compute_scene_rgb_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scene_rgb_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-blue-chrom-rms-error"))
         || (match_pattern(lc_option, "compute-scene-blue-chrom-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scene-blue-chrom-rms-error = %s\n",
                    fs_compute_scene_blue_chrom_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene blue chrom rms error results %s computed.\n",
                    fs_compute_scene_blue_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scene_blue_chrom_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-sum-chrom-rms-error"))
         || (match_pattern(lc_option, "compute-scene-sum-chrom-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scene-sum-chrom-rms-error = %s\n",
                    fs_compute_scene_sum_chrom_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene sum chrom rms error results %s computed.\n",
                    fs_compute_scene_sum_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scene_sum_chrom_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-lab-rms-error"))
         || (match_pattern(lc_option, "compute-scene-lab-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scene-lab-rms-error = %s\n",
                    fs_compute_scene_lab_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scene L*a*b error results %s computed.\n",
                    fs_compute_scene_lab_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scene_lab_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-rgb-rms-error"))
         || (match_pattern(lc_option, "compute-world-rgb-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-world-rgb-rms-error = %s\n",
                    fs_compute_world_rgb_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World RGB rms error results %s computed.\n",
                    fs_compute_world_rgb_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_world_rgb_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-blue-chrom-rms-error"))
         || (match_pattern(lc_option, "compute-world-blue-chrom-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-world-blue-chrom-rms-error = %s\n",
                    fs_compute_world_blue_chrom_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World blue chrom rms error results %s computed.\n",
                    fs_compute_world_blue_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_world_blue_chrom_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-sum-chrom-rms-error"))
         || (match_pattern(lc_option, "compute-world-sum-chrom-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-world-sum-chrom-rms-error = %s\n",
                    fs_compute_world_sum_chrom_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World sum chrom rms error results %s computed.\n",
                    fs_compute_world_sum_chrom_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_world_sum_chrom_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    do_all
         || (match_pattern(lc_option, "compute-lab-rms-error"))
         || (match_pattern(lc_option, "compute-world-lab-rms-error"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-world-lab-rms-error = %s\n",
                    fs_compute_world_lab_rms_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("World L*a*b error results %s computed.\n",
                    fs_compute_world_lab_rms_error ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_world_lab_rms_error = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') && (value[ 0 ] == '\0'))
    {
        ERE(pso("\n"));
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "actual-illum-lab-white"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("actual-illum-lab-white = %s\n",
                    fs_actual_illum_lab_white ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_actual_illum_lab_white)
            {
                ERE(pso("Actual illum is used for L*a*b white.\n"));
            }
            else
            {
                ERE(pso("(255, 255, 255) is used for L*a*b white.\n"));
            }
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_actual_illum_lab_white = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "normalize-3d-illum-estimate"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("normalize-3d-illum-estimate = %s\n",
                    fs_normalize_3D_illum_estimate ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D illum estimates %s normed before computing L*a*b.\n",
                    fs_normalize_3D_illum_estimate ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_normalize_3D_illum_estimate = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int compute_error
(
    int                     num_data_points,
    int*                    data_point_nums,
    Sharp_method            sharp_method,
    Result_by_algorithm*    cc_result_ptr,
    Colour_constancy_error* error_ptr
)
{
    Matrix* canon_data_mp                    = NULL;
    Matrix* selected_canon_data_mp           = NULL;
    Matrix* projected_canon_data_mp          = NULL;
    Matrix* projected_selected_canon_data_mp = NULL;
    Matrix* sum_chrom_canon_data_mp          = NULL;
    Matrix* sum_chrom_selected_canon_data_mp = NULL;
    Matrix* data_mp                          = NULL;
    Matrix* selected_data_mp                 = NULL;
    Matrix* sharp_data_mp                    = NULL;
    Matrix* sharp_selected_data_mp           = NULL;
    Matrix* derived_data_mp;
    Matrix* derived_selected_data_mp;
    Vector* actual_illum_vp                  = NULL;
    Vector* canon_illum_vp                   = NULL;
    Vector* sharp_canon_illum_vp             = NULL;
    const Vector* derived_canon_illum_vp = NULL;
    int     result                           = NO_ERROR;

#ifdef __GNUC__
    /* Reduce the noise from error checkers. */
    derived_data_mp = NULL;
    derived_selected_data_mp = NULL;
#endif

    dbi(num_data_points);

    if (    (fs_compute_scene_lab_rms_error)
         || (fs_compute_scene_rgb_rms_error)
         || (fs_compute_scene_blue_chrom_rms_error)
         || (fs_compute_scene_sum_chrom_rms_error)
       )
    {
        (void)get_selected_registered_data(&selected_data_mp, num_data_points,
                                           data_point_nums);
        (void)get_selected_canon_data(&selected_canon_data_mp, num_data_points,
                                      data_point_nums);
    }

    /*
    // These data items are optional, so ignore the returns. (If the get
    // routines fail, the data pointers remain NULL).
    */
    (void)get_registered_data(&data_mp);
    (void)get_canon_data(&canon_data_mp);

    (void)get_canon_illum(&canon_illum_vp);
    (void)get_actual_illum(&actual_illum_vp);

    if (data_mp != NULL)
    {
        dbi(data_mp->num_rows);
    }

    if (selected_data_mp != NULL)
    {
        dbi(selected_data_mp->num_rows);
    }

    if (canon_data_mp != NULL)
    {
        result = ow_normalize_canon_data(canon_data_mp, data_mp);

        if (result != ERROR)
        {
            result = get_projection_matrix(&projected_canon_data_mp,
                                           canon_data_mp, DIVIDE_BY_BLUE,
                                           TRUNCATE_INVALID_CHROMATICITY,
                                           MAX_X_DIV_B,
                                           DBL_NOT_SET, DBL_NOT_SET,
                                           (Matrix**)NULL, (Matrix**)NULL);
        }

        if (result != ERROR)
        {
            result = get_projection_matrix(&sum_chrom_canon_data_mp,
                                           canon_data_mp, DIVIDE_BY_SUM,
                                           ZERO_INVALID_CHROMATICITY,
                                           DBL_NOT_SET,
                                           DBL_NOT_SET, DBL_NOT_SET,
                                           (Matrix**)NULL, (Matrix**)NULL);
        }
    }

    if (    (result != ERROR)
         && (selected_canon_data_mp != NULL)
         && (selected_data_mp != NULL)
       )
    {
        result = ow_normalize_canon_data(selected_canon_data_mp,
                                         selected_data_mp);

        if (result != ERROR)
        {
            result = get_projection_matrix(&projected_selected_canon_data_mp,
                                           selected_canon_data_mp,
                                           DIVIDE_BY_BLUE,
                                           TRUNCATE_INVALID_CHROMATICITY,
                                           MAX_X_DIV_B,
                                           DBL_NOT_SET, DBL_NOT_SET,
                                           (Matrix**)NULL, (Matrix**)NULL);
        }

        if (result != ERROR)
        {
            result = get_projection_matrix(&sum_chrom_selected_canon_data_mp,
                                           selected_canon_data_mp,
                                           DIVIDE_BY_SUM,
                                           ZERO_INVALID_CHROMATICITY,
                                           DBL_NOT_SET,
                                           DBL_NOT_SET, DBL_NOT_SET,
                                           (Matrix**)NULL, (Matrix**)NULL);
        }
    }

    /*
    //  The logic guiding which parameters are sharp and which are not is
    //  tricky. We sharpen the data, and the in the case of maps which are in
    //  sharpened space, we sharpen the canonical illuminant RGB. We  do NOT
    //  sharpen the canonical data, nor the actual illuminant RGB.
    //
    //  We always want the error measurements to take place in un-sharpend
    //  space.   Thus to calculate the mapping error, we use the result (which
    //  is sharp) to map the sharpened data, unsharpen, and then compare with
    //  the non-sharp canonical data.
    //
    //  In the case of actual illuminant RGB based error measures, the mapping
    //  is used to map sharp canon illuminant RGB to sharp actual RGB,
    //  unsharpen, and then compare with the non-sharp actual illuminant RGB.
    //
    //  Note that the sharpening routines either set or preserve NULL pointers
    //  if data pointer was NULL.
    //
    //  Since sharp_method must coincide with the global sharpening state, we
    //  could get the derived data from sharpen_matrix_data() even when we are
    //  not sharpening, but instead we take advantage of knowing the
    //  sharp_method to avoid an unecessary matrix copy.
    */

    if (result != ERROR)
    {
        if (sharp_method != DONT_SHARPEN)
        {
            result = sharpen_matrix_data(&sharp_data_mp, data_mp);

            if (result != ERROR)
            {
                result = sharpen_matrix_data(&sharp_selected_data_mp,
                                             selected_data_mp);
            }

            derived_data_mp = sharp_data_mp;
            derived_selected_data_mp = sharp_selected_data_mp;
        }
        else
        {
            derived_data_mp = data_mp;
            derived_selected_data_mp = selected_data_mp;
        }
    }

    if (result != ERROR)
    {
        if (sharp_method != DONT_SHARPEN)
        {
            result = sharpen_vector_data(&sharp_canon_illum_vp, canon_illum_vp);
            derived_canon_illum_vp = sharp_canon_illum_vp;
        }
        else
        {
            derived_canon_illum_vp = canon_illum_vp;
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_angular_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_angular.sharp[ sharp_method ]),
                           get_illum_angular_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum angular error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_sum_chrom_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_sum_chrom.sharp[ sharp_method ]),
                           get_illum_sum_chrom_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum sum chrom error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_rgb_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_rgb.sharp[ sharp_method ]),
                           get_illum_rgb_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum rgb error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_r_div_sum_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_r_div_sum.sharp[ sharp_method ]),
                           get_illum_r_div_sum_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum R/(R+G+B) error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_g_div_sum_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_g_div_sum.sharp[ sharp_method ]),
                           get_illum_g_div_sum_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum G/(R+G+B) error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_sum_lum_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_sum_lum.sharp[ sharp_method ]),
                           get_illum_sum_lum_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum R+G+B error.");
        }
    }

    if (    (result != ERROR)
         && (actual_illum_vp != NULL)
         && (derived_canon_illum_vp != NULL)
         && (fs_compute_illum_lab_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->illum_lab.sharp[ sharp_method ]),
                           get_illum_lab_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing illum lab error.");
        }
    }

    if (    (result != ERROR)
         && (derived_selected_data_mp != NULL)
         && (sum_chrom_selected_canon_data_mp != NULL)
         && (fs_compute_scene_sum_chrom_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           selected_data_mp, derived_selected_data_mp,
                           sum_chrom_selected_canon_data_mp, cc_result_ptr,
                           &(error_ptr->scene_map_sum_chrom_rms.sharp[ sharp_method]),
                           get_sum_chrom_map_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing sum chrom map error.");
        }
    }

    if (    (result != ERROR)
         && (derived_selected_data_mp != NULL)
         && (projected_selected_canon_data_mp != NULL)
         && (fs_compute_scene_blue_chrom_rms_error))
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           selected_data_mp, derived_selected_data_mp,
                           projected_selected_canon_data_mp, cc_result_ptr,
                           &(error_ptr->scene_map_blue_chrom_rms.sharp[sharp_method]),
                           get_blue_chrom_map_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing blue chrom map error.");
        }
    }

    if (    (result != ERROR)
         && (selected_data_mp != NULL)
         && (selected_canon_data_mp != NULL)
         && (fs_compute_scene_rgb_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           selected_data_mp, derived_selected_data_mp,
                           selected_canon_data_mp, cc_result_ptr,
                           &(error_ptr->scene_map_3D_rms.sharp[ sharp_method ]),
                           get_map_rms_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing map rms error.");
        }
    }

    if (    (result != ERROR)
         && (selected_data_mp != NULL)
         && (selected_canon_data_mp != NULL)
         && (fs_compute_scene_lab_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           selected_data_mp, derived_selected_data_mp,
                           selected_canon_data_mp, cc_result_ptr,
                           &(error_ptr->scene_map_3D_lab.sharp[ sharp_method ]),
                           get_map_lab_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing map lab error.");
        }
    }

    if (    (result != ERROR)
         && (derived_data_mp != NULL)
         && (sum_chrom_canon_data_mp != NULL)
         && (fs_compute_world_sum_chrom_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           sum_chrom_canon_data_mp, cc_result_ptr,
                           &(error_ptr->world_map_sum_chrom_rms.sharp[ sharp_method]),
                           get_sum_chrom_map_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing sum chrom map error.");
        }
    }

    if (    (result != ERROR)
         && (derived_data_mp != NULL)
         && (projected_canon_data_mp != NULL)
         && (fs_compute_world_blue_chrom_rms_error))
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           projected_canon_data_mp, cc_result_ptr,
                           &(error_ptr->world_map_blue_chrom_rms.sharp[sharp_method]),
                           get_blue_chrom_map_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing blue chrom map error.");
        }
    }

    if (    (result != ERROR)
         && (data_mp != NULL)
         && (canon_data_mp != NULL)
         && (fs_compute_world_rgb_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->world_map_3D_rms.sharp[ sharp_method ]),
                           get_map_rms_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing map rms error.");
        }
    }

    if (    (result != ERROR)
         && (data_mp != NULL)
         && (canon_data_mp != NULL)
         && (fs_compute_world_lab_rms_error)
       )
    {
        result = do_errors(actual_illum_vp,
                           derived_canon_illum_vp,
                           data_mp, derived_data_mp,
                           canon_data_mp, cc_result_ptr,
                           &(error_ptr->world_map_3D_lab.sharp[ sharp_method ]),
                           get_map_lab_error);

        if (result == ERROR)
        {
            add_error("Error occured while computing map lab error.");
        }
    }

    free_vector(actual_illum_vp);
    free_vector(canon_illum_vp);
    free_matrix(data_mp);
    free_matrix(canon_data_mp);
    free_matrix(projected_canon_data_mp);
    free_matrix(sum_chrom_canon_data_mp);
    free_matrix(selected_data_mp);
    free_matrix(selected_canon_data_mp);
    free_matrix(projected_selected_canon_data_mp);
    free_matrix(sum_chrom_selected_canon_data_mp);

    free_matrix(sharp_data_mp);
    free_matrix(sharp_selected_data_mp);
    free_vector(sharp_canon_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_errors
(
    const Vector*        actual_illum_vp,
    const Vector*        canon_illum_vp,
    const Matrix*        data_mp,
    const Matrix*        sharp_data_mp,
    const Matrix*        canon_data_mp,
    Result_by_algorithm* cc_result_ptr,
    Error_by_algorithm*  error_ptr,
    int                  (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int result, i;

    ERE(do_best_fit_errors(actual_illum_vp, canon_illum_vp,
                           data_mp, sharp_data_mp, canon_data_mp,
                           &(cc_result_ptr->best_fit),
                           &(error_ptr->best_fit),
                           error_fn));

    ERE(do_standard_errors(actual_illum_vp, canon_illum_vp,
                           data_mp, sharp_data_mp, canon_data_mp,
                           &(cc_result_ptr->standard), &(error_ptr->standard),
                           error_fn));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        result = do_gamut_errors(actual_illum_vp, canon_illum_vp,
                                 data_mp, sharp_data_mp, canon_data_mp,
                                 &(cc_result_ptr->gamut_2D[ i ]),
                                 &(error_ptr->gamut_2D[ i ]),
                                 error_fn);

        if (result == ERROR)
        {
            add_error("Error occured while computing 2D gamut mapping %d.",
                      i+1);
            return ERROR;
        }
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        result = do_gamut_errors(actual_illum_vp, canon_illum_vp,
                                 data_mp, sharp_data_mp, canon_data_mp,
                                 &(cc_result_ptr->gamut_3D[ i ]),
                                 &(error_ptr->gamut_3D[ i ]),
                                 error_fn);

        if (result == ERROR)
        {
            add_error("Error occured while computing 3D gamut mapping %d.",
                      i+1);
            return ERROR;
        }
    }

/* Lindsay: Feb 17/98 */
    ERE(do_correlation_errors(actual_illum_vp, canon_illum_vp,
                              data_mp, sharp_data_mp, canon_data_mp,
                              &(cc_result_ptr->correlation),
                              &(error_ptr->correlation),
                              error_fn));
/* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
    ERE(do_simple_template_errors(actual_illum_vp, canon_illum_vp,
                                 data_mp, sharp_data_mp, canon_data_mp,
                                 &(cc_result_ptr->simple_template),
                                 &(error_ptr->simple_template),
                                 error_fn));
#endif

#ifdef TEMPLATE
    ERE(do_template_errors(actual_illum_vp, canon_illum_vp,
                           data_mp, sharp_data_mp, canon_data_mp,
                           &(cc_result_ptr->template), &(error_ptr->template),
                           error_fn));
#endif

#ifdef MULTI_TEMPLATE
    ERE(do_multi_template_errors(actual_illum_vp, canon_illum_vp,
                                 data_mp, sharp_data_mp, canon_data_mp,
                                 &(cc_result_ptr->multi_template),
                                 &(error_ptr->multi_template),
                                 error_fn));
#endif

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_best_fit_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Best_fit_result* best_fit_ptr,
    Best_fit_error*  error_ptr,
    int              (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int result;

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (Vector*)NULL,
                         (const Matrix*)(best_fit_ptr->world_best_3D_fit_mp),
                         &(error_ptr->world_best_3D_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing world best 3D fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)NULL,
                         (const Matrix*)(best_fit_ptr->world_best_2D_fit_mp),
                         &(error_ptr->world_best_2D_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing world best 2D fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)(best_fit_ptr->world_best_3D_diag_fit_vp),
                         (const Matrix*)NULL, &(error_ptr->world_best_3D_diag_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing world best 3D diag fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)(best_fit_ptr->world_best_2D_diag_fit_vp),
                         (const Matrix*)NULL,
                         &(error_ptr->world_best_2D_diag_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing world best 2D diag fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (Vector*)NULL,
                         (const Matrix*)(best_fit_ptr->scene_best_3D_fit_mp),
                         &(error_ptr->scene_best_3D_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing scene best 3D fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)NULL,
                         (const Matrix*)(best_fit_ptr->scene_best_2D_fit_mp),
                         &(error_ptr->scene_best_2D_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing scene best 2D fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)(best_fit_ptr->scene_best_3D_diag_fit_vp),
                         (const Matrix*)NULL, &(error_ptr->scene_best_3D_diag_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing scene best 3D diag fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)(best_fit_ptr->scene_best_2D_diag_fit_vp),
                         (const Matrix*)NULL,
                         &(error_ptr->scene_best_2D_diag_fit));

    if (result == ERROR)
    {
        add_error("Error occured while computing scene best 2D diag fit error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         (const Vector*)(best_fit_ptr->actual_vp),
                         (const Matrix*)NULL,
                         &(error_ptr->actual));

    if (result == ERROR)
    {
        add_error("Error occured while computing actual error.");
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_standard_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Standard_result* standard_result_ptr,
    Standard_error*  error_ptr,
    int              (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int     result;
    Vector* identity_diag_vp;

    NRE(identity_diag_vp = create_vector(3));

    ERE(ow_set_vector(identity_diag_vp, 1.0));

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         identity_diag_vp, (Matrix*)NULL,
                         &(error_ptr->nothing));

    free_vector(identity_diag_vp);

    if (result == ERROR)
    {
        add_error("Error occured while computing nothing error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         standard_result_ptr->ave_illum_vp,
                         (Matrix*)NULL, &(error_ptr->ave_illum));

    if (result == ERROR)
    {
        add_error("Error occured while computing ave illum error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         standard_result_ptr->grey_world_vp,
                         (Matrix*)NULL, &(error_ptr->grey_world));

    if (result == ERROR)
    {
        add_error("Error occured while computing grey world error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         standard_result_ptr->db_grey_vp,
                         (Matrix*)NULL, &(error_ptr->db_grey_world));

    if (result == ERROR)
    {
        add_error("Error occured while computing db grey world error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         standard_result_ptr->scale_by_max_vp,
                         (Matrix*)NULL, &(error_ptr->scale_by_max));

    if (result == ERROR)
    {
        add_error("Error occured while computing scale by max error.");
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_gamut_errors
(
    const Vector* actual_illum_vp,
    const Vector* canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    Gamut_result* gamut_result_ptr,
    Gamut_error*  error_ptr,
    int           (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int result, i;

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->mv_surface_vp,
                         (Matrix*)NULL,
                         &(error_ptr->mv_surface));

    if (result == ERROR)
    {
        add_error("Error occured while computing mv surface error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->ha_surface_vp,
                         (Matrix*)NULL,
                         &(error_ptr->ha_surface));

    if (result == ERROR)
    {
        add_error("Error occured while computing ha surface error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->wa_surface_vp,
                         (Matrix*)NULL,
                         &(error_ptr->wa_surface));

    if (result == ERROR)
    {
        add_error("Error occured while computing wa surface error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->mv_surface_illum_vp,
                         (Matrix*)NULL,
                         &(error_ptr->mv_surface_illum));

    if (result == ERROR)
    {
        add_error("Error occured while computing mv surface illum error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->ha_surface_illum_vp,
                         (Matrix*)NULL,
                         &(error_ptr->ha_surface_illum));

    if (result == ERROR)
    {
        add_error("Error occured while computing ha surface illum error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->wa_surface_illum_vp,
                         (Matrix*)NULL,
                         &(error_ptr->wa_surface_illum));

    if (result == ERROR)
    {
        add_error("Error occured while computing wa surface illum error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         gamut_result_ptr->icsa_surface_illum_vp,
                         (Matrix*)NULL,
                         &(error_ptr->icsa_surface_illum));

    if (result == ERROR)
    {
        add_error("Error occured while computing icsa surface illum error.");
        return ERROR;
    }

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->scia_surface_illum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->scia_surface_illum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing scia surface illum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->scca_surface_illum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->scca_surface_illum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing scca surface illum %d error.", i+1);
            return ERROR;
        }
    }

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->scla_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->scla_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing scla surface illum lum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->lcsa_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->lcsa_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing lcsa surface illum lum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->ha_surface_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->ha_surface_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing ha surface lum %d error.",
                      i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->mv_surface_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->mv_surface_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing mv surface lum %d error.",
                      i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->ha_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->ha_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing ha surface illum lum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->mv_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->mv_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing mv surface illum lum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->scia_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->scia_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing scia surface illum lum %d error.", i+1);
            return ERROR;
        }

        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             gamut_result_ptr->scca_surface_illum_lum_vp[ i ],
                             (Matrix*)NULL,
                             &(error_ptr->scca_surface_illum_lum[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing scca surface illum lum %d error.", i+1);
            return ERROR;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay: Sept 30/98 */
static int do_correlation_errors
(
    const Vector*       actual_illum_vp,
    const Vector*       canon_illum_vp,
    const Matrix*       data_mp,
    const Matrix*       sharp_data_mp,
    const Matrix*       canon_data_mp,
    Correlation_result* correlation_result_ptr,
    Correlation_error*  error_ptr,
    int                 (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int i; /* Modified Apr 14/99 - Arrays of errors and results */
    int result; /* Kobus: Added extra error messages. */

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->correlation_01_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_01[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation 01 %d error.",
                      i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->correlation_01_alt_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_01_alt[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation 01 alt %d error.",
                      i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                         correlation_result_ptr->correlation_gaussian_MAP_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_gaussian_MAP[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation gaussian MAP %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MAP_inv_dist_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform inv dist fill MAP %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
            correlation_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[i],
                            (Matrix*)NULL,
                            &(error_ptr->correlation_uniform_MAP_gaussian_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform gaussian fill MAP %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                 correlation_result_ptr->correlation_uniform_MAP_fitted_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MAP_fitted[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform fitted MAP %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                         correlation_result_ptr->correlation_gaussian_MLM_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_gaussian_MLM[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation gaussian MLM %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                        correlation_result_ptr->correlation_gaussian_MMSE_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_gaussian_MMSE[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation gaussian MMSE %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
            correlation_result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MLM_inv_dist_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform inv dist fill MLM %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->
                                 correlation_uniform_MMSE_inv_dist_fill_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MMSE_inv_dist_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform inv dist fill MMSE %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->
                                    correlation_uniform_MLM_gaussian_fill_vp[i],
                            (Matrix*)NULL,
                        &(error_ptr->correlation_uniform_MLM_gaussian_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform gaussian fill MLM %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->
                                   correlation_uniform_MMSE_gaussian_fill_vp[i],
                            (Matrix*)NULL,
                       &(error_ptr->correlation_uniform_MMSE_gaussian_fill[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform gaussian fill MMSE %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->
                             correlation_uniform_MLM_fitted_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MLM_fitted[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform fitted MLM %d error.", i+1);
            return ERROR;
        }
    }

    for (i = 0; i < NUM_CORMATS_PER_ALG; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             correlation_result_ptr->
                             correlation_uniform_MMSE_fitted_vp[i],
                             (Matrix*)NULL,
                             &(error_ptr->correlation_uniform_MMSE_fitted[i]));

        if (result == ERROR)
        {
            add_error("Error occured while computing correlation uniform fitted MMSE %d error.", i+1);
            return ERROR;
        }
    }

    return NO_ERROR;
}
/* End Lindsay: Sept 30/98 */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static int do_simple_template_errors
(
    const Vector*          actual_illum_vp,
    const Vector*          canon_illum_vp,
    const Matrix*          data_mp,
    const Matrix*          sharp_data_mp,
    const Matrix*          canon_data_mp,
    Simple_template_result* simple_template_result_ptr,
    Simple_template_error*  error_ptr,
    int                    (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int result;

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         simple_template_result_ptr->alg_vp,
                         (Matrix*)NULL, &(error_ptr->alg));

    if (result == ERROR)
    {
        add_error("Error occured while computing simple-template error.");
        return ERROR;
    }

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static int do_template_errors
(
    const Vector*    actual_illum_vp,
    const Vector*    canon_illum_vp,
    const Matrix*    data_mp,
    const Matrix*    sharp_data_mp,
    const Matrix*    canon_data_mp,
    Template_result* template_result_ptr,
    Template_error*  error_ptr,
    int              (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int result;

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         template_result_ptr->first_alg_vp,
                         (Matrix*)NULL, &(error_ptr->first_alg));

    if (result == ERROR)
    {
        add_error("Error occured while computing template first alg error.");
        return ERROR;
    }

    result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                         data_mp, sharp_data_mp, canon_data_mp,
                         template_result_ptr->second_alg_vp,
                         (Matrix*)NULL, &(error_ptr->second_alg));

    if (result == ERROR)
    {
        add_error("Error occured while computing template second alg error.");
        return ERROR;
    }

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE

static int do_multi_template_errors
(
    const Vector*          actual_illum_vp,
    const Vector*          canon_illum_vp,
    const Matrix*          data_mp,
    const Matrix*          sharp_data_mp,
    const Matrix*          canon_data_mp,
    Multi_template_result* multi_template_result_ptr,
    Multi_template_error*  error_ptr,
    int                    (*error_fn) (const Vector *, const Vector *, const Matrix *, const Matrix *, const Matrix *, const Vector *, const Matrix *, double *)
)
{
    int i, result;

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        result = (*error_fn)(actual_illum_vp, canon_illum_vp,
                             data_mp, sharp_data_mp, canon_data_mp,
                             multi_template_result_ptr->alg_vp[ i ],
                             (Matrix*)NULL, &(error_ptr->alg[ i ]));

        if (result == ERROR)
        {
            add_error("Error occured while computing multi-template alg %d error.",
                      i+1);
            return ERROR;
        }
    }

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_angular_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             estimate_vp       = NULL;
    Vector*             sharp_estimate_vp = NULL;
    double                error;
    int result = NO_ERROR;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* sharp_2D_canon_illum_vp = NULL;
        Vector* sharp_2D_estimate_vp    = NULL;

        ERE(get_projection_vector(&sharp_2D_canon_illum_vp,
                                  sharp_canon_illum_vp,
                                  DIVIDE_BY_BLUE,
                                  MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_2D_estimate_vp, sharp_2D_canon_illum_vp,
                           result_vp));

        free_vector(sharp_2D_canon_illum_vp);

        ERE(back_project_vector(&sharp_estimate_vp, sharp_2D_estimate_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_2D_estimate_vp);
    }
    else if (result_vp->length == 3)
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(unsharpen_vector_data(&estimate_vp, sharp_estimate_vp));

#ifdef TRUNCATE_ILLUMINANT_ESTIMATE_AT_ZERO
    /*
    // Truncating the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would put sharpening in
    // the best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing.
    */
    if (min_vector_element(estimate_vp) < -DBL_EPSILON)
    {
        verbose_pso(2, "Illuminant estimate has negative componants.\n");
        verbose_pso(2, "Setting those componants to zero.\n");
    }

    ERE(ow_min_thresh_vector(estimate_vp, 0.0));
#endif

    if (get_vector_angle_in_degrees(actual_illum_vp, estimate_vp, &error)
        == ERROR)
    {
        char mess[ 1000 ];

        kjb_print_error();
        ERE(kjb_sprintf(mess, sizeof(mess),
                        "Actual illum vector near line %d of file %s.",
                        __LINE__, __FILE__));
        fp_write_row_vector_with_title(actual_illum_vp, stderr, mess);
        ERE(kjb_sprintf(mess, sizeof(mess),
                        "Illum estimate vector near line %d of file %s.",
                        __LINE__, __FILE__));
        fp_write_row_vector_with_title(estimate_vp, stderr, mess);
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
#ifdef TEST
        result = ERROR;
#endif
    }
    else
    {
        *error_ptr = error;
    }

    free_vector(sharp_estimate_vp);
    free_vector(estimate_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_sum_chrom_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             rgb_estimate_vp       = NULL;
    Vector*             sharp_rgb_estimate_vp = NULL;
    Vector*             actual_sum_chrom_vp   = NULL;
    Vector*             estimate_sum_chrom_vp = NULL;
    Vector*             diff_vp               = NULL;
    int                 projection_result;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* temp_2D_vp           = NULL;
        Vector* sharp_estimate_2D_vp = NULL;

        ERE(get_projection_vector(&temp_2D_vp, sharp_canon_illum_vp,
                                  DIVIDE_BY_BLUE, MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_estimate_2D_vp, temp_2D_vp, result_vp));

        free_vector(temp_2D_vp);

        ERE(back_project_vector(&sharp_rgb_estimate_vp, sharp_estimate_2D_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_estimate_2D_vp);
    }
    else if (result_vp->length == 3)
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(unsharpen_vector_data(&rgb_estimate_vp, sharp_rgb_estimate_vp));

    /*
    // Truncating the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would sharpening in the
    // best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing. So we will only do this if we
    // have to: Specifically, when R+G+B is too small.
    */
    if (sum_vector_elements(rgb_estimate_vp) < 1.0 / MAX_X_DIV_SUM)
    {
        verbose_pso(2, "Illuminant estimate R+G+B is too small.\n");
        verbose_pso(2, "Setting negative componants to zero.\n");
        ERE(ow_min_thresh_vector(rgb_estimate_vp, 0.0));
    }

    ERE(get_projection_vector(&actual_sum_chrom_vp, actual_illum_vp,
                              DIVIDE_BY_SUM, MAX_X_DIV_SUM));

    projection_result = get_projection_vector(&estimate_sum_chrom_vp,
                                              rgb_estimate_vp,
                                              DIVIDE_BY_SUM, MAX_X_DIV_SUM);

    if (projection_result == ERROR)
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        add_error("Problem occured computing sum chromaticity error.");
        add_error("Vector to be projected was (%e, %e, %e).",
                  rgb_estimate_vp->elements[ 0 ],
                  rgb_estimate_vp->elements[ 1 ],
                  rgb_estimate_vp->elements[ 2 ]);
        add_error("Setting result as not available, and carrying on.");
        kjb_print_error();
    }
    else
    {
        ERE(subtract_vectors(&diff_vp, actual_sum_chrom_vp,
                             estimate_sum_chrom_vp));
        *error_ptr = vector_magnitude(diff_vp);
    }

    free_vector(rgb_estimate_vp);
    free_vector(sharp_rgb_estimate_vp);
    free_vector(estimate_sum_chrom_vp);
    free_vector(actual_sum_chrom_vp);
    free_vector(diff_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_r_div_sum_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             rgb_estimate_vp       = NULL;
    Vector*             sharp_rgb_estimate_vp = NULL;
    Vector*             actual_sum_chrom_vp   = NULL;
    Vector*             estimate_sum_chrom_vp = NULL;
    Vector*             diff_vp               = NULL;
    int                 projection_result;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* temp_2D_vp           = NULL;
        Vector* sharp_estimate_2D_vp = NULL;

        ERE(get_projection_vector(&temp_2D_vp, sharp_canon_illum_vp,
                                  DIVIDE_BY_BLUE, MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_estimate_2D_vp, temp_2D_vp, result_vp));

        free_vector(temp_2D_vp);

        ERE(back_project_vector(&sharp_rgb_estimate_vp, sharp_estimate_2D_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_estimate_2D_vp);
    }
    else if (result_vp->length == 3)
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(unsharpen_vector_data(&rgb_estimate_vp, sharp_rgb_estimate_vp));

    /*
    // Truncating the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would sharpening in the
    // best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing. So we will only do this if we
    // have to: Specifically, when R+G+B is too small.
    */
    if (sum_vector_elements(rgb_estimate_vp) < 1.0 / MAX_X_DIV_SUM)
    {
        verbose_pso(2, "Illuminant estimate R+G+B is too small.\n");
        verbose_pso(2, "Setting negative componants to zero.\n");
        ERE(ow_min_thresh_vector(rgb_estimate_vp, 0.0));
    }

    ERE(get_projection_vector(&actual_sum_chrom_vp, actual_illum_vp,
                              DIVIDE_BY_SUM, MAX_X_DIV_SUM));

    projection_result = get_projection_vector(&estimate_sum_chrom_vp,
                                              rgb_estimate_vp,
                                              DIVIDE_BY_SUM, MAX_X_DIV_SUM);

    if (projection_result == ERROR)
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        add_error("Problem occured computing G/(R+G+B) error.");
        add_error("Vector to be projected was (%e, %e, %e).",
                  rgb_estimate_vp->elements[ 0 ],
                  rgb_estimate_vp->elements[ 1 ],
                  rgb_estimate_vp->elements[ 2 ]);
        add_error("Setting result as not available, and carrying on.");
        kjb_print_error();
    }
    else
    {
        ERE(subtract_vectors(&diff_vp, actual_sum_chrom_vp,
                             estimate_sum_chrom_vp));
        *error_ptr = diff_vp->elements[ 0 ];
    }

    free_vector(rgb_estimate_vp);
    free_vector(sharp_rgb_estimate_vp);
    free_vector(estimate_sum_chrom_vp);
    free_vector(actual_sum_chrom_vp);
    free_vector(diff_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_g_div_sum_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             rgb_estimate_vp       = NULL;
    Vector*             sharp_rgb_estimate_vp = NULL;
    Vector*             actual_sum_chrom_vp   = NULL;
    Vector*             estimate_sum_chrom_vp = NULL;
    Vector*             diff_vp               = NULL;
    int                 projection_result;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* temp_2D_vp           = NULL;
        Vector* sharp_estimate_2D_vp = NULL;

        ERE(get_projection_vector(&temp_2D_vp, sharp_canon_illum_vp,
                                  DIVIDE_BY_BLUE, MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_estimate_2D_vp, temp_2D_vp, result_vp));

        free_vector(temp_2D_vp);

        ERE(back_project_vector(&sharp_rgb_estimate_vp, sharp_estimate_2D_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_estimate_2D_vp);
    }
    else if (result_vp->length == 3)
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(unsharpen_vector_data(&rgb_estimate_vp, sharp_rgb_estimate_vp));

    /*
    // Truncating the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would sharpening in the
    // best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing. So we will only do this if we
    // have to: Specifically, when R+G+B is too small.
    */
    if (sum_vector_elements(rgb_estimate_vp) < 1.0 / MAX_X_DIV_SUM)
    {
        verbose_pso(2, "Illuminant estimate R+G+B is too small.\n");
        verbose_pso(2, "Setting negative componants to zero.\n");
        ERE(ow_min_thresh_vector(rgb_estimate_vp, 0.0));
    }

    ERE(get_projection_vector(&actual_sum_chrom_vp, actual_illum_vp,
                              DIVIDE_BY_SUM, MAX_X_DIV_SUM));

    projection_result = get_projection_vector(&estimate_sum_chrom_vp,
                                              rgb_estimate_vp,
                                              DIVIDE_BY_SUM, MAX_X_DIV_SUM);

    if (projection_result == ERROR)
    {
        add_error("Problem occured computing G/(R+G+B) error.");
        add_error("Vector to be projected was (%e, %e, %e).",
                  rgb_estimate_vp->elements[ 0 ],
                  rgb_estimate_vp->elements[ 1 ],
                  rgb_estimate_vp->elements[ 2 ]);
        add_error("Setting result as not available, and carrying on.");
        kjb_print_error();
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
    }
    else
    {
        ERE(subtract_vectors(&diff_vp, actual_sum_chrom_vp,
                             estimate_sum_chrom_vp));
        *error_ptr = diff_vp->elements[ 1 ];
    }

    free_vector(rgb_estimate_vp);
    free_vector(sharp_rgb_estimate_vp);
    free_vector(estimate_sum_chrom_vp);
    free_vector(actual_sum_chrom_vp);
    free_vector(diff_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_rgb_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             estimate_vp                 = NULL;
    Vector*             sharp_estimate_vp           = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* sharp_2D_canon_illum_vp = NULL;
        Vector* sharp_2D_estimate_vp    = NULL;

        ERE(get_projection_vector(&sharp_2D_canon_illum_vp,
                                  sharp_canon_illum_vp, DIVIDE_BY_BLUE,
                                  MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_2D_estimate_vp, sharp_2D_canon_illum_vp,
                           result_vp));

        free_vector(sharp_2D_canon_illum_vp);

        ERE(back_project_vector(&sharp_estimate_vp, sharp_2D_estimate_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_2D_estimate_vp);
    }
    else
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }

    ERE(unsharpen_vector_data(&estimate_vp, sharp_estimate_vp));

    /*
    // Truncate the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would sharpening in the
    // best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing.
    */
#ifdef TRUNCATE_ILLUMINANT_ESTIMATE_AT_ZERO
    if (min_vector_element(estimate_vp) < -DBL_EPSILON)
    {
        verbose_pso(2, "Illuminant estimate has negative componants.\n");
        verbose_pso(2, "Setting those componants to zero.\n");
    }

    ERE(ow_min_thresh_vector(estimate_vp, 0.0));
#endif

#ifdef TRY_IT_DIFFERENTLY
    if (result_vp->length == 2)
    {
        double estimate_sum = sum_vector_elements(estimate_vp);
        double actual_sum   = sum_vector_elements(actual_illum_vp);

        ERE(ow_multiply_vector_by_scalar(estimate_vp,
                                         actual_sum / estimate_sum));
    }
    else if (fs_normalize_3D_illum_estimate)
    {
        ERE(ow_scale_vector_by_max(estimate_vp));
        ERE(ow_multiply_vector_by_scalar(estimate_vp, 255.0));
    }
#else
    if ((result_vp->length == 2) || (fs_normalize_3D_illum_estimate))
    {
        ERE(ow_scale_vector_by_max(estimate_vp));
        ERE(ow_multiply_vector_by_scalar(estimate_vp, 255.0));
    }
#endif

    *error_ptr = vector_distance(actual_illum_vp, estimate_vp);

    /*
    // I really don't know about this ... looks like an error to me on
    // 01/05/99. Most the time, this should be a NOP, as the maximum component
    // of the actual illuminant is usually set to 255. However, when it is not
    // set to 255, for example, when the option "aiw" is used, then we want to
    // know how well we are estimating this non-normalized value, and it now
    // looks to me like this piece of code would defeat that.
    //
    // Commented out on 01/05/99.
    //
    // max_actual_illum_rgb = max_vector_element(actual_illum_vp);
    // *error_ptr *= 255.0;
    // *error_ptr /= max_actual_illum_rgb;
    */

    free_vector(estimate_vp);
    free_vector(sharp_estimate_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_sum_lum_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             estimate_vp                 = NULL;
    Vector*             sharp_estimate_vp           = NULL;
    double                actual_illum_sum_lum;
    double                est_illum_sum_lum;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

#ifdef EMERGENCY_PATCH_FOR_THESIS
    if (divide_vectors(&sharp_estimate_vp, sharp_canon_illum_vp,
                       result_vp)
        == ERROR)
    {
        char mess[ 1000 ];

        kjb_print_error();
        ERE(kjb_sprintf(mess, sizeof(mess),
                        "Canon illum vector near line %d of file %s.",
                        __LINE__, __FILE__));
        fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
        ERE(kjb_sprintf(mess, sizeof(mess),
                        "Map result vector near line %d of file %s.",
                        __LINE__, __FILE__));
        fp_write_row_vector_with_title(result_vp, stderr, mess);
        *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
        return ERROR;
#else
        return NO_ERROR;
#endif
    }
#else
    ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                       result_vp));
#endif

    ERE(unsharpen_vector_data(&estimate_vp, sharp_estimate_vp));

    /*
    // Truncate the estimate at zero should give a better estimate since the
    // target is always positive. Since negative componants are usually due to
    // sharpening when combined with gamut mapping, this would sharpening in the
    // best possible light, but this should be acknowleged as part of the
    // algorithm. However, sticking with the native estimate as best as possible
    // gives a better handle on how we are doing.
    */
#ifdef TRUNCATE_ILLUMINANT_ESTIMATE_AT_ZERO
    if (min_vector_element(estimate_vp) < -DBL_EPSILON)
    {
        verbose_pso(2, "Illuminant estimate has negative componants.\n");
        verbose_pso(2, "Setting those componants to zero.\n");
    }

    ERE(ow_min_thresh_vector(estimate_vp, 0.0));
#endif

    if (fs_normalize_3D_illum_estimate)
    {
        ERE(ow_scale_vector_by_max(estimate_vp));
        ERE(ow_multiply_vector_by_scalar(estimate_vp, 255.0));
    }

    actual_illum_sum_lum = sum_vector_elements(actual_illum_vp);
    est_illum_sum_lum = sum_vector_elements(estimate_vp);

    *error_ptr = est_illum_sum_lum - actual_illum_sum_lum;

    free_vector(estimate_vp);
    free_vector(sharp_estimate_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_illum_lab_error
(
    const Vector* actual_illum_vp,
    const Vector* sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* __attribute__((unused)) dummy_sharp_data_mp,
    const Matrix* __attribute__((unused)) dummy_canon_data_mp,
    const Vector* result_vp,
    const Matrix* __attribute__((unused)) dummy_map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Vector*             estimate_vp                 = NULL;
    Vector*             sharp_estimate_vp           = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (result_vp == NULL)
         || (actual_illum_vp == NULL)
         || (sharp_canon_illum_vp == NULL)
       )
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (result_vp->length == 2)
    {
        Vector* sharp_2D_canon_illum_vp = NULL;
        Vector* sharp_2D_estimate_vp    = NULL;

        ERE(get_projection_vector(&sharp_2D_canon_illum_vp,
                                  sharp_canon_illum_vp, DIVIDE_BY_BLUE,
                                  MAX_X_DIV_B));

        ERE(divide_vectors(&sharp_2D_estimate_vp, sharp_2D_canon_illum_vp,
                           result_vp));

        free_vector(sharp_2D_canon_illum_vp);

        ERE(back_project_vector(&sharp_estimate_vp, sharp_2D_estimate_vp,
                                DIVIDE_BY_BLUE));
        free_vector(sharp_2D_estimate_vp);
    }
    else if (result_vp->length == 3)
    {
#ifdef EMERGENCY_PATCH_FOR_THESIS
        if (divide_vectors(&sharp_estimate_vp, sharp_canon_illum_vp,
                           result_vp)
            == ERROR)
        {
            char mess[ 1000 ];

            kjb_print_error();
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Canon illum vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(sharp_canon_illum_vp, stderr, mess);
            ERE(kjb_sprintf(mess, sizeof(mess),
                            "Map result vector near line %d of file %s.",
                            __LINE__, __FILE__));
            fp_write_row_vector_with_title(result_vp, stderr, mess);
            *error_ptr = DBL_HALF_MOST_NEGATIVE;

#ifdef TEST
            return ERROR;
#else
            return NO_ERROR;
#endif
        }
#else
        ERE(divide_vectors(&sharp_rgb_estimate_vp, sharp_canon_illum_vp,
                           result_vp));
#endif
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(unsharpen_vector_data(&estimate_vp, sharp_estimate_vp));

    if ((result_vp->length == 2) || fs_normalize_3D_illum_estimate)
    {
        ERE(ow_scale_vector_by_max(estimate_vp));
        ERE(ow_multiply_vector_by_scalar(estimate_vp, 255.0));
    }

    ERE(get_delta_lab(actual_illum_vp, estimate_vp, error_ptr));

    free_vector(sharp_estimate_vp);
    free_vector(estimate_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_sum_chrom_map_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* __attribute__((unused)) dummy_sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* sum_chrom_canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Matrix*             new_linear_map_mp = NULL;
    Vector*             sharp_diag_map_vp = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if ((diag_map_vp != NULL) && (map_mp != NULL))
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    else if ((diag_map_vp == NULL) && (map_mp == NULL))
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }
    else if (diag_map_vp != NULL)
    {
        NRE(new_linear_map_mp = create_diagonal_matrix(diag_map_vp));
        map_mp = new_linear_map_mp;
    }

    ERE(get_sum_chrom_map_error_guts(sharp_data_mp,
                                     sum_chrom_canon_data_mp,
                                     map_mp, error_ptr));

    free_matrix(new_linear_map_mp);
    free_vector(sharp_diag_map_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_sum_chrom_map_error_guts
(
    const Matrix* sharp_data_mp,
    const Matrix* sum_chrom_canon_data_mp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    Matrix* blue_chrom_data_mp     = NULL;
    Matrix* estimated_sum_chrom_mp = NULL;
    Matrix* rgb_mp;
    Matrix* sharp_rgb_mp           = NULL;

    if (map_mp == NULL)
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (map_mp->num_cols == 2)
    {
        Matrix* mapped_mp = NULL;

        ERE(get_projection_matrix(&blue_chrom_data_mp, sharp_data_mp,
                                  DIVIDE_BY_BLUE, TRUNCATE_INVALID_CHROMATICITY,
                                  MAX_X_DIV_B, DBL_NOT_SET, DBL_NOT_SET,
                                  (Matrix**) NULL, (Matrix**) NULL));

        ERE(multiply_matrices(&mapped_mp, blue_chrom_data_mp, map_mp));

        ERE(back_project_matrix(&sharp_rgb_mp, mapped_mp, DIVIDE_BY_BLUE));

        free_matrix(mapped_mp);
    }
    else if (map_mp->num_cols == 3)
    {
        ERE(multiply_matrices(&sharp_rgb_mp, sharp_data_mp, map_mp));
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(ow_unsharpen_matrix_data(sharp_rgb_mp));
    rgb_mp = sharp_rgb_mp;

    /*
    // Mapping can result in negative components, EVEN if we did not sharpen.
    // A best fit between the data and the canonical does not guarentee that
    // individual mapped RGB are all positive. Since projection is not well
    // defined for some combinations of negative RGB, we have to do something
    // about this. So, what we do is we assume that mapping has 2 steps. First,
    // we do the map, and then we apply the knowlege that we want positive RGB.
    // This is not entirely unreasonable. Thus we truncate the negative RGB to
    // zero. This procedure may result in more error or less error than other
    // procedures.
    */
    ERE(ow_min_thresh_matrix(rgb_mp, 0.0));

    ERE(get_projection_matrix(&estimated_sum_chrom_mp, rgb_mp,
                              DIVIDE_BY_SUM, ZERO_INVALID_CHROMATICITY,
                              DBL_NOT_SET, DBL_NOT_SET, DBL_NOT_SET,
                              (Matrix**) NULL, (Matrix**) NULL));

    *error_ptr = rms_matrix_row_difference(estimated_sum_chrom_mp,
                                           sum_chrom_canon_data_mp);

    free_matrix(sharp_rgb_mp);
    free_matrix(blue_chrom_data_mp);
    free_matrix(estimated_sum_chrom_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_blue_chrom_map_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* __attribute__((unused)) dummy_sharp_canon_illum_vp,
    const Matrix* __attribute__((unused)) dummy_data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* projected_canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Matrix*             new_linear_map_mp = NULL;
    Vector*             sharp_diag_map_vp = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if ((diag_map_vp != NULL) && (map_mp != NULL))
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    else if ((diag_map_vp == NULL) && (map_mp == NULL))
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }
    else if (diag_map_vp != NULL)
    {
        NRE(new_linear_map_mp = create_diagonal_matrix(diag_map_vp));
        map_mp = new_linear_map_mp;
    }

    ERE(get_blue_chrom_map_error_guts(sharp_data_mp,
                                      projected_canon_data_mp,
                                      map_mp, error_ptr));

    free_matrix(new_linear_map_mp);
    free_vector(sharp_diag_map_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_blue_chrom_map_error_guts
(
    const Matrix* sharp_data_mp,
    const Matrix* projected_canon_data_mp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    Matrix* estimated_2D_mp         = NULL;
    Matrix* estimated_3D_sharp_mp   = NULL;
    Matrix* estimated_3D_mp;

    if (map_mp == NULL)
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }

    if (map_mp->num_cols == 2)
    {
        Matrix* estimated_2D_sharp_mp = NULL;
        Matrix* projected_data_mp     = NULL;

        ERE(get_projection_matrix(&projected_data_mp, sharp_data_mp,
                                  DIVIDE_BY_BLUE, TRUNCATE_INVALID_CHROMATICITY,
                                  MAX_X_DIV_B, DBL_NOT_SET, DBL_NOT_SET,
                                  (Matrix**)NULL, (Matrix**)NULL));

        ERE(multiply_matrices(&estimated_2D_sharp_mp, projected_data_mp,
                              map_mp));

        free_matrix(projected_data_mp);

        ERE(back_project_matrix(&estimated_3D_sharp_mp, estimated_2D_sharp_mp,
                                DIVIDE_BY_BLUE));

        free_matrix(estimated_2D_sharp_mp);
    }
    else if (map_mp->num_cols == 3)
    {
        ERE(multiply_matrices(&estimated_3D_sharp_mp, sharp_data_mp, map_mp));
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(ow_unsharpen_matrix_data(estimated_3D_sharp_mp));
    estimated_3D_mp = estimated_3D_sharp_mp;

    /*
    // Note that mapping can result in negative components, EVEN if we did not
    // sharpen.  A best fit between the data and the canonical does not
    // guarentee that individual mapped RGB are all positive. If the blue
    // componant is negative or too close to zero, get_projection_matrix() uses
    // a small positive number instead. This underscores the reality that this
    // error metric is not really good for much except for verification that
    // certain things are roughly working (which is why I keep it).
    */
    ERE(get_projection_matrix(&estimated_2D_mp, estimated_3D_mp,
                              DIVIDE_BY_BLUE, TRUNCATE_INVALID_CHROMATICITY,
                              MAX_X_DIV_B, DBL_NOT_SET, DBL_NOT_SET,
                              (Matrix**)NULL, (Matrix**)NULL));

    *error_ptr = rms_matrix_row_difference(estimated_2D_mp,
                                           projected_canon_data_mp);

    free_matrix(estimated_3D_sharp_mp);
    free_matrix(estimated_2D_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_map_rms_error
(
    const Vector* __attribute__((unused)) dummy_actual_illum_vp,
    const Vector* __attribute__((unused)) dummy_sharp_canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Matrix*             new_linear_map_mp        = NULL;
    Matrix*             estimated_mp             = NULL;
    Vector*             sharp_diag_map_vp = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if ((diag_map_vp != NULL) && (map_mp != NULL))
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    else if ((diag_map_vp == NULL) && (map_mp == NULL))
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }
    else if (diag_map_vp != NULL)
    {
        NRE(new_linear_map_mp = create_diagonal_matrix(diag_map_vp));
        map_mp = new_linear_map_mp;
    }

    ERE(compute_result_data(&estimated_mp, data_mp, sharp_data_mp,
                            canon_data_mp, map_mp));

    *error_ptr = rms_matrix_row_difference(estimated_mp, canon_data_mp);

    /*
    // I really don't know about this ... looks like an error to me on
    // 01/05/99. Most the time, this should be a NOP, as the maximum component
    // of the actual illuminant is usually set to 255. However, when it is not
    // set to 255, for example, when the option "aiw" is used, then we want to
    // know how well we are estimating this non-normalized value, and it now
    // looks to me like this piece of code would defeat that.
    //
    // Commented out on 01/05/99.
    //
    // max_actual_illum_rgb = max_vector_element(actual_illum_vp);
    // *error_ptr *= 255.0;
    // *error_ptr /= max_actual_illum_rgb;
    */

    free_matrix(new_linear_map_mp);
    free_matrix(estimated_mp);
    free_vector(sharp_diag_map_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_map_lab_error
(
    const Vector* actual_illum_vp,
    const Vector* __attribute__((unused)) dummy_sharp_canon_illum_vp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Vector* diag_map_vp,
    const Matrix* map_mp,
    double*       error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Matrix*             new_linear_map_mp      = NULL;
    Matrix*             estimated_mp           = NULL;
    Matrix*             estimated_lab_mp       = NULL;
    Matrix*             canon_data_lab_mp      = NULL;
    const Vector*             white_rgb_vp;
    Vector*             sharp_diag_map_vp = NULL;

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if ((diag_map_vp != NULL) && (map_mp != NULL))
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
    else if ((diag_map_vp == NULL) && (map_mp == NULL))
    {
        *error_ptr = DBL_HALF_MOST_NEGATIVE;
        return NO_ERROR;
    }
    else if (diag_map_vp != NULL)
    {
        NRE(new_linear_map_mp = create_diagonal_matrix(diag_map_vp));
        map_mp = new_linear_map_mp;
    }

    ERE(compute_result_data(&estimated_mp, data_mp, sharp_data_mp, canon_data_mp,
                            map_mp));

    if (fs_actual_illum_lab_white)
    {
        white_rgb_vp = actual_illum_vp;
    }
    else
    {
        white_rgb_vp = NULL;
    }

    ERE(convert_matrix_rgb_to_lab(&estimated_lab_mp, estimated_mp,
                                  white_rgb_vp));

    ERE(convert_matrix_rgb_to_lab(&canon_data_lab_mp, canon_data_mp,
                                  white_rgb_vp));

    *error_ptr = rms_matrix_row_difference(estimated_lab_mp, canon_data_lab_mp);

    free_matrix(new_linear_map_mp);
    free_matrix(estimated_mp);
    free_matrix(estimated_lab_mp);
    free_matrix(canon_data_lab_mp);
    free_vector(sharp_diag_map_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_delta_lab
(
    const Vector* actual_illum_vp,
    const Vector* est_illum_vp,
    double*       error_ptr
)
{
    Vector*    actual_lab_vp = NULL;
    Vector*    est_lab_vp    = NULL;
    int        result        = NO_ERROR;
    const Vector*    white_rgb_vp;

    white_rgb_vp = fs_actual_illum_lab_white ? actual_illum_vp : NULL;

    if (result != ERROR)
        result = convert_vector_rgb_to_lab(&actual_lab_vp, actual_illum_vp,
                                           white_rgb_vp);

    if (result != ERROR)
        result = convert_vector_rgb_to_lab(&est_lab_vp, est_illum_vp,
                                           white_rgb_vp);

    if (result != ERROR)
    {
        *error_ptr = vector_distance(actual_lab_vp, est_lab_vp);
    }

    free_vector(actual_lab_vp);
    free_vector(est_lab_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

