
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef COMPUTE_OUTPUT_INCLUDED
#define COMPUTE_OUTPUT_INCLUDED

#include "m/m_incl.h"
#include "type.h"

/* -------------------------------------------------------------------------- */

typedef struct Image_map_data
{
    const KJB_image* image_ip;
    const KJB_image* input_ip;
    const KJB_image* canon_ip;
    Matrix*            map_mp;
    Vector*            map_vp;
    Sharp_method       sharp_method;
    int                illumination_image_flag;
    int                chromaticity_image_flag;
    char               request[ 100 ];
}
Image_map_data;

/* -------------------------------------------------------------------------- */

EXTERN_C        int                            
(
    set_output_image_options (const char* option, const char* value) 
);

int             print_output_dependence_message(void);

int             parse_image_request            
(
    char*                    request,
    Colour_constancy_result* result_ptr,
    int                      reduction_flag,
    int                      illumination_image_flag,
    int                      chromaticity_image_flag,
    Image_map_data*          image_map_data_ptr,
    Error_by_sharp*          error_ptr,
    double**                 error_item_ptr_ptr 
);

int             compute_result_image           
(
    KJB_image**     target_ipp,
    Image_map_data* image_map_data_ptr 
);

Image_map_data* create_image_map_data          (void);
void            queue_free_image_map_data      (void* image_map_data_ptr);

void            free_image_map_data            
(
    Image_map_data* image_map_data_ptr 
);

int             compute_result_data            
(
    Matrix**      result_data_mpp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Matrix* map_mp 
);

int             ow_normalize_canon_data        
(
    Matrix*       canon_data_mp,
    const Matrix* data_mp 
);

#endif

