
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef LIB_INCLUDED
#define LIB_INCLUDED

#include "l/l_incl.h"
#include "m/m_incl.h"
#include "s2/s2_fluorescence.h"
#include "type.h"

EXTERN_C int set_gamut_config_dir               
(
    const char* option,
    const char* value 
);

EXTERN_C int set_algorithm_options              
(
    const char* option,
    const char* value 
);

int   get_cc_methods                     
(
    char*                     arg,
    Colour_constancy_methods* cc_methods_ptr 
);

void  initialize_cc_result               
(
    Colour_constancy_result* result_ptr 
);

void  initialize_hull_result             (Hull_result*);
void  initialize_constraints             (Constraints*);
void  clear_result                       (Colour_constancy_result*);
void  clear_hull_result                  (Hull_result*);
void  free_constraints                   (Constraints*);
void  free_hull_list                     (Queue_element**);

int   ok_to_increase_pixel_error         
(
    int     prompt_to_increase_error_flag,
    double* absolute_pixel_error_ptr,
    double* relative_pixel_error_ptr 
);

int   read_real_from_gamut_config_file   
(
    double*     value_ptr,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
);

int   read_vector_from_gamut_config_file 
(
    Vector**    result_vpp,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
);

int   read_matrix_from_gamut_config_file 
(
    Matrix**    result_mpp,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
);

int   read_spectra_from_gamut_config_file
(
    Spectra**   result_sp_ptr,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
);

int   read_fl_db_from_gamut_config_file  
(
    Fluorescent_database** result_fl_db_ptr_ptr,
    const char*            env_var,
    const char*            file_name,
    const char*            message_name,
    char*                  config_file_name,
    size_t                 config_file_name_size 
);

FILE* open_gamut_config_file             
(
    const char* env_var,
    const char* file_name,
    const char* message_name 
);

int   select_data                        
(
    Matrix**      selected_mpp,
    const Matrix* data_mp,
    int           num_data_points,
    const int*    data_point_nums,
    const int*    invalid_data 
);

int   parse_result_request               
(
    char*                    request,
    Colour_constancy_result* result_ptr,
    Vector**                 map_vpp,
    Matrix**                 map_mpp,
    Sharp_method*            sharp_method_ptr,
    Error_by_sharp*          error_ptr,
    double**                 error_item_ptr_ptr 
);

int   remove_origin_from_data            
(
    Matrix** target_mpp,
    Matrix*  source_mp,
    double   tolerance 
);

int   parse_index                        
(
    char* str,
    int   max_index,
    int*  first_index_ptr,
    int*  second_index_ptr,
    int*  index_ptr,
    int*  sub_index_ptr 
);

int   check_index_range                  
(
    int         index,
    int         max_index,
    const char* title_str 
);

int   write_data_matrix                  
(
    const Matrix* output_mp,
    const char*   output_file,
    int           append_flag 
);

#endif

