
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "canon_gamut.h"
#include "correlation.h"     /* Lindsay: Feb 17/98 */

#include "data.h"

#include "grey.h"
#include "illum_data.h"
#include "luminance.h"
#include "make.h"
#include "sharp.h"
#include "synthetic_data.h"

#ifdef TEMPLATE
    /*
    // If your proposed algorithm group does not have any data which is
    // (easily) buildable from spectra, then it is not necessary to copy this
    // code.
    */
#   include "template.h"
#endif

#ifdef MULTI_TEMPLATE
    /*
    // If your proposed multi-variant algorithm group does not have any data
    // which is (easily) buildable from spectra, then it is not necessary to
    // copy this code.
    */
#   include "multi_template.h"
#endif

#define REQUIRE_EXPLICIT_CORR_MAKE

/* -------------------------------------------------------------------------- */

static int process_make_directive
(
    const char* make_directive,
    const char* target
);

/* -------------------------------------------------------------------------- */

int process_make_command(char* arg)
{
    char request[ 100 ];
    char make_directive[ 100 ];
    char target[ 100 ];
    int  result;

    target[ 0 ] = '\0';

    /* Null result can be meaninfull. */
    BUFF_GEN_GET_TOKEN(&arg, request, ">");

    if (*arg == '>')
    {
        arg++;

        if ( ! BUFF_GET_TOKEN_OK(&arg, target))
        {
            set_error("A file name is required after the \">\".");
            return ERROR;
        }
    }

    ERE(force_synthetic_data());

    arg = request;

    if ((target[ 0 ] != '\0') && (count_tokens(arg) != 1))
    {
        set_error("Exactly one make request must be used with a target file.");
        return ERROR;
    }

    if (is_interactive())
    {
        set_io_atn_trap();
    }

    BUFF_GET_TOKEN(&arg, make_directive);

    do
    {
        result = process_make_directive(make_directive, target);
    }
    while ((result != ERROR) && (BUFF_GET_TOKEN_OK(&arg, make_directive)));

    if (is_interactive())
    {
        unset_io_atn_trap();
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int process_make_directive
(
    const char* make_directive,
    const char* target
)
{
    int result;
    int found = FALSE;

    ERE(result = make_canon_gamut(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_illum_gamut(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_canon_illum(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_actual_illum(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_luminance(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_grey(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_data(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_canon_data(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    ERE(result = make_correlation(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

#ifdef TEMPLATE
    /*
    // If your proposed algorithm group does not have any data which is
    // (easily) buildable from spectra, then it is not necessary to copy
    // this code.
    */
    ERE(result = make_template(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;
#endif

#ifdef MULTI_TEMPLATE
    /*
    // Similar to TEMPLATE, but for algorithms which have multiple variants
    // which are indexed by the user.
    */
    ERE(result = make_multi_template(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;
#endif

    ERE(result = make_sharp(make_directive, target));
    if (result != NOT_FOUND) found = TRUE;

    if (result != NOT_FOUND) found = TRUE;

    if (! found)
    {
        set_error("%q is an invalid make directive.", make_directive);
        result = ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

