
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "output_error.h"

/* ------------------------------------------------------------------------- */

#define MAX_NUM_ERROR_OUTPUT_FILES  30

static int fs_num_illum_angular_error_output_files = 0;
static int fs_num_r_div_sum_error_output_files = 0;
static int fs_num_g_div_sum_error_output_files = 0;

/* Making use of zero initialization here! */
typedef struct
{
    char alg[ 50 ];
    FILE* fp;
}
Error_output_files;

static Error_output_files fs_illum_angular_error_output_files[ MAX_NUM_ERROR_OUTPUT_FILES ];
static Error_output_files fs_r_div_sum_error_output_files[ MAX_NUM_ERROR_OUTPUT_FILES ];
static Error_output_files fs_g_div_sum_error_output_files[ MAX_NUM_ERROR_OUTPUT_FILES ];

/* -------------------------------------------------------------------------- */

#ifdef TEST
    static void close_error_files(void);
    static void prepare_cleanup(void);
#endif

/* -------------------------------------------------------------------------- */

EXTERN_C int set_output_error_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    char* lc_option_pos;
    char measure[ 100 ];
    char alg[ 100 ];
    int  result             = NOT_FOUND;
    int  i;

#ifdef TEST
    prepare_cleanup();
#endif

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    lc_option_pos = lc_option;
    trim_beg(&lc_option_pos);

    if (    (BUFF_GEN_GET_TOKEN_OK(&lc_option_pos, alg, "|"))
         && (*lc_option_pos == '|')
       )
    {
        dbs(alg)
        lc_option_pos++;
        trim_end(alg);
        trim_beg(&lc_option_pos);
        trim_end(lc_option_pos);
        BUFF_CPY(measure, lc_option_pos);
    }
    else
    {
        return NOT_FOUND;
    }

    if (    (match_pattern(measure, "illum-angular-error-output-file"))
         || (match_pattern(measure, "iae-of"))
       )
    {
        int index = NOT_SET;

        ERE(parse_result_request(alg, (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL, (Error_by_sharp*)NULL,
                                 (double**)NULL));

        for (i=0; i<fs_num_illum_angular_error_output_files; i++)
        {
            if (STRCMP_EQ(alg, fs_illum_angular_error_output_files[ i ].alg))
            {
                index = i;
                break;
            }
        }

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (IS_SET(index))
            {
                ERE(pso("%s:%s = %F\n",
                        fs_illum_angular_error_output_files[ index ].alg,
                        measure, fs_illum_angular_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("%s:%s = off\n", alg, measure));
            }
            result = NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            if (IS_SET(index))
            {
                ERE(pso("Illum angular errror for %s is written to %F\n",
                        fs_illum_angular_error_output_files[ index ].alg,
                        fs_illum_angular_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("Illum angular errror for %s is not written to disk.\n",
                        alg));
            }
            result = NO_ERROR;
        }
        else if (IS_SET(index) && (is_no_value_word(value)))
        {
            kjb_fclose(fs_illum_angular_error_output_files[ index ].fp);

            fs_illum_angular_error_output_files[ index ].fp = NULL;
            fs_illum_angular_error_output_files[ index ].alg[ 0 ] = '\0';

            result = NO_ERROR;
        }
        else
        {
            if (IS_SET(index))
            {
                kjb_fclose(fs_illum_angular_error_output_files[ index ].fp);
                fs_illum_angular_error_output_files[ index ].fp = NULL;
            }
            else
            {
                if (fs_num_illum_angular_error_output_files >= MAX_NUM_ERROR_OUTPUT_FILES - 1)
                {
                    set_error("The number of error output files is limited to %d.",
                              MAX_NUM_ERROR_OUTPUT_FILES);
                    return ERROR;
                }

                index = fs_num_illum_angular_error_output_files;
            }

            NRE(fs_illum_angular_error_output_files[ index ].fp = kjb_fopen(value, "w"));

            BUFF_CPY(fs_illum_angular_error_output_files[ index ].alg, alg);

            fs_num_illum_angular_error_output_files++;

            result = NO_ERROR;
        }
    }

    if (    (match_pattern(measure, "r-div-sum-output-file"))
         || (match_pattern(measure, "rds-of"))
       )
    {
        int index = NOT_SET;

        ERE(parse_result_request(alg, (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL, (Error_by_sharp*)NULL,
                                 (double**)NULL));

        for (i=0; i<fs_num_r_div_sum_error_output_files; i++)
        {
            if (STRCMP_EQ(alg, fs_r_div_sum_error_output_files[ i ].alg))
            {
                index = i;
                break;
            }
        }

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (IS_SET(index))
            {
                ERE(pso("%s:%s = %F\n",
                        fs_r_div_sum_error_output_files[ index ].alg,
                        measure, fs_r_div_sum_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("%s:%s = off\n", alg, measure));
            }
            result = NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            if (IS_SET(index))
            {
                ERE(pso("Illum r-div-sum errror for %s is written to %F\n",
                        fs_r_div_sum_error_output_files[ index ].alg,
                        fs_r_div_sum_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("Illum r-div-sum errror for %s is not written to disk.\n",
                        alg));
            }
            result = NO_ERROR;
        }
        else if (IS_SET(index) && (is_no_value_word(value)))
        {
            kjb_fclose(fs_r_div_sum_error_output_files[ index ].fp);

            fs_r_div_sum_error_output_files[ index ].fp = NULL;
            fs_r_div_sum_error_output_files[ index ].alg[ 0 ] = '\0';

            result = NO_ERROR;
        }
        else
        {
            if (IS_SET(index))
            {
                kjb_fclose(fs_r_div_sum_error_output_files[ index ].fp);
                fs_r_div_sum_error_output_files[ index ].fp = NULL;
            }
            else
            {
                if (fs_num_r_div_sum_error_output_files >= MAX_NUM_ERROR_OUTPUT_FILES - 1)
                {
                    set_error("The number of error output files is limited to %d.",
                              MAX_NUM_ERROR_OUTPUT_FILES);
                    return ERROR;
                }

                index = fs_num_r_div_sum_error_output_files;
            }

            NRE(fs_r_div_sum_error_output_files[ index ].fp = kjb_fopen(value,
                                                                     "w"));

            BUFF_CPY(fs_r_div_sum_error_output_files[ index ].alg, alg);

            fs_num_r_div_sum_error_output_files++;

            result = NO_ERROR;
        }
    }

    if (    (match_pattern(measure, "g-div-sum-error-output-file"))
         || (match_pattern(measure, "gds-of"))
       )
    {
        int index = NOT_SET;

        ERE(parse_result_request(alg, (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL, (Error_by_sharp*)NULL,
                                 (double**)NULL));

        for (i=0; i<fs_num_g_div_sum_error_output_files; i++)
        {
            if (STRCMP_EQ(alg, fs_g_div_sum_error_output_files[ i ].alg))
            {
                index = i;
                break;
            }
        }

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (IS_SET(index))
            {
                ERE(pso("%s:%s = %F\n",
                        fs_g_div_sum_error_output_files[ index ].alg,
                        measure, fs_g_div_sum_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("%s:%s = off\n", alg, measure));
            }
            result = NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            if (IS_SET(index))
            {
                ERE(pso("Illum g-div-sum errror for %s is written to %F\n",
                        fs_g_div_sum_error_output_files[ index ].alg,
                        fs_g_div_sum_error_output_files[ index ].fp));
            }
            else
            {
                ERE(pso("Illum g-div-sum errror for %s is not written to disk.\n",
                        alg));
            }
            result = NO_ERROR;
        }
        else if (IS_SET(index) && (is_no_value_word(value)))
        {
            kjb_fclose(fs_g_div_sum_error_output_files[ index ].fp);

            fs_g_div_sum_error_output_files[ index ].fp = NULL;
            fs_g_div_sum_error_output_files[ index ].alg[ 0 ] = '\0';

            result = NO_ERROR;
        }
        else
        {
            if (IS_SET(index))
            {
                kjb_fclose(fs_g_div_sum_error_output_files[ index ].fp);
                fs_g_div_sum_error_output_files[ index ].fp = NULL;
            }
            else
            {
                if (fs_num_g_div_sum_error_output_files >= MAX_NUM_ERROR_OUTPUT_FILES - 1)
                {
                    set_error("The number of error output files is limited to %d.",
                              MAX_NUM_ERROR_OUTPUT_FILES);
                    return ERROR;
                }

                index = fs_num_g_div_sum_error_output_files;
            }

            NRE(fs_g_div_sum_error_output_files[ index ].fp = kjb_fopen(value, "w"));

            BUFF_CPY(fs_g_div_sum_error_output_files[ index ].alg, alg);

            fs_num_g_div_sum_error_output_files++;

            result = NO_ERROR;
        }
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int write_error_output_files(Colour_constancy_error* error_ptr)
{
    double* error_item_ptr;
    int i;

    for (i=0; i<fs_num_illum_angular_error_output_files; i++)
    {
        ERE(parse_result_request(fs_illum_angular_error_output_files[ i ].alg,
                                 (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL,
                                 &(error_ptr->illum_angular), &error_item_ptr));

        if (*error_item_ptr <= DBL_HALF_MOST_NEGATIVE/2.0) continue;

        ERE(kjb_fprintf(fs_illum_angular_error_output_files[ i ].fp, "%e\n",
                        *error_item_ptr));
        ERE(kjb_fflush(fs_illum_angular_error_output_files[ i ].fp));
    }

    for (i=0; i<fs_num_r_div_sum_error_output_files; i++)
    {
        ERE(parse_result_request(fs_r_div_sum_error_output_files[ i ].alg,
                                 (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL,
                                 &(error_ptr->illum_r_div_sum),
                                 &error_item_ptr));

        if (*error_item_ptr <= DBL_HALF_MOST_NEGATIVE/2.0) continue;

        ERE(kjb_fprintf(fs_r_div_sum_error_output_files[ i ].fp, "%e\n",
                        *error_item_ptr));
        ERE(kjb_fflush(fs_r_div_sum_error_output_files[ i ].fp));
    }

    for (i=0; i<fs_num_g_div_sum_error_output_files; i++)
    {
        ERE(parse_result_request(fs_g_div_sum_error_output_files[ i ].alg,
                                 (Colour_constancy_result*)NULL,
                                 (Vector**)NULL, (Matrix**)NULL,
                                 (Sharp_method*)NULL,
                                 &(error_ptr->illum_g_div_sum),
                                 &error_item_ptr));

        if (*error_item_ptr <= DBL_HALF_MOST_NEGATIVE/2.0) continue;

        ERE(kjb_fprintf(fs_g_div_sum_error_output_files[ i ].fp, "%e\n",
                        *error_item_ptr));
        ERE(kjb_fflush(fs_g_div_sum_error_output_files[ i ].fp));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

static void prepare_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(close_error_files);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

static void close_error_files(void)
{
    int i;

    for (i=0; i<fs_num_illum_angular_error_output_files; i++)
    {
        kjb_fclose(fs_illum_angular_error_output_files[ i ].fp);
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

