
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef SPECTRA_INCLUDED
#define SPECTRA_INCLUDED

#include "s/s_incl.h"

typedef struct World_info
{
    char illum_spectra_config_file[ MAX_FILE_NAME_SIZE ];
    char reflect_spectra_config_file[ MAX_FILE_NAME_SIZE ];
    char canon_spectra_config_file[ MAX_FILE_NAME_SIZE ];
    char specular_spectra_config_file[ MAX_FILE_NAME_SIZE ];
    char fluorescent_database_config_file[ MAX_FILE_NAME_SIZE ];
}
World_info;

EXTERN_C int set_spectra_options(const char* option, const char* value);

int get_num_illum_spectra(int* num_illum_spectra_ptr);

int get_synthetic_RGB(Matrix** rgb_data_mpp, int illum);

int generate_synthetic_RGB_data
(
    Matrix**              rgb_data_mpp,
    Spectra*              illum_spectra_sp,
    int                   illum,
    Spectra*              reflect_spectra_sp,
    Fluorescent_database* fl_db_ptr
);

int world_is_up_to_date(int proposed_world_version);

int get_illum_spectra(Spectra** target_sp_ptr);

int get_reflect_spectra(Spectra** target_sp_ptr);

int get_reflect_spectra_with_specularities(Spectra** target_sp_ptr);

int make_reflect_spectra_with_specularities(void);

int get_specular_spectra(Spectra** target_sp_ptr);

void invalidate_reflectance_with_spec(void);
int set_canon_spectra_by_number(int number);
int set_random_canon_spectra(void);

int get_canon_spectra(Spectra** target_sp_ptr);

int set_random_actual_spectra(void);
int set_actual_spectra_by_number(int number);

int get_actual_spectra(Spectra** target_sp_ptr);

int get_optional_fluorescent_database
(
    Fluorescent_database** fl_db_ptr_ptr
);

int illum_spectra_is_up_to_date(int proposed_version);
int reflect_spectra_is_up_to_date(int proposed_version);
int specular_spectra_is_up_to_date(int proposed_version);
int reflect_spectra_with_spec_is_up_to_date(int proposed_version);
int actual_spectra_is_up_to_date(int proposed_version);
int canon_spectra_is_up_to_date(int proposed_version);
int fluorescent_db_is_up_to_date(int proposed_version);

int get_current_world_info(World_info* world_info_ptr);
int echo_world_info(World_info* world_info_ptr);
int write_world_info(FILE* fp, World_info* world_info_ptr);
int read_world_info(FILE* fp, World_info* world_info_ptr);
int initialize_world_info(World_info* world_info_ptr);

#endif

